/*
Copyright (C) 2018-2019 de4dot@gmail.com

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using Generator.Enums;
using Generator.Enums.Decoder;

namespace Generator.Decoder {
	static class DecoderTable_Legacy {
		public const string OneByteHandlers = nameof(OneByteHandlers);

		public static (string name, object?[] handlers)[] CreateHandlers(GenTypes genTypes) {
			var legacyEnum = genTypes[TypeIds.OpCodeHandlerKind];
			var codeEnum = genTypes[TypeIds.Code];
			var decoderOptionsEnum = genTypes[TypeIds.DecoderOptions];
			var registerEnum = genTypes[TypeIds.Register];
			var handlerFlagsEnum = genTypes[TypeIds.HandlerFlags];
			var legacyHandlerFlagsEnum = genTypes[TypeIds.LegacyHandlerFlags];

			var invalid = new object[] { legacyEnum[nameof(OpCodeHandlerKind.Invalid)] };
			var invalid_NoModRM = new object[] { legacyEnum[nameof(OpCodeHandlerKind.Invalid_NoModRM)] };
			var handlers = new (string name, object?[] handlers)[] {
				("handlers_FPU_D8_low",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fadd_m32fp)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fmul_m32fp)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fcom_m32fp)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fcomp_m32fp)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fsub_m32fp)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fsubr_m32fp)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fdiv_m32fp)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fdivr_m32fp)] },
				}),

				("handlers_FPU_D8_high",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fadd_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fmul_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcom_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomp_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fsub_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fsubr_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fdiv_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fdivr_st0_sti)] },
				}),

				("handlers_FPU_D9_low",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fld_m32fp)] },
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fst_m32fp)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fstp_m32fp)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_2a)], codeEnum[nameof(Code.Fldenv_m14byte)], codeEnum[nameof(Code.Fldenv_m28byte)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fldcw_m2byte)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_2b)], codeEnum[nameof(Code.Fnstenv_m14byte)], codeEnum[nameof(Code.Fnstenv_m28byte)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fnstcw_m2byte)] },
				}),

				("handlers_FPU_D9_high",
				new object[0x40] {
					// C0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fld_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fld_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fld_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fld_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fld_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fld_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fld_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fld_st0_sti)] },

					// C8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti)] },

					// D0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fnop)] },
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Cyrix_D9D7)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},

					// D8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstpnce_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstpnce_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstpnce_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstpnce_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstpnce_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstpnce_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstpnce_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstpnce_sti)] },

					// E0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fchs)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fabs)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Cyrix_D9E2)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Ftst)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fxam)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Ftstp)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Cyrix_D9E7)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},

					// E8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fld1)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fldl2t)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fldl2e)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fldpi)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fldlg2)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fldln2)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fldz)] },
					invalid,

					// F0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.F2xm1)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fyl2x)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fptan)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fpatan)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fxtract)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fprem1)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fdecstp)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fincstp)] },

					// F8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fprem)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fyl2xp1)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fsqrt)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fsincos)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Frndint)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fscale)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fsin)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fcos)] },
				}),

				("handlers_FPU_DA_low",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fiadd_m32int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fimul_m32int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Ficom_m32int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Ficomp_m32int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fisub_m32int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fisubr_m32int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fidiv_m32int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fidivr_m32int)] },
				}),

				("handlers_FPU_DA_high",
				new object[0x40] {
					// C0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovb_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovb_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovb_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovb_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovb_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovb_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovb_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovb_st0_sti)] },

					// C8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmove_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmove_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmove_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmove_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmove_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmove_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmove_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmove_st0_sti)] },

					// D0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovbe_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovbe_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovbe_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovbe_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovbe_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovbe_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovbe_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovbe_st0_sti)] },

					// D8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovu_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovu_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovu_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovu_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovu_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovu_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovu_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovu_st0_sti)] },

					// E0
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// E8
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fucompp)] },
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// F0
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// F8
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
				}),

				("handlers_FPU_DB_low",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fild_m32int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fisttp_m32int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fist_m32int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fistp_m32int)] },
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fld_m80fp)] },
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fstp_m80fp)] },
				}),

				("handlers_FPU_DB_high",
				new object[0x40] {
					// C0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnb_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnb_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnb_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnb_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnb_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnb_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnb_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnb_st0_sti)] },

					// C8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovne_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovne_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovne_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovne_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovne_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovne_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovne_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovne_st0_sti)] },

					// D0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnbe_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnbe_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnbe_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnbe_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnbe_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnbe_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnbe_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnbe_st0_sti)] },

					// D8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnu_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnu_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnu_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnu_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnu_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnu_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnu_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcmovnu_st0_sti)] },

					// E0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fneni)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fndisi)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fnclex)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fninit)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fnsetpm)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options3)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Frstpm)] }, decoderOptionsEnum[nameof(DecoderOptions.OldFpu)]
						},
						invalid,
					},
					invalid,
					invalid,

					// E8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomi_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomi_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomi_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomi_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomi_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomi_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomi_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomi_st0_sti)] },

					// F0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomi_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomi_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomi_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomi_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomi_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomi_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomi_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomi_st0_sti)] },

					// F8
					invalid,
					invalid,
					invalid,
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Frint2)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					invalid,
					invalid,
					invalid,
				}),

				("handlers_FPU_DC_low",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fadd_m64fp)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fmul_m64fp)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fcom_m64fp)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fcomp_m64fp)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fsub_m64fp)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fsubr_m64fp)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fdiv_m64fp)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fdivr_m64fp)] },
				}),

				("handlers_FPU_DC_high",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fadd_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fmul_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcom_st0_sti_DCD0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomp_st0_sti_DCD8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fsubr_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fsub_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fdivr_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fdiv_sti_st0)] },
				}),

				("handlers_FPU_DD_low",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fld_m64fp)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fisttp_m64int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fst_m64fp)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fstp_m64fp)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_2a)], codeEnum[nameof(Code.Frstor_m94byte)], codeEnum[nameof(Code.Frstor_m108byte)] },
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_2b)], codeEnum[nameof(Code.Fnsave_m94byte)], codeEnum[nameof(Code.Fnsave_m108byte)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fnstsw_m2byte)] },
				}),

				("handlers_FPU_DD_high",
				new object[0x40] {
					// C0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Ffree_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Ffree_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Ffree_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Ffree_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Ffree_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Ffree_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Ffree_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Ffree_sti)] },

					// C8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti_DDC8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti_DDC8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti_DDC8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti_DDC8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti_DDC8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti_DDC8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti_DDC8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti_DDC8)] },

					// D0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fst_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fst_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fst_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fst_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fst_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fst_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fst_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fst_sti)] },

					// D8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti)] },

					// E0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucom_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucom_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucom_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucom_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucom_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucom_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucom_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucom_st0_sti)] },

					// E8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomp_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomp_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomp_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomp_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomp_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomp_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomp_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomp_st0_sti)] },

					// F0
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// F8
					invalid,
					invalid,
					invalid,
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Frichop)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					invalid,
					invalid,
					invalid,
				}),

				("handlers_FPU_DE_low",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fiadd_m16int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fimul_m16int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Ficom_m16int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Ficomp_m16int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fisub_m16int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fisubr_m16int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fidiv_m16int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fidivr_m16int)] },
				}),

				("handlers_FPU_DE_high",
				new object[0x40] {
					// C0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Faddp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Faddp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Faddp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Faddp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Faddp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Faddp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Faddp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Faddp_sti_st0)] },

					// C8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fmulp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fmulp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fmulp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fmulp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fmulp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fmulp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fmulp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fmulp_sti_st0)] },

					// D0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomp_st0_sti_DED0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomp_st0_sti_DED0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomp_st0_sti_DED0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomp_st0_sti_DED0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomp_st0_sti_DED0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomp_st0_sti_DED0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomp_st0_sti_DED0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomp_st0_sti_DED0)] },

					// D8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Cyrix_DED8)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Fcompp)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Cyrix_DEDA)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Cyrix_DEDC)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Cyrix_DEDD)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Cyrix_DEDE)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					invalid,

					// E0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fsubrp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fsubrp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fsubrp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fsubrp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fsubrp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fsubrp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fsubrp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fsubrp_sti_st0)] },

					// E8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fsubp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fsubp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fsubp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fsubp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fsubp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fsubp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fsubp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fsubp_sti_st0)] },

					// F0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fdivrp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fdivrp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fdivrp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fdivrp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fdivrp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fdivrp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fdivrp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fdivrp_sti_st0)] },

					// F8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fdivp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fdivp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fdivp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fdivp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fdivp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fdivp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fdivp_sti_st0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi_ST)], codeEnum[nameof(Code.Fdivp_sti_st0)] },
				}),

				("handlers_FPU_DF_low",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fild_m16int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fisttp_m16int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fist_m16int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fistp_m16int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fbld_m80bcd)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fild_m64int)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fbstp_m80bcd)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mf_1)], codeEnum[nameof(Code.Fistp_m64int)] },
				}),

				("handlers_FPU_DF_high",
				new object[0x40] {
					// C0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Ffreep_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Ffreep_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Ffreep_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Ffreep_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Ffreep_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Ffreep_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Ffreep_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Ffreep_sti)] },

					// C8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti_DFC8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti_DFC8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti_DFC8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti_DFC8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti_DFC8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti_DFC8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti_DFC8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fxch_st0_sti_DFC8)] },

					// D0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti_DFD0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti_DFD0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti_DFD0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti_DFD0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti_DFD0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti_DFD0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti_DFD0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti_DFD0)] },

					// D8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti_DFD8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti_DFD8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti_DFD8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti_DFD8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti_DFD8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti_DFD8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti_DFD8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.STi)], codeEnum[nameof(Code.Fstp_sti_DFD8)] },

					// E0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reg)], codeEnum[nameof(Code.Fnstsw_AX)], registerEnum[nameof(Register.AX)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options3)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reg)], codeEnum[nameof(Code.Fnstdw_AX)], registerEnum[nameof(Register.AX)] }, decoderOptionsEnum[nameof(DecoderOptions.OldFpu)]
						},
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options3)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reg)], codeEnum[nameof(Code.Fnstsg_AX)], registerEnum[nameof(Register.AX)] }, decoderOptionsEnum[nameof(DecoderOptions.OldFpu)]
						},
						invalid,
					},
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// E8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomip_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomip_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomip_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomip_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomip_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomip_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomip_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fucomip_st0_sti)] },

					// F0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomip_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomip_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomip_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomip_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomip_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomip_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomip_st0_sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.ST_STi)], codeEnum[nameof(Code.Fcomip_st0_sti)] },

					// F8
					invalid,
					invalid,
					invalid,
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Frinear)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					invalid,
					invalid,
					invalid,
				}),

				("handlers_Grp_80",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_2)], codeEnum[nameof(Code.Add_rm8_imm8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_2)], codeEnum[nameof(Code.Or_rm8_imm8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_2)], codeEnum[nameof(Code.Adc_rm8_imm8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_2)], codeEnum[nameof(Code.Sbb_rm8_imm8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_2)], codeEnum[nameof(Code.And_rm8_imm8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_2)], codeEnum[nameof(Code.Sub_rm8_imm8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_2)], codeEnum[nameof(Code.Xor_rm8_imm8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_1)], codeEnum[nameof(Code.Cmp_rm8_imm8)] },
				}),

				("handlers_Grp_81",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Iz_4)], codeEnum[nameof(Code.Add_rm16_imm16)], codeEnum[nameof(Code.Add_rm32_imm32)], codeEnum[nameof(Code.Add_rm64_imm32)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Iz_4)], codeEnum[nameof(Code.Or_rm16_imm16)], codeEnum[nameof(Code.Or_rm32_imm32)], codeEnum[nameof(Code.Or_rm64_imm32)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Iz_4)], codeEnum[nameof(Code.Adc_rm16_imm16)], codeEnum[nameof(Code.Adc_rm32_imm32)], codeEnum[nameof(Code.Adc_rm64_imm32)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Iz_4)], codeEnum[nameof(Code.Sbb_rm16_imm16)], codeEnum[nameof(Code.Sbb_rm32_imm32)], codeEnum[nameof(Code.Sbb_rm64_imm32)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Iz_4)], codeEnum[nameof(Code.And_rm16_imm16)], codeEnum[nameof(Code.And_rm32_imm32)], codeEnum[nameof(Code.And_rm64_imm32)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Iz_4)], codeEnum[nameof(Code.Sub_rm16_imm16)], codeEnum[nameof(Code.Sub_rm32_imm32)], codeEnum[nameof(Code.Sub_rm64_imm32)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Iz_4)], codeEnum[nameof(Code.Xor_rm16_imm16)], codeEnum[nameof(Code.Xor_rm32_imm32)], codeEnum[nameof(Code.Xor_rm64_imm32)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Iz_3)], codeEnum[nameof(Code.Cmp_rm16_imm16)], codeEnum[nameof(Code.Cmp_rm32_imm32)], codeEnum[nameof(Code.Cmp_rm64_imm32)] },
				}),

				("handlers_Grp_82",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_2)], codeEnum[nameof(Code.Add_rm8_imm8_82)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_2)], codeEnum[nameof(Code.Or_rm8_imm8_82)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_2)], codeEnum[nameof(Code.Adc_rm8_imm8_82)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_2)], codeEnum[nameof(Code.Sbb_rm8_imm8_82)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_2)], codeEnum[nameof(Code.And_rm8_imm8_82)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_2)], codeEnum[nameof(Code.Sub_rm8_imm8_82)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_2)], codeEnum[nameof(Code.Xor_rm8_imm8_82)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_1)], codeEnum[nameof(Code.Cmp_rm8_imm8_82)] },
				}),

				("handlers_Grp_83",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Ib_4)], codeEnum[nameof(Code.Add_rm16_imm8)], codeEnum[nameof(Code.Add_rm32_imm8)], codeEnum[nameof(Code.Add_rm64_imm8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Ib_4)], codeEnum[nameof(Code.Or_rm16_imm8)], codeEnum[nameof(Code.Or_rm32_imm8)], codeEnum[nameof(Code.Or_rm64_imm8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Ib_4)], codeEnum[nameof(Code.Adc_rm16_imm8)], codeEnum[nameof(Code.Adc_rm32_imm8)], codeEnum[nameof(Code.Adc_rm64_imm8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Ib_4)], codeEnum[nameof(Code.Sbb_rm16_imm8)], codeEnum[nameof(Code.Sbb_rm32_imm8)], codeEnum[nameof(Code.Sbb_rm64_imm8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Ib_4)], codeEnum[nameof(Code.And_rm16_imm8)], codeEnum[nameof(Code.And_rm32_imm8)], codeEnum[nameof(Code.And_rm64_imm8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Ib_4)], codeEnum[nameof(Code.Sub_rm16_imm8)], codeEnum[nameof(Code.Sub_rm32_imm8)], codeEnum[nameof(Code.Sub_rm64_imm8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Ib_4)], codeEnum[nameof(Code.Xor_rm16_imm8)], codeEnum[nameof(Code.Xor_rm32_imm8)], codeEnum[nameof(Code.Xor_rm64_imm8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Ib_3)], codeEnum[nameof(Code.Cmp_rm16_imm8)], codeEnum[nameof(Code.Cmp_rm32_imm8)], codeEnum[nameof(Code.Cmp_rm64_imm8)] },
				}),

				("handlers_Grp_8F",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushEv)], codeEnum[nameof(Code.Pop_rm16)], codeEnum[nameof(Code.Pop_rm32)], codeEnum[nameof(Code.Pop_rm64)] },
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
				}),

				("handlers_Grp_C0",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_1)], codeEnum[nameof(Code.Rol_rm8_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_1)], codeEnum[nameof(Code.Ror_rm8_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_1)], codeEnum[nameof(Code.Rcl_rm8_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_1)], codeEnum[nameof(Code.Rcr_rm8_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_1)], codeEnum[nameof(Code.Shl_rm8_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_1)], codeEnum[nameof(Code.Shr_rm8_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_1)], codeEnum[nameof(Code.Sal_rm8_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_1)], codeEnum[nameof(Code.Sar_rm8_imm8)] },
				}),

				("handlers_Grp_C1",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Ib2_3)], codeEnum[nameof(Code.Rol_rm16_imm8)], codeEnum[nameof(Code.Rol_rm32_imm8)], codeEnum[nameof(Code.Rol_rm64_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Ib2_3)], codeEnum[nameof(Code.Ror_rm16_imm8)], codeEnum[nameof(Code.Ror_rm32_imm8)], codeEnum[nameof(Code.Ror_rm64_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Ib2_3)], codeEnum[nameof(Code.Rcl_rm16_imm8)], codeEnum[nameof(Code.Rcl_rm32_imm8)], codeEnum[nameof(Code.Rcl_rm64_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Ib2_3)], codeEnum[nameof(Code.Rcr_rm16_imm8)], codeEnum[nameof(Code.Rcr_rm32_imm8)], codeEnum[nameof(Code.Rcr_rm64_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Ib2_3)], codeEnum[nameof(Code.Shl_rm16_imm8)], codeEnum[nameof(Code.Shl_rm32_imm8)], codeEnum[nameof(Code.Shl_rm64_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Ib2_3)], codeEnum[nameof(Code.Shr_rm16_imm8)], codeEnum[nameof(Code.Shr_rm32_imm8)], codeEnum[nameof(Code.Shr_rm64_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Ib2_3)], codeEnum[nameof(Code.Sal_rm16_imm8)], codeEnum[nameof(Code.Sal_rm32_imm8)], codeEnum[nameof(Code.Sal_rm64_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Ib2_3)], codeEnum[nameof(Code.Sar_rm16_imm8)], codeEnum[nameof(Code.Sar_rm32_imm8)], codeEnum[nameof(Code.Sar_rm64_imm8)] },
				}),

				("handlers_Grp_D0",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb1)], codeEnum[nameof(Code.Rol_rm8_1)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb1)], codeEnum[nameof(Code.Ror_rm8_1)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb1)], codeEnum[nameof(Code.Rcl_rm8_1)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb1)], codeEnum[nameof(Code.Rcr_rm8_1)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb1)], codeEnum[nameof(Code.Shl_rm8_1)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb1)], codeEnum[nameof(Code.Shr_rm8_1)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb1)], codeEnum[nameof(Code.Sal_rm8_1)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb1)], codeEnum[nameof(Code.Sar_rm8_1)] },
				}),

				("handlers_Grp_D1",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev1)], codeEnum[nameof(Code.Rol_rm16_1)], codeEnum[nameof(Code.Rol_rm32_1)], codeEnum[nameof(Code.Rol_rm64_1)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev1)], codeEnum[nameof(Code.Ror_rm16_1)], codeEnum[nameof(Code.Ror_rm32_1)], codeEnum[nameof(Code.Ror_rm64_1)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev1)], codeEnum[nameof(Code.Rcl_rm16_1)], codeEnum[nameof(Code.Rcl_rm32_1)], codeEnum[nameof(Code.Rcl_rm64_1)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev1)], codeEnum[nameof(Code.Rcr_rm16_1)], codeEnum[nameof(Code.Rcr_rm32_1)], codeEnum[nameof(Code.Rcr_rm64_1)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev1)], codeEnum[nameof(Code.Shl_rm16_1)], codeEnum[nameof(Code.Shl_rm32_1)], codeEnum[nameof(Code.Shl_rm64_1)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev1)], codeEnum[nameof(Code.Shr_rm16_1)], codeEnum[nameof(Code.Shr_rm32_1)], codeEnum[nameof(Code.Shr_rm64_1)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev1)], codeEnum[nameof(Code.Sal_rm16_1)], codeEnum[nameof(Code.Sal_rm32_1)], codeEnum[nameof(Code.Sal_rm64_1)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev1)], codeEnum[nameof(Code.Sar_rm16_1)], codeEnum[nameof(Code.Sar_rm32_1)], codeEnum[nameof(Code.Sar_rm64_1)] },
				}),

				("handlers_Grp_D2",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_CL)], codeEnum[nameof(Code.Rol_rm8_CL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_CL)], codeEnum[nameof(Code.Ror_rm8_CL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_CL)], codeEnum[nameof(Code.Rcl_rm8_CL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_CL)], codeEnum[nameof(Code.Rcr_rm8_CL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_CL)], codeEnum[nameof(Code.Shl_rm8_CL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_CL)], codeEnum[nameof(Code.Shr_rm8_CL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_CL)], codeEnum[nameof(Code.Sal_rm8_CL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_CL)], codeEnum[nameof(Code.Sar_rm8_CL)] },
				}),

				("handlers_Grp_D3",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_CL)], codeEnum[nameof(Code.Rol_rm16_CL)], codeEnum[nameof(Code.Rol_rm32_CL)], codeEnum[nameof(Code.Rol_rm64_CL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_CL)], codeEnum[nameof(Code.Ror_rm16_CL)], codeEnum[nameof(Code.Ror_rm32_CL)], codeEnum[nameof(Code.Ror_rm64_CL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_CL)], codeEnum[nameof(Code.Rcl_rm16_CL)], codeEnum[nameof(Code.Rcl_rm32_CL)], codeEnum[nameof(Code.Rcl_rm64_CL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_CL)], codeEnum[nameof(Code.Rcr_rm16_CL)], codeEnum[nameof(Code.Rcr_rm32_CL)], codeEnum[nameof(Code.Rcr_rm64_CL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_CL)], codeEnum[nameof(Code.Shl_rm16_CL)], codeEnum[nameof(Code.Shl_rm32_CL)], codeEnum[nameof(Code.Shl_rm64_CL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_CL)], codeEnum[nameof(Code.Shr_rm16_CL)], codeEnum[nameof(Code.Shr_rm32_CL)], codeEnum[nameof(Code.Shr_rm64_CL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_CL)], codeEnum[nameof(Code.Sal_rm16_CL)], codeEnum[nameof(Code.Sal_rm32_CL)], codeEnum[nameof(Code.Sal_rm64_CL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_CL)], codeEnum[nameof(Code.Sar_rm16_CL)], codeEnum[nameof(Code.Sar_rm32_CL)], codeEnum[nameof(Code.Sar_rm64_CL)] },
				}),

				("handlers_Grp_F6",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_1)], codeEnum[nameof(Code.Test_rm8_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_1)], codeEnum[nameof(Code.Test_rm8_imm8_F6r1)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_2)], codeEnum[nameof(Code.Not_rm8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_2)], codeEnum[nameof(Code.Neg_rm8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_1)], codeEnum[nameof(Code.Mul_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_1)], codeEnum[nameof(Code.Imul_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_1)], codeEnum[nameof(Code.Div_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_1)], codeEnum[nameof(Code.Idiv_rm8)] },
				}),

				("handlers_Grp_F7",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Iz_3)], codeEnum[nameof(Code.Test_rm16_imm16)], codeEnum[nameof(Code.Test_rm32_imm32)], codeEnum[nameof(Code.Test_rm64_imm32)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Iz_3)], codeEnum[nameof(Code.Test_rm16_imm16_F7r1)], codeEnum[nameof(Code.Test_rm32_imm32_F7r1)], codeEnum[nameof(Code.Test_rm64_imm32_F7r1)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_4)], codeEnum[nameof(Code.Not_rm16)], codeEnum[nameof(Code.Not_rm32)], codeEnum[nameof(Code.Not_rm64)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_4)], codeEnum[nameof(Code.Neg_rm16)], codeEnum[nameof(Code.Neg_rm32)], codeEnum[nameof(Code.Neg_rm64)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_3a)], codeEnum[nameof(Code.Mul_rm16)], codeEnum[nameof(Code.Mul_rm32)], codeEnum[nameof(Code.Mul_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_3a)], codeEnum[nameof(Code.Imul_rm16)], codeEnum[nameof(Code.Imul_rm32)], codeEnum[nameof(Code.Imul_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_3a)], codeEnum[nameof(Code.Div_rm16)], codeEnum[nameof(Code.Div_rm32)], codeEnum[nameof(Code.Div_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_3a)], codeEnum[nameof(Code.Idiv_rm16)], codeEnum[nameof(Code.Idiv_rm32)], codeEnum[nameof(Code.Idiv_rm64)] },
				}),

				("handlers_Grp_FE",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_2)], codeEnum[nameof(Code.Inc_rm8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_2)], codeEnum[nameof(Code.Dec_rm8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
				}),

				("handlers_Grp_FF",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_4)], codeEnum[nameof(Code.Inc_rm16)], codeEnum[nameof(Code.Inc_rm32)], codeEnum[nameof(Code.Inc_rm64)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_4)], codeEnum[nameof(Code.Dec_rm16)], codeEnum[nameof(Code.Dec_rm32)], codeEnum[nameof(Code.Dec_rm64)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Evj)], codeEnum[nameof(Code.Call_rm16)], codeEnum[nameof(Code.Call_rm32)], codeEnum[nameof(Code.Call_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ep)], codeEnum[nameof(Code.Call_m1616)], codeEnum[nameof(Code.Call_m1632)], codeEnum[nameof(Code.Call_m1664)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Evj)], codeEnum[nameof(Code.Jmp_rm16)], codeEnum[nameof(Code.Jmp_rm32)], codeEnum[nameof(Code.Jmp_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ep)], codeEnum[nameof(Code.Jmp_m1616)], codeEnum[nameof(Code.Jmp_m1632)], codeEnum[nameof(Code.Jmp_m1664)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushEv)], codeEnum[nameof(Code.Push_rm16)], codeEnum[nameof(Code.Push_rm32)], codeEnum[nameof(Code.Push_rm64)] },
					invalid,
				}),

				("handlers_Grp_0F00",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Evw)], codeEnum[nameof(Code.Sldt_rm16)], codeEnum[nameof(Code.Sldt_r32m16)], codeEnum[nameof(Code.Sldt_r64m16)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Evw)], codeEnum[nameof(Code.Str_rm16)], codeEnum[nameof(Code.Str_r32m16)], codeEnum[nameof(Code.Str_r64m16)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ew)], codeEnum[nameof(Code.Lldt_rm16)], codeEnum[nameof(Code.Lldt_r32m16)], codeEnum[nameof(Code.Lldt_r64m16)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ew)], codeEnum[nameof(Code.Ltr_rm16)], codeEnum[nameof(Code.Ltr_r32m16)], codeEnum[nameof(Code.Ltr_r64m16)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ew)], codeEnum[nameof(Code.Verr_rm16)], codeEnum[nameof(Code.Verr_r32m16)], codeEnum[nameof(Code.Verr_r64m16)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ew)], codeEnum[nameof(Code.Verw_rm16)], codeEnum[nameof(Code.Verw_r32m16)], codeEnum[nameof(Code.Verw_r64m16)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options_DontReadModRM)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness_DontReadModRM)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_3b)], codeEnum[nameof(Code.Jmpe_rm16)], codeEnum[nameof(Code.Jmpe_rm32)] },
							invalid,
						}, decoderOptionsEnum[nameof(DecoderOptions.Jmpe)]
					},
					invalid,
				}),

				("handlers_Grp_0F01_lo",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ms)], codeEnum[nameof(Code.Sgdt_m1632_16)], codeEnum[nameof(Code.Sgdt_m1632)], codeEnum[nameof(Code.Sgdt_m1664)] },
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ms)], codeEnum[nameof(Code.Sidt_m1632_16)], codeEnum[nameof(Code.Sidt_m1632)], codeEnum[nameof(Code.Sidt_m1664)] },
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ms)], codeEnum[nameof(Code.Lgdt_m1632_16)], codeEnum[nameof(Code.Lgdt_m1632)], codeEnum[nameof(Code.Lgdt_m1664)] },
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ms)], codeEnum[nameof(Code.Lidt_m1632_16)], codeEnum[nameof(Code.Lidt_m1632)], codeEnum[nameof(Code.Lidt_m1664)] },
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Evw)], codeEnum[nameof(Code.Smsw_rm16)], codeEnum[nameof(Code.Smsw_r32m16)], codeEnum[nameof(Code.Smsw_r64m16)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Rstorssp_m64)] },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Evw)], codeEnum[nameof(Code.Lmsw_rm16)], codeEnum[nameof(Code.Lmsw_r32m16)], codeEnum[nameof(Code.Lmsw_r64m16)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Invlpg_m)] },
				}),

				("handlers_Grp_0F01_hi",
				new object?[0x40] {
					// C0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Enclv)] },
						invalid_NoModRM,
						invalid_NoModRM,
						invalid_NoModRM,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Vmcall)] },
						invalid_NoModRM,
						invalid_NoModRM,
						invalid_NoModRM,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Vmlaunch)] },
						invalid_NoModRM,
						invalid_NoModRM,
						invalid_NoModRM,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Vmresume)] },
						invalid_NoModRM,
						invalid_NoModRM,
						invalid_NoModRM,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Vmxoff)] },
						invalid_NoModRM,
						invalid_NoModRM,
						invalid_NoModRM,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Pconfig)] },
						invalid_NoModRM,
						invalid_NoModRM,
						invalid_NoModRM,
					},
					null,
					null,

					// C8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5)], codeEnum[nameof(Code.Monitorw)], codeEnum[nameof(Code.Monitord)], codeEnum[nameof(Code.Monitorq)] },
						invalid_NoModRM,
						invalid_NoModRM,
						invalid_NoModRM,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Mwait)] },
						invalid_NoModRM,
						invalid_NoModRM,
						invalid_NoModRM,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Clac)] },
						invalid_NoModRM,
						invalid_NoModRM,
						invalid_NoModRM,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Stac)] },
						invalid_NoModRM,
						invalid_NoModRM,
						invalid_NoModRM,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						invalid_NoModRM,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Tdcall)] },
						invalid_NoModRM,
						invalid_NoModRM,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						invalid_NoModRM,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
							invalid_NoModRM,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Seamret)] },
						},
						invalid_NoModRM,
						invalid_NoModRM,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						invalid_NoModRM,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
							invalid_NoModRM,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Seamops)] },
						},
						invalid_NoModRM,
						invalid_NoModRM,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Encls)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
							invalid_NoModRM,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Seamcall)] },
						},
						invalid_NoModRM,
						invalid_NoModRM,
					},

					// D0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Xgetbv)] },
						invalid_NoModRM,
						invalid_NoModRM,
						invalid_NoModRM,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Xsetbv)] },
						invalid_NoModRM,
						invalid_NoModRM,
						invalid_NoModRM,
					},
					null,
					null,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Vmfunc)] },
						invalid_NoModRM,
						invalid_NoModRM,
						invalid_NoModRM,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Xend)] },
						invalid_NoModRM,
						invalid_NoModRM,
						invalid_NoModRM,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Xtest)] },
						invalid_NoModRM,
						invalid_NoModRM,
						invalid_NoModRM,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Enclu)] },
						invalid_NoModRM,
						invalid_NoModRM,
						invalid_NoModRM,
					},

					// D8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5)], codeEnum[nameof(Code.Vmrunw)], codeEnum[nameof(Code.Vmrund)], codeEnum[nameof(Code.Vmrunq)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Vmmcall)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Vmmcall)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Vmgexit)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Vmgexit_F2)] },
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5)], codeEnum[nameof(Code.Vmloadw)], codeEnum[nameof(Code.Vmloadd)], codeEnum[nameof(Code.Vmloadq)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5)], codeEnum[nameof(Code.Vmsavew)], codeEnum[nameof(Code.Vmsaved)], codeEnum[nameof(Code.Vmsaveq)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Stgi)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Clgi)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Skinit)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5)], codeEnum[nameof(Code.Invlpgaw)], codeEnum[nameof(Code.Invlpgad)], codeEnum[nameof(Code.Invlpgaq)] },

					// E0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// E8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Serialize)] },
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Setssbsy)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Xsusldtrk)] },
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						invalid,
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Xresldtrk)] },
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Saveprevssp)] },
						invalid,
					},
					null,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						invalid_NoModRM,
						invalid_NoModRM,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
							invalid_NoModRM,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Uiret)] },
						},
						invalid_NoModRM,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						invalid_NoModRM,
						invalid_NoModRM,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
							invalid_NoModRM,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Testui)] },
						},
						invalid_NoModRM,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Rdpkru)] },
						invalid_NoModRM,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						invalid_NoModRM,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Clui)] },
						},
						invalid_NoModRM,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Wrpkru)] },
						invalid_NoModRM,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
							invalid_NoModRM,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Stui)] },
						},
						invalid_NoModRM,
					},

					// F0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// F8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness_DontReadModRM)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Swapgs)] },
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Rdtscp)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5)], codeEnum[nameof(Code.Monitorxw)], codeEnum[nameof(Code.Monitorxd)], codeEnum[nameof(Code.Monitorxq)] },
						invalid_NoModRM,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Mcommit)] },
						invalid_NoModRM,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Mwaitx)] },
						invalid_NoModRM,
						invalid_NoModRM,
						invalid_NoModRM,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5)], codeEnum[nameof(Code.Clzerow)], codeEnum[nameof(Code.Clzerod)], codeEnum[nameof(Code.Clzeroq)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Rdpru)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5)], codeEnum[nameof(Code.Invlpgbw)], codeEnum[nameof(Code.Invlpgbd)], codeEnum[nameof(Code.Invlpgbq)] },
						invalid_NoModRM,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
							invalid_NoModRM,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Rmpadjust)] },
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
							invalid_NoModRM,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Rmpupdate)] },
						},
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Tlbsync)] },
						invalid_NoModRM,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
							invalid_NoModRM,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Psmash)] },
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5)], codeEnum[nameof(Code.Pvalidatew)], codeEnum[nameof(Code.Pvalidated)], codeEnum[nameof(Code.Pvalidateq)] },
					},
				}),

				("handlers_Grp_0F36_Cyrix",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW_1a)], codeEnum[nameof(Code.Rdshr_rm32)], 0x03 },
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
				}),

				("handlers_Grp_0F37_Cyrix",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW_1a)], codeEnum[nameof(Code.Wrshr_rm32)], 0x03 },
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
				}),

				("handlers_Grp_0F7A_Cyrix",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW_1a)], codeEnum[nameof(Code.Svldt_m80)], 0x02 },
					},
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
				}),

				("handlers_Grp_0F7B_Cyrix",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW_1a)], codeEnum[nameof(Code.Rsldt_m80)], 0x02 },
					},
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
				}),

				("handlers_Grp_0F7C_Cyrix",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW_1a)], codeEnum[nameof(Code.Svts_m80)], 0x02 },
					},
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
				}),

				("handlers_Grp_0F7D_Cyrix",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW_1a)], codeEnum[nameof(Code.Rsts_m80)], 0x02 },
					},
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
				}),

				("handlers_Grp_0FA6_lo",
				new object[8] {
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
				}),

				("handlers_Grp_0FA6_hi",
				new object?[0x40] {
					// C0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix4)],
						invalid_NoModRM,
						invalid_NoModRM,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5)], codeEnum[nameof(Code.Montmul_16)], codeEnum[nameof(Code.Montmul_32)], codeEnum[nameof(Code.Montmul_64)] },
						invalid_NoModRM,
						0x00,
					},
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// C8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix4)],
						invalid_NoModRM,
						invalid_NoModRM,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5)], codeEnum[nameof(Code.Xsha1_16)], codeEnum[nameof(Code.Xsha1_32)], codeEnum[nameof(Code.Xsha1_64)] },
						invalid_NoModRM,
						0x00,
					},
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// D0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix4)],
						invalid_NoModRM,
						invalid_NoModRM,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5)], codeEnum[nameof(Code.Xsha256_16)], codeEnum[nameof(Code.Xsha256_32)], codeEnum[nameof(Code.Xsha256_64)] },
						invalid_NoModRM,
						0x00,
					},
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// D8
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// E0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// E8
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// F0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// F8
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,
				}),

				("handlers_Grp_0FA7_lo",
				new object[8] {
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
				}),

				("handlers_Grp_0FA7_hi",
				new object?[0x40] {
					// C0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5)], codeEnum[nameof(Code.Xstore_16)], codeEnum[nameof(Code.Xstore_32)], codeEnum[nameof(Code.Xstore_64)] },
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// C8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix4)],
						invalid_NoModRM,
						invalid_NoModRM,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5)], codeEnum[nameof(Code.Xcryptecb_16)], codeEnum[nameof(Code.Xcryptecb_32)], codeEnum[nameof(Code.Xcryptecb_64)] },
						invalid_NoModRM,
						0x00,
					},
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// D0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix4)],
						invalid_NoModRM,
						invalid_NoModRM,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5)], codeEnum[nameof(Code.Xcryptcbc_16)], codeEnum[nameof(Code.Xcryptcbc_32)], codeEnum[nameof(Code.Xcryptcbc_64)] },
						invalid_NoModRM,
						0x00,
					},
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// D8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix4)],
						invalid_NoModRM,
						invalid_NoModRM,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5)], codeEnum[nameof(Code.Xcryptctr_16)], codeEnum[nameof(Code.Xcryptctr_32)], codeEnum[nameof(Code.Xcryptctr_64)] },
						invalid_NoModRM,
						0x00,
					},
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// E0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix4)],
						invalid_NoModRM,
						invalid_NoModRM,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5)], codeEnum[nameof(Code.Xcryptcfb_16)], codeEnum[nameof(Code.Xcryptcfb_32)], codeEnum[nameof(Code.Xcryptcfb_64)] },
						invalid_NoModRM,
						0x00,
					},
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// E8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix4)],
						invalid_NoModRM,
						invalid_NoModRM,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5)], codeEnum[nameof(Code.Xcryptofb_16)], codeEnum[nameof(Code.Xcryptofb_32)], codeEnum[nameof(Code.Xcryptofb_64)] },
						invalid_NoModRM,
						0x00,
					},
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// F0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// F8
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,
				}),

				("handlers_Grp_0FBA",
				new object[8] {
					invalid,
					invalid,
					invalid,
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Ib2_3)], codeEnum[nameof(Code.Bt_rm16_imm8)], codeEnum[nameof(Code.Bt_rm32_imm8)], codeEnum[nameof(Code.Bt_rm64_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Ib2_4)], codeEnum[nameof(Code.Bts_rm16_imm8)], codeEnum[nameof(Code.Bts_rm32_imm8)], codeEnum[nameof(Code.Bts_rm64_imm8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Ib2_4)], codeEnum[nameof(Code.Btr_rm16_imm8)], codeEnum[nameof(Code.Btr_rm32_imm8)], codeEnum[nameof(Code.Btr_rm64_imm8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Ib2_4)], codeEnum[nameof(Code.Btc_rm16_imm8)], codeEnum[nameof(Code.Btc_rm32_imm8)], codeEnum[nameof(Code.Btc_rm64_imm8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
				}),

				("handlers_Grp_0FC7",
				new object[8] {
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_REXW_4)], codeEnum[nameof(Code.Cmpxchg8b_m64)], codeEnum[nameof(Code.Cmpxchg16b_m128)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)), handlerFlagsEnum[nameof(HandlerFlags.Lock)] },
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_REXW_2)], codeEnum[nameof(Code.Xrstors_mem)], codeEnum[nameof(Code.Xrstors64_mem)] },
						invalid,
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_REXW_2)], codeEnum[nameof(Code.Xsavec_mem)], codeEnum[nameof(Code.Xsavec64_mem)] },
						invalid,
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Xsaves_mem)], codeEnum[nameof(Code.Xsaves64_mem)], 0x02 },
						},
						invalid,
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix3)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Rv)], codeEnum[nameof(Code.Rdrand_r16)], codeEnum[nameof(Code.Rdrand_r32)], codeEnum[nameof(Code.Rdrand_r64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Vmptrld_m64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Rv)], codeEnum[nameof(Code.Rdrand_r16)], codeEnum[nameof(Code.Rdrand_r32)], codeEnum[nameof(Code.Rdrand_r64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Vmclear_m64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness_DontReadModRM)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Rq)], codeEnum[nameof(Code.Senduipi_r64)] },
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Vmxon_m64)] },
						invalid,
						invalid,
						new OrEnumValue(legacyHandlerFlagsEnum, nameof(LegacyHandlerFlags.HandlerReg), nameof(LegacyHandlerFlags.Handler66Reg))
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix3)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Rv)], codeEnum[nameof(Code.Rdseed_r16)], codeEnum[nameof(Code.Rdseed_r32)], codeEnum[nameof(Code.Rdseed_r64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Vmptrst_m64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Rv)], codeEnum[nameof(Code.Rdseed_r16)], codeEnum[nameof(Code.Rdseed_r32)], codeEnum[nameof(Code.Rdseed_r64)] },
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Rv_32_64)], codeEnum[nameof(Code.Rdpid_r32)], codeEnum[nameof(Code.Rdpid_r64)] },
						invalid,
						invalid,
						invalid,
						new OrEnumValue(legacyHandlerFlagsEnum, nameof(LegacyHandlerFlags.HandlerReg), nameof(LegacyHandlerFlags.Handler66Reg))
					},
				}),

				("handlers_Grp_C6_lo",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Ib_2)], codeEnum[nameof(Code.Mov_rm8_imm8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.XacquireXreleaseNoLock)) },
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
				}),

				("handlers_Grp_C6_hi",
				new object?[0x40] {
					// C0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// C8
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// D0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// D8
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// E0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// E8
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// F0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// F8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ib3)], codeEnum[nameof(Code.Xabort_imm8)] },
					null,
					null,
					null,
					null,
					null,
					null,
					null,
				}),

				("handlers_Grp_C7_lo",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Iz_4)], codeEnum[nameof(Code.Mov_rm16_imm16)], codeEnum[nameof(Code.Mov_rm32_imm32)], codeEnum[nameof(Code.Mov_rm64_imm32)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.XacquireXreleaseNoLock)) },
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
				}),

				("handlers_Grp_C7_hi",
				new object?[0x40] {
					// C0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// C8
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// D0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// D8
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// E0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// E8
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// F0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// F8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jx)], codeEnum[nameof(Code.Xbegin_rel16)], codeEnum[nameof(Code.Xbegin_rel32)], codeEnum[nameof(Code.Xbegin_rel32)] },
					null,
					null,
					null,
					null,
					null,
					null,
					null,
				}),

				("handlers_Grp_0F71",
				new object[8] {
					invalid,
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.NIb)], codeEnum[nameof(Code.Psrlw_mm_imm8)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RIb)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psrlw_xmm_imm8)] },
						invalid,
						invalid,
					},
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.NIb)], codeEnum[nameof(Code.Psraw_mm_imm8)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RIb)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psraw_xmm_imm8)] },
						invalid,
						invalid,
					},
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.NIb)], codeEnum[nameof(Code.Psllw_mm_imm8)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RIb)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psllw_xmm_imm8)] },
						invalid,
						invalid,
					},
					invalid,
				}),

				("handlers_Grp_0F72",
				new object[8] {
					invalid,
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.NIb)], codeEnum[nameof(Code.Psrld_mm_imm8)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RIb)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psrld_xmm_imm8)] },
						invalid,
						invalid,
					},
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.NIb)], codeEnum[nameof(Code.Psrad_mm_imm8)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RIb)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psrad_xmm_imm8)] },
						invalid,
						invalid,
					},
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.NIb)], codeEnum[nameof(Code.Pslld_mm_imm8)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RIb)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pslld_xmm_imm8)] },
						invalid,
						invalid,
					},
					invalid,
				}),

				("handlers_Grp_0F73",
				new object[8] {
					invalid,
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.NIb)], codeEnum[nameof(Code.Psrlq_mm_imm8)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RIb)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psrlq_xmm_imm8)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RIb)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psrldq_xmm_imm8)] },
						invalid,
						invalid,
					},
					invalid,
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.NIb)], codeEnum[nameof(Code.Psllq_mm_imm8)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RIb)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psllq_xmm_imm8)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RIb)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pslldq_xmm_imm8)] },
						invalid,
						invalid,
					},
				}),

				("handlers_Grp_0FAE_lo",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_2)], codeEnum[nameof(Code.Fxsave_m512byte)], codeEnum[nameof(Code.Fxsave64_m512byte)] },
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness_DontReadModRM)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Rdfsbase_r32)], codeEnum[nameof(Code.Rdfsbase_r64)], 0x01 },
								invalid,
							},
						},
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_2)], codeEnum[nameof(Code.Fxrstor_m512byte)], codeEnum[nameof(Code.Fxrstor64_m512byte)] },
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness_DontReadModRM)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Rdgsbase_r32)], codeEnum[nameof(Code.Rdgsbase_r64)], 0x01 },
								invalid,
							},
						},
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Ldmxcsr_m32)] },
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness_DontReadModRM)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Wrfsbase_r32)], codeEnum[nameof(Code.Wrfsbase_r64)], 0x01 },
								invalid,
							},
						},
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Stmxcsr_m32)] },
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness_DontReadModRM)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Wrgsbase_r32)], codeEnum[nameof(Code.Wrgsbase_r64)], 0x01 },
								invalid,
							},
						},
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_2)], codeEnum[nameof(Code.Xsave_mem)], codeEnum[nameof(Code.Xsave64_mem)] },
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Ptwrite_rm32)], codeEnum[nameof(Code.Ptwrite_rm64)], 0x07 },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_2)], codeEnum[nameof(Code.Xrstor_mem)], codeEnum[nameof(Code.Xrstor64_mem)] },
						invalid,
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_2)], codeEnum[nameof(Code.Xsaveopt_mem)], codeEnum[nameof(Code.Xsaveopt64_mem)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Clwb_m8)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Clrssbsy_m64)] },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Clflush_m8)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Clflushopt_m8)] },
						invalid,
						invalid,
					},
				}),

				("handlers_Grp_0FAE_hi",
				new object?[0x40] {
					// C0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// C8
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// D0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// D8
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// E0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// E8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Lfence)] },
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Incsspd_r32)], codeEnum[nameof(Code.Incsspq_r64)], 0x01 },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Lfence_E9)] },
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Incsspd_r32)], codeEnum[nameof(Code.Incsspq_r64)], 0x01 },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Lfence_EA)] },
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Incsspd_r32)], codeEnum[nameof(Code.Incsspq_r64)], 0x01 },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Lfence_EB)] },
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Incsspd_r32)], codeEnum[nameof(Code.Incsspq_r64)], 0x01 },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Lfence_EC)] },
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Incsspd_r32)], codeEnum[nameof(Code.Incsspq_r64)], 0x01 },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Lfence_ED)] },
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Incsspd_r32)], codeEnum[nameof(Code.Incsspq_r64)], 0x01 },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Lfence_EE)] },
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Incsspd_r32)], codeEnum[nameof(Code.Incsspq_r64)], 0x01 },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Lfence_EF)] },
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Incsspd_r32)], codeEnum[nameof(Code.Incsspq_r64)], 0x01 },
						invalid,
					},

					// F0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Mfence)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Tpause_r32)], codeEnum[nameof(Code.Tpause_r64)], 0x01 },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5_ModRM_as)], codeEnum[nameof(Code.Umonitor_r16)], codeEnum[nameof(Code.Umonitor_r32)], codeEnum[nameof(Code.Umonitor_r64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Umwait_r32)], codeEnum[nameof(Code.Umwait_r64)], 0x01 }
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Mfence_F1)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Tpause_r32)], codeEnum[nameof(Code.Tpause_r64)], 0x01 },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5_ModRM_as)], codeEnum[nameof(Code.Umonitor_r16)], codeEnum[nameof(Code.Umonitor_r32)], codeEnum[nameof(Code.Umonitor_r64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Umwait_r32)], codeEnum[nameof(Code.Umwait_r64)], 0x01 }
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Mfence_F2)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Tpause_r32)], codeEnum[nameof(Code.Tpause_r64)], 0x01 },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5_ModRM_as)], codeEnum[nameof(Code.Umonitor_r16)], codeEnum[nameof(Code.Umonitor_r32)], codeEnum[nameof(Code.Umonitor_r64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Umwait_r32)], codeEnum[nameof(Code.Umwait_r64)], 0x01 }
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Mfence_F3)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Tpause_r32)], codeEnum[nameof(Code.Tpause_r64)], 0x01 },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5_ModRM_as)], codeEnum[nameof(Code.Umonitor_r16)], codeEnum[nameof(Code.Umonitor_r32)], codeEnum[nameof(Code.Umonitor_r64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Umwait_r32)], codeEnum[nameof(Code.Umwait_r64)], 0x01 }
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Mfence_F4)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Tpause_r32)], codeEnum[nameof(Code.Tpause_r64)], 0x01 },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5_ModRM_as)], codeEnum[nameof(Code.Umonitor_r16)], codeEnum[nameof(Code.Umonitor_r32)], codeEnum[nameof(Code.Umonitor_r64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Umwait_r32)], codeEnum[nameof(Code.Umwait_r64)], 0x01 }
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Mfence_F5)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Tpause_r32)], codeEnum[nameof(Code.Tpause_r64)], 0x01 },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5_ModRM_as)], codeEnum[nameof(Code.Umonitor_r16)], codeEnum[nameof(Code.Umonitor_r32)], codeEnum[nameof(Code.Umonitor_r64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Umwait_r32)], codeEnum[nameof(Code.Umwait_r64)], 0x01 }
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Mfence_F6)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Tpause_r32)], codeEnum[nameof(Code.Tpause_r64)], 0x01 },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5_ModRM_as)], codeEnum[nameof(Code.Umonitor_r16)], codeEnum[nameof(Code.Umonitor_r32)], codeEnum[nameof(Code.Umonitor_r64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Umwait_r32)], codeEnum[nameof(Code.Umwait_r64)], 0x01 }
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Mfence_F7)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Tpause_r32)], codeEnum[nameof(Code.Tpause_r64)], 0x01 },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple5_ModRM_as)], codeEnum[nameof(Code.Umonitor_r16)], codeEnum[nameof(Code.Umonitor_r32)], codeEnum[nameof(Code.Umonitor_r64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Umwait_r32)], codeEnum[nameof(Code.Umwait_r64)], 0x01 }
					},

					// F8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Sfence)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options_DontReadModRM)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Pcommit)] }, decoderOptionsEnum[nameof(DecoderOptions.Pcommit)]
						},
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Sfence_F9)] },
						invalid,
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Sfence_FA)] },
						invalid,
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Sfence_FB)] },
						invalid,
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Sfence_FC)] },
						invalid,
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Sfence_FD)] },
						invalid,
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Sfence_FE)] },
						invalid,
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Sfence_FF)] },
						invalid,
						invalid,
						invalid,
					},
				}),

				("reservedNop_0F0D", new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_3a)], codeEnum[nameof(Code.Reservednop_rm16_r16_0F0D)], codeEnum[nameof(Code.Reservednop_rm32_r32_0F0D)], codeEnum[nameof(Code.Reservednop_rm64_r64_0F0D)] }),
				("reservedNop_0F18", new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_3a)], codeEnum[nameof(Code.Reservednop_rm16_r16_0F18)], codeEnum[nameof(Code.Reservednop_rm32_r32_0F18)], codeEnum[nameof(Code.Reservednop_rm64_r64_0F18)] }),
				("reservedNop_0F19", new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_3a)], codeEnum[nameof(Code.Reservednop_rm16_r16_0F19)], codeEnum[nameof(Code.Reservednop_rm32_r32_0F19)], codeEnum[nameof(Code.Reservednop_rm64_r64_0F19)] }),
				("reservedNop_0F1A", new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_3a)], codeEnum[nameof(Code.Reservednop_rm16_r16_0F1A)], codeEnum[nameof(Code.Reservednop_rm32_r32_0F1A)], codeEnum[nameof(Code.Reservednop_rm64_r64_0F1A)] }),
				("reservedNop_0F1B", new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_3a)], codeEnum[nameof(Code.Reservednop_rm16_r16_0F1B)], codeEnum[nameof(Code.Reservednop_rm32_r32_0F1B)], codeEnum[nameof(Code.Reservednop_rm64_r64_0F1B)] }),
				("reservedNop_0F1C", new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_3a)], codeEnum[nameof(Code.Reservednop_rm16_r16_0F1C)], codeEnum[nameof(Code.Reservednop_rm32_r32_0F1C)], codeEnum[nameof(Code.Reservednop_rm64_r64_0F1C)] }),
				("reservedNop_0F1D", new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_3a)], codeEnum[nameof(Code.Reservednop_rm16_r16_0F1D)], codeEnum[nameof(Code.Reservednop_rm32_r32_0F1D)], codeEnum[nameof(Code.Reservednop_rm64_r64_0F1D)] }),
				("reservedNop_0F1E", new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_3a)], codeEnum[nameof(Code.Reservednop_rm16_r16_0F1E)], codeEnum[nameof(Code.Reservednop_rm32_r32_0F1E)], codeEnum[nameof(Code.Reservednop_rm64_r64_0F1E)] }),
				("reservedNop_0F1F", new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_3a)], codeEnum[nameof(Code.Reservednop_rm16_r16_0F1F)], codeEnum[nameof(Code.Reservednop_rm32_r32_0F1F)], codeEnum[nameof(Code.Reservednop_rm64_r64_0F1F)] }),

				("handlers_Grp_0F0D_mem",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Prefetch_m8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Prefetchw_m8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Prefetchwt1_m8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Prefetchreserved3_m8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Prefetchreserved4_m8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Prefetchreserved5_m8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Prefetchreserved6_m8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Prefetchreserved7_m8)] },
				}),
				("grp0F0D", new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
					"reservedNop_0F0D",
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_0F0D_mem" }
				}),

				("handlers_Grp_0F18_mem",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Prefetchnta_m8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Prefetcht0_m8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Prefetcht1_m8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Prefetcht2_m8)] },
					"reservedNop_0F18",
					"reservedNop_0F18",
					"reservedNop_0F18",
					"reservedNop_0F18",
				}),
				("grp0F18", new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reservednop)],
					"reservedNop_0F18",
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
						"reservedNop_0F18",
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_0F18_mem" }
					}
				}),

				("handlers_Grp_0F1C_mem",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix4)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Cldemote_m8)] },
						"reservedNop_0F1C",
						"reservedNop_0F1C",
						"reservedNop_0F1C",
						0,
					},
					"reservedNop_0F1C",
					"reservedNop_0F1C",
					"reservedNop_0F1C",
					"reservedNop_0F1C",
					"reservedNop_0F1C",
					"reservedNop_0F1C",
					"reservedNop_0F1C",
				}),
				("grp0F1C", new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reservednop)],
					"reservedNop_0F1C",
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
						"reservedNop_0F1C",
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_0F1C_mem" }
					}
				}),

				("handlers_Grp_0F1E_reg_lo",
				new object[8] {
					"reservedNop_0F1E",
					"reservedNop_0F1E",
					"reservedNop_0F1E",
					"reservedNop_0F1E",
					"reservedNop_0F1E",
					"reservedNop_0F1E",
					"reservedNop_0F1E",
					"reservedNop_0F1E",
				}),
				("grp0F1E_1", new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
					"reservedNop_0F1E",
					"reservedNop_0F1E",
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_REXW)], codeEnum[nameof(Code.Rdsspd_r32)], codeEnum[nameof(Code.Rdsspq_r64)], 0x01 },
						"reservedNop_0F1E"
					},
					"reservedNop_0F1E"
				}),
				("handlers_Grp_0F1E_reg_hi",
				new object?[0x40] {
					// C0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// C8
					"grp0F1E_1",
					"grp0F1E_1",
					"grp0F1E_1",
					"grp0F1E_1",
					"grp0F1E_1",
					"grp0F1E_1",
					"grp0F1E_1",
					"grp0F1E_1",

					// D0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// D8
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// E0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// E8
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// F0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// F8
					null,
					null,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						"reservedNop_0F1E",
						"reservedNop_0F1E",
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Endbr64)] },
						"reservedNop_0F1E",
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						"reservedNop_0F1E",
						"reservedNop_0F1E",
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple_ModRM)], codeEnum[nameof(Code.Endbr32)] },
						"reservedNop_0F1E",
					},
					null,
					null,
					null,
					null,
				}),
				("grp0F1E", new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reservednop)],
					"reservedNop_0F1E",
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group8x64)], "handlers_Grp_0F1E_reg_lo", "handlers_Grp_0F1E_reg_hi" },
						"reservedNop_0F1E"
					}
				}),

				("handlers_Grp_0F1F",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_3a)], codeEnum[nameof(Code.Nop_rm16)], codeEnum[nameof(Code.Nop_rm32)], codeEnum[nameof(Code.Nop_rm64)] },
					"reservedNop_0F1F",
					"reservedNop_0F1F",
					"reservedNop_0F1F",
					"reservedNop_0F1F",
					"reservedNop_0F1F",
					"reservedNop_0F1F",
					"reservedNop_0F1F",
				}),
				("grp0F1F", new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reservednop)],
					"reservedNop_0F1F",
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_0F1F" }
				}),

				("handlers_Grp_660F78",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RIbIb)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Extrq_xmm_imm8_imm8)] },
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
				}),

				("handlers_Grp_F30F38D8",
				new object[8] {
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Aesencwide128kl_m384)] },
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Aesdecwide128kl_m384)] },
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Aesencwide256kl_m512)] },
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_1)], codeEnum[nameof(Code.Aesdecwide256kl_m512)] },
					},
					invalid,
					invalid,
					invalid,
					invalid,
				}),

				("handlers_Grp_0F3AF0_lo",
				new object[8] {
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
				}),

				("handlers_Grp_0F3AF0_hi",
				new object?[0x40] {
					// C0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						invalid_NoModRM,
						invalid_NoModRM,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ib)], codeEnum[nameof(Code.Hreset_imm8)] },
						invalid_NoModRM,
					},
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// C8
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// D0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// D8
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// E0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// E8
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// F0
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,

					// F8
					null,
					null,
					null,
					null,
					null,
					null,
					null,
					null,
				}),

				("ThreeByteHandlers_0F38XX",
				new object[0x100] {
					// 00
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pshufb_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pshufb_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Phaddw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Phaddw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Phaddd_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Phaddd_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Phaddsw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Phaddsw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pmaddubsw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmaddubsw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Phsubw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Phsubw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Phsubd_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Phsubd_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Phsubsw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Phsubsw_xmm_xmmm128)] },
						invalid,
						invalid,
					},

					// 08
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Psignb_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psignb_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Psignw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psignw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Psignd_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psignd_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pmulhrsw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmulhrsw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					invalid,
					invalid,
					invalid,
					invalid,

					// 10
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pblendvb_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					invalid,
					invalid,
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Blendvps_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Blendvpd_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Ptest_xmm_xmmm128)] },
						invalid,
						invalid,
					},

					// 18
					invalid,
					invalid,
					invalid,
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pabsb_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pabsb_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pabsw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pabsw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pabsd_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pabsd_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					invalid,

					// 20
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmovsxbw_xmm_xmmm64)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmovsxbd_xmm_xmmm32)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmovsxbq_xmm_xmmm16)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmovsxwd_xmm_xmmm64)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmovsxwq_xmm_xmmm32)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmovsxdq_xmm_xmmm64)] },
						invalid,
						invalid,
					},
					invalid,
					invalid,

					// 28
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmuldq_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pcmpeqq_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VM)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movntdqa_xmm_m128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Packusdw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					invalid,
					invalid,
					invalid,
					invalid,

					// 30
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmovzxbw_xmm_xmmm64)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmovzxbd_xmm_xmmm32)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmovzxbq_xmm_xmmm16)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmovzxwd_xmm_xmmm64)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmovzxwq_xmm_xmmm32)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmovzxdq_xmm_xmmm64)] },
						invalid,
						invalid,
					},
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pcmpgtq_xmm_xmmm128)] },
						invalid,
						invalid,
					},

					// 38
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pminsb_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pminsd_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pminuw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pminud_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmaxsb_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmaxsd_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmaxuw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmaxud_xmm_xmmm128)] },
						invalid,
						invalid,
					},

					// 40
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmulld_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Phminposuw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 48
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 50
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 58
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 60
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 68
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 70
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 78
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 80
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_32_64)], codeEnum[nameof(Code.Invept_r32_m128)], codeEnum[nameof(Code.Invept_r64_m128)], false, true },
						},
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_32_64)], codeEnum[nameof(Code.Invvpid_r32_m128)], codeEnum[nameof(Code.Invvpid_r64_m128)], false, true },
						},
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_32_64)], codeEnum[nameof(Code.Invpcid_r32_m128)], codeEnum[nameof(Code.Invpcid_r64_m128)], false, true },
						},
						invalid,
						invalid,
					},
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 88
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 90
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 98
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// A0
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// A8
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// B0
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// B8
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// C0
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// C8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Sha1nexte_xmm_xmmm128)] },
						invalid,
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Sha1msg1_xmm_xmmm128)] },
						invalid,
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Sha1msg2_xmm_xmmm128)] },
						invalid,
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Sha256rnds2_xmm_xmmm128)] },
						invalid,
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Sha256msg1_xmm_xmmm128)] },
						invalid,
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Sha256msg2_xmm_xmmm128)] },
						invalid,
						invalid,
						invalid,
					},
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Gf2p8mulb_xmm_xmmm128)] },
						invalid,
						invalid,
					},

					// D0
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// D8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_F30F38D8" },
						invalid,
					},
					invalid,
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Aesimc_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Aesenc_xmm_xmmm128)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_3)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Loadiwkey_xmm_xmm)], codeEnum[nameof(Code.Aesenc128kl_xmm_m384)] },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Aesenclast_xmm_xmmm128)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Aesdec128kl_xmm_m384)] },
						},
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Aesdec_xmm_xmmm128)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Aesenc256kl_xmm_m512)] },
						},
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Aesdeclast_xmm_xmmm128)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Aesdec256kl_xmm_m512)] },
						},
						invalid,
					},

					// E0
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// E8
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// F0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix4)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Mv)], codeEnum[nameof(Code.Movbe_r16_m16)], codeEnum[nameof(Code.Movbe_r32_m32)], codeEnum[nameof(Code.Movbe_r64_m64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Mv)], codeEnum[nameof(Code.Movbe_r16_m16)], codeEnum[nameof(Code.Movbe_r32_m32)], codeEnum[nameof(Code.Movbe_r64_m64)] },
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Eb_REX)], codeEnum[nameof(Code.Crc32_r32_rm8)], codeEnum[nameof(Code.Crc32_r64_rm8)] },
						0x1C,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix4)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mv_Gv)], codeEnum[nameof(Code.Movbe_m16_r16)], codeEnum[nameof(Code.Movbe_m32_r32)], codeEnum[nameof(Code.Movbe_m64_r64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mv_Gv)], codeEnum[nameof(Code.Movbe_m16_r16)], codeEnum[nameof(Code.Movbe_m32_r32)], codeEnum[nameof(Code.Movbe_m64_r64)] },
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gdq_Ev)], codeEnum[nameof(Code.Crc32_r32_rm16)], codeEnum[nameof(Code.Crc32_r32_rm32)], codeEnum[nameof(Code.Crc32_r64_rm64)] },
						0x1C,
					},
					invalid,
					invalid,
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_REX)], codeEnum[nameof(Code.Wrussd_m32_r32)], codeEnum[nameof(Code.Wrussq_m64_r64)] }
						},
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_REX)], codeEnum[nameof(Code.Wrssd_m32_r32)], codeEnum[nameof(Code.Wrssq_m64_r64)] }
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_REX)], codeEnum[nameof(Code.Adcx_r32_rm32)], codeEnum[nameof(Code.Adcx_r64_rm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_REX)], codeEnum[nameof(Code.Adox_r32_rm32)], codeEnum[nameof(Code.Adox_r64_rm64)] },
						invalid,
					},
					invalid,

					// F8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_M_as)], codeEnum[nameof(Code.Movdir64b_r16_m512)], codeEnum[nameof(Code.Movdir64b_r32_m512)], codeEnum[nameof(Code.Movdir64b_r64_m512)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_M_as)], codeEnum[nameof(Code.Enqcmds_r16_m512)], codeEnum[nameof(Code.Enqcmds_r32_m512)], codeEnum[nameof(Code.Enqcmds_r64_m512)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_M_as)], codeEnum[nameof(Code.Enqcmd_r16_m512)], codeEnum[nameof(Code.Enqcmd_r32_m512)], codeEnum[nameof(Code.Enqcmd_r64_m512)] }
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_REX)], codeEnum[nameof(Code.Movdiri_m32_r32)], codeEnum[nameof(Code.Movdiri_m64_r64)] }
						},
						invalid,
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gd_Rd)], codeEnum[nameof(Code.Encodekey128_r32_r32)] },
							invalid,
						},
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gd_Rd)], codeEnum[nameof(Code.Encodekey256_r32_r32)] },
							invalid,
						},
						invalid,
					},
					invalid,
					invalid,
					invalid,
					invalid,
				}),

				("ThreeByteHandlers_0F3AXX",
				new object[0x100] {
					// 00
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 08
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Roundps_xmm_xmmm128_imm8)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Roundpd_xmm_xmmm128_imm8)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Roundss_xmm_xmmm32_imm8)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Roundsd_xmm_xmmm64_imm8)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Blendps_xmm_xmmm128_imm8)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Blendpd_xmm_xmmm128_imm8)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pblendw_xmm_xmmm128_imm8)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q_Ib)], codeEnum[nameof(Code.Palignr_mm_mmm64_imm8)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Palignr_xmm_xmmm128_imm8)] },
						invalid,
						invalid,
					},

					// 10
					invalid,
					invalid,
					invalid,
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.GvM_VX_Ib)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pextrb_r32m8_xmm_imm8)], codeEnum[nameof(Code.Pextrb_r64m8_xmm_imm8)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.GvM_VX_Ib)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pextrw_r32m16_xmm_imm8)], codeEnum[nameof(Code.Pextrw_r64m16_xmm_imm8)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.GvM_VX_Ib)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pextrd_rm32_xmm_imm8)], codeEnum[nameof(Code.Pextrq_rm64_xmm_imm8)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ed_V_Ib)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Extractps_rm32_xmm_imm8)], codeEnum[nameof(Code.Extractps_r64m32_xmm_imm8)] },
						invalid,
						invalid,
					},

					// 18
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 20
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VX_E_Ib)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pinsrb_xmm_r32m8_imm8)], codeEnum[nameof(Code.Pinsrb_xmm_r64m8_imm8)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Insertps_xmm_xmmm32_imm8)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VX_E_Ib)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pinsrd_xmm_rm32_imm8)], codeEnum[nameof(Code.Pinsrq_xmm_rm64_imm8)] },
						invalid,
						invalid,
					},
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 28
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 30
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 38
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 40
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Dpps_xmm_xmmm128_imm8)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Dppd_xmm_xmmm128_imm8)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Mpsadbw_xmm_xmmm128_imm8)] },
						invalid,
						invalid,
					},
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pclmulqdq_xmm_xmmm128_imm8)] },
						invalid,
						invalid,
					},
					invalid,
					invalid,
					invalid,

					// 48
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 50
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 58
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 60
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_3)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pcmpestrm_xmm_xmmm128_imm8)], codeEnum[nameof(Code.Pcmpestrm64_xmm_xmmm128_imm8)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_3)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pcmpestri_xmm_xmmm128_imm8)], codeEnum[nameof(Code.Pcmpestri64_xmm_xmmm128_imm8)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pcmpistrm_xmm_xmmm128_imm8)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pcmpistri_xmm_xmmm128_imm8)] },
						invalid,
						invalid,
					},
					invalid,
					invalid,
					invalid,
					invalid,

					// 68
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 70
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 78
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 80
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 88
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 90
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// 98
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// A0
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// A8
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// B0
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// B8
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// C0
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// C8
					invalid,
					invalid,
					invalid,
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Sha1rnds4_xmm_xmmm128_imm8)] },
						invalid,
						invalid,
						invalid,
					},
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Gf2p8affineqb_xmm_xmmm128_imm8)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Gf2p8affineinvqb_xmm_xmmm128_imm8)] },
						invalid,
						invalid,
					},

					// D0
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// D8
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Aeskeygenassist_xmm_xmmm128_imm8)] },
						invalid,
						invalid,
					},

					// E0
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// E8
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// F0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group8x64)], "handlers_Grp_0F3AF0_lo", "handlers_Grp_0F3AF0_hi" },
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,

					// F8
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
					invalid,
				}),

				("TwoByteHandlers_0FXX",
				new object[0x100] {
					// 00
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_0F00" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group8x64)], "handlers_Grp_0F01_lo", "handlers_Grp_0F01_hi" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev3)], codeEnum[nameof(Code.Lar_r16_rm16)], codeEnum[nameof(Code.Lar_r32_r32m16)], codeEnum[nameof(Code.Lar_r64_r64m16)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev3)], codeEnum[nameof(Code.Lsl_r16_rm16)], codeEnum[nameof(Code.Lsl_r32_r32m16)], codeEnum[nameof(Code.Lsl_r64_r64m16)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options3)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Loadallreset286)] }, decoderOptionsEnum[nameof(DecoderOptions.Loadall286)]
						},
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options3)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Syscall)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Loadall286)] }, decoderOptionsEnum[nameof(DecoderOptions.Loadall286)]
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Syscall)] },
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Clts)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options3)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple2_3b)], codeEnum[nameof(Code.Sysretd)], codeEnum[nameof(Code.Sysretd)], codeEnum[nameof(Code.Sysretq)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Loadall386)] }, decoderOptionsEnum[nameof(DecoderOptions.Loadall386)]
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple2_3b)], codeEnum[nameof(Code.Sysretd)], codeEnum[nameof(Code.Sysretd)], codeEnum[nameof(Code.Sysretq)] },
					},
					// 08
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Invd)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Wbinvd)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options3)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Cl1invmb)] }, decoderOptionsEnum[nameof(DecoderOptions.Cl1invmb)]
						},
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Ud2)] },
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reservednop)],
						"reservedNop_0F0D",
						"grp0F0D"
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Femms)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.D3NOW)] },

					// 10
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options3)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movups_xmm_xmmm128)] },
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movupd_xmm_xmmm128)] },
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movss_xmm_xmmm32)] },
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movsd_xmm_xmmm64)] }
							},
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Gb_1)], codeEnum[nameof(Code.Umov_rm8_r8)] }, decoderOptionsEnum[nameof(DecoderOptions.Umov)]
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movups_xmm_xmmm128)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movupd_xmm_xmmm128)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movss_xmm_xmmm32)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movsd_xmm_xmmm64)] }
						},
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options3)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.WV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movups_xmmm128_xmm)] },
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.WV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movupd_xmmm128_xmm)] },
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.WV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movss_xmmm32_xmm)] },
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.WV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movsd_xmmm64_xmm)] }
							},
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_3b)], codeEnum[nameof(Code.Umov_rm16_r16)], codeEnum[nameof(Code.Umov_rm32_r32)] }, decoderOptionsEnum[nameof(DecoderOptions.Umov)]
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.WV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movups_xmmm128_xmm)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.WV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movupd_xmmm128_xmm)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.WV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movss_xmmm32_xmm)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.WV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movsd_xmmm64_xmm)] }
						},
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options3)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_3)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movhlps_xmm_xmm)], codeEnum[nameof(Code.Movlps_xmm_m64)] },
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.VM)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movlpd_xmm_m64)] },
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movsldup_xmm_xmmm128)] },
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movddup_xmm_xmmm64)] }
							},
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gb_Eb)], codeEnum[nameof(Code.Umov_r8_rm8)] }, decoderOptionsEnum[nameof(DecoderOptions.Umov)]
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_3)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movhlps_xmm_xmm)], codeEnum[nameof(Code.Movlps_xmm_m64)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VM)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movlpd_xmm_m64)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movsldup_xmm_xmmm128)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movddup_xmm_xmmm64)] }
						},
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options3)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.MV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movlps_m64_xmm)] },
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.MV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movlpd_m64_xmm)] },
								invalid,
								invalid,
							},
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3b)], codeEnum[nameof(Code.Umov_r16_rm16)], codeEnum[nameof(Code.Umov_r32_rm32)] }, decoderOptionsEnum[nameof(DecoderOptions.Umov)]
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.MV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movlps_m64_xmm)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.MV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movlpd_m64_xmm)] },
							invalid,
							invalid,
						},
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Unpcklps_xmm_xmmm128)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Unpcklpd_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Unpckhps_xmm_xmmm128)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Unpckhpd_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_3)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movlhps_xmm_xmm)], codeEnum[nameof(Code.Movhps_xmm_m64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VM)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movhpd_xmm_m64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movshdup_xmm_xmmm128)] },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movhps_m64_xmm)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movhpd_m64_xmm)] },
						invalid,
						invalid,
					},

					// 18
					"grp0F18",
					"reservedNop_0F19",
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reservednop)],
						"reservedNop_0F1A",
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options_DontReadModRM)],
							"reservedNop_0F1A",
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
									"reservedNop_0F1A",
									new object[] { legacyEnum[nameof(OpCodeHandlerKind.B_MIB)], codeEnum[nameof(Code.Bndldx_bnd_mib)] }
								},
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.B_BM)], codeEnum[nameof(Code.Bndmov_bnd_bndm64)], codeEnum[nameof(Code.Bndmov_bnd_bndm128)] },
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.B_Ev)], codeEnum[nameof(Code.Bndcl_bnd_rm32)], codeEnum[nameof(Code.Bndcl_bnd_rm64)], true },
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.B_Ev)], codeEnum[nameof(Code.Bndcu_bnd_rm32)], codeEnum[nameof(Code.Bndcu_bnd_rm64)], true }
							}, decoderOptionsEnum[nameof(DecoderOptions.MPX)]
						}
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reservednop)],
						"reservedNop_0F1B",
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options_DontReadModRM)],
							"reservedNop_0F1B",
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
									"reservedNop_0F1B",
									new object[] { legacyEnum[nameof(OpCodeHandlerKind.MIB_B)], codeEnum[nameof(Code.Bndstx_mib_bnd)] }
								},
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.BM_B)], codeEnum[nameof(Code.Bndmov_bndm64_bnd)], codeEnum[nameof(Code.Bndmov_bndm128_bnd)] },
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
									"reservedNop_0F1B",
									new object[] { legacyEnum[nameof(OpCodeHandlerKind.B_Ev)], codeEnum[nameof(Code.Bndmk_bnd_m32)], codeEnum[nameof(Code.Bndmk_bnd_m64)], false }
								},
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.B_Ev)], codeEnum[nameof(Code.Bndcn_bnd_rm32)], codeEnum[nameof(Code.Bndcn_bnd_rm64)], true }
							}, decoderOptionsEnum[nameof(DecoderOptions.MPX)]
						}
					},
					"grp0F1C",
					"reservedNop_0F1D",
					"grp0F1E",
					"grp0F1F",

					// 20
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.R_C_3a)], codeEnum[nameof(Code.Mov_r32_cr)], codeEnum[nameof(Code.Mov_r64_cr)], registerEnum[nameof(Register.CR0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.R_C_3a)], codeEnum[nameof(Code.Mov_r32_dr)], codeEnum[nameof(Code.Mov_r64_dr)], registerEnum[nameof(Register.DR0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.C_R_3a)], codeEnum[nameof(Code.Mov_cr_r32)], codeEnum[nameof(Code.Mov_cr_r64)], registerEnum[nameof(Register.CR0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.C_R_3a)], codeEnum[nameof(Code.Mov_dr_r32)], codeEnum[nameof(Code.Mov_dr_r64)], registerEnum[nameof(Register.DR0)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options3)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.R_C_3b)], codeEnum[nameof(Code.Mov_r32_tr)], registerEnum[nameof(Register.TR0)] }, decoderOptionsEnum[nameof(DecoderOptions.MovTr)]
						},
						invalid,
					},
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options3)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.C_R_3b)], codeEnum[nameof(Code.Mov_tr_r32)], registerEnum[nameof(Register.TR0)] }, decoderOptionsEnum[nameof(DecoderOptions.MovTr)]
						},
						invalid,
					},
					invalid,

					// 28
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movaps_xmm_xmmm128)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movapd_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.WV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movaps_xmmm128_xmm)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.WV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movapd_xmmm128_xmm)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VQ)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cvtpi2ps_xmm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VQ)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cvtpi2pd_xmm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.V_Ev)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cvtsi2ss_xmm_rm32)], codeEnum[nameof(Code.Cvtsi2ss_xmm_rm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.V_Ev)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cvtsi2sd_xmm_rm32)], codeEnum[nameof(Code.Cvtsi2sd_xmm_rm64)] }
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movntps_m128_xmm)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movntpd_m128_xmm)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movntss_m32_xmm)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movntsd_m64_xmm)] }
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_W)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cvttps2pi_mm_xmmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_W)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cvttpd2pi_mm_xmmm128)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_W)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cvttss2si_r32_xmmm32)], codeEnum[nameof(Code.Cvttss2si_r64_xmmm32)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_W)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cvttsd2si_r32_xmmm64)], codeEnum[nameof(Code.Cvttsd2si_r64_xmmm64)] }
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_W)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cvtps2pi_mm_xmmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_W)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cvtpd2pi_mm_xmmm128)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_W)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cvtss2si_r32_xmmm32)], codeEnum[nameof(Code.Cvtss2si_r64_xmmm32)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_W)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cvtsd2si_r32_xmmm64)], codeEnum[nameof(Code.Cvtsd2si_r64_xmmm64)] }
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Ucomiss_xmm_xmmm32)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Ucomisd_xmm_xmmm64)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Comiss_xmm_xmmm32)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Comisd_xmm_xmmm64)] },
						invalid,
						invalid,
					},

					// 30
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Wrmsr)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Rdtsc)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Rdmsr)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Rdpmc)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Sysenter)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple4)], codeEnum[nameof(Code.Sysexitd)], codeEnum[nameof(Code.Sysexitq)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_0F36_Cyrix" }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple4b)], codeEnum[nameof(Code.Getsecd)], codeEnum[nameof(Code.Getsecq)] },
							invalid_NoModRM,
							invalid_NoModRM,
							invalid_NoModRM,
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_0F37_Cyrix" }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},

					// 38
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.AnotherTable)], "ThreeByteHandlers_0F38XX" },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Smint)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Dmint)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix_DMI)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_2)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.AnotherTable)], "ThreeByteHandlers_0F3AXX" },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Rdm)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix_DMI)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Bb0_reset)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)],
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Bb1_reset)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Cpu_write)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Cpu_read)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					invalid,
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Altinst)] }, decoderOptionsEnum[nameof(DecoderOptions.ALTINST)]
					},

					// 40
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Cmovo_r16_rm16)], codeEnum[nameof(Code.Cmovo_r32_rm32)], codeEnum[nameof(Code.Cmovo_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Cmovno_r16_rm16)], codeEnum[nameof(Code.Cmovno_r32_rm32)], codeEnum[nameof(Code.Cmovno_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Cmovb_r16_rm16)], codeEnum[nameof(Code.Cmovb_r32_rm32)], codeEnum[nameof(Code.Cmovb_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Cmovae_r16_rm16)], codeEnum[nameof(Code.Cmovae_r32_rm32)], codeEnum[nameof(Code.Cmovae_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Cmove_r16_rm16)], codeEnum[nameof(Code.Cmove_r32_rm32)], codeEnum[nameof(Code.Cmove_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Cmovne_r16_rm16)], codeEnum[nameof(Code.Cmovne_r32_rm32)], codeEnum[nameof(Code.Cmovne_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Cmovbe_r16_rm16)], codeEnum[nameof(Code.Cmovbe_r32_rm32)], codeEnum[nameof(Code.Cmovbe_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Cmova_r16_rm16)], codeEnum[nameof(Code.Cmova_r32_rm32)], codeEnum[nameof(Code.Cmova_r64_rm64)] },

					// 48
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Cmovs_r16_rm16)], codeEnum[nameof(Code.Cmovs_r32_rm32)], codeEnum[nameof(Code.Cmovs_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Cmovns_r16_rm16)], codeEnum[nameof(Code.Cmovns_r32_rm32)], codeEnum[nameof(Code.Cmovns_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Cmovp_r16_rm16)], codeEnum[nameof(Code.Cmovp_r32_rm32)], codeEnum[nameof(Code.Cmovp_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Cmovnp_r16_rm16)], codeEnum[nameof(Code.Cmovnp_r32_rm32)], codeEnum[nameof(Code.Cmovnp_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Cmovl_r16_rm16)], codeEnum[nameof(Code.Cmovl_r32_rm32)], codeEnum[nameof(Code.Cmovl_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Cmovge_r16_rm16)], codeEnum[nameof(Code.Cmovge_r32_rm32)], codeEnum[nameof(Code.Cmovge_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Cmovle_r16_rm16)], codeEnum[nameof(Code.Cmovle_r32_rm32)], codeEnum[nameof(Code.Cmovle_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Cmovg_r16_rm16)], codeEnum[nameof(Code.Cmovg_r32_rm32)], codeEnum[nameof(Code.Cmovg_r64_rm64)] },

					// 50
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_RX)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movmskps_r32_xmm)], codeEnum[nameof(Code.Movmskps_r64_xmm)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_RX)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movmskpd_r32_xmm)], codeEnum[nameof(Code.Movmskpd_r64_xmm)] },
							invalid,
							invalid,
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Paveb_mm_mmm64)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Sqrtps_xmm_xmmm128)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Sqrtpd_xmm_xmmm128)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Sqrtss_xmm_xmmm32)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Sqrtsd_xmm_xmmm64)] }
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Paddsiw_mm_mmm64)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Rsqrtps_xmm_xmmm128)] },
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Rsqrtss_xmm_xmmm32)] },
							invalid,
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pmagw_mm_mmm64)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Rcpps_xmm_xmmm128)] },
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Rcpss_xmm_xmmm32)] },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Andps_xmm_xmmm128)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Andpd_xmm_xmmm128)] },
							invalid,
							invalid,
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pdistib_mm_m64)] }
						}, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Andnps_xmm_xmmm128)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Andnpd_xmm_xmmm128)] },
							invalid,
							invalid,
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Psubsiw_mm_mmm64)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Orps_xmm_xmmm128)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Orpd_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Xorps_xmm_xmmm128)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Xorpd_xmm_xmmm128)] },
						invalid,
						invalid,
					},

					// 58
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Addps_xmm_xmmm128)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Addpd_xmm_xmmm128)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Addss_xmm_xmmm32)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Addsd_xmm_xmmm64)] }
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pmvzb_mm_m64)] }
						}, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Mulps_xmm_xmmm128)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Mulpd_xmm_xmmm128)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Mulss_xmm_xmmm32)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Mulsd_xmm_xmmm64)] }
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pmulhrw_mm_mmm64)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cvtps2pd_xmm_xmmm64)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cvtpd2ps_xmm_xmmm128)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cvtss2sd_xmm_xmmm32)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cvtsd2ss_xmm_xmmm64)] }
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pmvnzb_mm_m64)] }
						}, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cvtdq2ps_xmm_xmmm128)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cvtps2dq_xmm_xmmm128)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cvttps2dq_xmm_xmmm128)] },
							invalid,
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pmvlzb_mm_m64)] }
						}, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Subps_xmm_xmmm128)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Subpd_xmm_xmmm128)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Subss_xmm_xmmm32)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Subsd_xmm_xmmm64)] }
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pmvgezb_mm_m64)] }
						}, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Minps_xmm_xmmm128)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Minpd_xmm_xmmm128)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Minss_xmm_xmmm32)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Minsd_xmm_xmmm64)] }
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pmulhriw_mm_mmm64)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Divps_xmm_xmmm128)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Divpd_xmm_xmmm128)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Divss_xmm_xmmm32)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Divsd_xmm_xmmm64)] }
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pmachriw_mm_m64)] }
						}, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Maxps_xmm_xmmm128)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Maxpd_xmm_xmmm128)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Maxss_xmm_xmmm32)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Maxsd_xmm_xmmm64)] }
					},

					// 60
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Punpcklbw_mm_mmm32)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Punpcklbw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Punpcklwd_mm_mmm32)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Punpcklwd_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Punpckldq_mm_mmm32)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Punpckldq_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Packsswb_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Packsswb_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pcmpgtb_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pcmpgtb_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pcmpgtw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pcmpgtw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pcmpgtd_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pcmpgtd_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Packuswb_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Packuswb_xmm_xmmm128)] },
						invalid,
						invalid,
					},

					// 68
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Punpckhbw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Punpckhbw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Punpckhwd_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Punpckhwd_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Punpckhdq_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Punpckhdq_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Packssdw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Packssdw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Punpcklqdq_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Punpckhqdq_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Ev)], codeEnum[nameof(Code.Movd_mm_rm32)], codeEnum[nameof(Code.Movq_mm_rm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VX_Ev)], codeEnum[nameof(Code.Movd_xmm_rm32)], codeEnum[nameof(Code.Movq_xmm_rm64)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Movq_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movdqa_xmm_xmmm128)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movdqu_xmm_xmmm128)] },
						invalid,
					},

					// 70
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q_Ib)], codeEnum[nameof(Code.Pshufw_mm_mmm64_imm8)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pshufd_xmm_xmmm128_imm8)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pshufhw_xmm_xmmm128_imm8)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pshuflw_xmm_xmmm128_imm8)] }
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_0F71" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_0F72" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_0F73" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pcmpeqb_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pcmpeqb_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pcmpeqw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pcmpeqw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pcmpeqd_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pcmpeqd_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix_NoModRM)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Emms)] },
						invalid_NoModRM,
						invalid_NoModRM,
						invalid_NoModRM,
					},

					// 78
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_32_64)], codeEnum[nameof(Code.Vmread_rm32_r32)], codeEnum[nameof(Code.Vmread_rm64_r64)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_660F78" },
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VRIbIb)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Insertq_xmm_xmm_imm8_imm8)] }
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.M_Sw)], codeEnum[nameof(Code.Svdc_m80_Sreg)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_32_64)], codeEnum[nameof(Code.Vmwrite_r32_rm32)], codeEnum[nameof(Code.Vmwrite_r64_rm64)], true, true },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_3)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Extrq_xmm_xmm)], codeEnum[nameof(Code.INVALID)] },
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_3)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Insertq_xmm_xmm)], codeEnum[nameof(Code.INVALID)] }
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Sw_M)], codeEnum[nameof(Code.Rsdc_Sreg_m80)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_0F7A_Cyrix" }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_0F7B_Cyrix" }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Haddpd_xmm_xmmm128)] },
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Haddps_xmm_xmmm128)] }
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_0F7C_Cyrix" }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Hsubpd_xmm_xmmm128)] },
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Hsubps_xmm_xmmm128)] }
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_0F7D_Cyrix" }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options1632_1)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_P)], codeEnum[nameof(Code.Movd_rm32_mm)], codeEnum[nameof(Code.Movq_rm64_mm)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_VX)], codeEnum[nameof(Code.Movd_rm32_xmm)], codeEnum[nameof(Code.Movq_rm64_xmm)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movq_xmm_xmmm64)] },
							invalid,
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Smint_0F7E)] }, decoderOptionsEnum[nameof(DecoderOptions.Cyrix_SMINT_0F7E)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Q_P)], codeEnum[nameof(Code.Movq_mmm64_mm)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.WV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movdqa_xmmm128_xmm)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.WV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movdqu_xmmm128_xmm)] },
						invalid,
					},

					// 80
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jz)], codeEnum[nameof(Code.Jo_rel16)], codeEnum[nameof(Code.Jo_rel32_32)], codeEnum[nameof(Code.Jo_rel32_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jz)], codeEnum[nameof(Code.Jno_rel16)], codeEnum[nameof(Code.Jno_rel32_32)], codeEnum[nameof(Code.Jno_rel32_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jz)], codeEnum[nameof(Code.Jb_rel16)], codeEnum[nameof(Code.Jb_rel32_32)], codeEnum[nameof(Code.Jb_rel32_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jz)], codeEnum[nameof(Code.Jae_rel16)], codeEnum[nameof(Code.Jae_rel32_32)], codeEnum[nameof(Code.Jae_rel32_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jz)], codeEnum[nameof(Code.Je_rel16)], codeEnum[nameof(Code.Je_rel32_32)], codeEnum[nameof(Code.Je_rel32_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jz)], codeEnum[nameof(Code.Jne_rel16)], codeEnum[nameof(Code.Jne_rel32_32)], codeEnum[nameof(Code.Jne_rel32_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jz)], codeEnum[nameof(Code.Jbe_rel16)], codeEnum[nameof(Code.Jbe_rel32_32)], codeEnum[nameof(Code.Jbe_rel32_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jz)], codeEnum[nameof(Code.Ja_rel16)], codeEnum[nameof(Code.Ja_rel32_32)], codeEnum[nameof(Code.Ja_rel32_64)] },

					// 88
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jz)], codeEnum[nameof(Code.Js_rel16)], codeEnum[nameof(Code.Js_rel32_32)], codeEnum[nameof(Code.Js_rel32_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jz)], codeEnum[nameof(Code.Jns_rel16)], codeEnum[nameof(Code.Jns_rel32_32)], codeEnum[nameof(Code.Jns_rel32_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jz)], codeEnum[nameof(Code.Jp_rel16)], codeEnum[nameof(Code.Jp_rel32_32)], codeEnum[nameof(Code.Jp_rel32_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jz)], codeEnum[nameof(Code.Jnp_rel16)], codeEnum[nameof(Code.Jnp_rel32_32)], codeEnum[nameof(Code.Jnp_rel32_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jz)], codeEnum[nameof(Code.Jl_rel16)], codeEnum[nameof(Code.Jl_rel32_32)], codeEnum[nameof(Code.Jl_rel32_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jz)], codeEnum[nameof(Code.Jge_rel16)], codeEnum[nameof(Code.Jge_rel32_32)], codeEnum[nameof(Code.Jge_rel32_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jz)], codeEnum[nameof(Code.Jle_rel16)], codeEnum[nameof(Code.Jle_rel32_32)], codeEnum[nameof(Code.Jle_rel32_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jz)], codeEnum[nameof(Code.Jg_rel16)], codeEnum[nameof(Code.Jg_rel32_32)], codeEnum[nameof(Code.Jg_rel32_64)] },

					// 90
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_1)], codeEnum[nameof(Code.Seto_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_1)], codeEnum[nameof(Code.Setno_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_1)], codeEnum[nameof(Code.Setb_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_1)], codeEnum[nameof(Code.Setae_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_1)], codeEnum[nameof(Code.Sete_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_1)], codeEnum[nameof(Code.Setne_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_1)], codeEnum[nameof(Code.Setbe_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_1)], codeEnum[nameof(Code.Seta_rm8)] },

					// 98
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_1)], codeEnum[nameof(Code.Sets_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_1)], codeEnum[nameof(Code.Setns_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_1)], codeEnum[nameof(Code.Setp_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_1)], codeEnum[nameof(Code.Setnp_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_1)], codeEnum[nameof(Code.Setl_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_1)], codeEnum[nameof(Code.Setge_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_1)], codeEnum[nameof(Code.Setle_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_1)], codeEnum[nameof(Code.Setg_rm8)] },

					// A0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushOpSizeReg_4a)], codeEnum[nameof(Code.Pushw_FS)], codeEnum[nameof(Code.Pushd_FS)], codeEnum[nameof(Code.Pushq_FS)], registerEnum[nameof(Register.FS)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushOpSizeReg_4a)], codeEnum[nameof(Code.Popw_FS)], codeEnum[nameof(Code.Popd_FS)], codeEnum[nameof(Code.Popq_FS)], registerEnum[nameof(Register.FS)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Cpuid)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_3a)], codeEnum[nameof(Code.Bt_rm16_r16)], codeEnum[nameof(Code.Bt_rm32_r32)], codeEnum[nameof(Code.Bt_rm64_r64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_Ib)], codeEnum[nameof(Code.Shld_rm16_r16_imm8)], codeEnum[nameof(Code.Shld_rm32_r32_imm8)], codeEnum[nameof(Code.Shld_rm64_r64_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_CL)], codeEnum[nameof(Code.Shld_rm16_r16_CL)], codeEnum[nameof(Code.Shld_rm32_r32_CL)], codeEnum[nameof(Code.Shld_rm64_r64_CL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options5)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group8x64)], "handlers_Grp_0FA6_lo", "handlers_Grp_0FA6_hi" },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3b)], codeEnum[nameof(Code.Xbts_r16_rm16)], codeEnum[nameof(Code.Xbts_r32_rm32)] }, decoderOptionsEnum[nameof(DecoderOptions.Xbts)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Gb_1)], codeEnum[nameof(Code.Cmpxchg486_rm8_r8)] }, decoderOptionsEnum[nameof(DecoderOptions.Cmpxchg486A)]
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group8x64)], "handlers_Grp_0FA6_lo", "handlers_Grp_0FA6_hi" },
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options5)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group8x64)], "handlers_Grp_0FA7_lo", "handlers_Grp_0FA7_hi" },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_3b)], codeEnum[nameof(Code.Ibts_rm16_r16)], codeEnum[nameof(Code.Ibts_rm32_r32)] }, decoderOptionsEnum[nameof(DecoderOptions.Xbts)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_3b)], codeEnum[nameof(Code.Cmpxchg486_rm16_r16)], codeEnum[nameof(Code.Cmpxchg486_rm32_r32)] }, decoderOptionsEnum[nameof(DecoderOptions.Cmpxchg486A)]
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group8x64)], "handlers_Grp_0FA7_lo", "handlers_Grp_0FA7_hi" },
					},

					// A8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushOpSizeReg_4a)], codeEnum[nameof(Code.Pushw_GS)], codeEnum[nameof(Code.Pushd_GS)], codeEnum[nameof(Code.Pushq_GS)], registerEnum[nameof(Register.GS)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushOpSizeReg_4a)], codeEnum[nameof(Code.Popw_GS)], codeEnum[nameof(Code.Popd_GS)], codeEnum[nameof(Code.Popq_GS)], registerEnum[nameof(Register.GS)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Rsm)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_4)], codeEnum[nameof(Code.Bts_rm16_r16)], codeEnum[nameof(Code.Bts_rm32_r32)], codeEnum[nameof(Code.Bts_rm64_r64)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_Ib)], codeEnum[nameof(Code.Shrd_rm16_r16_imm8)], codeEnum[nameof(Code.Shrd_rm32_r32_imm8)], codeEnum[nameof(Code.Shrd_rm64_r64_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_CL)], codeEnum[nameof(Code.Shrd_rm16_r16_CL)], codeEnum[nameof(Code.Shrd_rm32_r32_CL)], codeEnum[nameof(Code.Shrd_rm64_r64_CL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group8x64)], "handlers_Grp_0FAE_lo", "handlers_Grp_0FAE_hi" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Imul_r16_rm16)], codeEnum[nameof(Code.Imul_r32_rm32)], codeEnum[nameof(Code.Imul_r64_rm64)] },

					// B0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Gb_2)], codeEnum[nameof(Code.Cmpxchg_rm8_r8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_4)], codeEnum[nameof(Code.Cmpxchg_rm16_r16)], codeEnum[nameof(Code.Cmpxchg_rm32_r32)], codeEnum[nameof(Code.Cmpxchg_rm64_r64)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Mp_3)], codeEnum[nameof(Code.Lss_r16_m1616)], codeEnum[nameof(Code.Lss_r32_m1632)], codeEnum[nameof(Code.Lss_r64_m1664)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_4)], codeEnum[nameof(Code.Btr_rm16_r16)], codeEnum[nameof(Code.Btr_rm32_r32)], codeEnum[nameof(Code.Btr_rm64_r64)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Mp_3)], codeEnum[nameof(Code.Lfs_r16_m1616)], codeEnum[nameof(Code.Lfs_r32_m1632)], codeEnum[nameof(Code.Lfs_r64_m1664)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Mp_3)], codeEnum[nameof(Code.Lgs_r16_m1616)], codeEnum[nameof(Code.Lgs_r32_m1632)], codeEnum[nameof(Code.Lgs_r64_m1664)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Eb)], codeEnum[nameof(Code.Movzx_r16_rm8)], codeEnum[nameof(Code.Movzx_r32_rm8)], codeEnum[nameof(Code.Movzx_r64_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ew)], codeEnum[nameof(Code.Movzx_r16_rm16)], codeEnum[nameof(Code.Movzx_r32_rm16)], codeEnum[nameof(Code.Movzx_r64_rm16)] },

					// B8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options3)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix4)],
							invalid,
							invalid,
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Popcnt_r16_rm16)], codeEnum[nameof(Code.Popcnt_r32_rm32)], codeEnum[nameof(Code.Popcnt_r64_rm64)] },
							invalid,
							0x1C,
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jdisp)], codeEnum[nameof(Code.Jmpe_disp16)], codeEnum[nameof(Code.Jmpe_disp32)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix4)],
								invalid,
								invalid,
								new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Popcnt_r16_rm16)], codeEnum[nameof(Code.Popcnt_r32_rm32)], codeEnum[nameof(Code.Popcnt_r64_rm64)] },
								invalid,
								0x1C,
							},
						}, decoderOptionsEnum[nameof(DecoderOptions.Jmpe)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Ud1_r16_rm16)], codeEnum[nameof(Code.Ud1_r32_rm32)], codeEnum[nameof(Code.Ud1_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_0FBA" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_4)], codeEnum[nameof(Code.Btc_rm16_r16)], codeEnum[nameof(Code.Btc_rm32_r32)], codeEnum[nameof(Code.Btc_rm64_r64)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options3)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix4)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Bsf_r16_rm16)], codeEnum[nameof(Code.Bsf_r32_rm32)], codeEnum[nameof(Code.Bsf_r64_rm64)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Bsf_r16_rm16)], codeEnum[nameof(Code.Bsf_r32_rm32)], codeEnum[nameof(Code.Bsf_r64_rm64)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Tzcnt_r16_rm16)], codeEnum[nameof(Code.Tzcnt_r32_rm32)], codeEnum[nameof(Code.Tzcnt_r64_rm64)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Bsf_r16_rm16)], codeEnum[nameof(Code.Bsf_r32_rm32)], codeEnum[nameof(Code.Bsf_r64_rm64)] },
							0x14,
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Bsf_r16_rm16)], codeEnum[nameof(Code.Bsf_r32_rm32)], codeEnum[nameof(Code.Bsf_r64_rm64)] }, decoderOptionsEnum[nameof(DecoderOptions.NoMPFX_0FBC)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options3)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix4)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Bsr_r16_rm16)], codeEnum[nameof(Code.Bsr_r32_rm32)], codeEnum[nameof(Code.Bsr_r64_rm64)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Bsr_r16_rm16)], codeEnum[nameof(Code.Bsr_r32_rm32)], codeEnum[nameof(Code.Bsr_r64_rm64)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Lzcnt_r16_rm16)], codeEnum[nameof(Code.Lzcnt_r32_rm32)], codeEnum[nameof(Code.Lzcnt_r64_rm64)] },
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Bsr_r16_rm16)], codeEnum[nameof(Code.Bsr_r32_rm32)], codeEnum[nameof(Code.Bsr_r64_rm64)] },
							0x14,
						},
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Bsr_r16_rm16)], codeEnum[nameof(Code.Bsr_r32_rm32)], codeEnum[nameof(Code.Bsr_r64_rm64)] }, decoderOptionsEnum[nameof(DecoderOptions.NoMPFX_0FBD)]
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Eb)], codeEnum[nameof(Code.Movsx_r16_rm8)], codeEnum[nameof(Code.Movsx_r32_rm8)], codeEnum[nameof(Code.Movsx_r64_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ew)], codeEnum[nameof(Code.Movsx_r16_rm16)], codeEnum[nameof(Code.Movsx_r32_rm16)], codeEnum[nameof(Code.Movsx_r64_rm16)] },

					// C0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Gb_2)], codeEnum[nameof(Code.Xadd_rm8_r8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_4)], codeEnum[nameof(Code.Xadd_rm16_r16)], codeEnum[nameof(Code.Xadd_rm32_r32)], codeEnum[nameof(Code.Xadd_rm64_r64)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cmpps_xmm_xmmm128_imm8)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cmppd_xmm_xmmm128_imm8)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cmpss_xmm_xmmm32_imm8)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cmpsd_xmm_xmmm64_imm8)] }
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Mv_Gv_REXW)], codeEnum[nameof(Code.Movnti_m32_r32)], codeEnum[nameof(Code.Movnti_m64_r64)] },
						invalid,
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Ev_Ib)], codeEnum[nameof(Code.Pinsrw_mm_r32m16_imm8)], codeEnum[nameof(Code.Pinsrw_mm_r64m16_imm8)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VX_E_Ib)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pinsrw_xmm_r32m16_imm8)], codeEnum[nameof(Code.Pinsrw_xmm_r64m16_imm8)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_N_Ib_REX)], codeEnum[nameof(Code.Pextrw_r32_mm_imm8)], codeEnum[nameof(Code.Pextrw_r64_mm_imm8)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RM)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_Ib_REX)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pextrw_r32_xmm_imm8)], codeEnum[nameof(Code.Pextrw_r64_xmm_imm8)] },
							invalid,
						},
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Shufps_xmm_xmmm128_imm8)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VWIb_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Shufpd_xmm_xmmm128_imm8)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_0FC7" },

					// C8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Bswap_r16)], 0 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Bswap_r16)], 1 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Bswap_r16)], 2 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Bswap_r16)], 3 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Bswap_r16)], 4 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Bswap_r16)], 5 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Bswap_r16)], 6 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Bswap_r16)], 7 },

					// D0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Addsubpd_xmm_xmmm128)] },
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Addsubps_xmm_xmmm128)] }
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Psrlw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psrlw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Psrld_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psrld_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Psrlq_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psrlq_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Paddq_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Paddq_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pmullw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmullw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.WV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movq_xmmm64_xmm)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VN)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movq2dq_xmm_mm)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_R)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movdq2q_mm_xmm)] }
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_N)], codeEnum[nameof(Code.Pmovmskb_r32_mm)], codeEnum[nameof(Code.Pmovmskb_r64_mm)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_RX)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmovmskb_r32_xmm)], codeEnum[nameof(Code.Pmovmskb_r64_xmm)] },
						invalid,
						invalid,
					},

					// D8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Psubusb_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psubusb_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Psubusw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psubusw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pminub_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pminub_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pand_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pand_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Paddusb_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Paddusb_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Paddusw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Paddusw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pmaxub_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmaxub_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pandn_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pandn_xmm_xmmm128)] },
						invalid,
						invalid,
					},

					// E0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pavgb_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pavgb_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Psraw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psraw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Psrad_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psrad_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pavgw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pavgw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pmulhuw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmulhuw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pmulhw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmulhw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cvttpd2dq_xmm_xmmm128)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cvtdq2pd_xmm_xmmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Cvtpd2dq_xmm_xmmm128)] }
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MP)], codeEnum[nameof(Code.Movntq_m64_mm)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.MV)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Movntdq_m128_xmm)] },
						invalid,
						invalid,
					},

					// E8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Psubsb_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psubsb_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Psubsw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psubsw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pminsw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pminsw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Por_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Por_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Paddsb_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Paddsb_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Paddsw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Paddsw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pmaxsw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmaxsw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pxor_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pxor_xmm_xmmm128)] },
						invalid,
						invalid,
					},

					// F0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						invalid,
						invalid,
						invalid,
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VM)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Lddqu_xmm_m128)] }
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Psllw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psllw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pslld_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pslld_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Psllq_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psllq_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pmuludq_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmuludq_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Pmaddwd_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Pmaddwd_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Psadbw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psadbw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.rDI_P_N)], codeEnum[nameof(Code.Maskmovq_rDI_mm_mm)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.rDI_VX_RX)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Maskmovdqu_rDI_xmm_xmm)] },
						invalid,
						invalid,
					},

					// F8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Psubb_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psubb_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Psubw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psubw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Psubd_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psubd_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Psubq_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Psubq_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Paddb_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Paddb_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Paddw_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Paddw_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MandatoryPrefix)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.P_Q)], codeEnum[nameof(Code.Paddd_mm_mmm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.VW_2)], registerEnum[nameof(Register.XMM0)], codeEnum[nameof(Code.Paddd_xmm_xmmm128)] },
						invalid,
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options3)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Ud0_r16_rm16)], codeEnum[nameof(Code.Ud0_r32_rm32)], codeEnum[nameof(Code.Ud0_r64_rm64)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Ud0)] }, decoderOptionsEnum[nameof(DecoderOptions.AMD)]
					},
				}),

				(OneByteHandlers,
				new object[0x100] {
					// 00
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Gb_2)], codeEnum[nameof(Code.Add_rm8_r8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_4)], codeEnum[nameof(Code.Add_rm16_r16)], codeEnum[nameof(Code.Add_rm32_r32)], codeEnum[nameof(Code.Add_rm64_r64)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gb_Eb)], codeEnum[nameof(Code.Add_r8_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Add_r16_rm16)], codeEnum[nameof(Code.Add_r32_rm32)], codeEnum[nameof(Code.Add_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIb)], codeEnum[nameof(Code.Add_AL_imm8)], registerEnum[nameof(Register.AL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reg_Iz)], codeEnum[nameof(Code.Add_AX_imm16)], codeEnum[nameof(Code.Add_EAX_imm32)], codeEnum[nameof(Code.Add_RAX_imm32)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushOpSizeReg_4b)], codeEnum[nameof(Code.Pushw_ES)], codeEnum[nameof(Code.Pushd_ES)], registerEnum[nameof(Register.ES)] },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushOpSizeReg_4b)], codeEnum[nameof(Code.Popw_ES)], codeEnum[nameof(Code.Popd_ES)], registerEnum[nameof(Register.ES)] },
						invalid,
					},

					// 08
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Gb_2)], codeEnum[nameof(Code.Or_rm8_r8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_4)], codeEnum[nameof(Code.Or_rm16_r16)], codeEnum[nameof(Code.Or_rm32_r32)], codeEnum[nameof(Code.Or_rm64_r64)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gb_Eb)], codeEnum[nameof(Code.Or_r8_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Or_r16_rm16)], codeEnum[nameof(Code.Or_r32_rm32)], codeEnum[nameof(Code.Or_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIb)], codeEnum[nameof(Code.Or_AL_imm8)], registerEnum[nameof(Register.AL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reg_Iz)], codeEnum[nameof(Code.Or_AX_imm16)], codeEnum[nameof(Code.Or_EAX_imm32)], codeEnum[nameof(Code.Or_RAX_imm32)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushOpSizeReg_4b)], codeEnum[nameof(Code.Pushw_CS)], codeEnum[nameof(Code.Pushd_CS)], registerEnum[nameof(Register.CS)] },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.AnotherTable)], "TwoByteHandlers_0FXX" },

					// 10
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Gb_2)], codeEnum[nameof(Code.Adc_rm8_r8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_4)], codeEnum[nameof(Code.Adc_rm16_r16)], codeEnum[nameof(Code.Adc_rm32_r32)], codeEnum[nameof(Code.Adc_rm64_r64)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gb_Eb)], codeEnum[nameof(Code.Adc_r8_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Adc_r16_rm16)], codeEnum[nameof(Code.Adc_r32_rm32)], codeEnum[nameof(Code.Adc_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIb)], codeEnum[nameof(Code.Adc_AL_imm8)], registerEnum[nameof(Register.AL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reg_Iz)], codeEnum[nameof(Code.Adc_AX_imm16)], codeEnum[nameof(Code.Adc_EAX_imm32)], codeEnum[nameof(Code.Adc_RAX_imm32)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushOpSizeReg_4b)], codeEnum[nameof(Code.Pushw_SS)], codeEnum[nameof(Code.Pushd_SS)], registerEnum[nameof(Register.SS)] },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushOpSizeReg_4b)], codeEnum[nameof(Code.Popw_SS)], codeEnum[nameof(Code.Popd_SS)], registerEnum[nameof(Register.SS)] },
						invalid,
					},

					// 18
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Gb_2)], codeEnum[nameof(Code.Sbb_rm8_r8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_4)], codeEnum[nameof(Code.Sbb_rm16_r16)], codeEnum[nameof(Code.Sbb_rm32_r32)], codeEnum[nameof(Code.Sbb_rm64_r64)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gb_Eb)], codeEnum[nameof(Code.Sbb_r8_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Sbb_r16_rm16)], codeEnum[nameof(Code.Sbb_r32_rm32)], codeEnum[nameof(Code.Sbb_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIb)], codeEnum[nameof(Code.Sbb_AL_imm8)], registerEnum[nameof(Register.AL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reg_Iz)], codeEnum[nameof(Code.Sbb_AX_imm16)], codeEnum[nameof(Code.Sbb_EAX_imm32)], codeEnum[nameof(Code.Sbb_RAX_imm32)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushOpSizeReg_4b)], codeEnum[nameof(Code.Pushw_DS)], codeEnum[nameof(Code.Pushd_DS)], registerEnum[nameof(Register.DS)] },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushOpSizeReg_4b)], codeEnum[nameof(Code.Popw_DS)], codeEnum[nameof(Code.Popd_DS)], registerEnum[nameof(Register.DS)] },
						invalid,
					},

					// 20
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Gb_2)], codeEnum[nameof(Code.And_rm8_r8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_4)], codeEnum[nameof(Code.And_rm16_r16)], codeEnum[nameof(Code.And_rm32_r32)], codeEnum[nameof(Code.And_rm64_r64)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gb_Eb)], codeEnum[nameof(Code.And_r8_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.And_r16_rm16)], codeEnum[nameof(Code.And_r32_rm32)], codeEnum[nameof(Code.And_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIb)], codeEnum[nameof(Code.And_AL_imm8)], registerEnum[nameof(Register.AL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reg_Iz)], codeEnum[nameof(Code.And_AX_imm16)], codeEnum[nameof(Code.And_EAX_imm32)], codeEnum[nameof(Code.And_RAX_imm32)] },
					invalid,// ES:
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Daa)] },
						invalid,
					},

					// 28
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Gb_2)], codeEnum[nameof(Code.Sub_rm8_r8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_4)], codeEnum[nameof(Code.Sub_rm16_r16)], codeEnum[nameof(Code.Sub_rm32_r32)], codeEnum[nameof(Code.Sub_rm64_r64)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gb_Eb)], codeEnum[nameof(Code.Sub_r8_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Sub_r16_rm16)], codeEnum[nameof(Code.Sub_r32_rm32)], codeEnum[nameof(Code.Sub_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIb)], codeEnum[nameof(Code.Sub_AL_imm8)], registerEnum[nameof(Register.AL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reg_Iz)], codeEnum[nameof(Code.Sub_AX_imm16)], codeEnum[nameof(Code.Sub_EAX_imm32)], codeEnum[nameof(Code.Sub_RAX_imm32)] },
					invalid,// CS:
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Das)] },
						invalid,
					},

					// 30
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Gb_2)], codeEnum[nameof(Code.Xor_rm8_r8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_4)], codeEnum[nameof(Code.Xor_rm16_r16)], codeEnum[nameof(Code.Xor_rm32_r32)], codeEnum[nameof(Code.Xor_rm64_r64)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gb_Eb)], codeEnum[nameof(Code.Xor_r8_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Xor_r16_rm16)], codeEnum[nameof(Code.Xor_r32_rm32)], codeEnum[nameof(Code.Xor_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIb)], codeEnum[nameof(Code.Xor_AL_imm8)], registerEnum[nameof(Register.AL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reg_Iz)], codeEnum[nameof(Code.Xor_AX_imm16)], codeEnum[nameof(Code.Xor_EAX_imm32)], codeEnum[nameof(Code.Xor_RAX_imm32)] },
					invalid,// SS:
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Aaa)] },
						invalid,
					},

					// 38
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Gb_1)], codeEnum[nameof(Code.Cmp_rm8_r8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_3a)], codeEnum[nameof(Code.Cmp_rm16_r16)], codeEnum[nameof(Code.Cmp_rm32_r32)], codeEnum[nameof(Code.Cmp_rm64_r64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gb_Eb)], codeEnum[nameof(Code.Cmp_r8_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Cmp_r16_rm16)], codeEnum[nameof(Code.Cmp_r32_rm32)], codeEnum[nameof(Code.Cmp_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIb)], codeEnum[nameof(Code.Cmp_AL_imm8)], registerEnum[nameof(Register.AL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reg_Iz)], codeEnum[nameof(Code.Cmp_AX_imm16)], codeEnum[nameof(Code.Cmp_EAX_imm32)], codeEnum[nameof(Code.Cmp_RAX_imm32)] },
					invalid,// DS:
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Aas)] },
						invalid,
					},

					// 40
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Inc_r16)], 0 },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Inc_r16)], 1 },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Inc_r16)], 2 },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Inc_r16)], 3 },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Inc_r16)], 4 },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Inc_r16)], 5 },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Inc_r16)], 6 },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Inc_r16)], 7 },
						invalid,
					},

					// 48
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Dec_r16)], 0 },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Dec_r16)], 1 },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Dec_r16)], 2 },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Dec_r16)], 3 },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Dec_r16)], 4 },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Dec_r16)], 5 },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Dec_r16)], 6 },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.SimpleReg)], codeEnum[nameof(Code.Dec_r16)], 7 },
						invalid,
					},

					// 50
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushSimpleReg)], 0, codeEnum[nameof(Code.Push_r16)], codeEnum[nameof(Code.Push_r32)], codeEnum[nameof(Code.Push_r64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushSimpleReg)], 1, codeEnum[nameof(Code.Push_r16)], codeEnum[nameof(Code.Push_r32)], codeEnum[nameof(Code.Push_r64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushSimpleReg)], 2, codeEnum[nameof(Code.Push_r16)], codeEnum[nameof(Code.Push_r32)], codeEnum[nameof(Code.Push_r64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushSimpleReg)], 3, codeEnum[nameof(Code.Push_r16)], codeEnum[nameof(Code.Push_r32)], codeEnum[nameof(Code.Push_r64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushSimpleReg)], 4, codeEnum[nameof(Code.Push_r16)], codeEnum[nameof(Code.Push_r32)], codeEnum[nameof(Code.Push_r64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushSimpleReg)], 5, codeEnum[nameof(Code.Push_r16)], codeEnum[nameof(Code.Push_r32)], codeEnum[nameof(Code.Push_r64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushSimpleReg)], 6, codeEnum[nameof(Code.Push_r16)], codeEnum[nameof(Code.Push_r32)], codeEnum[nameof(Code.Push_r64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushSimpleReg)], 7, codeEnum[nameof(Code.Push_r16)], codeEnum[nameof(Code.Push_r32)], codeEnum[nameof(Code.Push_r64)] },

					// 58
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushSimpleReg)], 0, codeEnum[nameof(Code.Pop_r16)], codeEnum[nameof(Code.Pop_r32)], codeEnum[nameof(Code.Pop_r64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushSimpleReg)], 1, codeEnum[nameof(Code.Pop_r16)], codeEnum[nameof(Code.Pop_r32)], codeEnum[nameof(Code.Pop_r64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushSimpleReg)], 2, codeEnum[nameof(Code.Pop_r16)], codeEnum[nameof(Code.Pop_r32)], codeEnum[nameof(Code.Pop_r64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushSimpleReg)], 3, codeEnum[nameof(Code.Pop_r16)], codeEnum[nameof(Code.Pop_r32)], codeEnum[nameof(Code.Pop_r64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushSimpleReg)], 4, codeEnum[nameof(Code.Pop_r16)], codeEnum[nameof(Code.Pop_r32)], codeEnum[nameof(Code.Pop_r64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushSimpleReg)], 5, codeEnum[nameof(Code.Pop_r16)], codeEnum[nameof(Code.Pop_r32)], codeEnum[nameof(Code.Pop_r64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushSimpleReg)], 6, codeEnum[nameof(Code.Pop_r16)], codeEnum[nameof(Code.Pop_r32)], codeEnum[nameof(Code.Pop_r64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushSimpleReg)], 7, codeEnum[nameof(Code.Pop_r16)], codeEnum[nameof(Code.Pop_r32)], codeEnum[nameof(Code.Pop_r64)] },

					// 60
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple2_3b)], codeEnum[nameof(Code.Pushaw)], codeEnum[nameof(Code.Pushad)], codeEnum[nameof(Code.Pushad)] },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple2_3b)], codeEnum[nameof(Code.Popaw)], codeEnum[nameof(Code.Popad)], codeEnum[nameof(Code.Popad)] },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.EVEX)], new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ma)], codeEnum[nameof(Code.Bound_r16_m1616)], codeEnum[nameof(Code.Bound_r32_m3232)] } },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.RvMw_Gw)], codeEnum[nameof(Code.Arpl_rm16_r16)], codeEnum[nameof(Code.Arpl_r32m16_r32)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev2)], codeEnum[nameof(Code.Movsxd_r16_rm16)], codeEnum[nameof(Code.Movsxd_r32_rm32)], codeEnum[nameof(Code.Movsxd_r64_rm32)] },
					},
					invalid,// FS:
					invalid,// GS:
					invalid,// os
					invalid,// as

					// 68
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushIz)], codeEnum[nameof(Code.Push_imm16)], codeEnum[nameof(Code.Pushd_imm32)], codeEnum[nameof(Code.Pushq_imm32)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_Iz)], codeEnum[nameof(Code.Imul_r16_rm16_imm16)], codeEnum[nameof(Code.Imul_r32_rm32_imm32)], codeEnum[nameof(Code.Imul_r64_rm64_imm32)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushIb2)], codeEnum[nameof(Code.Pushw_imm8)], codeEnum[nameof(Code.Pushd_imm8)], codeEnum[nameof(Code.Pushq_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_Ib)], codeEnum[nameof(Code.Imul_r16_rm16_imm8)], codeEnum[nameof(Code.Imul_r32_rm32_imm8)], codeEnum[nameof(Code.Imul_r64_rm64_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Yb_Reg)], codeEnum[nameof(Code.Insb_m8_DX)], registerEnum[nameof(Register.DX)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Yv_Reg2)], codeEnum[nameof(Code.Insw_m16_DX)], codeEnum[nameof(Code.Insd_m32_DX)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reg_Xb)], codeEnum[nameof(Code.Outsb_DX_m8)], registerEnum[nameof(Register.DX)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reg_Xv2)], codeEnum[nameof(Code.Outsw_DX_m16)], codeEnum[nameof(Code.Outsd_DX_m32)] },

					// 70
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jb)], codeEnum[nameof(Code.Jo_rel8_16)], codeEnum[nameof(Code.Jo_rel8_32)], codeEnum[nameof(Code.Jo_rel8_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jb)], codeEnum[nameof(Code.Jno_rel8_16)], codeEnum[nameof(Code.Jno_rel8_32)], codeEnum[nameof(Code.Jno_rel8_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jb)], codeEnum[nameof(Code.Jb_rel8_16)], codeEnum[nameof(Code.Jb_rel8_32)], codeEnum[nameof(Code.Jb_rel8_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jb)], codeEnum[nameof(Code.Jae_rel8_16)], codeEnum[nameof(Code.Jae_rel8_32)], codeEnum[nameof(Code.Jae_rel8_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jb)], codeEnum[nameof(Code.Je_rel8_16)], codeEnum[nameof(Code.Je_rel8_32)], codeEnum[nameof(Code.Je_rel8_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jb)], codeEnum[nameof(Code.Jne_rel8_16)], codeEnum[nameof(Code.Jne_rel8_32)], codeEnum[nameof(Code.Jne_rel8_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jb)], codeEnum[nameof(Code.Jbe_rel8_16)], codeEnum[nameof(Code.Jbe_rel8_32)], codeEnum[nameof(Code.Jbe_rel8_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jb)], codeEnum[nameof(Code.Ja_rel8_16)], codeEnum[nameof(Code.Ja_rel8_32)], codeEnum[nameof(Code.Ja_rel8_64)] },

					// 78
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jb)], codeEnum[nameof(Code.Js_rel8_16)], codeEnum[nameof(Code.Js_rel8_32)], codeEnum[nameof(Code.Js_rel8_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jb)], codeEnum[nameof(Code.Jns_rel8_16)], codeEnum[nameof(Code.Jns_rel8_32)], codeEnum[nameof(Code.Jns_rel8_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jb)], codeEnum[nameof(Code.Jp_rel8_16)], codeEnum[nameof(Code.Jp_rel8_32)], codeEnum[nameof(Code.Jp_rel8_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jb)], codeEnum[nameof(Code.Jnp_rel8_16)], codeEnum[nameof(Code.Jnp_rel8_32)], codeEnum[nameof(Code.Jnp_rel8_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jb)], codeEnum[nameof(Code.Jl_rel8_16)], codeEnum[nameof(Code.Jl_rel8_32)], codeEnum[nameof(Code.Jl_rel8_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jb)], codeEnum[nameof(Code.Jge_rel8_16)], codeEnum[nameof(Code.Jge_rel8_32)], codeEnum[nameof(Code.Jge_rel8_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jb)], codeEnum[nameof(Code.Jle_rel8_16)], codeEnum[nameof(Code.Jle_rel8_32)], codeEnum[nameof(Code.Jle_rel8_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jb)], codeEnum[nameof(Code.Jg_rel8_16)], codeEnum[nameof(Code.Jg_rel8_32)], codeEnum[nameof(Code.Jg_rel8_64)] },

					// 80
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_80" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_81" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_82" },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_83" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Gb_1)], codeEnum[nameof(Code.Test_rm8_r8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_3a)], codeEnum[nameof(Code.Test_rm16_r16)], codeEnum[nameof(Code.Test_rm32_r32)], codeEnum[nameof(Code.Test_rm64_r64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Gb_2)], codeEnum[nameof(Code.Xchg_rm8_r8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.XacquireXreleaseNoLock), nameof(HandlerFlags.Lock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_4)], codeEnum[nameof(Code.Xchg_rm16_r16)], codeEnum[nameof(Code.Xchg_rm32_r32)], codeEnum[nameof(Code.Xchg_rm64_r64)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xacquire), nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.XacquireXreleaseNoLock), nameof(HandlerFlags.Lock)) },

					// 88
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Eb_Gb_2)], codeEnum[nameof(Code.Mov_rm8_r8)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.XacquireXreleaseNoLock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Gv_4)], codeEnum[nameof(Code.Mov_rm16_r16)], codeEnum[nameof(Code.Mov_rm32_r32)], codeEnum[nameof(Code.Mov_rm64_r64)], new OrEnumValue(handlerFlagsEnum, nameof(HandlerFlags.Xrelease), nameof(HandlerFlags.XacquireXreleaseNoLock)) },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gb_Eb)], codeEnum[nameof(Code.Mov_r8_rm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Ev_3a)], codeEnum[nameof(Code.Mov_r16_rm16)], codeEnum[nameof(Code.Mov_r32_rm32)], codeEnum[nameof(Code.Mov_r64_rm64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ev_Sw)], codeEnum[nameof(Code.Mov_rm16_Sreg)], codeEnum[nameof(Code.Mov_r32m16_Sreg)], codeEnum[nameof(Code.Mov_r64m16_Sreg)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_M)], codeEnum[nameof(Code.Lea_r16_m)], codeEnum[nameof(Code.Lea_r32_m)], codeEnum[nameof(Code.Lea_r64_m)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Sw_Ev)], codeEnum[nameof(Code.Mov_Sreg_rm16)], codeEnum[nameof(Code.Mov_Sreg_r32m16)], codeEnum[nameof(Code.Mov_Sreg_r64m16)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.XOP)], new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_8F" } },

					// 90
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Xchg_Reg_rAX)], 0 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Xchg_Reg_rAX)], 1 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Xchg_Reg_rAX)], 2 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Xchg_Reg_rAX)], 3 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Xchg_Reg_rAX)], 4 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Xchg_Reg_rAX)], 5 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Xchg_Reg_rAX)], 6 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Xchg_Reg_rAX)], 7 },

					// 98
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple2_3a)], codeEnum[nameof(Code.Cbw)], codeEnum[nameof(Code.Cwde)], codeEnum[nameof(Code.Cdqe)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple2_3a)], codeEnum[nameof(Code.Cwd)], codeEnum[nameof(Code.Cdq)], codeEnum[nameof(Code.Cqo)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ap)], codeEnum[nameof(Code.Call_ptr1616)], codeEnum[nameof(Code.Call_ptr1632)] },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Wait)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushSimple2)], codeEnum[nameof(Code.Pushfw)], codeEnum[nameof(Code.Pushfd)], codeEnum[nameof(Code.Pushfq)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.PushSimple2)], codeEnum[nameof(Code.Popfw)], codeEnum[nameof(Code.Popfd)], codeEnum[nameof(Code.Popfq)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Sahf)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options3)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Sahf)] },
							invalid, decoderOptionsEnum[nameof(DecoderOptions.NoLahfSahf64)]
						},
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Lahf)] },
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Options3)],
							new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Lahf)] },
							invalid, decoderOptionsEnum[nameof(DecoderOptions.NoLahfSahf64)]
						},
					},

					// A0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reg_Ob)], codeEnum[nameof(Code.Mov_AL_moffs8)], registerEnum[nameof(Register.AL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reg_Ov)], codeEnum[nameof(Code.Mov_AX_moffs16)], codeEnum[nameof(Code.Mov_EAX_moffs32)], codeEnum[nameof(Code.Mov_RAX_moffs64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ob_Reg)], codeEnum[nameof(Code.Mov_moffs8_AL)], registerEnum[nameof(Register.AL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ov_Reg)], codeEnum[nameof(Code.Mov_moffs16_AX)], codeEnum[nameof(Code.Mov_moffs32_EAX)], codeEnum[nameof(Code.Mov_moffs64_RAX)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Yb_Xb)], codeEnum[nameof(Code.Movsb_m8_m8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Yv_Xv)], codeEnum[nameof(Code.Movsw_m16_m16)], codeEnum[nameof(Code.Movsd_m32_m32)], codeEnum[nameof(Code.Movsq_m64_m64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Xb_Yb)], codeEnum[nameof(Code.Cmpsb_m8_m8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Xv_Yv)], codeEnum[nameof(Code.Cmpsw_m16_m16)], codeEnum[nameof(Code.Cmpsd_m32_m32)], codeEnum[nameof(Code.Cmpsq_m64_m64)] },

					// A8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIb)], codeEnum[nameof(Code.Test_AL_imm8)], registerEnum[nameof(Register.AL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reg_Iz)], codeEnum[nameof(Code.Test_AX_imm16)], codeEnum[nameof(Code.Test_EAX_imm32)], codeEnum[nameof(Code.Test_RAX_imm32)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Yb_Reg)], codeEnum[nameof(Code.Stosb_m8_AL)], registerEnum[nameof(Register.AL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Yv_Reg)], codeEnum[nameof(Code.Stosw_m16_AX)], codeEnum[nameof(Code.Stosd_m32_EAX)], codeEnum[nameof(Code.Stosq_m64_RAX)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reg_Xb)], codeEnum[nameof(Code.Lodsb_AL_m8)], registerEnum[nameof(Register.AL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reg_Xv)], codeEnum[nameof(Code.Lodsw_AX_m16)], codeEnum[nameof(Code.Lodsd_EAX_m32)], codeEnum[nameof(Code.Lodsq_RAX_m64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reg_Yb)], codeEnum[nameof(Code.Scasb_AL_m8)], registerEnum[nameof(Register.AL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reg_Yv)], codeEnum[nameof(Code.Scasw_AX_m16)], codeEnum[nameof(Code.Scasd_EAX_m32)], codeEnum[nameof(Code.Scasq_RAX_m64)] },

					// B0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIb3)], 0 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIb3)], 1 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIb3)], 2 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIb3)], 3 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIb3)], 4 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIb3)], 5 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIb3)], 6 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIb3)], 7 },

					// B8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIz2)], 0 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIz2)], 1 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIz2)], 2 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIz2)], 3 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIz2)], 4 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIz2)], 5 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIz2)], 6 },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIz2)], 7 },

					// C0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_C0" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_C1" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.BranchIw)], codeEnum[nameof(Code.Retnw_imm16)], codeEnum[nameof(Code.Retnd_imm16)], codeEnum[nameof(Code.Retnq_imm16)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.BranchSimple)], codeEnum[nameof(Code.Retnw)], codeEnum[nameof(Code.Retnd)], codeEnum[nameof(Code.Retnq)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.VEX3)], new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Mp_2)], codeEnum[nameof(Code.Les_r16_m1616)], codeEnum[nameof(Code.Les_r32_m1632)] } },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.VEX2)], new object[] { legacyEnum[nameof(OpCodeHandlerKind.Gv_Mp_2)], codeEnum[nameof(Code.Lds_r16_m1616)], codeEnum[nameof(Code.Lds_r32_m1632)] } },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group8x64)], "handlers_Grp_C6_lo", "handlers_Grp_C6_hi" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group8x64)], "handlers_Grp_C7_lo", "handlers_Grp_C7_hi" },

					// C8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Iw_Ib)], codeEnum[nameof(Code.Enterw_imm16_imm8)], codeEnum[nameof(Code.Enterd_imm16_imm8)], codeEnum[nameof(Code.Enterq_imm16_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple3)], codeEnum[nameof(Code.Leavew)], codeEnum[nameof(Code.Leaved)], codeEnum[nameof(Code.Leaveq)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple2Iw)], codeEnum[nameof(Code.Retfw_imm16)], codeEnum[nameof(Code.Retfd_imm16)], codeEnum[nameof(Code.Retfq_imm16)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple2_3a)], codeEnum[nameof(Code.Retfw)], codeEnum[nameof(Code.Retfd)], codeEnum[nameof(Code.Retfq)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Int3)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ib)], codeEnum[nameof(Code.Int_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Into)] },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple2_3a)], codeEnum[nameof(Code.Iretw)], codeEnum[nameof(Code.Iretd)], codeEnum[nameof(Code.Iretq)] },

					// D0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_D0" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_D1" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_D2" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_D3" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ib)], codeEnum[nameof(Code.Aam_imm8)] },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ib)], codeEnum[nameof(Code.Aad_imm8)] },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Salc)] },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.MemBx)], codeEnum[nameof(Code.Xlat_m8)] },

					// D8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group8x8)], "handlers_FPU_D8_low", "handlers_FPU_D8_high" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group8x64)], "handlers_FPU_D9_low", "handlers_FPU_D9_high" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group8x64)], "handlers_FPU_DA_low", "handlers_FPU_DA_high" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group8x64)], "handlers_FPU_DB_low", "handlers_FPU_DB_high" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group8x8)], "handlers_FPU_DC_low", "handlers_FPU_DC_high" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group8x64)], "handlers_FPU_DD_low", "handlers_FPU_DD_high" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group8x64)], "handlers_FPU_DE_low", "handlers_FPU_DE_high" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group8x64)], "handlers_FPU_DF_low", "handlers_FPU_DF_high" },

					// E0
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jb2)], codeEnum[nameof(Code.Loopne_rel8_16_CX)], codeEnum[nameof(Code.Loopne_rel8_16_ECX)], codeEnum[nameof(Code.Loopne_rel8_16_RCX)], codeEnum[nameof(Code.Loopne_rel8_32_CX)], codeEnum[nameof(Code.Loopne_rel8_32_ECX)], codeEnum[nameof(Code.Loopne_rel8_64_ECX)], codeEnum[nameof(Code.Loopne_rel8_64_RCX)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jb2)], codeEnum[nameof(Code.Loope_rel8_16_CX)], codeEnum[nameof(Code.Loope_rel8_16_ECX)], codeEnum[nameof(Code.Loope_rel8_16_RCX)], codeEnum[nameof(Code.Loope_rel8_32_CX)], codeEnum[nameof(Code.Loope_rel8_32_ECX)], codeEnum[nameof(Code.Loope_rel8_64_ECX)], codeEnum[nameof(Code.Loope_rel8_64_RCX)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jb2)], codeEnum[nameof(Code.Loop_rel8_16_CX)], codeEnum[nameof(Code.Loop_rel8_16_ECX)], codeEnum[nameof(Code.Loop_rel8_16_RCX)], codeEnum[nameof(Code.Loop_rel8_32_CX)], codeEnum[nameof(Code.Loop_rel8_32_ECX)], codeEnum[nameof(Code.Loop_rel8_64_ECX)], codeEnum[nameof(Code.Loop_rel8_64_RCX)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jb2)], codeEnum[nameof(Code.Jcxz_rel8_16)], codeEnum[nameof(Code.Jecxz_rel8_16)], codeEnum[nameof(Code.Jrcxz_rel8_16)], codeEnum[nameof(Code.Jcxz_rel8_32)], codeEnum[nameof(Code.Jecxz_rel8_32)], codeEnum[nameof(Code.Jecxz_rel8_64)], codeEnum[nameof(Code.Jrcxz_rel8_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.RegIb)], codeEnum[nameof(Code.In_AL_imm8)], registerEnum[nameof(Register.AL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Reg_Ib2)], codeEnum[nameof(Code.In_AX_imm8)], codeEnum[nameof(Code.In_EAX_imm8)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.IbReg)], codeEnum[nameof(Code.Out_imm8_AL)], registerEnum[nameof(Register.AL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.IbReg2)], codeEnum[nameof(Code.Out_imm8_AX)], codeEnum[nameof(Code.Out_imm8_EAX)] },

					// E8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jz)], codeEnum[nameof(Code.Call_rel16)], codeEnum[nameof(Code.Call_rel32_32)], codeEnum[nameof(Code.Call_rel32_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jz)], codeEnum[nameof(Code.Jmp_rel16)], codeEnum[nameof(Code.Jmp_rel32_32)], codeEnum[nameof(Code.Jmp_rel32_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Bitness)],
						new object[] { legacyEnum[nameof(OpCodeHandlerKind.Ap)], codeEnum[nameof(Code.Jmp_ptr1616)], codeEnum[nameof(Code.Jmp_ptr1632)] },
						invalid,
					},
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Jb)], codeEnum[nameof(Code.Jmp_rel8_16)], codeEnum[nameof(Code.Jmp_rel8_32)], codeEnum[nameof(Code.Jmp_rel8_64)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.AL_DX)], codeEnum[nameof(Code.In_AL_DX)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.eAX_DX)], codeEnum[nameof(Code.In_AX_DX)], codeEnum[nameof(Code.In_EAX_DX)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.DX_AL)], codeEnum[nameof(Code.Out_DX_AL)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.DX_eAX)], codeEnum[nameof(Code.Out_DX_AX)], codeEnum[nameof(Code.Out_DX_EAX)] },

					// F0
					invalid,// lock
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Int1)] },
					invalid,// repne
					invalid,// rep
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Hlt)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Cmc)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_F6" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_F7" },

					// F8
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Clc)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Stc)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Cli)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Sti)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Cld)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Simple)], codeEnum[nameof(Code.Std)] },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_FE" },
					new object[] { legacyEnum[nameof(OpCodeHandlerKind.Group)], "handlers_Grp_FF" },
				}),
			};
			return handlers;
		}
	}
}
