/*
Copyright (C) 2018-2019 de4dot@gmail.com

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using Iced.Intel;
using System.Linq;

namespace Iced.UnitTests.Intel {
	static partial class ToEnumConverter {
		public static bool TryCode(string value, out Code code) => codeDict.TryGetValue(value, out code);
		public static Code GetCode(string value) => TryCode(value, out var code) ? code : throw new InvalidOperationException($"Invalid Code value: {value}");
		public static string[] GetCodeNames() => codeDict.OrderBy(a => a.Value).Select(a => a.Key).ToArray();

		static readonly Dictionary<string, Code> codeDict =
			// GENERATOR-BEGIN: CodeHash
			// ⚠️This was generated by GENERATOR!🦹‍♂️
			new Dictionary<string, Code>(4312, StringComparer.Ordinal) {
				{ "INVALID", Code.INVALID },
				{ "DeclareByte", Code.DeclareByte },
				{ "DeclareWord", Code.DeclareWord },
				{ "DeclareDword", Code.DeclareDword },
				{ "DeclareQword", Code.DeclareQword },
				{ "Add_rm8_r8", Code.Add_rm8_r8 },
				{ "Add_rm16_r16", Code.Add_rm16_r16 },
				{ "Add_rm32_r32", Code.Add_rm32_r32 },
				{ "Add_rm64_r64", Code.Add_rm64_r64 },
				{ "Add_r8_rm8", Code.Add_r8_rm8 },
				{ "Add_r16_rm16", Code.Add_r16_rm16 },
				{ "Add_r32_rm32", Code.Add_r32_rm32 },
				{ "Add_r64_rm64", Code.Add_r64_rm64 },
				{ "Add_AL_imm8", Code.Add_AL_imm8 },
				{ "Add_AX_imm16", Code.Add_AX_imm16 },
				{ "Add_EAX_imm32", Code.Add_EAX_imm32 },
				{ "Add_RAX_imm32", Code.Add_RAX_imm32 },
				{ "Pushw_ES", Code.Pushw_ES },
				{ "Pushd_ES", Code.Pushd_ES },
				{ "Popw_ES", Code.Popw_ES },
				{ "Popd_ES", Code.Popd_ES },
				{ "Or_rm8_r8", Code.Or_rm8_r8 },
				{ "Or_rm16_r16", Code.Or_rm16_r16 },
				{ "Or_rm32_r32", Code.Or_rm32_r32 },
				{ "Or_rm64_r64", Code.Or_rm64_r64 },
				{ "Or_r8_rm8", Code.Or_r8_rm8 },
				{ "Or_r16_rm16", Code.Or_r16_rm16 },
				{ "Or_r32_rm32", Code.Or_r32_rm32 },
				{ "Or_r64_rm64", Code.Or_r64_rm64 },
				{ "Or_AL_imm8", Code.Or_AL_imm8 },
				{ "Or_AX_imm16", Code.Or_AX_imm16 },
				{ "Or_EAX_imm32", Code.Or_EAX_imm32 },
				{ "Or_RAX_imm32", Code.Or_RAX_imm32 },
				{ "Pushw_CS", Code.Pushw_CS },
				{ "Pushd_CS", Code.Pushd_CS },
				{ "Popw_CS", Code.Popw_CS },
				{ "Adc_rm8_r8", Code.Adc_rm8_r8 },
				{ "Adc_rm16_r16", Code.Adc_rm16_r16 },
				{ "Adc_rm32_r32", Code.Adc_rm32_r32 },
				{ "Adc_rm64_r64", Code.Adc_rm64_r64 },
				{ "Adc_r8_rm8", Code.Adc_r8_rm8 },
				{ "Adc_r16_rm16", Code.Adc_r16_rm16 },
				{ "Adc_r32_rm32", Code.Adc_r32_rm32 },
				{ "Adc_r64_rm64", Code.Adc_r64_rm64 },
				{ "Adc_AL_imm8", Code.Adc_AL_imm8 },
				{ "Adc_AX_imm16", Code.Adc_AX_imm16 },
				{ "Adc_EAX_imm32", Code.Adc_EAX_imm32 },
				{ "Adc_RAX_imm32", Code.Adc_RAX_imm32 },
				{ "Pushw_SS", Code.Pushw_SS },
				{ "Pushd_SS", Code.Pushd_SS },
				{ "Popw_SS", Code.Popw_SS },
				{ "Popd_SS", Code.Popd_SS },
				{ "Sbb_rm8_r8", Code.Sbb_rm8_r8 },
				{ "Sbb_rm16_r16", Code.Sbb_rm16_r16 },
				{ "Sbb_rm32_r32", Code.Sbb_rm32_r32 },
				{ "Sbb_rm64_r64", Code.Sbb_rm64_r64 },
				{ "Sbb_r8_rm8", Code.Sbb_r8_rm8 },
				{ "Sbb_r16_rm16", Code.Sbb_r16_rm16 },
				{ "Sbb_r32_rm32", Code.Sbb_r32_rm32 },
				{ "Sbb_r64_rm64", Code.Sbb_r64_rm64 },
				{ "Sbb_AL_imm8", Code.Sbb_AL_imm8 },
				{ "Sbb_AX_imm16", Code.Sbb_AX_imm16 },
				{ "Sbb_EAX_imm32", Code.Sbb_EAX_imm32 },
				{ "Sbb_RAX_imm32", Code.Sbb_RAX_imm32 },
				{ "Pushw_DS", Code.Pushw_DS },
				{ "Pushd_DS", Code.Pushd_DS },
				{ "Popw_DS", Code.Popw_DS },
				{ "Popd_DS", Code.Popd_DS },
				{ "And_rm8_r8", Code.And_rm8_r8 },
				{ "And_rm16_r16", Code.And_rm16_r16 },
				{ "And_rm32_r32", Code.And_rm32_r32 },
				{ "And_rm64_r64", Code.And_rm64_r64 },
				{ "And_r8_rm8", Code.And_r8_rm8 },
				{ "And_r16_rm16", Code.And_r16_rm16 },
				{ "And_r32_rm32", Code.And_r32_rm32 },
				{ "And_r64_rm64", Code.And_r64_rm64 },
				{ "And_AL_imm8", Code.And_AL_imm8 },
				{ "And_AX_imm16", Code.And_AX_imm16 },
				{ "And_EAX_imm32", Code.And_EAX_imm32 },
				{ "And_RAX_imm32", Code.And_RAX_imm32 },
				{ "Daa", Code.Daa },
				{ "Sub_rm8_r8", Code.Sub_rm8_r8 },
				{ "Sub_rm16_r16", Code.Sub_rm16_r16 },
				{ "Sub_rm32_r32", Code.Sub_rm32_r32 },
				{ "Sub_rm64_r64", Code.Sub_rm64_r64 },
				{ "Sub_r8_rm8", Code.Sub_r8_rm8 },
				{ "Sub_r16_rm16", Code.Sub_r16_rm16 },
				{ "Sub_r32_rm32", Code.Sub_r32_rm32 },
				{ "Sub_r64_rm64", Code.Sub_r64_rm64 },
				{ "Sub_AL_imm8", Code.Sub_AL_imm8 },
				{ "Sub_AX_imm16", Code.Sub_AX_imm16 },
				{ "Sub_EAX_imm32", Code.Sub_EAX_imm32 },
				{ "Sub_RAX_imm32", Code.Sub_RAX_imm32 },
				{ "Das", Code.Das },
				{ "Xor_rm8_r8", Code.Xor_rm8_r8 },
				{ "Xor_rm16_r16", Code.Xor_rm16_r16 },
				{ "Xor_rm32_r32", Code.Xor_rm32_r32 },
				{ "Xor_rm64_r64", Code.Xor_rm64_r64 },
				{ "Xor_r8_rm8", Code.Xor_r8_rm8 },
				{ "Xor_r16_rm16", Code.Xor_r16_rm16 },
				{ "Xor_r32_rm32", Code.Xor_r32_rm32 },
				{ "Xor_r64_rm64", Code.Xor_r64_rm64 },
				{ "Xor_AL_imm8", Code.Xor_AL_imm8 },
				{ "Xor_AX_imm16", Code.Xor_AX_imm16 },
				{ "Xor_EAX_imm32", Code.Xor_EAX_imm32 },
				{ "Xor_RAX_imm32", Code.Xor_RAX_imm32 },
				{ "Aaa", Code.Aaa },
				{ "Cmp_rm8_r8", Code.Cmp_rm8_r8 },
				{ "Cmp_rm16_r16", Code.Cmp_rm16_r16 },
				{ "Cmp_rm32_r32", Code.Cmp_rm32_r32 },
				{ "Cmp_rm64_r64", Code.Cmp_rm64_r64 },
				{ "Cmp_r8_rm8", Code.Cmp_r8_rm8 },
				{ "Cmp_r16_rm16", Code.Cmp_r16_rm16 },
				{ "Cmp_r32_rm32", Code.Cmp_r32_rm32 },
				{ "Cmp_r64_rm64", Code.Cmp_r64_rm64 },
				{ "Cmp_AL_imm8", Code.Cmp_AL_imm8 },
				{ "Cmp_AX_imm16", Code.Cmp_AX_imm16 },
				{ "Cmp_EAX_imm32", Code.Cmp_EAX_imm32 },
				{ "Cmp_RAX_imm32", Code.Cmp_RAX_imm32 },
				{ "Aas", Code.Aas },
				{ "Inc_r16", Code.Inc_r16 },
				{ "Inc_r32", Code.Inc_r32 },
				{ "Dec_r16", Code.Dec_r16 },
				{ "Dec_r32", Code.Dec_r32 },
				{ "Push_r16", Code.Push_r16 },
				{ "Push_r32", Code.Push_r32 },
				{ "Push_r64", Code.Push_r64 },
				{ "Pop_r16", Code.Pop_r16 },
				{ "Pop_r32", Code.Pop_r32 },
				{ "Pop_r64", Code.Pop_r64 },
				{ "Pushaw", Code.Pushaw },
				{ "Pushad", Code.Pushad },
				{ "Popaw", Code.Popaw },
				{ "Popad", Code.Popad },
				{ "Bound_r16_m1616", Code.Bound_r16_m1616 },
				{ "Bound_r32_m3232", Code.Bound_r32_m3232 },
				{ "Arpl_rm16_r16", Code.Arpl_rm16_r16 },
				{ "Arpl_r32m16_r32", Code.Arpl_r32m16_r32 },
				{ "Movsxd_r16_rm16", Code.Movsxd_r16_rm16 },
				{ "Movsxd_r32_rm32", Code.Movsxd_r32_rm32 },
				{ "Movsxd_r64_rm32", Code.Movsxd_r64_rm32 },
				{ "Push_imm16", Code.Push_imm16 },
				{ "Pushd_imm32", Code.Pushd_imm32 },
				{ "Pushq_imm32", Code.Pushq_imm32 },
				{ "Imul_r16_rm16_imm16", Code.Imul_r16_rm16_imm16 },
				{ "Imul_r32_rm32_imm32", Code.Imul_r32_rm32_imm32 },
				{ "Imul_r64_rm64_imm32", Code.Imul_r64_rm64_imm32 },
				{ "Pushw_imm8", Code.Pushw_imm8 },
				{ "Pushd_imm8", Code.Pushd_imm8 },
				{ "Pushq_imm8", Code.Pushq_imm8 },
				{ "Imul_r16_rm16_imm8", Code.Imul_r16_rm16_imm8 },
				{ "Imul_r32_rm32_imm8", Code.Imul_r32_rm32_imm8 },
				{ "Imul_r64_rm64_imm8", Code.Imul_r64_rm64_imm8 },
				{ "Insb_m8_DX", Code.Insb_m8_DX },
				{ "Insw_m16_DX", Code.Insw_m16_DX },
				{ "Insd_m32_DX", Code.Insd_m32_DX },
				{ "Outsb_DX_m8", Code.Outsb_DX_m8 },
				{ "Outsw_DX_m16", Code.Outsw_DX_m16 },
				{ "Outsd_DX_m32", Code.Outsd_DX_m32 },
				{ "Jo_rel8_16", Code.Jo_rel8_16 },
				{ "Jo_rel8_32", Code.Jo_rel8_32 },
				{ "Jo_rel8_64", Code.Jo_rel8_64 },
				{ "Jno_rel8_16", Code.Jno_rel8_16 },
				{ "Jno_rel8_32", Code.Jno_rel8_32 },
				{ "Jno_rel8_64", Code.Jno_rel8_64 },
				{ "Jb_rel8_16", Code.Jb_rel8_16 },
				{ "Jb_rel8_32", Code.Jb_rel8_32 },
				{ "Jb_rel8_64", Code.Jb_rel8_64 },
				{ "Jae_rel8_16", Code.Jae_rel8_16 },
				{ "Jae_rel8_32", Code.Jae_rel8_32 },
				{ "Jae_rel8_64", Code.Jae_rel8_64 },
				{ "Je_rel8_16", Code.Je_rel8_16 },
				{ "Je_rel8_32", Code.Je_rel8_32 },
				{ "Je_rel8_64", Code.Je_rel8_64 },
				{ "Jne_rel8_16", Code.Jne_rel8_16 },
				{ "Jne_rel8_32", Code.Jne_rel8_32 },
				{ "Jne_rel8_64", Code.Jne_rel8_64 },
				{ "Jbe_rel8_16", Code.Jbe_rel8_16 },
				{ "Jbe_rel8_32", Code.Jbe_rel8_32 },
				{ "Jbe_rel8_64", Code.Jbe_rel8_64 },
				{ "Ja_rel8_16", Code.Ja_rel8_16 },
				{ "Ja_rel8_32", Code.Ja_rel8_32 },
				{ "Ja_rel8_64", Code.Ja_rel8_64 },
				{ "Js_rel8_16", Code.Js_rel8_16 },
				{ "Js_rel8_32", Code.Js_rel8_32 },
				{ "Js_rel8_64", Code.Js_rel8_64 },
				{ "Jns_rel8_16", Code.Jns_rel8_16 },
				{ "Jns_rel8_32", Code.Jns_rel8_32 },
				{ "Jns_rel8_64", Code.Jns_rel8_64 },
				{ "Jp_rel8_16", Code.Jp_rel8_16 },
				{ "Jp_rel8_32", Code.Jp_rel8_32 },
				{ "Jp_rel8_64", Code.Jp_rel8_64 },
				{ "Jnp_rel8_16", Code.Jnp_rel8_16 },
				{ "Jnp_rel8_32", Code.Jnp_rel8_32 },
				{ "Jnp_rel8_64", Code.Jnp_rel8_64 },
				{ "Jl_rel8_16", Code.Jl_rel8_16 },
				{ "Jl_rel8_32", Code.Jl_rel8_32 },
				{ "Jl_rel8_64", Code.Jl_rel8_64 },
				{ "Jge_rel8_16", Code.Jge_rel8_16 },
				{ "Jge_rel8_32", Code.Jge_rel8_32 },
				{ "Jge_rel8_64", Code.Jge_rel8_64 },
				{ "Jle_rel8_16", Code.Jle_rel8_16 },
				{ "Jle_rel8_32", Code.Jle_rel8_32 },
				{ "Jle_rel8_64", Code.Jle_rel8_64 },
				{ "Jg_rel8_16", Code.Jg_rel8_16 },
				{ "Jg_rel8_32", Code.Jg_rel8_32 },
				{ "Jg_rel8_64", Code.Jg_rel8_64 },
				{ "Add_rm8_imm8", Code.Add_rm8_imm8 },
				{ "Or_rm8_imm8", Code.Or_rm8_imm8 },
				{ "Adc_rm8_imm8", Code.Adc_rm8_imm8 },
				{ "Sbb_rm8_imm8", Code.Sbb_rm8_imm8 },
				{ "And_rm8_imm8", Code.And_rm8_imm8 },
				{ "Sub_rm8_imm8", Code.Sub_rm8_imm8 },
				{ "Xor_rm8_imm8", Code.Xor_rm8_imm8 },
				{ "Cmp_rm8_imm8", Code.Cmp_rm8_imm8 },
				{ "Add_rm16_imm16", Code.Add_rm16_imm16 },
				{ "Add_rm32_imm32", Code.Add_rm32_imm32 },
				{ "Add_rm64_imm32", Code.Add_rm64_imm32 },
				{ "Or_rm16_imm16", Code.Or_rm16_imm16 },
				{ "Or_rm32_imm32", Code.Or_rm32_imm32 },
				{ "Or_rm64_imm32", Code.Or_rm64_imm32 },
				{ "Adc_rm16_imm16", Code.Adc_rm16_imm16 },
				{ "Adc_rm32_imm32", Code.Adc_rm32_imm32 },
				{ "Adc_rm64_imm32", Code.Adc_rm64_imm32 },
				{ "Sbb_rm16_imm16", Code.Sbb_rm16_imm16 },
				{ "Sbb_rm32_imm32", Code.Sbb_rm32_imm32 },
				{ "Sbb_rm64_imm32", Code.Sbb_rm64_imm32 },
				{ "And_rm16_imm16", Code.And_rm16_imm16 },
				{ "And_rm32_imm32", Code.And_rm32_imm32 },
				{ "And_rm64_imm32", Code.And_rm64_imm32 },
				{ "Sub_rm16_imm16", Code.Sub_rm16_imm16 },
				{ "Sub_rm32_imm32", Code.Sub_rm32_imm32 },
				{ "Sub_rm64_imm32", Code.Sub_rm64_imm32 },
				{ "Xor_rm16_imm16", Code.Xor_rm16_imm16 },
				{ "Xor_rm32_imm32", Code.Xor_rm32_imm32 },
				{ "Xor_rm64_imm32", Code.Xor_rm64_imm32 },
				{ "Cmp_rm16_imm16", Code.Cmp_rm16_imm16 },
				{ "Cmp_rm32_imm32", Code.Cmp_rm32_imm32 },
				{ "Cmp_rm64_imm32", Code.Cmp_rm64_imm32 },
				{ "Add_rm8_imm8_82", Code.Add_rm8_imm8_82 },
				{ "Or_rm8_imm8_82", Code.Or_rm8_imm8_82 },
				{ "Adc_rm8_imm8_82", Code.Adc_rm8_imm8_82 },
				{ "Sbb_rm8_imm8_82", Code.Sbb_rm8_imm8_82 },
				{ "And_rm8_imm8_82", Code.And_rm8_imm8_82 },
				{ "Sub_rm8_imm8_82", Code.Sub_rm8_imm8_82 },
				{ "Xor_rm8_imm8_82", Code.Xor_rm8_imm8_82 },
				{ "Cmp_rm8_imm8_82", Code.Cmp_rm8_imm8_82 },
				{ "Add_rm16_imm8", Code.Add_rm16_imm8 },
				{ "Add_rm32_imm8", Code.Add_rm32_imm8 },
				{ "Add_rm64_imm8", Code.Add_rm64_imm8 },
				{ "Or_rm16_imm8", Code.Or_rm16_imm8 },
				{ "Or_rm32_imm8", Code.Or_rm32_imm8 },
				{ "Or_rm64_imm8", Code.Or_rm64_imm8 },
				{ "Adc_rm16_imm8", Code.Adc_rm16_imm8 },
				{ "Adc_rm32_imm8", Code.Adc_rm32_imm8 },
				{ "Adc_rm64_imm8", Code.Adc_rm64_imm8 },
				{ "Sbb_rm16_imm8", Code.Sbb_rm16_imm8 },
				{ "Sbb_rm32_imm8", Code.Sbb_rm32_imm8 },
				{ "Sbb_rm64_imm8", Code.Sbb_rm64_imm8 },
				{ "And_rm16_imm8", Code.And_rm16_imm8 },
				{ "And_rm32_imm8", Code.And_rm32_imm8 },
				{ "And_rm64_imm8", Code.And_rm64_imm8 },
				{ "Sub_rm16_imm8", Code.Sub_rm16_imm8 },
				{ "Sub_rm32_imm8", Code.Sub_rm32_imm8 },
				{ "Sub_rm64_imm8", Code.Sub_rm64_imm8 },
				{ "Xor_rm16_imm8", Code.Xor_rm16_imm8 },
				{ "Xor_rm32_imm8", Code.Xor_rm32_imm8 },
				{ "Xor_rm64_imm8", Code.Xor_rm64_imm8 },
				{ "Cmp_rm16_imm8", Code.Cmp_rm16_imm8 },
				{ "Cmp_rm32_imm8", Code.Cmp_rm32_imm8 },
				{ "Cmp_rm64_imm8", Code.Cmp_rm64_imm8 },
				{ "Test_rm8_r8", Code.Test_rm8_r8 },
				{ "Test_rm16_r16", Code.Test_rm16_r16 },
				{ "Test_rm32_r32", Code.Test_rm32_r32 },
				{ "Test_rm64_r64", Code.Test_rm64_r64 },
				{ "Xchg_rm8_r8", Code.Xchg_rm8_r8 },
				{ "Xchg_rm16_r16", Code.Xchg_rm16_r16 },
				{ "Xchg_rm32_r32", Code.Xchg_rm32_r32 },
				{ "Xchg_rm64_r64", Code.Xchg_rm64_r64 },
				{ "Mov_rm8_r8", Code.Mov_rm8_r8 },
				{ "Mov_rm16_r16", Code.Mov_rm16_r16 },
				{ "Mov_rm32_r32", Code.Mov_rm32_r32 },
				{ "Mov_rm64_r64", Code.Mov_rm64_r64 },
				{ "Mov_r8_rm8", Code.Mov_r8_rm8 },
				{ "Mov_r16_rm16", Code.Mov_r16_rm16 },
				{ "Mov_r32_rm32", Code.Mov_r32_rm32 },
				{ "Mov_r64_rm64", Code.Mov_r64_rm64 },
				{ "Mov_rm16_Sreg", Code.Mov_rm16_Sreg },
				{ "Mov_r32m16_Sreg", Code.Mov_r32m16_Sreg },
				{ "Mov_r64m16_Sreg", Code.Mov_r64m16_Sreg },
				{ "Lea_r16_m", Code.Lea_r16_m },
				{ "Lea_r32_m", Code.Lea_r32_m },
				{ "Lea_r64_m", Code.Lea_r64_m },
				{ "Mov_Sreg_rm16", Code.Mov_Sreg_rm16 },
				{ "Mov_Sreg_r32m16", Code.Mov_Sreg_r32m16 },
				{ "Mov_Sreg_r64m16", Code.Mov_Sreg_r64m16 },
				{ "Pop_rm16", Code.Pop_rm16 },
				{ "Pop_rm32", Code.Pop_rm32 },
				{ "Pop_rm64", Code.Pop_rm64 },
				{ "Nopw", Code.Nopw },
				{ "Nopd", Code.Nopd },
				{ "Nopq", Code.Nopq },
				{ "Xchg_r16_AX", Code.Xchg_r16_AX },
				{ "Xchg_r32_EAX", Code.Xchg_r32_EAX },
				{ "Xchg_r64_RAX", Code.Xchg_r64_RAX },
				{ "Pause", Code.Pause },
				{ "Cbw", Code.Cbw },
				{ "Cwde", Code.Cwde },
				{ "Cdqe", Code.Cdqe },
				{ "Cwd", Code.Cwd },
				{ "Cdq", Code.Cdq },
				{ "Cqo", Code.Cqo },
				{ "Call_ptr1616", Code.Call_ptr1616 },
				{ "Call_ptr1632", Code.Call_ptr1632 },
				{ "Wait", Code.Wait },
				{ "Pushfw", Code.Pushfw },
				{ "Pushfd", Code.Pushfd },
				{ "Pushfq", Code.Pushfq },
				{ "Popfw", Code.Popfw },
				{ "Popfd", Code.Popfd },
				{ "Popfq", Code.Popfq },
				{ "Sahf", Code.Sahf },
				{ "Lahf", Code.Lahf },
				{ "Mov_AL_moffs8", Code.Mov_AL_moffs8 },
				{ "Mov_AX_moffs16", Code.Mov_AX_moffs16 },
				{ "Mov_EAX_moffs32", Code.Mov_EAX_moffs32 },
				{ "Mov_RAX_moffs64", Code.Mov_RAX_moffs64 },
				{ "Mov_moffs8_AL", Code.Mov_moffs8_AL },
				{ "Mov_moffs16_AX", Code.Mov_moffs16_AX },
				{ "Mov_moffs32_EAX", Code.Mov_moffs32_EAX },
				{ "Mov_moffs64_RAX", Code.Mov_moffs64_RAX },
				{ "Movsb_m8_m8", Code.Movsb_m8_m8 },
				{ "Movsw_m16_m16", Code.Movsw_m16_m16 },
				{ "Movsd_m32_m32", Code.Movsd_m32_m32 },
				{ "Movsq_m64_m64", Code.Movsq_m64_m64 },
				{ "Cmpsb_m8_m8", Code.Cmpsb_m8_m8 },
				{ "Cmpsw_m16_m16", Code.Cmpsw_m16_m16 },
				{ "Cmpsd_m32_m32", Code.Cmpsd_m32_m32 },
				{ "Cmpsq_m64_m64", Code.Cmpsq_m64_m64 },
				{ "Test_AL_imm8", Code.Test_AL_imm8 },
				{ "Test_AX_imm16", Code.Test_AX_imm16 },
				{ "Test_EAX_imm32", Code.Test_EAX_imm32 },
				{ "Test_RAX_imm32", Code.Test_RAX_imm32 },
				{ "Stosb_m8_AL", Code.Stosb_m8_AL },
				{ "Stosw_m16_AX", Code.Stosw_m16_AX },
				{ "Stosd_m32_EAX", Code.Stosd_m32_EAX },
				{ "Stosq_m64_RAX", Code.Stosq_m64_RAX },
				{ "Lodsb_AL_m8", Code.Lodsb_AL_m8 },
				{ "Lodsw_AX_m16", Code.Lodsw_AX_m16 },
				{ "Lodsd_EAX_m32", Code.Lodsd_EAX_m32 },
				{ "Lodsq_RAX_m64", Code.Lodsq_RAX_m64 },
				{ "Scasb_AL_m8", Code.Scasb_AL_m8 },
				{ "Scasw_AX_m16", Code.Scasw_AX_m16 },
				{ "Scasd_EAX_m32", Code.Scasd_EAX_m32 },
				{ "Scasq_RAX_m64", Code.Scasq_RAX_m64 },
				{ "Mov_r8_imm8", Code.Mov_r8_imm8 },
				{ "Mov_r16_imm16", Code.Mov_r16_imm16 },
				{ "Mov_r32_imm32", Code.Mov_r32_imm32 },
				{ "Mov_r64_imm64", Code.Mov_r64_imm64 },
				{ "Rol_rm8_imm8", Code.Rol_rm8_imm8 },
				{ "Ror_rm8_imm8", Code.Ror_rm8_imm8 },
				{ "Rcl_rm8_imm8", Code.Rcl_rm8_imm8 },
				{ "Rcr_rm8_imm8", Code.Rcr_rm8_imm8 },
				{ "Shl_rm8_imm8", Code.Shl_rm8_imm8 },
				{ "Shr_rm8_imm8", Code.Shr_rm8_imm8 },
				{ "Sal_rm8_imm8", Code.Sal_rm8_imm8 },
				{ "Sar_rm8_imm8", Code.Sar_rm8_imm8 },
				{ "Rol_rm16_imm8", Code.Rol_rm16_imm8 },
				{ "Rol_rm32_imm8", Code.Rol_rm32_imm8 },
				{ "Rol_rm64_imm8", Code.Rol_rm64_imm8 },
				{ "Ror_rm16_imm8", Code.Ror_rm16_imm8 },
				{ "Ror_rm32_imm8", Code.Ror_rm32_imm8 },
				{ "Ror_rm64_imm8", Code.Ror_rm64_imm8 },
				{ "Rcl_rm16_imm8", Code.Rcl_rm16_imm8 },
				{ "Rcl_rm32_imm8", Code.Rcl_rm32_imm8 },
				{ "Rcl_rm64_imm8", Code.Rcl_rm64_imm8 },
				{ "Rcr_rm16_imm8", Code.Rcr_rm16_imm8 },
				{ "Rcr_rm32_imm8", Code.Rcr_rm32_imm8 },
				{ "Rcr_rm64_imm8", Code.Rcr_rm64_imm8 },
				{ "Shl_rm16_imm8", Code.Shl_rm16_imm8 },
				{ "Shl_rm32_imm8", Code.Shl_rm32_imm8 },
				{ "Shl_rm64_imm8", Code.Shl_rm64_imm8 },
				{ "Shr_rm16_imm8", Code.Shr_rm16_imm8 },
				{ "Shr_rm32_imm8", Code.Shr_rm32_imm8 },
				{ "Shr_rm64_imm8", Code.Shr_rm64_imm8 },
				{ "Sal_rm16_imm8", Code.Sal_rm16_imm8 },
				{ "Sal_rm32_imm8", Code.Sal_rm32_imm8 },
				{ "Sal_rm64_imm8", Code.Sal_rm64_imm8 },
				{ "Sar_rm16_imm8", Code.Sar_rm16_imm8 },
				{ "Sar_rm32_imm8", Code.Sar_rm32_imm8 },
				{ "Sar_rm64_imm8", Code.Sar_rm64_imm8 },
				{ "Retnw_imm16", Code.Retnw_imm16 },
				{ "Retnd_imm16", Code.Retnd_imm16 },
				{ "Retnq_imm16", Code.Retnq_imm16 },
				{ "Retnw", Code.Retnw },
				{ "Retnd", Code.Retnd },
				{ "Retnq", Code.Retnq },
				{ "Les_r16_m1616", Code.Les_r16_m1616 },
				{ "Les_r32_m1632", Code.Les_r32_m1632 },
				{ "Lds_r16_m1616", Code.Lds_r16_m1616 },
				{ "Lds_r32_m1632", Code.Lds_r32_m1632 },
				{ "Mov_rm8_imm8", Code.Mov_rm8_imm8 },
				{ "Xabort_imm8", Code.Xabort_imm8 },
				{ "Mov_rm16_imm16", Code.Mov_rm16_imm16 },
				{ "Mov_rm32_imm32", Code.Mov_rm32_imm32 },
				{ "Mov_rm64_imm32", Code.Mov_rm64_imm32 },
				{ "Xbegin_rel16", Code.Xbegin_rel16 },
				{ "Xbegin_rel32", Code.Xbegin_rel32 },
				{ "Enterw_imm16_imm8", Code.Enterw_imm16_imm8 },
				{ "Enterd_imm16_imm8", Code.Enterd_imm16_imm8 },
				{ "Enterq_imm16_imm8", Code.Enterq_imm16_imm8 },
				{ "Leavew", Code.Leavew },
				{ "Leaved", Code.Leaved },
				{ "Leaveq", Code.Leaveq },
				{ "Retfw_imm16", Code.Retfw_imm16 },
				{ "Retfd_imm16", Code.Retfd_imm16 },
				{ "Retfq_imm16", Code.Retfq_imm16 },
				{ "Retfw", Code.Retfw },
				{ "Retfd", Code.Retfd },
				{ "Retfq", Code.Retfq },
				{ "Int3", Code.Int3 },
				{ "Int_imm8", Code.Int_imm8 },
				{ "Into", Code.Into },
				{ "Iretw", Code.Iretw },
				{ "Iretd", Code.Iretd },
				{ "Iretq", Code.Iretq },
				{ "Rol_rm8_1", Code.Rol_rm8_1 },
				{ "Ror_rm8_1", Code.Ror_rm8_1 },
				{ "Rcl_rm8_1", Code.Rcl_rm8_1 },
				{ "Rcr_rm8_1", Code.Rcr_rm8_1 },
				{ "Shl_rm8_1", Code.Shl_rm8_1 },
				{ "Shr_rm8_1", Code.Shr_rm8_1 },
				{ "Sal_rm8_1", Code.Sal_rm8_1 },
				{ "Sar_rm8_1", Code.Sar_rm8_1 },
				{ "Rol_rm16_1", Code.Rol_rm16_1 },
				{ "Rol_rm32_1", Code.Rol_rm32_1 },
				{ "Rol_rm64_1", Code.Rol_rm64_1 },
				{ "Ror_rm16_1", Code.Ror_rm16_1 },
				{ "Ror_rm32_1", Code.Ror_rm32_1 },
				{ "Ror_rm64_1", Code.Ror_rm64_1 },
				{ "Rcl_rm16_1", Code.Rcl_rm16_1 },
				{ "Rcl_rm32_1", Code.Rcl_rm32_1 },
				{ "Rcl_rm64_1", Code.Rcl_rm64_1 },
				{ "Rcr_rm16_1", Code.Rcr_rm16_1 },
				{ "Rcr_rm32_1", Code.Rcr_rm32_1 },
				{ "Rcr_rm64_1", Code.Rcr_rm64_1 },
				{ "Shl_rm16_1", Code.Shl_rm16_1 },
				{ "Shl_rm32_1", Code.Shl_rm32_1 },
				{ "Shl_rm64_1", Code.Shl_rm64_1 },
				{ "Shr_rm16_1", Code.Shr_rm16_1 },
				{ "Shr_rm32_1", Code.Shr_rm32_1 },
				{ "Shr_rm64_1", Code.Shr_rm64_1 },
				{ "Sal_rm16_1", Code.Sal_rm16_1 },
				{ "Sal_rm32_1", Code.Sal_rm32_1 },
				{ "Sal_rm64_1", Code.Sal_rm64_1 },
				{ "Sar_rm16_1", Code.Sar_rm16_1 },
				{ "Sar_rm32_1", Code.Sar_rm32_1 },
				{ "Sar_rm64_1", Code.Sar_rm64_1 },
				{ "Rol_rm8_CL", Code.Rol_rm8_CL },
				{ "Ror_rm8_CL", Code.Ror_rm8_CL },
				{ "Rcl_rm8_CL", Code.Rcl_rm8_CL },
				{ "Rcr_rm8_CL", Code.Rcr_rm8_CL },
				{ "Shl_rm8_CL", Code.Shl_rm8_CL },
				{ "Shr_rm8_CL", Code.Shr_rm8_CL },
				{ "Sal_rm8_CL", Code.Sal_rm8_CL },
				{ "Sar_rm8_CL", Code.Sar_rm8_CL },
				{ "Rol_rm16_CL", Code.Rol_rm16_CL },
				{ "Rol_rm32_CL", Code.Rol_rm32_CL },
				{ "Rol_rm64_CL", Code.Rol_rm64_CL },
				{ "Ror_rm16_CL", Code.Ror_rm16_CL },
				{ "Ror_rm32_CL", Code.Ror_rm32_CL },
				{ "Ror_rm64_CL", Code.Ror_rm64_CL },
				{ "Rcl_rm16_CL", Code.Rcl_rm16_CL },
				{ "Rcl_rm32_CL", Code.Rcl_rm32_CL },
				{ "Rcl_rm64_CL", Code.Rcl_rm64_CL },
				{ "Rcr_rm16_CL", Code.Rcr_rm16_CL },
				{ "Rcr_rm32_CL", Code.Rcr_rm32_CL },
				{ "Rcr_rm64_CL", Code.Rcr_rm64_CL },
				{ "Shl_rm16_CL", Code.Shl_rm16_CL },
				{ "Shl_rm32_CL", Code.Shl_rm32_CL },
				{ "Shl_rm64_CL", Code.Shl_rm64_CL },
				{ "Shr_rm16_CL", Code.Shr_rm16_CL },
				{ "Shr_rm32_CL", Code.Shr_rm32_CL },
				{ "Shr_rm64_CL", Code.Shr_rm64_CL },
				{ "Sal_rm16_CL", Code.Sal_rm16_CL },
				{ "Sal_rm32_CL", Code.Sal_rm32_CL },
				{ "Sal_rm64_CL", Code.Sal_rm64_CL },
				{ "Sar_rm16_CL", Code.Sar_rm16_CL },
				{ "Sar_rm32_CL", Code.Sar_rm32_CL },
				{ "Sar_rm64_CL", Code.Sar_rm64_CL },
				{ "Aam_imm8", Code.Aam_imm8 },
				{ "Aad_imm8", Code.Aad_imm8 },
				{ "Salc", Code.Salc },
				{ "Xlat_m8", Code.Xlat_m8 },
				{ "Fadd_m32fp", Code.Fadd_m32fp },
				{ "Fmul_m32fp", Code.Fmul_m32fp },
				{ "Fcom_m32fp", Code.Fcom_m32fp },
				{ "Fcomp_m32fp", Code.Fcomp_m32fp },
				{ "Fsub_m32fp", Code.Fsub_m32fp },
				{ "Fsubr_m32fp", Code.Fsubr_m32fp },
				{ "Fdiv_m32fp", Code.Fdiv_m32fp },
				{ "Fdivr_m32fp", Code.Fdivr_m32fp },
				{ "Fadd_st0_sti", Code.Fadd_st0_sti },
				{ "Fmul_st0_sti", Code.Fmul_st0_sti },
				{ "Fcom_st0_sti", Code.Fcom_st0_sti },
				{ "Fcomp_st0_sti", Code.Fcomp_st0_sti },
				{ "Fsub_st0_sti", Code.Fsub_st0_sti },
				{ "Fsubr_st0_sti", Code.Fsubr_st0_sti },
				{ "Fdiv_st0_sti", Code.Fdiv_st0_sti },
				{ "Fdivr_st0_sti", Code.Fdivr_st0_sti },
				{ "Fld_m32fp", Code.Fld_m32fp },
				{ "Fst_m32fp", Code.Fst_m32fp },
				{ "Fstp_m32fp", Code.Fstp_m32fp },
				{ "Fldenv_m14byte", Code.Fldenv_m14byte },
				{ "Fldenv_m28byte", Code.Fldenv_m28byte },
				{ "Fldcw_m2byte", Code.Fldcw_m2byte },
				{ "Fnstenv_m14byte", Code.Fnstenv_m14byte },
				{ "Fstenv_m14byte", Code.Fstenv_m14byte },
				{ "Fnstenv_m28byte", Code.Fnstenv_m28byte },
				{ "Fstenv_m28byte", Code.Fstenv_m28byte },
				{ "Fnstcw_m2byte", Code.Fnstcw_m2byte },
				{ "Fstcw_m2byte", Code.Fstcw_m2byte },
				{ "Fld_st0_sti", Code.Fld_st0_sti },
				{ "Fxch_st0_sti", Code.Fxch_st0_sti },
				{ "Fnop", Code.Fnop },
				{ "Fstpnce_sti", Code.Fstpnce_sti },
				{ "Fchs", Code.Fchs },
				{ "Fabs", Code.Fabs },
				{ "Ftst", Code.Ftst },
				{ "Fxam", Code.Fxam },
				{ "Fld1", Code.Fld1 },
				{ "Fldl2t", Code.Fldl2t },
				{ "Fldl2e", Code.Fldl2e },
				{ "Fldpi", Code.Fldpi },
				{ "Fldlg2", Code.Fldlg2 },
				{ "Fldln2", Code.Fldln2 },
				{ "Fldz", Code.Fldz },
				{ "F2xm1", Code.F2xm1 },
				{ "Fyl2x", Code.Fyl2x },
				{ "Fptan", Code.Fptan },
				{ "Fpatan", Code.Fpatan },
				{ "Fxtract", Code.Fxtract },
				{ "Fprem1", Code.Fprem1 },
				{ "Fdecstp", Code.Fdecstp },
				{ "Fincstp", Code.Fincstp },
				{ "Fprem", Code.Fprem },
				{ "Fyl2xp1", Code.Fyl2xp1 },
				{ "Fsqrt", Code.Fsqrt },
				{ "Fsincos", Code.Fsincos },
				{ "Frndint", Code.Frndint },
				{ "Fscale", Code.Fscale },
				{ "Fsin", Code.Fsin },
				{ "Fcos", Code.Fcos },
				{ "Fiadd_m32int", Code.Fiadd_m32int },
				{ "Fimul_m32int", Code.Fimul_m32int },
				{ "Ficom_m32int", Code.Ficom_m32int },
				{ "Ficomp_m32int", Code.Ficomp_m32int },
				{ "Fisub_m32int", Code.Fisub_m32int },
				{ "Fisubr_m32int", Code.Fisubr_m32int },
				{ "Fidiv_m32int", Code.Fidiv_m32int },
				{ "Fidivr_m32int", Code.Fidivr_m32int },
				{ "Fcmovb_st0_sti", Code.Fcmovb_st0_sti },
				{ "Fcmove_st0_sti", Code.Fcmove_st0_sti },
				{ "Fcmovbe_st0_sti", Code.Fcmovbe_st0_sti },
				{ "Fcmovu_st0_sti", Code.Fcmovu_st0_sti },
				{ "Fucompp", Code.Fucompp },
				{ "Fild_m32int", Code.Fild_m32int },
				{ "Fisttp_m32int", Code.Fisttp_m32int },
				{ "Fist_m32int", Code.Fist_m32int },
				{ "Fistp_m32int", Code.Fistp_m32int },
				{ "Fld_m80fp", Code.Fld_m80fp },
				{ "Fstp_m80fp", Code.Fstp_m80fp },
				{ "Fcmovnb_st0_sti", Code.Fcmovnb_st0_sti },
				{ "Fcmovne_st0_sti", Code.Fcmovne_st0_sti },
				{ "Fcmovnbe_st0_sti", Code.Fcmovnbe_st0_sti },
				{ "Fcmovnu_st0_sti", Code.Fcmovnu_st0_sti },
				{ "Fneni", Code.Fneni },
				{ "Feni", Code.Feni },
				{ "Fndisi", Code.Fndisi },
				{ "Fdisi", Code.Fdisi },
				{ "Fnclex", Code.Fnclex },
				{ "Fclex", Code.Fclex },
				{ "Fninit", Code.Fninit },
				{ "Finit", Code.Finit },
				{ "Fnsetpm", Code.Fnsetpm },
				{ "Fsetpm", Code.Fsetpm },
				{ "Frstpm", Code.Frstpm },
				{ "Fucomi_st0_sti", Code.Fucomi_st0_sti },
				{ "Fcomi_st0_sti", Code.Fcomi_st0_sti },
				{ "Fadd_m64fp", Code.Fadd_m64fp },
				{ "Fmul_m64fp", Code.Fmul_m64fp },
				{ "Fcom_m64fp", Code.Fcom_m64fp },
				{ "Fcomp_m64fp", Code.Fcomp_m64fp },
				{ "Fsub_m64fp", Code.Fsub_m64fp },
				{ "Fsubr_m64fp", Code.Fsubr_m64fp },
				{ "Fdiv_m64fp", Code.Fdiv_m64fp },
				{ "Fdivr_m64fp", Code.Fdivr_m64fp },
				{ "Fadd_sti_st0", Code.Fadd_sti_st0 },
				{ "Fmul_sti_st0", Code.Fmul_sti_st0 },
				{ "Fcom_st0_sti_DCD0", Code.Fcom_st0_sti_DCD0 },
				{ "Fcomp_st0_sti_DCD8", Code.Fcomp_st0_sti_DCD8 },
				{ "Fsubr_sti_st0", Code.Fsubr_sti_st0 },
				{ "Fsub_sti_st0", Code.Fsub_sti_st0 },
				{ "Fdivr_sti_st0", Code.Fdivr_sti_st0 },
				{ "Fdiv_sti_st0", Code.Fdiv_sti_st0 },
				{ "Fld_m64fp", Code.Fld_m64fp },
				{ "Fisttp_m64int", Code.Fisttp_m64int },
				{ "Fst_m64fp", Code.Fst_m64fp },
				{ "Fstp_m64fp", Code.Fstp_m64fp },
				{ "Frstor_m94byte", Code.Frstor_m94byte },
				{ "Frstor_m108byte", Code.Frstor_m108byte },
				{ "Fnsave_m94byte", Code.Fnsave_m94byte },
				{ "Fsave_m94byte", Code.Fsave_m94byte },
				{ "Fnsave_m108byte", Code.Fnsave_m108byte },
				{ "Fsave_m108byte", Code.Fsave_m108byte },
				{ "Fnstsw_m2byte", Code.Fnstsw_m2byte },
				{ "Fstsw_m2byte", Code.Fstsw_m2byte },
				{ "Ffree_sti", Code.Ffree_sti },
				{ "Fxch_st0_sti_DDC8", Code.Fxch_st0_sti_DDC8 },
				{ "Fst_sti", Code.Fst_sti },
				{ "Fstp_sti", Code.Fstp_sti },
				{ "Fucom_st0_sti", Code.Fucom_st0_sti },
				{ "Fucomp_st0_sti", Code.Fucomp_st0_sti },
				{ "Fiadd_m16int", Code.Fiadd_m16int },
				{ "Fimul_m16int", Code.Fimul_m16int },
				{ "Ficom_m16int", Code.Ficom_m16int },
				{ "Ficomp_m16int", Code.Ficomp_m16int },
				{ "Fisub_m16int", Code.Fisub_m16int },
				{ "Fisubr_m16int", Code.Fisubr_m16int },
				{ "Fidiv_m16int", Code.Fidiv_m16int },
				{ "Fidivr_m16int", Code.Fidivr_m16int },
				{ "Faddp_sti_st0", Code.Faddp_sti_st0 },
				{ "Fmulp_sti_st0", Code.Fmulp_sti_st0 },
				{ "Fcomp_st0_sti_DED0", Code.Fcomp_st0_sti_DED0 },
				{ "Fcompp", Code.Fcompp },
				{ "Fsubrp_sti_st0", Code.Fsubrp_sti_st0 },
				{ "Fsubp_sti_st0", Code.Fsubp_sti_st0 },
				{ "Fdivrp_sti_st0", Code.Fdivrp_sti_st0 },
				{ "Fdivp_sti_st0", Code.Fdivp_sti_st0 },
				{ "Fild_m16int", Code.Fild_m16int },
				{ "Fisttp_m16int", Code.Fisttp_m16int },
				{ "Fist_m16int", Code.Fist_m16int },
				{ "Fistp_m16int", Code.Fistp_m16int },
				{ "Fbld_m80bcd", Code.Fbld_m80bcd },
				{ "Fild_m64int", Code.Fild_m64int },
				{ "Fbstp_m80bcd", Code.Fbstp_m80bcd },
				{ "Fistp_m64int", Code.Fistp_m64int },
				{ "Ffreep_sti", Code.Ffreep_sti },
				{ "Fxch_st0_sti_DFC8", Code.Fxch_st0_sti_DFC8 },
				{ "Fstp_sti_DFD0", Code.Fstp_sti_DFD0 },
				{ "Fstp_sti_DFD8", Code.Fstp_sti_DFD8 },
				{ "Fnstsw_AX", Code.Fnstsw_AX },
				{ "Fstsw_AX", Code.Fstsw_AX },
				{ "Fstdw_AX", Code.Fstdw_AX },
				{ "Fstsg_AX", Code.Fstsg_AX },
				{ "Fucomip_st0_sti", Code.Fucomip_st0_sti },
				{ "Fcomip_st0_sti", Code.Fcomip_st0_sti },
				{ "Loopne_rel8_16_CX", Code.Loopne_rel8_16_CX },
				{ "Loopne_rel8_32_CX", Code.Loopne_rel8_32_CX },
				{ "Loopne_rel8_16_ECX", Code.Loopne_rel8_16_ECX },
				{ "Loopne_rel8_32_ECX", Code.Loopne_rel8_32_ECX },
				{ "Loopne_rel8_64_ECX", Code.Loopne_rel8_64_ECX },
				{ "Loopne_rel8_16_RCX", Code.Loopne_rel8_16_RCX },
				{ "Loopne_rel8_64_RCX", Code.Loopne_rel8_64_RCX },
				{ "Loope_rel8_16_CX", Code.Loope_rel8_16_CX },
				{ "Loope_rel8_32_CX", Code.Loope_rel8_32_CX },
				{ "Loope_rel8_16_ECX", Code.Loope_rel8_16_ECX },
				{ "Loope_rel8_32_ECX", Code.Loope_rel8_32_ECX },
				{ "Loope_rel8_64_ECX", Code.Loope_rel8_64_ECX },
				{ "Loope_rel8_16_RCX", Code.Loope_rel8_16_RCX },
				{ "Loope_rel8_64_RCX", Code.Loope_rel8_64_RCX },
				{ "Loop_rel8_16_CX", Code.Loop_rel8_16_CX },
				{ "Loop_rel8_32_CX", Code.Loop_rel8_32_CX },
				{ "Loop_rel8_16_ECX", Code.Loop_rel8_16_ECX },
				{ "Loop_rel8_32_ECX", Code.Loop_rel8_32_ECX },
				{ "Loop_rel8_64_ECX", Code.Loop_rel8_64_ECX },
				{ "Loop_rel8_16_RCX", Code.Loop_rel8_16_RCX },
				{ "Loop_rel8_64_RCX", Code.Loop_rel8_64_RCX },
				{ "Jcxz_rel8_16", Code.Jcxz_rel8_16 },
				{ "Jcxz_rel8_32", Code.Jcxz_rel8_32 },
				{ "Jecxz_rel8_16", Code.Jecxz_rel8_16 },
				{ "Jecxz_rel8_32", Code.Jecxz_rel8_32 },
				{ "Jecxz_rel8_64", Code.Jecxz_rel8_64 },
				{ "Jrcxz_rel8_16", Code.Jrcxz_rel8_16 },
				{ "Jrcxz_rel8_64", Code.Jrcxz_rel8_64 },
				{ "In_AL_imm8", Code.In_AL_imm8 },
				{ "In_AX_imm8", Code.In_AX_imm8 },
				{ "In_EAX_imm8", Code.In_EAX_imm8 },
				{ "Out_imm8_AL", Code.Out_imm8_AL },
				{ "Out_imm8_AX", Code.Out_imm8_AX },
				{ "Out_imm8_EAX", Code.Out_imm8_EAX },
				{ "Call_rel16", Code.Call_rel16 },
				{ "Call_rel32_32", Code.Call_rel32_32 },
				{ "Call_rel32_64", Code.Call_rel32_64 },
				{ "Jmp_rel16", Code.Jmp_rel16 },
				{ "Jmp_rel32_32", Code.Jmp_rel32_32 },
				{ "Jmp_rel32_64", Code.Jmp_rel32_64 },
				{ "Jmp_ptr1616", Code.Jmp_ptr1616 },
				{ "Jmp_ptr1632", Code.Jmp_ptr1632 },
				{ "Jmp_rel8_16", Code.Jmp_rel8_16 },
				{ "Jmp_rel8_32", Code.Jmp_rel8_32 },
				{ "Jmp_rel8_64", Code.Jmp_rel8_64 },
				{ "In_AL_DX", Code.In_AL_DX },
				{ "In_AX_DX", Code.In_AX_DX },
				{ "In_EAX_DX", Code.In_EAX_DX },
				{ "Out_DX_AL", Code.Out_DX_AL },
				{ "Out_DX_AX", Code.Out_DX_AX },
				{ "Out_DX_EAX", Code.Out_DX_EAX },
				{ "Int1", Code.Int1 },
				{ "Hlt", Code.Hlt },
				{ "Cmc", Code.Cmc },
				{ "Test_rm8_imm8", Code.Test_rm8_imm8 },
				{ "Test_rm8_imm8_F6r1", Code.Test_rm8_imm8_F6r1 },
				{ "Not_rm8", Code.Not_rm8 },
				{ "Neg_rm8", Code.Neg_rm8 },
				{ "Mul_rm8", Code.Mul_rm8 },
				{ "Imul_rm8", Code.Imul_rm8 },
				{ "Div_rm8", Code.Div_rm8 },
				{ "Idiv_rm8", Code.Idiv_rm8 },
				{ "Test_rm16_imm16", Code.Test_rm16_imm16 },
				{ "Test_rm32_imm32", Code.Test_rm32_imm32 },
				{ "Test_rm64_imm32", Code.Test_rm64_imm32 },
				{ "Test_rm16_imm16_F7r1", Code.Test_rm16_imm16_F7r1 },
				{ "Test_rm32_imm32_F7r1", Code.Test_rm32_imm32_F7r1 },
				{ "Test_rm64_imm32_F7r1", Code.Test_rm64_imm32_F7r1 },
				{ "Not_rm16", Code.Not_rm16 },
				{ "Not_rm32", Code.Not_rm32 },
				{ "Not_rm64", Code.Not_rm64 },
				{ "Neg_rm16", Code.Neg_rm16 },
				{ "Neg_rm32", Code.Neg_rm32 },
				{ "Neg_rm64", Code.Neg_rm64 },
				{ "Mul_rm16", Code.Mul_rm16 },
				{ "Mul_rm32", Code.Mul_rm32 },
				{ "Mul_rm64", Code.Mul_rm64 },
				{ "Imul_rm16", Code.Imul_rm16 },
				{ "Imul_rm32", Code.Imul_rm32 },
				{ "Imul_rm64", Code.Imul_rm64 },
				{ "Div_rm16", Code.Div_rm16 },
				{ "Div_rm32", Code.Div_rm32 },
				{ "Div_rm64", Code.Div_rm64 },
				{ "Idiv_rm16", Code.Idiv_rm16 },
				{ "Idiv_rm32", Code.Idiv_rm32 },
				{ "Idiv_rm64", Code.Idiv_rm64 },
				{ "Clc", Code.Clc },
				{ "Stc", Code.Stc },
				{ "Cli", Code.Cli },
				{ "Sti", Code.Sti },
				{ "Cld", Code.Cld },
				{ "Std", Code.Std },
				{ "Inc_rm8", Code.Inc_rm8 },
				{ "Dec_rm8", Code.Dec_rm8 },
				{ "Inc_rm16", Code.Inc_rm16 },
				{ "Inc_rm32", Code.Inc_rm32 },
				{ "Inc_rm64", Code.Inc_rm64 },
				{ "Dec_rm16", Code.Dec_rm16 },
				{ "Dec_rm32", Code.Dec_rm32 },
				{ "Dec_rm64", Code.Dec_rm64 },
				{ "Call_rm16", Code.Call_rm16 },
				{ "Call_rm32", Code.Call_rm32 },
				{ "Call_rm64", Code.Call_rm64 },
				{ "Call_m1616", Code.Call_m1616 },
				{ "Call_m1632", Code.Call_m1632 },
				{ "Call_m1664", Code.Call_m1664 },
				{ "Jmp_rm16", Code.Jmp_rm16 },
				{ "Jmp_rm32", Code.Jmp_rm32 },
				{ "Jmp_rm64", Code.Jmp_rm64 },
				{ "Jmp_m1616", Code.Jmp_m1616 },
				{ "Jmp_m1632", Code.Jmp_m1632 },
				{ "Jmp_m1664", Code.Jmp_m1664 },
				{ "Push_rm16", Code.Push_rm16 },
				{ "Push_rm32", Code.Push_rm32 },
				{ "Push_rm64", Code.Push_rm64 },
				{ "Sldt_rm16", Code.Sldt_rm16 },
				{ "Sldt_r32m16", Code.Sldt_r32m16 },
				{ "Sldt_r64m16", Code.Sldt_r64m16 },
				{ "Str_rm16", Code.Str_rm16 },
				{ "Str_r32m16", Code.Str_r32m16 },
				{ "Str_r64m16", Code.Str_r64m16 },
				{ "Lldt_rm16", Code.Lldt_rm16 },
				{ "Lldt_r32m16", Code.Lldt_r32m16 },
				{ "Lldt_r64m16", Code.Lldt_r64m16 },
				{ "Ltr_rm16", Code.Ltr_rm16 },
				{ "Ltr_r32m16", Code.Ltr_r32m16 },
				{ "Ltr_r64m16", Code.Ltr_r64m16 },
				{ "Verr_rm16", Code.Verr_rm16 },
				{ "Verr_r32m16", Code.Verr_r32m16 },
				{ "Verr_r64m16", Code.Verr_r64m16 },
				{ "Verw_rm16", Code.Verw_rm16 },
				{ "Verw_r32m16", Code.Verw_r32m16 },
				{ "Verw_r64m16", Code.Verw_r64m16 },
				{ "Jmpe_rm16", Code.Jmpe_rm16 },
				{ "Jmpe_rm32", Code.Jmpe_rm32 },
				{ "Sgdt_m1632_16", Code.Sgdt_m1632_16 },
				{ "Sgdt_m1632", Code.Sgdt_m1632 },
				{ "Sgdt_m1664", Code.Sgdt_m1664 },
				{ "Sidt_m1632_16", Code.Sidt_m1632_16 },
				{ "Sidt_m1632", Code.Sidt_m1632 },
				{ "Sidt_m1664", Code.Sidt_m1664 },
				{ "Lgdt_m1632_16", Code.Lgdt_m1632_16 },
				{ "Lgdt_m1632", Code.Lgdt_m1632 },
				{ "Lgdt_m1664", Code.Lgdt_m1664 },
				{ "Lidt_m1632_16", Code.Lidt_m1632_16 },
				{ "Lidt_m1632", Code.Lidt_m1632 },
				{ "Lidt_m1664", Code.Lidt_m1664 },
				{ "Smsw_rm16", Code.Smsw_rm16 },
				{ "Smsw_r32m16", Code.Smsw_r32m16 },
				{ "Smsw_r64m16", Code.Smsw_r64m16 },
				{ "Rstorssp_m64", Code.Rstorssp_m64 },
				{ "Lmsw_rm16", Code.Lmsw_rm16 },
				{ "Lmsw_r32m16", Code.Lmsw_r32m16 },
				{ "Lmsw_r64m16", Code.Lmsw_r64m16 },
				{ "Invlpg_m", Code.Invlpg_m },
				{ "Enclv", Code.Enclv },
				{ "Vmcall", Code.Vmcall },
				{ "Vmlaunch", Code.Vmlaunch },
				{ "Vmresume", Code.Vmresume },
				{ "Vmxoff", Code.Vmxoff },
				{ "Pconfig", Code.Pconfig },
				{ "Monitorw", Code.Monitorw },
				{ "Monitord", Code.Monitord },
				{ "Monitorq", Code.Monitorq },
				{ "Mwait", Code.Mwait },
				{ "Clac", Code.Clac },
				{ "Stac", Code.Stac },
				{ "Encls", Code.Encls },
				{ "Xgetbv", Code.Xgetbv },
				{ "Xsetbv", Code.Xsetbv },
				{ "Vmfunc", Code.Vmfunc },
				{ "Xend", Code.Xend },
				{ "Xtest", Code.Xtest },
				{ "Enclu", Code.Enclu },
				{ "Vmrunw", Code.Vmrunw },
				{ "Vmrund", Code.Vmrund },
				{ "Vmrunq", Code.Vmrunq },
				{ "Vmmcall", Code.Vmmcall },
				{ "Vmloadw", Code.Vmloadw },
				{ "Vmloadd", Code.Vmloadd },
				{ "Vmloadq", Code.Vmloadq },
				{ "Vmsavew", Code.Vmsavew },
				{ "Vmsaved", Code.Vmsaved },
				{ "Vmsaveq", Code.Vmsaveq },
				{ "Stgi", Code.Stgi },
				{ "Clgi", Code.Clgi },
				{ "Skinit", Code.Skinit },
				{ "Invlpgaw", Code.Invlpgaw },
				{ "Invlpgad", Code.Invlpgad },
				{ "Invlpgaq", Code.Invlpgaq },
				{ "Setssbsy", Code.Setssbsy },
				{ "Saveprevssp", Code.Saveprevssp },
				{ "Rdpkru", Code.Rdpkru },
				{ "Wrpkru", Code.Wrpkru },
				{ "Swapgs", Code.Swapgs },
				{ "Rdtscp", Code.Rdtscp },
				{ "Monitorxw", Code.Monitorxw },
				{ "Monitorxd", Code.Monitorxd },
				{ "Monitorxq", Code.Monitorxq },
				{ "Mcommit", Code.Mcommit },
				{ "Mwaitx", Code.Mwaitx },
				{ "Clzerow", Code.Clzerow },
				{ "Clzerod", Code.Clzerod },
				{ "Clzeroq", Code.Clzeroq },
				{ "Rdpru", Code.Rdpru },
				{ "Lar_r16_rm16", Code.Lar_r16_rm16 },
				{ "Lar_r32_r32m16", Code.Lar_r32_r32m16 },
				{ "Lar_r64_r64m16", Code.Lar_r64_r64m16 },
				{ "Lsl_r16_rm16", Code.Lsl_r16_rm16 },
				{ "Lsl_r32_r32m16", Code.Lsl_r32_r32m16 },
				{ "Lsl_r64_r64m16", Code.Lsl_r64_r64m16 },
				{ "Loadallreset286", Code.Loadallreset286 },
				{ "Loadall286", Code.Loadall286 },
				{ "Syscall", Code.Syscall },
				{ "Clts", Code.Clts },
				{ "Loadall386", Code.Loadall386 },
				{ "Sysretd", Code.Sysretd },
				{ "Sysretq", Code.Sysretq },
				{ "Invd", Code.Invd },
				{ "Wbinvd", Code.Wbinvd },
				{ "Wbnoinvd", Code.Wbnoinvd },
				{ "Cl1invmb", Code.Cl1invmb },
				{ "Ud2", Code.Ud2 },
				{ "Reservednop_rm16_r16_0F0D", Code.Reservednop_rm16_r16_0F0D },
				{ "Reservednop_rm32_r32_0F0D", Code.Reservednop_rm32_r32_0F0D },
				{ "Reservednop_rm64_r64_0F0D", Code.Reservednop_rm64_r64_0F0D },
				{ "Prefetch_m8", Code.Prefetch_m8 },
				{ "Prefetchw_m8", Code.Prefetchw_m8 },
				{ "Prefetchwt1_m8", Code.Prefetchwt1_m8 },
				{ "Femms", Code.Femms },
				{ "Umov_rm8_r8", Code.Umov_rm8_r8 },
				{ "Umov_rm16_r16", Code.Umov_rm16_r16 },
				{ "Umov_rm32_r32", Code.Umov_rm32_r32 },
				{ "Umov_r8_rm8", Code.Umov_r8_rm8 },
				{ "Umov_r16_rm16", Code.Umov_r16_rm16 },
				{ "Umov_r32_rm32", Code.Umov_r32_rm32 },
				{ "Movups_xmm_xmmm128", Code.Movups_xmm_xmmm128 },
				{ "VEX_Vmovups_xmm_xmmm128", Code.VEX_Vmovups_xmm_xmmm128 },
				{ "VEX_Vmovups_ymm_ymmm256", Code.VEX_Vmovups_ymm_ymmm256 },
				{ "EVEX_Vmovups_xmm_k1z_xmmm128", Code.EVEX_Vmovups_xmm_k1z_xmmm128 },
				{ "EVEX_Vmovups_ymm_k1z_ymmm256", Code.EVEX_Vmovups_ymm_k1z_ymmm256 },
				{ "EVEX_Vmovups_zmm_k1z_zmmm512", Code.EVEX_Vmovups_zmm_k1z_zmmm512 },
				{ "Movupd_xmm_xmmm128", Code.Movupd_xmm_xmmm128 },
				{ "VEX_Vmovupd_xmm_xmmm128", Code.VEX_Vmovupd_xmm_xmmm128 },
				{ "VEX_Vmovupd_ymm_ymmm256", Code.VEX_Vmovupd_ymm_ymmm256 },
				{ "EVEX_Vmovupd_xmm_k1z_xmmm128", Code.EVEX_Vmovupd_xmm_k1z_xmmm128 },
				{ "EVEX_Vmovupd_ymm_k1z_ymmm256", Code.EVEX_Vmovupd_ymm_k1z_ymmm256 },
				{ "EVEX_Vmovupd_zmm_k1z_zmmm512", Code.EVEX_Vmovupd_zmm_k1z_zmmm512 },
				{ "Movss_xmm_xmmm32", Code.Movss_xmm_xmmm32 },
				{ "VEX_Vmovss_xmm_xmm_xmm", Code.VEX_Vmovss_xmm_xmm_xmm },
				{ "VEX_Vmovss_xmm_m32", Code.VEX_Vmovss_xmm_m32 },
				{ "EVEX_Vmovss_xmm_k1z_xmm_xmm", Code.EVEX_Vmovss_xmm_k1z_xmm_xmm },
				{ "EVEX_Vmovss_xmm_k1z_m32", Code.EVEX_Vmovss_xmm_k1z_m32 },
				{ "Movsd_xmm_xmmm64", Code.Movsd_xmm_xmmm64 },
				{ "VEX_Vmovsd_xmm_xmm_xmm", Code.VEX_Vmovsd_xmm_xmm_xmm },
				{ "VEX_Vmovsd_xmm_m64", Code.VEX_Vmovsd_xmm_m64 },
				{ "EVEX_Vmovsd_xmm_k1z_xmm_xmm", Code.EVEX_Vmovsd_xmm_k1z_xmm_xmm },
				{ "EVEX_Vmovsd_xmm_k1z_m64", Code.EVEX_Vmovsd_xmm_k1z_m64 },
				{ "Movups_xmmm128_xmm", Code.Movups_xmmm128_xmm },
				{ "VEX_Vmovups_xmmm128_xmm", Code.VEX_Vmovups_xmmm128_xmm },
				{ "VEX_Vmovups_ymmm256_ymm", Code.VEX_Vmovups_ymmm256_ymm },
				{ "EVEX_Vmovups_xmmm128_k1z_xmm", Code.EVEX_Vmovups_xmmm128_k1z_xmm },
				{ "EVEX_Vmovups_ymmm256_k1z_ymm", Code.EVEX_Vmovups_ymmm256_k1z_ymm },
				{ "EVEX_Vmovups_zmmm512_k1z_zmm", Code.EVEX_Vmovups_zmmm512_k1z_zmm },
				{ "Movupd_xmmm128_xmm", Code.Movupd_xmmm128_xmm },
				{ "VEX_Vmovupd_xmmm128_xmm", Code.VEX_Vmovupd_xmmm128_xmm },
				{ "VEX_Vmovupd_ymmm256_ymm", Code.VEX_Vmovupd_ymmm256_ymm },
				{ "EVEX_Vmovupd_xmmm128_k1z_xmm", Code.EVEX_Vmovupd_xmmm128_k1z_xmm },
				{ "EVEX_Vmovupd_ymmm256_k1z_ymm", Code.EVEX_Vmovupd_ymmm256_k1z_ymm },
				{ "EVEX_Vmovupd_zmmm512_k1z_zmm", Code.EVEX_Vmovupd_zmmm512_k1z_zmm },
				{ "Movss_xmmm32_xmm", Code.Movss_xmmm32_xmm },
				{ "VEX_Vmovss_xmm_xmm_xmm_0F11", Code.VEX_Vmovss_xmm_xmm_xmm_0F11 },
				{ "VEX_Vmovss_m32_xmm", Code.VEX_Vmovss_m32_xmm },
				{ "EVEX_Vmovss_xmm_k1z_xmm_xmm_0F11", Code.EVEX_Vmovss_xmm_k1z_xmm_xmm_0F11 },
				{ "EVEX_Vmovss_m32_k1_xmm", Code.EVEX_Vmovss_m32_k1_xmm },
				{ "Movsd_xmmm64_xmm", Code.Movsd_xmmm64_xmm },
				{ "VEX_Vmovsd_xmm_xmm_xmm_0F11", Code.VEX_Vmovsd_xmm_xmm_xmm_0F11 },
				{ "VEX_Vmovsd_m64_xmm", Code.VEX_Vmovsd_m64_xmm },
				{ "EVEX_Vmovsd_xmm_k1z_xmm_xmm_0F11", Code.EVEX_Vmovsd_xmm_k1z_xmm_xmm_0F11 },
				{ "EVEX_Vmovsd_m64_k1_xmm", Code.EVEX_Vmovsd_m64_k1_xmm },
				{ "Movhlps_xmm_xmm", Code.Movhlps_xmm_xmm },
				{ "Movlps_xmm_m64", Code.Movlps_xmm_m64 },
				{ "VEX_Vmovhlps_xmm_xmm_xmm", Code.VEX_Vmovhlps_xmm_xmm_xmm },
				{ "VEX_Vmovlps_xmm_xmm_m64", Code.VEX_Vmovlps_xmm_xmm_m64 },
				{ "EVEX_Vmovhlps_xmm_xmm_xmm", Code.EVEX_Vmovhlps_xmm_xmm_xmm },
				{ "EVEX_Vmovlps_xmm_xmm_m64", Code.EVEX_Vmovlps_xmm_xmm_m64 },
				{ "Movlpd_xmm_m64", Code.Movlpd_xmm_m64 },
				{ "VEX_Vmovlpd_xmm_xmm_m64", Code.VEX_Vmovlpd_xmm_xmm_m64 },
				{ "EVEX_Vmovlpd_xmm_xmm_m64", Code.EVEX_Vmovlpd_xmm_xmm_m64 },
				{ "Movsldup_xmm_xmmm128", Code.Movsldup_xmm_xmmm128 },
				{ "VEX_Vmovsldup_xmm_xmmm128", Code.VEX_Vmovsldup_xmm_xmmm128 },
				{ "VEX_Vmovsldup_ymm_ymmm256", Code.VEX_Vmovsldup_ymm_ymmm256 },
				{ "EVEX_Vmovsldup_xmm_k1z_xmmm128", Code.EVEX_Vmovsldup_xmm_k1z_xmmm128 },
				{ "EVEX_Vmovsldup_ymm_k1z_ymmm256", Code.EVEX_Vmovsldup_ymm_k1z_ymmm256 },
				{ "EVEX_Vmovsldup_zmm_k1z_zmmm512", Code.EVEX_Vmovsldup_zmm_k1z_zmmm512 },
				{ "Movddup_xmm_xmmm64", Code.Movddup_xmm_xmmm64 },
				{ "VEX_Vmovddup_xmm_xmmm64", Code.VEX_Vmovddup_xmm_xmmm64 },
				{ "VEX_Vmovddup_ymm_ymmm256", Code.VEX_Vmovddup_ymm_ymmm256 },
				{ "EVEX_Vmovddup_xmm_k1z_xmmm64", Code.EVEX_Vmovddup_xmm_k1z_xmmm64 },
				{ "EVEX_Vmovddup_ymm_k1z_ymmm256", Code.EVEX_Vmovddup_ymm_k1z_ymmm256 },
				{ "EVEX_Vmovddup_zmm_k1z_zmmm512", Code.EVEX_Vmovddup_zmm_k1z_zmmm512 },
				{ "Movlps_m64_xmm", Code.Movlps_m64_xmm },
				{ "VEX_Vmovlps_m64_xmm", Code.VEX_Vmovlps_m64_xmm },
				{ "EVEX_Vmovlps_m64_xmm", Code.EVEX_Vmovlps_m64_xmm },
				{ "Movlpd_m64_xmm", Code.Movlpd_m64_xmm },
				{ "VEX_Vmovlpd_m64_xmm", Code.VEX_Vmovlpd_m64_xmm },
				{ "EVEX_Vmovlpd_m64_xmm", Code.EVEX_Vmovlpd_m64_xmm },
				{ "Unpcklps_xmm_xmmm128", Code.Unpcklps_xmm_xmmm128 },
				{ "VEX_Vunpcklps_xmm_xmm_xmmm128", Code.VEX_Vunpcklps_xmm_xmm_xmmm128 },
				{ "VEX_Vunpcklps_ymm_ymm_ymmm256", Code.VEX_Vunpcklps_ymm_ymm_ymmm256 },
				{ "EVEX_Vunpcklps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vunpcklps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vunpcklps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vunpcklps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vunpcklps_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vunpcklps_zmm_k1z_zmm_zmmm512b32 },
				{ "Unpcklpd_xmm_xmmm128", Code.Unpcklpd_xmm_xmmm128 },
				{ "VEX_Vunpcklpd_xmm_xmm_xmmm128", Code.VEX_Vunpcklpd_xmm_xmm_xmmm128 },
				{ "VEX_Vunpcklpd_ymm_ymm_ymmm256", Code.VEX_Vunpcklpd_ymm_ymm_ymmm256 },
				{ "EVEX_Vunpcklpd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vunpcklpd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vunpcklpd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vunpcklpd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vunpcklpd_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vunpcklpd_zmm_k1z_zmm_zmmm512b64 },
				{ "Unpckhps_xmm_xmmm128", Code.Unpckhps_xmm_xmmm128 },
				{ "VEX_Vunpckhps_xmm_xmm_xmmm128", Code.VEX_Vunpckhps_xmm_xmm_xmmm128 },
				{ "VEX_Vunpckhps_ymm_ymm_ymmm256", Code.VEX_Vunpckhps_ymm_ymm_ymmm256 },
				{ "EVEX_Vunpckhps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vunpckhps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vunpckhps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vunpckhps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vunpckhps_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vunpckhps_zmm_k1z_zmm_zmmm512b32 },
				{ "Unpckhpd_xmm_xmmm128", Code.Unpckhpd_xmm_xmmm128 },
				{ "VEX_Vunpckhpd_xmm_xmm_xmmm128", Code.VEX_Vunpckhpd_xmm_xmm_xmmm128 },
				{ "VEX_Vunpckhpd_ymm_ymm_ymmm256", Code.VEX_Vunpckhpd_ymm_ymm_ymmm256 },
				{ "EVEX_Vunpckhpd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vunpckhpd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vunpckhpd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vunpckhpd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vunpckhpd_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vunpckhpd_zmm_k1z_zmm_zmmm512b64 },
				{ "Movlhps_xmm_xmm", Code.Movlhps_xmm_xmm },
				{ "VEX_Vmovlhps_xmm_xmm_xmm", Code.VEX_Vmovlhps_xmm_xmm_xmm },
				{ "EVEX_Vmovlhps_xmm_xmm_xmm", Code.EVEX_Vmovlhps_xmm_xmm_xmm },
				{ "Movhps_xmm_m64", Code.Movhps_xmm_m64 },
				{ "VEX_Vmovhps_xmm_xmm_m64", Code.VEX_Vmovhps_xmm_xmm_m64 },
				{ "EVEX_Vmovhps_xmm_xmm_m64", Code.EVEX_Vmovhps_xmm_xmm_m64 },
				{ "Movhpd_xmm_m64", Code.Movhpd_xmm_m64 },
				{ "VEX_Vmovhpd_xmm_xmm_m64", Code.VEX_Vmovhpd_xmm_xmm_m64 },
				{ "EVEX_Vmovhpd_xmm_xmm_m64", Code.EVEX_Vmovhpd_xmm_xmm_m64 },
				{ "Movshdup_xmm_xmmm128", Code.Movshdup_xmm_xmmm128 },
				{ "VEX_Vmovshdup_xmm_xmmm128", Code.VEX_Vmovshdup_xmm_xmmm128 },
				{ "VEX_Vmovshdup_ymm_ymmm256", Code.VEX_Vmovshdup_ymm_ymmm256 },
				{ "EVEX_Vmovshdup_xmm_k1z_xmmm128", Code.EVEX_Vmovshdup_xmm_k1z_xmmm128 },
				{ "EVEX_Vmovshdup_ymm_k1z_ymmm256", Code.EVEX_Vmovshdup_ymm_k1z_ymmm256 },
				{ "EVEX_Vmovshdup_zmm_k1z_zmmm512", Code.EVEX_Vmovshdup_zmm_k1z_zmmm512 },
				{ "Movhps_m64_xmm", Code.Movhps_m64_xmm },
				{ "VEX_Vmovhps_m64_xmm", Code.VEX_Vmovhps_m64_xmm },
				{ "EVEX_Vmovhps_m64_xmm", Code.EVEX_Vmovhps_m64_xmm },
				{ "Movhpd_m64_xmm", Code.Movhpd_m64_xmm },
				{ "VEX_Vmovhpd_m64_xmm", Code.VEX_Vmovhpd_m64_xmm },
				{ "EVEX_Vmovhpd_m64_xmm", Code.EVEX_Vmovhpd_m64_xmm },
				{ "Reservednop_rm16_r16_0F18", Code.Reservednop_rm16_r16_0F18 },
				{ "Reservednop_rm32_r32_0F18", Code.Reservednop_rm32_r32_0F18 },
				{ "Reservednop_rm64_r64_0F18", Code.Reservednop_rm64_r64_0F18 },
				{ "Reservednop_rm16_r16_0F19", Code.Reservednop_rm16_r16_0F19 },
				{ "Reservednop_rm32_r32_0F19", Code.Reservednop_rm32_r32_0F19 },
				{ "Reservednop_rm64_r64_0F19", Code.Reservednop_rm64_r64_0F19 },
				{ "Reservednop_rm16_r16_0F1A", Code.Reservednop_rm16_r16_0F1A },
				{ "Reservednop_rm32_r32_0F1A", Code.Reservednop_rm32_r32_0F1A },
				{ "Reservednop_rm64_r64_0F1A", Code.Reservednop_rm64_r64_0F1A },
				{ "Reservednop_rm16_r16_0F1B", Code.Reservednop_rm16_r16_0F1B },
				{ "Reservednop_rm32_r32_0F1B", Code.Reservednop_rm32_r32_0F1B },
				{ "Reservednop_rm64_r64_0F1B", Code.Reservednop_rm64_r64_0F1B },
				{ "Reservednop_rm16_r16_0F1C", Code.Reservednop_rm16_r16_0F1C },
				{ "Reservednop_rm32_r32_0F1C", Code.Reservednop_rm32_r32_0F1C },
				{ "Reservednop_rm64_r64_0F1C", Code.Reservednop_rm64_r64_0F1C },
				{ "Reservednop_rm16_r16_0F1D", Code.Reservednop_rm16_r16_0F1D },
				{ "Reservednop_rm32_r32_0F1D", Code.Reservednop_rm32_r32_0F1D },
				{ "Reservednop_rm64_r64_0F1D", Code.Reservednop_rm64_r64_0F1D },
				{ "Reservednop_rm16_r16_0F1E", Code.Reservednop_rm16_r16_0F1E },
				{ "Reservednop_rm32_r32_0F1E", Code.Reservednop_rm32_r32_0F1E },
				{ "Reservednop_rm64_r64_0F1E", Code.Reservednop_rm64_r64_0F1E },
				{ "Reservednop_rm16_r16_0F1F", Code.Reservednop_rm16_r16_0F1F },
				{ "Reservednop_rm32_r32_0F1F", Code.Reservednop_rm32_r32_0F1F },
				{ "Reservednop_rm64_r64_0F1F", Code.Reservednop_rm64_r64_0F1F },
				{ "Prefetchnta_m8", Code.Prefetchnta_m8 },
				{ "Prefetcht0_m8", Code.Prefetcht0_m8 },
				{ "Prefetcht1_m8", Code.Prefetcht1_m8 },
				{ "Prefetcht2_m8", Code.Prefetcht2_m8 },
				{ "Bndldx_bnd_mib", Code.Bndldx_bnd_mib },
				{ "Bndmov_bnd_bndm64", Code.Bndmov_bnd_bndm64 },
				{ "Bndmov_bnd_bndm128", Code.Bndmov_bnd_bndm128 },
				{ "Bndcl_bnd_rm32", Code.Bndcl_bnd_rm32 },
				{ "Bndcl_bnd_rm64", Code.Bndcl_bnd_rm64 },
				{ "Bndcu_bnd_rm32", Code.Bndcu_bnd_rm32 },
				{ "Bndcu_bnd_rm64", Code.Bndcu_bnd_rm64 },
				{ "Bndstx_mib_bnd", Code.Bndstx_mib_bnd },
				{ "Bndmov_bndm64_bnd", Code.Bndmov_bndm64_bnd },
				{ "Bndmov_bndm128_bnd", Code.Bndmov_bndm128_bnd },
				{ "Bndmk_bnd_m32", Code.Bndmk_bnd_m32 },
				{ "Bndmk_bnd_m64", Code.Bndmk_bnd_m64 },
				{ "Bndcn_bnd_rm32", Code.Bndcn_bnd_rm32 },
				{ "Bndcn_bnd_rm64", Code.Bndcn_bnd_rm64 },
				{ "Cldemote_m8", Code.Cldemote_m8 },
				{ "Rdsspd_r32", Code.Rdsspd_r32 },
				{ "Rdsspq_r64", Code.Rdsspq_r64 },
				{ "Endbr64", Code.Endbr64 },
				{ "Endbr32", Code.Endbr32 },
				{ "Nop_rm16", Code.Nop_rm16 },
				{ "Nop_rm32", Code.Nop_rm32 },
				{ "Nop_rm64", Code.Nop_rm64 },
				{ "Mov_r32_cr", Code.Mov_r32_cr },
				{ "Mov_r64_cr", Code.Mov_r64_cr },
				{ "Mov_r32_dr", Code.Mov_r32_dr },
				{ "Mov_r64_dr", Code.Mov_r64_dr },
				{ "Mov_cr_r32", Code.Mov_cr_r32 },
				{ "Mov_cr_r64", Code.Mov_cr_r64 },
				{ "Mov_dr_r32", Code.Mov_dr_r32 },
				{ "Mov_dr_r64", Code.Mov_dr_r64 },
				{ "Mov_r32_tr", Code.Mov_r32_tr },
				{ "Mov_tr_r32", Code.Mov_tr_r32 },
				{ "Movaps_xmm_xmmm128", Code.Movaps_xmm_xmmm128 },
				{ "VEX_Vmovaps_xmm_xmmm128", Code.VEX_Vmovaps_xmm_xmmm128 },
				{ "VEX_Vmovaps_ymm_ymmm256", Code.VEX_Vmovaps_ymm_ymmm256 },
				{ "EVEX_Vmovaps_xmm_k1z_xmmm128", Code.EVEX_Vmovaps_xmm_k1z_xmmm128 },
				{ "EVEX_Vmovaps_ymm_k1z_ymmm256", Code.EVEX_Vmovaps_ymm_k1z_ymmm256 },
				{ "EVEX_Vmovaps_zmm_k1z_zmmm512", Code.EVEX_Vmovaps_zmm_k1z_zmmm512 },
				{ "Movapd_xmm_xmmm128", Code.Movapd_xmm_xmmm128 },
				{ "VEX_Vmovapd_xmm_xmmm128", Code.VEX_Vmovapd_xmm_xmmm128 },
				{ "VEX_Vmovapd_ymm_ymmm256", Code.VEX_Vmovapd_ymm_ymmm256 },
				{ "EVEX_Vmovapd_xmm_k1z_xmmm128", Code.EVEX_Vmovapd_xmm_k1z_xmmm128 },
				{ "EVEX_Vmovapd_ymm_k1z_ymmm256", Code.EVEX_Vmovapd_ymm_k1z_ymmm256 },
				{ "EVEX_Vmovapd_zmm_k1z_zmmm512", Code.EVEX_Vmovapd_zmm_k1z_zmmm512 },
				{ "Movaps_xmmm128_xmm", Code.Movaps_xmmm128_xmm },
				{ "VEX_Vmovaps_xmmm128_xmm", Code.VEX_Vmovaps_xmmm128_xmm },
				{ "VEX_Vmovaps_ymmm256_ymm", Code.VEX_Vmovaps_ymmm256_ymm },
				{ "EVEX_Vmovaps_xmmm128_k1z_xmm", Code.EVEX_Vmovaps_xmmm128_k1z_xmm },
				{ "EVEX_Vmovaps_ymmm256_k1z_ymm", Code.EVEX_Vmovaps_ymmm256_k1z_ymm },
				{ "EVEX_Vmovaps_zmmm512_k1z_zmm", Code.EVEX_Vmovaps_zmmm512_k1z_zmm },
				{ "Movapd_xmmm128_xmm", Code.Movapd_xmmm128_xmm },
				{ "VEX_Vmovapd_xmmm128_xmm", Code.VEX_Vmovapd_xmmm128_xmm },
				{ "VEX_Vmovapd_ymmm256_ymm", Code.VEX_Vmovapd_ymmm256_ymm },
				{ "EVEX_Vmovapd_xmmm128_k1z_xmm", Code.EVEX_Vmovapd_xmmm128_k1z_xmm },
				{ "EVEX_Vmovapd_ymmm256_k1z_ymm", Code.EVEX_Vmovapd_ymmm256_k1z_ymm },
				{ "EVEX_Vmovapd_zmmm512_k1z_zmm", Code.EVEX_Vmovapd_zmmm512_k1z_zmm },
				{ "Cvtpi2ps_xmm_mmm64", Code.Cvtpi2ps_xmm_mmm64 },
				{ "Cvtpi2pd_xmm_mmm64", Code.Cvtpi2pd_xmm_mmm64 },
				{ "Cvtsi2ss_xmm_rm32", Code.Cvtsi2ss_xmm_rm32 },
				{ "Cvtsi2ss_xmm_rm64", Code.Cvtsi2ss_xmm_rm64 },
				{ "VEX_Vcvtsi2ss_xmm_xmm_rm32", Code.VEX_Vcvtsi2ss_xmm_xmm_rm32 },
				{ "VEX_Vcvtsi2ss_xmm_xmm_rm64", Code.VEX_Vcvtsi2ss_xmm_xmm_rm64 },
				{ "EVEX_Vcvtsi2ss_xmm_xmm_rm32_er", Code.EVEX_Vcvtsi2ss_xmm_xmm_rm32_er },
				{ "EVEX_Vcvtsi2ss_xmm_xmm_rm64_er", Code.EVEX_Vcvtsi2ss_xmm_xmm_rm64_er },
				{ "Cvtsi2sd_xmm_rm32", Code.Cvtsi2sd_xmm_rm32 },
				{ "Cvtsi2sd_xmm_rm64", Code.Cvtsi2sd_xmm_rm64 },
				{ "VEX_Vcvtsi2sd_xmm_xmm_rm32", Code.VEX_Vcvtsi2sd_xmm_xmm_rm32 },
				{ "VEX_Vcvtsi2sd_xmm_xmm_rm64", Code.VEX_Vcvtsi2sd_xmm_xmm_rm64 },
				{ "EVEX_Vcvtsi2sd_xmm_xmm_rm32_er", Code.EVEX_Vcvtsi2sd_xmm_xmm_rm32_er },
				{ "EVEX_Vcvtsi2sd_xmm_xmm_rm64_er", Code.EVEX_Vcvtsi2sd_xmm_xmm_rm64_er },
				{ "Movntps_m128_xmm", Code.Movntps_m128_xmm },
				{ "VEX_Vmovntps_m128_xmm", Code.VEX_Vmovntps_m128_xmm },
				{ "VEX_Vmovntps_m256_ymm", Code.VEX_Vmovntps_m256_ymm },
				{ "EVEX_Vmovntps_m128_xmm", Code.EVEX_Vmovntps_m128_xmm },
				{ "EVEX_Vmovntps_m256_ymm", Code.EVEX_Vmovntps_m256_ymm },
				{ "EVEX_Vmovntps_m512_zmm", Code.EVEX_Vmovntps_m512_zmm },
				{ "Movntpd_m128_xmm", Code.Movntpd_m128_xmm },
				{ "VEX_Vmovntpd_m128_xmm", Code.VEX_Vmovntpd_m128_xmm },
				{ "VEX_Vmovntpd_m256_ymm", Code.VEX_Vmovntpd_m256_ymm },
				{ "EVEX_Vmovntpd_m128_xmm", Code.EVEX_Vmovntpd_m128_xmm },
				{ "EVEX_Vmovntpd_m256_ymm", Code.EVEX_Vmovntpd_m256_ymm },
				{ "EVEX_Vmovntpd_m512_zmm", Code.EVEX_Vmovntpd_m512_zmm },
				{ "Movntss_m32_xmm", Code.Movntss_m32_xmm },
				{ "Movntsd_m64_xmm", Code.Movntsd_m64_xmm },
				{ "Cvttps2pi_mm_xmmm64", Code.Cvttps2pi_mm_xmmm64 },
				{ "Cvttpd2pi_mm_xmmm128", Code.Cvttpd2pi_mm_xmmm128 },
				{ "Cvttss2si_r32_xmmm32", Code.Cvttss2si_r32_xmmm32 },
				{ "Cvttss2si_r64_xmmm32", Code.Cvttss2si_r64_xmmm32 },
				{ "VEX_Vcvttss2si_r32_xmmm32", Code.VEX_Vcvttss2si_r32_xmmm32 },
				{ "VEX_Vcvttss2si_r64_xmmm32", Code.VEX_Vcvttss2si_r64_xmmm32 },
				{ "EVEX_Vcvttss2si_r32_xmmm32_sae", Code.EVEX_Vcvttss2si_r32_xmmm32_sae },
				{ "EVEX_Vcvttss2si_r64_xmmm32_sae", Code.EVEX_Vcvttss2si_r64_xmmm32_sae },
				{ "Cvttsd2si_r32_xmmm64", Code.Cvttsd2si_r32_xmmm64 },
				{ "Cvttsd2si_r64_xmmm64", Code.Cvttsd2si_r64_xmmm64 },
				{ "VEX_Vcvttsd2si_r32_xmmm64", Code.VEX_Vcvttsd2si_r32_xmmm64 },
				{ "VEX_Vcvttsd2si_r64_xmmm64", Code.VEX_Vcvttsd2si_r64_xmmm64 },
				{ "EVEX_Vcvttsd2si_r32_xmmm64_sae", Code.EVEX_Vcvttsd2si_r32_xmmm64_sae },
				{ "EVEX_Vcvttsd2si_r64_xmmm64_sae", Code.EVEX_Vcvttsd2si_r64_xmmm64_sae },
				{ "Cvtps2pi_mm_xmmm64", Code.Cvtps2pi_mm_xmmm64 },
				{ "Cvtpd2pi_mm_xmmm128", Code.Cvtpd2pi_mm_xmmm128 },
				{ "Cvtss2si_r32_xmmm32", Code.Cvtss2si_r32_xmmm32 },
				{ "Cvtss2si_r64_xmmm32", Code.Cvtss2si_r64_xmmm32 },
				{ "VEX_Vcvtss2si_r32_xmmm32", Code.VEX_Vcvtss2si_r32_xmmm32 },
				{ "VEX_Vcvtss2si_r64_xmmm32", Code.VEX_Vcvtss2si_r64_xmmm32 },
				{ "EVEX_Vcvtss2si_r32_xmmm32_er", Code.EVEX_Vcvtss2si_r32_xmmm32_er },
				{ "EVEX_Vcvtss2si_r64_xmmm32_er", Code.EVEX_Vcvtss2si_r64_xmmm32_er },
				{ "Cvtsd2si_r32_xmmm64", Code.Cvtsd2si_r32_xmmm64 },
				{ "Cvtsd2si_r64_xmmm64", Code.Cvtsd2si_r64_xmmm64 },
				{ "VEX_Vcvtsd2si_r32_xmmm64", Code.VEX_Vcvtsd2si_r32_xmmm64 },
				{ "VEX_Vcvtsd2si_r64_xmmm64", Code.VEX_Vcvtsd2si_r64_xmmm64 },
				{ "EVEX_Vcvtsd2si_r32_xmmm64_er", Code.EVEX_Vcvtsd2si_r32_xmmm64_er },
				{ "EVEX_Vcvtsd2si_r64_xmmm64_er", Code.EVEX_Vcvtsd2si_r64_xmmm64_er },
				{ "Ucomiss_xmm_xmmm32", Code.Ucomiss_xmm_xmmm32 },
				{ "VEX_Vucomiss_xmm_xmmm32", Code.VEX_Vucomiss_xmm_xmmm32 },
				{ "EVEX_Vucomiss_xmm_xmmm32_sae", Code.EVEX_Vucomiss_xmm_xmmm32_sae },
				{ "Ucomisd_xmm_xmmm64", Code.Ucomisd_xmm_xmmm64 },
				{ "VEX_Vucomisd_xmm_xmmm64", Code.VEX_Vucomisd_xmm_xmmm64 },
				{ "EVEX_Vucomisd_xmm_xmmm64_sae", Code.EVEX_Vucomisd_xmm_xmmm64_sae },
				{ "Comiss_xmm_xmmm32", Code.Comiss_xmm_xmmm32 },
				{ "Comisd_xmm_xmmm64", Code.Comisd_xmm_xmmm64 },
				{ "VEX_Vcomiss_xmm_xmmm32", Code.VEX_Vcomiss_xmm_xmmm32 },
				{ "VEX_Vcomisd_xmm_xmmm64", Code.VEX_Vcomisd_xmm_xmmm64 },
				{ "EVEX_Vcomiss_xmm_xmmm32_sae", Code.EVEX_Vcomiss_xmm_xmmm32_sae },
				{ "EVEX_Vcomisd_xmm_xmmm64_sae", Code.EVEX_Vcomisd_xmm_xmmm64_sae },
				{ "Wrmsr", Code.Wrmsr },
				{ "Rdtsc", Code.Rdtsc },
				{ "Rdmsr", Code.Rdmsr },
				{ "Rdpmc", Code.Rdpmc },
				{ "Sysenter", Code.Sysenter },
				{ "Sysexitd", Code.Sysexitd },
				{ "Sysexitq", Code.Sysexitq },
				{ "Getsecd", Code.Getsecd },
				{ "Cmovo_r16_rm16", Code.Cmovo_r16_rm16 },
				{ "Cmovo_r32_rm32", Code.Cmovo_r32_rm32 },
				{ "Cmovo_r64_rm64", Code.Cmovo_r64_rm64 },
				{ "Cmovno_r16_rm16", Code.Cmovno_r16_rm16 },
				{ "Cmovno_r32_rm32", Code.Cmovno_r32_rm32 },
				{ "Cmovno_r64_rm64", Code.Cmovno_r64_rm64 },
				{ "Cmovb_r16_rm16", Code.Cmovb_r16_rm16 },
				{ "Cmovb_r32_rm32", Code.Cmovb_r32_rm32 },
				{ "Cmovb_r64_rm64", Code.Cmovb_r64_rm64 },
				{ "Cmovae_r16_rm16", Code.Cmovae_r16_rm16 },
				{ "Cmovae_r32_rm32", Code.Cmovae_r32_rm32 },
				{ "Cmovae_r64_rm64", Code.Cmovae_r64_rm64 },
				{ "Cmove_r16_rm16", Code.Cmove_r16_rm16 },
				{ "Cmove_r32_rm32", Code.Cmove_r32_rm32 },
				{ "Cmove_r64_rm64", Code.Cmove_r64_rm64 },
				{ "Cmovne_r16_rm16", Code.Cmovne_r16_rm16 },
				{ "Cmovne_r32_rm32", Code.Cmovne_r32_rm32 },
				{ "Cmovne_r64_rm64", Code.Cmovne_r64_rm64 },
				{ "Cmovbe_r16_rm16", Code.Cmovbe_r16_rm16 },
				{ "Cmovbe_r32_rm32", Code.Cmovbe_r32_rm32 },
				{ "Cmovbe_r64_rm64", Code.Cmovbe_r64_rm64 },
				{ "Cmova_r16_rm16", Code.Cmova_r16_rm16 },
				{ "Cmova_r32_rm32", Code.Cmova_r32_rm32 },
				{ "Cmova_r64_rm64", Code.Cmova_r64_rm64 },
				{ "Cmovs_r16_rm16", Code.Cmovs_r16_rm16 },
				{ "Cmovs_r32_rm32", Code.Cmovs_r32_rm32 },
				{ "Cmovs_r64_rm64", Code.Cmovs_r64_rm64 },
				{ "Cmovns_r16_rm16", Code.Cmovns_r16_rm16 },
				{ "Cmovns_r32_rm32", Code.Cmovns_r32_rm32 },
				{ "Cmovns_r64_rm64", Code.Cmovns_r64_rm64 },
				{ "Cmovp_r16_rm16", Code.Cmovp_r16_rm16 },
				{ "Cmovp_r32_rm32", Code.Cmovp_r32_rm32 },
				{ "Cmovp_r64_rm64", Code.Cmovp_r64_rm64 },
				{ "Cmovnp_r16_rm16", Code.Cmovnp_r16_rm16 },
				{ "Cmovnp_r32_rm32", Code.Cmovnp_r32_rm32 },
				{ "Cmovnp_r64_rm64", Code.Cmovnp_r64_rm64 },
				{ "Cmovl_r16_rm16", Code.Cmovl_r16_rm16 },
				{ "Cmovl_r32_rm32", Code.Cmovl_r32_rm32 },
				{ "Cmovl_r64_rm64", Code.Cmovl_r64_rm64 },
				{ "Cmovge_r16_rm16", Code.Cmovge_r16_rm16 },
				{ "Cmovge_r32_rm32", Code.Cmovge_r32_rm32 },
				{ "Cmovge_r64_rm64", Code.Cmovge_r64_rm64 },
				{ "Cmovle_r16_rm16", Code.Cmovle_r16_rm16 },
				{ "Cmovle_r32_rm32", Code.Cmovle_r32_rm32 },
				{ "Cmovle_r64_rm64", Code.Cmovle_r64_rm64 },
				{ "Cmovg_r16_rm16", Code.Cmovg_r16_rm16 },
				{ "Cmovg_r32_rm32", Code.Cmovg_r32_rm32 },
				{ "Cmovg_r64_rm64", Code.Cmovg_r64_rm64 },
				{ "VEX_Kandw_kr_kr_kr", Code.VEX_Kandw_kr_kr_kr },
				{ "VEX_Kandq_kr_kr_kr", Code.VEX_Kandq_kr_kr_kr },
				{ "VEX_Kandb_kr_kr_kr", Code.VEX_Kandb_kr_kr_kr },
				{ "VEX_Kandd_kr_kr_kr", Code.VEX_Kandd_kr_kr_kr },
				{ "VEX_Kandnw_kr_kr_kr", Code.VEX_Kandnw_kr_kr_kr },
				{ "VEX_Kandnq_kr_kr_kr", Code.VEX_Kandnq_kr_kr_kr },
				{ "VEX_Kandnb_kr_kr_kr", Code.VEX_Kandnb_kr_kr_kr },
				{ "VEX_Kandnd_kr_kr_kr", Code.VEX_Kandnd_kr_kr_kr },
				{ "VEX_Knotw_kr_kr", Code.VEX_Knotw_kr_kr },
				{ "VEX_Knotq_kr_kr", Code.VEX_Knotq_kr_kr },
				{ "VEX_Knotb_kr_kr", Code.VEX_Knotb_kr_kr },
				{ "VEX_Knotd_kr_kr", Code.VEX_Knotd_kr_kr },
				{ "VEX_Korw_kr_kr_kr", Code.VEX_Korw_kr_kr_kr },
				{ "VEX_Korq_kr_kr_kr", Code.VEX_Korq_kr_kr_kr },
				{ "VEX_Korb_kr_kr_kr", Code.VEX_Korb_kr_kr_kr },
				{ "VEX_Kord_kr_kr_kr", Code.VEX_Kord_kr_kr_kr },
				{ "VEX_Kxnorw_kr_kr_kr", Code.VEX_Kxnorw_kr_kr_kr },
				{ "VEX_Kxnorq_kr_kr_kr", Code.VEX_Kxnorq_kr_kr_kr },
				{ "VEX_Kxnorb_kr_kr_kr", Code.VEX_Kxnorb_kr_kr_kr },
				{ "VEX_Kxnord_kr_kr_kr", Code.VEX_Kxnord_kr_kr_kr },
				{ "VEX_Kxorw_kr_kr_kr", Code.VEX_Kxorw_kr_kr_kr },
				{ "VEX_Kxorq_kr_kr_kr", Code.VEX_Kxorq_kr_kr_kr },
				{ "VEX_Kxorb_kr_kr_kr", Code.VEX_Kxorb_kr_kr_kr },
				{ "VEX_Kxord_kr_kr_kr", Code.VEX_Kxord_kr_kr_kr },
				{ "VEX_Kaddw_kr_kr_kr", Code.VEX_Kaddw_kr_kr_kr },
				{ "VEX_Kaddq_kr_kr_kr", Code.VEX_Kaddq_kr_kr_kr },
				{ "VEX_Kaddb_kr_kr_kr", Code.VEX_Kaddb_kr_kr_kr },
				{ "VEX_Kaddd_kr_kr_kr", Code.VEX_Kaddd_kr_kr_kr },
				{ "VEX_Kunpckwd_kr_kr_kr", Code.VEX_Kunpckwd_kr_kr_kr },
				{ "VEX_Kunpckdq_kr_kr_kr", Code.VEX_Kunpckdq_kr_kr_kr },
				{ "VEX_Kunpckbw_kr_kr_kr", Code.VEX_Kunpckbw_kr_kr_kr },
				{ "Movmskps_r32_xmm", Code.Movmskps_r32_xmm },
				{ "Movmskps_r64_xmm", Code.Movmskps_r64_xmm },
				{ "VEX_Vmovmskps_r32_xmm", Code.VEX_Vmovmskps_r32_xmm },
				{ "VEX_Vmovmskps_r64_xmm", Code.VEX_Vmovmskps_r64_xmm },
				{ "VEX_Vmovmskps_r32_ymm", Code.VEX_Vmovmskps_r32_ymm },
				{ "VEX_Vmovmskps_r64_ymm", Code.VEX_Vmovmskps_r64_ymm },
				{ "Movmskpd_r32_xmm", Code.Movmskpd_r32_xmm },
				{ "Movmskpd_r64_xmm", Code.Movmskpd_r64_xmm },
				{ "VEX_Vmovmskpd_r32_xmm", Code.VEX_Vmovmskpd_r32_xmm },
				{ "VEX_Vmovmskpd_r64_xmm", Code.VEX_Vmovmskpd_r64_xmm },
				{ "VEX_Vmovmskpd_r32_ymm", Code.VEX_Vmovmskpd_r32_ymm },
				{ "VEX_Vmovmskpd_r64_ymm", Code.VEX_Vmovmskpd_r64_ymm },
				{ "Sqrtps_xmm_xmmm128", Code.Sqrtps_xmm_xmmm128 },
				{ "VEX_Vsqrtps_xmm_xmmm128", Code.VEX_Vsqrtps_xmm_xmmm128 },
				{ "VEX_Vsqrtps_ymm_ymmm256", Code.VEX_Vsqrtps_ymm_ymmm256 },
				{ "EVEX_Vsqrtps_xmm_k1z_xmmm128b32", Code.EVEX_Vsqrtps_xmm_k1z_xmmm128b32 },
				{ "EVEX_Vsqrtps_ymm_k1z_ymmm256b32", Code.EVEX_Vsqrtps_ymm_k1z_ymmm256b32 },
				{ "EVEX_Vsqrtps_zmm_k1z_zmmm512b32_er", Code.EVEX_Vsqrtps_zmm_k1z_zmmm512b32_er },
				{ "Sqrtpd_xmm_xmmm128", Code.Sqrtpd_xmm_xmmm128 },
				{ "VEX_Vsqrtpd_xmm_xmmm128", Code.VEX_Vsqrtpd_xmm_xmmm128 },
				{ "VEX_Vsqrtpd_ymm_ymmm256", Code.VEX_Vsqrtpd_ymm_ymmm256 },
				{ "EVEX_Vsqrtpd_xmm_k1z_xmmm128b64", Code.EVEX_Vsqrtpd_xmm_k1z_xmmm128b64 },
				{ "EVEX_Vsqrtpd_ymm_k1z_ymmm256b64", Code.EVEX_Vsqrtpd_ymm_k1z_ymmm256b64 },
				{ "EVEX_Vsqrtpd_zmm_k1z_zmmm512b64_er", Code.EVEX_Vsqrtpd_zmm_k1z_zmmm512b64_er },
				{ "Sqrtss_xmm_xmmm32", Code.Sqrtss_xmm_xmmm32 },
				{ "VEX_Vsqrtss_xmm_xmm_xmmm32", Code.VEX_Vsqrtss_xmm_xmm_xmmm32 },
				{ "EVEX_Vsqrtss_xmm_k1z_xmm_xmmm32_er", Code.EVEX_Vsqrtss_xmm_k1z_xmm_xmmm32_er },
				{ "Sqrtsd_xmm_xmmm64", Code.Sqrtsd_xmm_xmmm64 },
				{ "VEX_Vsqrtsd_xmm_xmm_xmmm64", Code.VEX_Vsqrtsd_xmm_xmm_xmmm64 },
				{ "EVEX_Vsqrtsd_xmm_k1z_xmm_xmmm64_er", Code.EVEX_Vsqrtsd_xmm_k1z_xmm_xmmm64_er },
				{ "Rsqrtps_xmm_xmmm128", Code.Rsqrtps_xmm_xmmm128 },
				{ "VEX_Vrsqrtps_xmm_xmmm128", Code.VEX_Vrsqrtps_xmm_xmmm128 },
				{ "VEX_Vrsqrtps_ymm_ymmm256", Code.VEX_Vrsqrtps_ymm_ymmm256 },
				{ "Rsqrtss_xmm_xmmm32", Code.Rsqrtss_xmm_xmmm32 },
				{ "VEX_Vrsqrtss_xmm_xmm_xmmm32", Code.VEX_Vrsqrtss_xmm_xmm_xmmm32 },
				{ "Rcpps_xmm_xmmm128", Code.Rcpps_xmm_xmmm128 },
				{ "VEX_Vrcpps_xmm_xmmm128", Code.VEX_Vrcpps_xmm_xmmm128 },
				{ "VEX_Vrcpps_ymm_ymmm256", Code.VEX_Vrcpps_ymm_ymmm256 },
				{ "Rcpss_xmm_xmmm32", Code.Rcpss_xmm_xmmm32 },
				{ "VEX_Vrcpss_xmm_xmm_xmmm32", Code.VEX_Vrcpss_xmm_xmm_xmmm32 },
				{ "Andps_xmm_xmmm128", Code.Andps_xmm_xmmm128 },
				{ "VEX_Vandps_xmm_xmm_xmmm128", Code.VEX_Vandps_xmm_xmm_xmmm128 },
				{ "VEX_Vandps_ymm_ymm_ymmm256", Code.VEX_Vandps_ymm_ymm_ymmm256 },
				{ "EVEX_Vandps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vandps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vandps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vandps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vandps_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vandps_zmm_k1z_zmm_zmmm512b32 },
				{ "Andpd_xmm_xmmm128", Code.Andpd_xmm_xmmm128 },
				{ "VEX_Vandpd_xmm_xmm_xmmm128", Code.VEX_Vandpd_xmm_xmm_xmmm128 },
				{ "VEX_Vandpd_ymm_ymm_ymmm256", Code.VEX_Vandpd_ymm_ymm_ymmm256 },
				{ "EVEX_Vandpd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vandpd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vandpd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vandpd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vandpd_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vandpd_zmm_k1z_zmm_zmmm512b64 },
				{ "Andnps_xmm_xmmm128", Code.Andnps_xmm_xmmm128 },
				{ "VEX_Vandnps_xmm_xmm_xmmm128", Code.VEX_Vandnps_xmm_xmm_xmmm128 },
				{ "VEX_Vandnps_ymm_ymm_ymmm256", Code.VEX_Vandnps_ymm_ymm_ymmm256 },
				{ "EVEX_Vandnps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vandnps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vandnps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vandnps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vandnps_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vandnps_zmm_k1z_zmm_zmmm512b32 },
				{ "Andnpd_xmm_xmmm128", Code.Andnpd_xmm_xmmm128 },
				{ "VEX_Vandnpd_xmm_xmm_xmmm128", Code.VEX_Vandnpd_xmm_xmm_xmmm128 },
				{ "VEX_Vandnpd_ymm_ymm_ymmm256", Code.VEX_Vandnpd_ymm_ymm_ymmm256 },
				{ "EVEX_Vandnpd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vandnpd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vandnpd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vandnpd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vandnpd_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vandnpd_zmm_k1z_zmm_zmmm512b64 },
				{ "Orps_xmm_xmmm128", Code.Orps_xmm_xmmm128 },
				{ "VEX_Vorps_xmm_xmm_xmmm128", Code.VEX_Vorps_xmm_xmm_xmmm128 },
				{ "VEX_Vorps_ymm_ymm_ymmm256", Code.VEX_Vorps_ymm_ymm_ymmm256 },
				{ "EVEX_Vorps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vorps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vorps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vorps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vorps_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vorps_zmm_k1z_zmm_zmmm512b32 },
				{ "Orpd_xmm_xmmm128", Code.Orpd_xmm_xmmm128 },
				{ "VEX_Vorpd_xmm_xmm_xmmm128", Code.VEX_Vorpd_xmm_xmm_xmmm128 },
				{ "VEX_Vorpd_ymm_ymm_ymmm256", Code.VEX_Vorpd_ymm_ymm_ymmm256 },
				{ "EVEX_Vorpd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vorpd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vorpd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vorpd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vorpd_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vorpd_zmm_k1z_zmm_zmmm512b64 },
				{ "Xorps_xmm_xmmm128", Code.Xorps_xmm_xmmm128 },
				{ "VEX_Vxorps_xmm_xmm_xmmm128", Code.VEX_Vxorps_xmm_xmm_xmmm128 },
				{ "VEX_Vxorps_ymm_ymm_ymmm256", Code.VEX_Vxorps_ymm_ymm_ymmm256 },
				{ "EVEX_Vxorps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vxorps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vxorps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vxorps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vxorps_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vxorps_zmm_k1z_zmm_zmmm512b32 },
				{ "Xorpd_xmm_xmmm128", Code.Xorpd_xmm_xmmm128 },
				{ "VEX_Vxorpd_xmm_xmm_xmmm128", Code.VEX_Vxorpd_xmm_xmm_xmmm128 },
				{ "VEX_Vxorpd_ymm_ymm_ymmm256", Code.VEX_Vxorpd_ymm_ymm_ymmm256 },
				{ "EVEX_Vxorpd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vxorpd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vxorpd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vxorpd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vxorpd_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vxorpd_zmm_k1z_zmm_zmmm512b64 },
				{ "Addps_xmm_xmmm128", Code.Addps_xmm_xmmm128 },
				{ "VEX_Vaddps_xmm_xmm_xmmm128", Code.VEX_Vaddps_xmm_xmm_xmmm128 },
				{ "VEX_Vaddps_ymm_ymm_ymmm256", Code.VEX_Vaddps_ymm_ymm_ymmm256 },
				{ "EVEX_Vaddps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vaddps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vaddps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vaddps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vaddps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vaddps_zmm_k1z_zmm_zmmm512b32_er },
				{ "Addpd_xmm_xmmm128", Code.Addpd_xmm_xmmm128 },
				{ "VEX_Vaddpd_xmm_xmm_xmmm128", Code.VEX_Vaddpd_xmm_xmm_xmmm128 },
				{ "VEX_Vaddpd_ymm_ymm_ymmm256", Code.VEX_Vaddpd_ymm_ymm_ymmm256 },
				{ "EVEX_Vaddpd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vaddpd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vaddpd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vaddpd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vaddpd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vaddpd_zmm_k1z_zmm_zmmm512b64_er },
				{ "Addss_xmm_xmmm32", Code.Addss_xmm_xmmm32 },
				{ "VEX_Vaddss_xmm_xmm_xmmm32", Code.VEX_Vaddss_xmm_xmm_xmmm32 },
				{ "EVEX_Vaddss_xmm_k1z_xmm_xmmm32_er", Code.EVEX_Vaddss_xmm_k1z_xmm_xmmm32_er },
				{ "Addsd_xmm_xmmm64", Code.Addsd_xmm_xmmm64 },
				{ "VEX_Vaddsd_xmm_xmm_xmmm64", Code.VEX_Vaddsd_xmm_xmm_xmmm64 },
				{ "EVEX_Vaddsd_xmm_k1z_xmm_xmmm64_er", Code.EVEX_Vaddsd_xmm_k1z_xmm_xmmm64_er },
				{ "Mulps_xmm_xmmm128", Code.Mulps_xmm_xmmm128 },
				{ "VEX_Vmulps_xmm_xmm_xmmm128", Code.VEX_Vmulps_xmm_xmm_xmmm128 },
				{ "VEX_Vmulps_ymm_ymm_ymmm256", Code.VEX_Vmulps_ymm_ymm_ymmm256 },
				{ "EVEX_Vmulps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vmulps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vmulps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vmulps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vmulps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vmulps_zmm_k1z_zmm_zmmm512b32_er },
				{ "Mulpd_xmm_xmmm128", Code.Mulpd_xmm_xmmm128 },
				{ "VEX_Vmulpd_xmm_xmm_xmmm128", Code.VEX_Vmulpd_xmm_xmm_xmmm128 },
				{ "VEX_Vmulpd_ymm_ymm_ymmm256", Code.VEX_Vmulpd_ymm_ymm_ymmm256 },
				{ "EVEX_Vmulpd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vmulpd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vmulpd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vmulpd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vmulpd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vmulpd_zmm_k1z_zmm_zmmm512b64_er },
				{ "Mulss_xmm_xmmm32", Code.Mulss_xmm_xmmm32 },
				{ "VEX_Vmulss_xmm_xmm_xmmm32", Code.VEX_Vmulss_xmm_xmm_xmmm32 },
				{ "EVEX_Vmulss_xmm_k1z_xmm_xmmm32_er", Code.EVEX_Vmulss_xmm_k1z_xmm_xmmm32_er },
				{ "Mulsd_xmm_xmmm64", Code.Mulsd_xmm_xmmm64 },
				{ "VEX_Vmulsd_xmm_xmm_xmmm64", Code.VEX_Vmulsd_xmm_xmm_xmmm64 },
				{ "EVEX_Vmulsd_xmm_k1z_xmm_xmmm64_er", Code.EVEX_Vmulsd_xmm_k1z_xmm_xmmm64_er },
				{ "Cvtps2pd_xmm_xmmm64", Code.Cvtps2pd_xmm_xmmm64 },
				{ "VEX_Vcvtps2pd_xmm_xmmm64", Code.VEX_Vcvtps2pd_xmm_xmmm64 },
				{ "VEX_Vcvtps2pd_ymm_xmmm128", Code.VEX_Vcvtps2pd_ymm_xmmm128 },
				{ "EVEX_Vcvtps2pd_xmm_k1z_xmmm64b32", Code.EVEX_Vcvtps2pd_xmm_k1z_xmmm64b32 },
				{ "EVEX_Vcvtps2pd_ymm_k1z_xmmm128b32", Code.EVEX_Vcvtps2pd_ymm_k1z_xmmm128b32 },
				{ "EVEX_Vcvtps2pd_zmm_k1z_ymmm256b32_sae", Code.EVEX_Vcvtps2pd_zmm_k1z_ymmm256b32_sae },
				{ "Cvtpd2ps_xmm_xmmm128", Code.Cvtpd2ps_xmm_xmmm128 },
				{ "VEX_Vcvtpd2ps_xmm_xmmm128", Code.VEX_Vcvtpd2ps_xmm_xmmm128 },
				{ "VEX_Vcvtpd2ps_xmm_ymmm256", Code.VEX_Vcvtpd2ps_xmm_ymmm256 },
				{ "EVEX_Vcvtpd2ps_xmm_k1z_xmmm128b64", Code.EVEX_Vcvtpd2ps_xmm_k1z_xmmm128b64 },
				{ "EVEX_Vcvtpd2ps_xmm_k1z_ymmm256b64", Code.EVEX_Vcvtpd2ps_xmm_k1z_ymmm256b64 },
				{ "EVEX_Vcvtpd2ps_ymm_k1z_zmmm512b64_er", Code.EVEX_Vcvtpd2ps_ymm_k1z_zmmm512b64_er },
				{ "Cvtss2sd_xmm_xmmm32", Code.Cvtss2sd_xmm_xmmm32 },
				{ "VEX_Vcvtss2sd_xmm_xmm_xmmm32", Code.VEX_Vcvtss2sd_xmm_xmm_xmmm32 },
				{ "EVEX_Vcvtss2sd_xmm_k1z_xmm_xmmm32_sae", Code.EVEX_Vcvtss2sd_xmm_k1z_xmm_xmmm32_sae },
				{ "Cvtsd2ss_xmm_xmmm64", Code.Cvtsd2ss_xmm_xmmm64 },
				{ "VEX_Vcvtsd2ss_xmm_xmm_xmmm64", Code.VEX_Vcvtsd2ss_xmm_xmm_xmmm64 },
				{ "EVEX_Vcvtsd2ss_xmm_k1z_xmm_xmmm64_er", Code.EVEX_Vcvtsd2ss_xmm_k1z_xmm_xmmm64_er },
				{ "Cvtdq2ps_xmm_xmmm128", Code.Cvtdq2ps_xmm_xmmm128 },
				{ "VEX_Vcvtdq2ps_xmm_xmmm128", Code.VEX_Vcvtdq2ps_xmm_xmmm128 },
				{ "VEX_Vcvtdq2ps_ymm_ymmm256", Code.VEX_Vcvtdq2ps_ymm_ymmm256 },
				{ "EVEX_Vcvtdq2ps_xmm_k1z_xmmm128b32", Code.EVEX_Vcvtdq2ps_xmm_k1z_xmmm128b32 },
				{ "EVEX_Vcvtdq2ps_ymm_k1z_ymmm256b32", Code.EVEX_Vcvtdq2ps_ymm_k1z_ymmm256b32 },
				{ "EVEX_Vcvtdq2ps_zmm_k1z_zmmm512b32_er", Code.EVEX_Vcvtdq2ps_zmm_k1z_zmmm512b32_er },
				{ "EVEX_Vcvtqq2ps_xmm_k1z_xmmm128b64", Code.EVEX_Vcvtqq2ps_xmm_k1z_xmmm128b64 },
				{ "EVEX_Vcvtqq2ps_xmm_k1z_ymmm256b64", Code.EVEX_Vcvtqq2ps_xmm_k1z_ymmm256b64 },
				{ "EVEX_Vcvtqq2ps_ymm_k1z_zmmm512b64_er", Code.EVEX_Vcvtqq2ps_ymm_k1z_zmmm512b64_er },
				{ "Cvtps2dq_xmm_xmmm128", Code.Cvtps2dq_xmm_xmmm128 },
				{ "VEX_Vcvtps2dq_xmm_xmmm128", Code.VEX_Vcvtps2dq_xmm_xmmm128 },
				{ "VEX_Vcvtps2dq_ymm_ymmm256", Code.VEX_Vcvtps2dq_ymm_ymmm256 },
				{ "EVEX_Vcvtps2dq_xmm_k1z_xmmm128b32", Code.EVEX_Vcvtps2dq_xmm_k1z_xmmm128b32 },
				{ "EVEX_Vcvtps2dq_ymm_k1z_ymmm256b32", Code.EVEX_Vcvtps2dq_ymm_k1z_ymmm256b32 },
				{ "EVEX_Vcvtps2dq_zmm_k1z_zmmm512b32_er", Code.EVEX_Vcvtps2dq_zmm_k1z_zmmm512b32_er },
				{ "Cvttps2dq_xmm_xmmm128", Code.Cvttps2dq_xmm_xmmm128 },
				{ "VEX_Vcvttps2dq_xmm_xmmm128", Code.VEX_Vcvttps2dq_xmm_xmmm128 },
				{ "VEX_Vcvttps2dq_ymm_ymmm256", Code.VEX_Vcvttps2dq_ymm_ymmm256 },
				{ "EVEX_Vcvttps2dq_xmm_k1z_xmmm128b32", Code.EVEX_Vcvttps2dq_xmm_k1z_xmmm128b32 },
				{ "EVEX_Vcvttps2dq_ymm_k1z_ymmm256b32", Code.EVEX_Vcvttps2dq_ymm_k1z_ymmm256b32 },
				{ "EVEX_Vcvttps2dq_zmm_k1z_zmmm512b32_sae", Code.EVEX_Vcvttps2dq_zmm_k1z_zmmm512b32_sae },
				{ "Subps_xmm_xmmm128", Code.Subps_xmm_xmmm128 },
				{ "VEX_Vsubps_xmm_xmm_xmmm128", Code.VEX_Vsubps_xmm_xmm_xmmm128 },
				{ "VEX_Vsubps_ymm_ymm_ymmm256", Code.VEX_Vsubps_ymm_ymm_ymmm256 },
				{ "EVEX_Vsubps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vsubps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vsubps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vsubps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vsubps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vsubps_zmm_k1z_zmm_zmmm512b32_er },
				{ "Subpd_xmm_xmmm128", Code.Subpd_xmm_xmmm128 },
				{ "VEX_Vsubpd_xmm_xmm_xmmm128", Code.VEX_Vsubpd_xmm_xmm_xmmm128 },
				{ "VEX_Vsubpd_ymm_ymm_ymmm256", Code.VEX_Vsubpd_ymm_ymm_ymmm256 },
				{ "EVEX_Vsubpd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vsubpd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vsubpd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vsubpd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vsubpd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vsubpd_zmm_k1z_zmm_zmmm512b64_er },
				{ "Subss_xmm_xmmm32", Code.Subss_xmm_xmmm32 },
				{ "VEX_Vsubss_xmm_xmm_xmmm32", Code.VEX_Vsubss_xmm_xmm_xmmm32 },
				{ "EVEX_Vsubss_xmm_k1z_xmm_xmmm32_er", Code.EVEX_Vsubss_xmm_k1z_xmm_xmmm32_er },
				{ "Subsd_xmm_xmmm64", Code.Subsd_xmm_xmmm64 },
				{ "VEX_Vsubsd_xmm_xmm_xmmm64", Code.VEX_Vsubsd_xmm_xmm_xmmm64 },
				{ "EVEX_Vsubsd_xmm_k1z_xmm_xmmm64_er", Code.EVEX_Vsubsd_xmm_k1z_xmm_xmmm64_er },
				{ "Minps_xmm_xmmm128", Code.Minps_xmm_xmmm128 },
				{ "VEX_Vminps_xmm_xmm_xmmm128", Code.VEX_Vminps_xmm_xmm_xmmm128 },
				{ "VEX_Vminps_ymm_ymm_ymmm256", Code.VEX_Vminps_ymm_ymm_ymmm256 },
				{ "EVEX_Vminps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vminps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vminps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vminps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vminps_zmm_k1z_zmm_zmmm512b32_sae", Code.EVEX_Vminps_zmm_k1z_zmm_zmmm512b32_sae },
				{ "Minpd_xmm_xmmm128", Code.Minpd_xmm_xmmm128 },
				{ "VEX_Vminpd_xmm_xmm_xmmm128", Code.VEX_Vminpd_xmm_xmm_xmmm128 },
				{ "VEX_Vminpd_ymm_ymm_ymmm256", Code.VEX_Vminpd_ymm_ymm_ymmm256 },
				{ "EVEX_Vminpd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vminpd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vminpd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vminpd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vminpd_zmm_k1z_zmm_zmmm512b64_sae", Code.EVEX_Vminpd_zmm_k1z_zmm_zmmm512b64_sae },
				{ "Minss_xmm_xmmm32", Code.Minss_xmm_xmmm32 },
				{ "VEX_Vminss_xmm_xmm_xmmm32", Code.VEX_Vminss_xmm_xmm_xmmm32 },
				{ "EVEX_Vminss_xmm_k1z_xmm_xmmm32_sae", Code.EVEX_Vminss_xmm_k1z_xmm_xmmm32_sae },
				{ "Minsd_xmm_xmmm64", Code.Minsd_xmm_xmmm64 },
				{ "VEX_Vminsd_xmm_xmm_xmmm64", Code.VEX_Vminsd_xmm_xmm_xmmm64 },
				{ "EVEX_Vminsd_xmm_k1z_xmm_xmmm64_sae", Code.EVEX_Vminsd_xmm_k1z_xmm_xmmm64_sae },
				{ "Divps_xmm_xmmm128", Code.Divps_xmm_xmmm128 },
				{ "VEX_Vdivps_xmm_xmm_xmmm128", Code.VEX_Vdivps_xmm_xmm_xmmm128 },
				{ "VEX_Vdivps_ymm_ymm_ymmm256", Code.VEX_Vdivps_ymm_ymm_ymmm256 },
				{ "EVEX_Vdivps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vdivps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vdivps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vdivps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vdivps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vdivps_zmm_k1z_zmm_zmmm512b32_er },
				{ "Divpd_xmm_xmmm128", Code.Divpd_xmm_xmmm128 },
				{ "VEX_Vdivpd_xmm_xmm_xmmm128", Code.VEX_Vdivpd_xmm_xmm_xmmm128 },
				{ "VEX_Vdivpd_ymm_ymm_ymmm256", Code.VEX_Vdivpd_ymm_ymm_ymmm256 },
				{ "EVEX_Vdivpd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vdivpd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vdivpd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vdivpd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vdivpd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vdivpd_zmm_k1z_zmm_zmmm512b64_er },
				{ "Divss_xmm_xmmm32", Code.Divss_xmm_xmmm32 },
				{ "VEX_Vdivss_xmm_xmm_xmmm32", Code.VEX_Vdivss_xmm_xmm_xmmm32 },
				{ "EVEX_Vdivss_xmm_k1z_xmm_xmmm32_er", Code.EVEX_Vdivss_xmm_k1z_xmm_xmmm32_er },
				{ "Divsd_xmm_xmmm64", Code.Divsd_xmm_xmmm64 },
				{ "VEX_Vdivsd_xmm_xmm_xmmm64", Code.VEX_Vdivsd_xmm_xmm_xmmm64 },
				{ "EVEX_Vdivsd_xmm_k1z_xmm_xmmm64_er", Code.EVEX_Vdivsd_xmm_k1z_xmm_xmmm64_er },
				{ "Maxps_xmm_xmmm128", Code.Maxps_xmm_xmmm128 },
				{ "VEX_Vmaxps_xmm_xmm_xmmm128", Code.VEX_Vmaxps_xmm_xmm_xmmm128 },
				{ "VEX_Vmaxps_ymm_ymm_ymmm256", Code.VEX_Vmaxps_ymm_ymm_ymmm256 },
				{ "EVEX_Vmaxps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vmaxps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vmaxps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vmaxps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vmaxps_zmm_k1z_zmm_zmmm512b32_sae", Code.EVEX_Vmaxps_zmm_k1z_zmm_zmmm512b32_sae },
				{ "Maxpd_xmm_xmmm128", Code.Maxpd_xmm_xmmm128 },
				{ "VEX_Vmaxpd_xmm_xmm_xmmm128", Code.VEX_Vmaxpd_xmm_xmm_xmmm128 },
				{ "VEX_Vmaxpd_ymm_ymm_ymmm256", Code.VEX_Vmaxpd_ymm_ymm_ymmm256 },
				{ "EVEX_Vmaxpd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vmaxpd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vmaxpd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vmaxpd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vmaxpd_zmm_k1z_zmm_zmmm512b64_sae", Code.EVEX_Vmaxpd_zmm_k1z_zmm_zmmm512b64_sae },
				{ "Maxss_xmm_xmmm32", Code.Maxss_xmm_xmmm32 },
				{ "VEX_Vmaxss_xmm_xmm_xmmm32", Code.VEX_Vmaxss_xmm_xmm_xmmm32 },
				{ "EVEX_Vmaxss_xmm_k1z_xmm_xmmm32_sae", Code.EVEX_Vmaxss_xmm_k1z_xmm_xmmm32_sae },
				{ "Maxsd_xmm_xmmm64", Code.Maxsd_xmm_xmmm64 },
				{ "VEX_Vmaxsd_xmm_xmm_xmmm64", Code.VEX_Vmaxsd_xmm_xmm_xmmm64 },
				{ "EVEX_Vmaxsd_xmm_k1z_xmm_xmmm64_sae", Code.EVEX_Vmaxsd_xmm_k1z_xmm_xmmm64_sae },
				{ "Punpcklbw_mm_mmm32", Code.Punpcklbw_mm_mmm32 },
				{ "Punpcklbw_xmm_xmmm128", Code.Punpcklbw_xmm_xmmm128 },
				{ "VEX_Vpunpcklbw_xmm_xmm_xmmm128", Code.VEX_Vpunpcklbw_xmm_xmm_xmmm128 },
				{ "VEX_Vpunpcklbw_ymm_ymm_ymmm256", Code.VEX_Vpunpcklbw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpunpcklbw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpunpcklbw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpunpcklbw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpunpcklbw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpunpcklbw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpunpcklbw_zmm_k1z_zmm_zmmm512 },
				{ "Punpcklwd_mm_mmm32", Code.Punpcklwd_mm_mmm32 },
				{ "Punpcklwd_xmm_xmmm128", Code.Punpcklwd_xmm_xmmm128 },
				{ "VEX_Vpunpcklwd_xmm_xmm_xmmm128", Code.VEX_Vpunpcklwd_xmm_xmm_xmmm128 },
				{ "VEX_Vpunpcklwd_ymm_ymm_ymmm256", Code.VEX_Vpunpcklwd_ymm_ymm_ymmm256 },
				{ "EVEX_Vpunpcklwd_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpunpcklwd_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpunpcklwd_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpunpcklwd_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpunpcklwd_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpunpcklwd_zmm_k1z_zmm_zmmm512 },
				{ "Punpckldq_mm_mmm32", Code.Punpckldq_mm_mmm32 },
				{ "Punpckldq_xmm_xmmm128", Code.Punpckldq_xmm_xmmm128 },
				{ "VEX_Vpunpckldq_xmm_xmm_xmmm128", Code.VEX_Vpunpckldq_xmm_xmm_xmmm128 },
				{ "VEX_Vpunpckldq_ymm_ymm_ymmm256", Code.VEX_Vpunpckldq_ymm_ymm_ymmm256 },
				{ "EVEX_Vpunpckldq_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpunpckldq_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpunpckldq_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpunpckldq_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpunpckldq_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpunpckldq_zmm_k1z_zmm_zmmm512b32 },
				{ "Packsswb_mm_mmm64", Code.Packsswb_mm_mmm64 },
				{ "Packsswb_xmm_xmmm128", Code.Packsswb_xmm_xmmm128 },
				{ "VEX_Vpacksswb_xmm_xmm_xmmm128", Code.VEX_Vpacksswb_xmm_xmm_xmmm128 },
				{ "VEX_Vpacksswb_ymm_ymm_ymmm256", Code.VEX_Vpacksswb_ymm_ymm_ymmm256 },
				{ "EVEX_Vpacksswb_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpacksswb_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpacksswb_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpacksswb_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpacksswb_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpacksswb_zmm_k1z_zmm_zmmm512 },
				{ "Pcmpgtb_mm_mmm64", Code.Pcmpgtb_mm_mmm64 },
				{ "Pcmpgtb_xmm_xmmm128", Code.Pcmpgtb_xmm_xmmm128 },
				{ "VEX_Vpcmpgtb_xmm_xmm_xmmm128", Code.VEX_Vpcmpgtb_xmm_xmm_xmmm128 },
				{ "VEX_Vpcmpgtb_ymm_ymm_ymmm256", Code.VEX_Vpcmpgtb_ymm_ymm_ymmm256 },
				{ "EVEX_Vpcmpgtb_kr_k1_xmm_xmmm128", Code.EVEX_Vpcmpgtb_kr_k1_xmm_xmmm128 },
				{ "EVEX_Vpcmpgtb_kr_k1_ymm_ymmm256", Code.EVEX_Vpcmpgtb_kr_k1_ymm_ymmm256 },
				{ "EVEX_Vpcmpgtb_kr_k1_zmm_zmmm512", Code.EVEX_Vpcmpgtb_kr_k1_zmm_zmmm512 },
				{ "Pcmpgtw_mm_mmm64", Code.Pcmpgtw_mm_mmm64 },
				{ "Pcmpgtw_xmm_xmmm128", Code.Pcmpgtw_xmm_xmmm128 },
				{ "VEX_Vpcmpgtw_xmm_xmm_xmmm128", Code.VEX_Vpcmpgtw_xmm_xmm_xmmm128 },
				{ "VEX_Vpcmpgtw_ymm_ymm_ymmm256", Code.VEX_Vpcmpgtw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpcmpgtw_kr_k1_xmm_xmmm128", Code.EVEX_Vpcmpgtw_kr_k1_xmm_xmmm128 },
				{ "EVEX_Vpcmpgtw_kr_k1_ymm_ymmm256", Code.EVEX_Vpcmpgtw_kr_k1_ymm_ymmm256 },
				{ "EVEX_Vpcmpgtw_kr_k1_zmm_zmmm512", Code.EVEX_Vpcmpgtw_kr_k1_zmm_zmmm512 },
				{ "Pcmpgtd_mm_mmm64", Code.Pcmpgtd_mm_mmm64 },
				{ "Pcmpgtd_xmm_xmmm128", Code.Pcmpgtd_xmm_xmmm128 },
				{ "VEX_Vpcmpgtd_xmm_xmm_xmmm128", Code.VEX_Vpcmpgtd_xmm_xmm_xmmm128 },
				{ "VEX_Vpcmpgtd_ymm_ymm_ymmm256", Code.VEX_Vpcmpgtd_ymm_ymm_ymmm256 },
				{ "EVEX_Vpcmpgtd_kr_k1_xmm_xmmm128b32", Code.EVEX_Vpcmpgtd_kr_k1_xmm_xmmm128b32 },
				{ "EVEX_Vpcmpgtd_kr_k1_ymm_ymmm256b32", Code.EVEX_Vpcmpgtd_kr_k1_ymm_ymmm256b32 },
				{ "EVEX_Vpcmpgtd_kr_k1_zmm_zmmm512b32", Code.EVEX_Vpcmpgtd_kr_k1_zmm_zmmm512b32 },
				{ "Packuswb_mm_mmm64", Code.Packuswb_mm_mmm64 },
				{ "Packuswb_xmm_xmmm128", Code.Packuswb_xmm_xmmm128 },
				{ "VEX_Vpackuswb_xmm_xmm_xmmm128", Code.VEX_Vpackuswb_xmm_xmm_xmmm128 },
				{ "VEX_Vpackuswb_ymm_ymm_ymmm256", Code.VEX_Vpackuswb_ymm_ymm_ymmm256 },
				{ "EVEX_Vpackuswb_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpackuswb_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpackuswb_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpackuswb_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpackuswb_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpackuswb_zmm_k1z_zmm_zmmm512 },
				{ "Punpckhbw_mm_mmm64", Code.Punpckhbw_mm_mmm64 },
				{ "Punpckhbw_xmm_xmmm128", Code.Punpckhbw_xmm_xmmm128 },
				{ "VEX_Vpunpckhbw_xmm_xmm_xmmm128", Code.VEX_Vpunpckhbw_xmm_xmm_xmmm128 },
				{ "VEX_Vpunpckhbw_ymm_ymm_ymmm256", Code.VEX_Vpunpckhbw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpunpckhbw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpunpckhbw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpunpckhbw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpunpckhbw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpunpckhbw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpunpckhbw_zmm_k1z_zmm_zmmm512 },
				{ "Punpckhwd_mm_mmm64", Code.Punpckhwd_mm_mmm64 },
				{ "Punpckhwd_xmm_xmmm128", Code.Punpckhwd_xmm_xmmm128 },
				{ "VEX_Vpunpckhwd_xmm_xmm_xmmm128", Code.VEX_Vpunpckhwd_xmm_xmm_xmmm128 },
				{ "VEX_Vpunpckhwd_ymm_ymm_ymmm256", Code.VEX_Vpunpckhwd_ymm_ymm_ymmm256 },
				{ "EVEX_Vpunpckhwd_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpunpckhwd_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpunpckhwd_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpunpckhwd_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpunpckhwd_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpunpckhwd_zmm_k1z_zmm_zmmm512 },
				{ "Punpckhdq_mm_mmm64", Code.Punpckhdq_mm_mmm64 },
				{ "Punpckhdq_xmm_xmmm128", Code.Punpckhdq_xmm_xmmm128 },
				{ "VEX_Vpunpckhdq_xmm_xmm_xmmm128", Code.VEX_Vpunpckhdq_xmm_xmm_xmmm128 },
				{ "VEX_Vpunpckhdq_ymm_ymm_ymmm256", Code.VEX_Vpunpckhdq_ymm_ymm_ymmm256 },
				{ "EVEX_Vpunpckhdq_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpunpckhdq_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpunpckhdq_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpunpckhdq_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpunpckhdq_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpunpckhdq_zmm_k1z_zmm_zmmm512b32 },
				{ "Packssdw_mm_mmm64", Code.Packssdw_mm_mmm64 },
				{ "Packssdw_xmm_xmmm128", Code.Packssdw_xmm_xmmm128 },
				{ "VEX_Vpackssdw_xmm_xmm_xmmm128", Code.VEX_Vpackssdw_xmm_xmm_xmmm128 },
				{ "VEX_Vpackssdw_ymm_ymm_ymmm256", Code.VEX_Vpackssdw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpackssdw_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpackssdw_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpackssdw_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpackssdw_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpackssdw_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpackssdw_zmm_k1z_zmm_zmmm512b32 },
				{ "Punpcklqdq_xmm_xmmm128", Code.Punpcklqdq_xmm_xmmm128 },
				{ "VEX_Vpunpcklqdq_xmm_xmm_xmmm128", Code.VEX_Vpunpcklqdq_xmm_xmm_xmmm128 },
				{ "VEX_Vpunpcklqdq_ymm_ymm_ymmm256", Code.VEX_Vpunpcklqdq_ymm_ymm_ymmm256 },
				{ "EVEX_Vpunpcklqdq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpunpcklqdq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpunpcklqdq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpunpcklqdq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpunpcklqdq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpunpcklqdq_zmm_k1z_zmm_zmmm512b64 },
				{ "Punpckhqdq_xmm_xmmm128", Code.Punpckhqdq_xmm_xmmm128 },
				{ "VEX_Vpunpckhqdq_xmm_xmm_xmmm128", Code.VEX_Vpunpckhqdq_xmm_xmm_xmmm128 },
				{ "VEX_Vpunpckhqdq_ymm_ymm_ymmm256", Code.VEX_Vpunpckhqdq_ymm_ymm_ymmm256 },
				{ "EVEX_Vpunpckhqdq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpunpckhqdq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpunpckhqdq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpunpckhqdq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpunpckhqdq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpunpckhqdq_zmm_k1z_zmm_zmmm512b64 },
				{ "Movd_mm_rm32", Code.Movd_mm_rm32 },
				{ "Movq_mm_rm64", Code.Movq_mm_rm64 },
				{ "Movd_xmm_rm32", Code.Movd_xmm_rm32 },
				{ "Movq_xmm_rm64", Code.Movq_xmm_rm64 },
				{ "VEX_Vmovd_xmm_rm32", Code.VEX_Vmovd_xmm_rm32 },
				{ "VEX_Vmovq_xmm_rm64", Code.VEX_Vmovq_xmm_rm64 },
				{ "EVEX_Vmovd_xmm_rm32", Code.EVEX_Vmovd_xmm_rm32 },
				{ "EVEX_Vmovq_xmm_rm64", Code.EVEX_Vmovq_xmm_rm64 },
				{ "Movq_mm_mmm64", Code.Movq_mm_mmm64 },
				{ "Movdqa_xmm_xmmm128", Code.Movdqa_xmm_xmmm128 },
				{ "VEX_Vmovdqa_xmm_xmmm128", Code.VEX_Vmovdqa_xmm_xmmm128 },
				{ "VEX_Vmovdqa_ymm_ymmm256", Code.VEX_Vmovdqa_ymm_ymmm256 },
				{ "EVEX_Vmovdqa32_xmm_k1z_xmmm128", Code.EVEX_Vmovdqa32_xmm_k1z_xmmm128 },
				{ "EVEX_Vmovdqa32_ymm_k1z_ymmm256", Code.EVEX_Vmovdqa32_ymm_k1z_ymmm256 },
				{ "EVEX_Vmovdqa32_zmm_k1z_zmmm512", Code.EVEX_Vmovdqa32_zmm_k1z_zmmm512 },
				{ "EVEX_Vmovdqa64_xmm_k1z_xmmm128", Code.EVEX_Vmovdqa64_xmm_k1z_xmmm128 },
				{ "EVEX_Vmovdqa64_ymm_k1z_ymmm256", Code.EVEX_Vmovdqa64_ymm_k1z_ymmm256 },
				{ "EVEX_Vmovdqa64_zmm_k1z_zmmm512", Code.EVEX_Vmovdqa64_zmm_k1z_zmmm512 },
				{ "Movdqu_xmm_xmmm128", Code.Movdqu_xmm_xmmm128 },
				{ "VEX_Vmovdqu_xmm_xmmm128", Code.VEX_Vmovdqu_xmm_xmmm128 },
				{ "VEX_Vmovdqu_ymm_ymmm256", Code.VEX_Vmovdqu_ymm_ymmm256 },
				{ "EVEX_Vmovdqu32_xmm_k1z_xmmm128", Code.EVEX_Vmovdqu32_xmm_k1z_xmmm128 },
				{ "EVEX_Vmovdqu32_ymm_k1z_ymmm256", Code.EVEX_Vmovdqu32_ymm_k1z_ymmm256 },
				{ "EVEX_Vmovdqu32_zmm_k1z_zmmm512", Code.EVEX_Vmovdqu32_zmm_k1z_zmmm512 },
				{ "EVEX_Vmovdqu64_xmm_k1z_xmmm128", Code.EVEX_Vmovdqu64_xmm_k1z_xmmm128 },
				{ "EVEX_Vmovdqu64_ymm_k1z_ymmm256", Code.EVEX_Vmovdqu64_ymm_k1z_ymmm256 },
				{ "EVEX_Vmovdqu64_zmm_k1z_zmmm512", Code.EVEX_Vmovdqu64_zmm_k1z_zmmm512 },
				{ "EVEX_Vmovdqu8_xmm_k1z_xmmm128", Code.EVEX_Vmovdqu8_xmm_k1z_xmmm128 },
				{ "EVEX_Vmovdqu8_ymm_k1z_ymmm256", Code.EVEX_Vmovdqu8_ymm_k1z_ymmm256 },
				{ "EVEX_Vmovdqu8_zmm_k1z_zmmm512", Code.EVEX_Vmovdqu8_zmm_k1z_zmmm512 },
				{ "EVEX_Vmovdqu16_xmm_k1z_xmmm128", Code.EVEX_Vmovdqu16_xmm_k1z_xmmm128 },
				{ "EVEX_Vmovdqu16_ymm_k1z_ymmm256", Code.EVEX_Vmovdqu16_ymm_k1z_ymmm256 },
				{ "EVEX_Vmovdqu16_zmm_k1z_zmmm512", Code.EVEX_Vmovdqu16_zmm_k1z_zmmm512 },
				{ "Pshufw_mm_mmm64_imm8", Code.Pshufw_mm_mmm64_imm8 },
				{ "Pshufd_xmm_xmmm128_imm8", Code.Pshufd_xmm_xmmm128_imm8 },
				{ "VEX_Vpshufd_xmm_xmmm128_imm8", Code.VEX_Vpshufd_xmm_xmmm128_imm8 },
				{ "VEX_Vpshufd_ymm_ymmm256_imm8", Code.VEX_Vpshufd_ymm_ymmm256_imm8 },
				{ "EVEX_Vpshufd_xmm_k1z_xmmm128b32_imm8", Code.EVEX_Vpshufd_xmm_k1z_xmmm128b32_imm8 },
				{ "EVEX_Vpshufd_ymm_k1z_ymmm256b32_imm8", Code.EVEX_Vpshufd_ymm_k1z_ymmm256b32_imm8 },
				{ "EVEX_Vpshufd_zmm_k1z_zmmm512b32_imm8", Code.EVEX_Vpshufd_zmm_k1z_zmmm512b32_imm8 },
				{ "Pshufhw_xmm_xmmm128_imm8", Code.Pshufhw_xmm_xmmm128_imm8 },
				{ "VEX_Vpshufhw_xmm_xmmm128_imm8", Code.VEX_Vpshufhw_xmm_xmmm128_imm8 },
				{ "VEX_Vpshufhw_ymm_ymmm256_imm8", Code.VEX_Vpshufhw_ymm_ymmm256_imm8 },
				{ "EVEX_Vpshufhw_xmm_k1z_xmmm128_imm8", Code.EVEX_Vpshufhw_xmm_k1z_xmmm128_imm8 },
				{ "EVEX_Vpshufhw_ymm_k1z_ymmm256_imm8", Code.EVEX_Vpshufhw_ymm_k1z_ymmm256_imm8 },
				{ "EVEX_Vpshufhw_zmm_k1z_zmmm512_imm8", Code.EVEX_Vpshufhw_zmm_k1z_zmmm512_imm8 },
				{ "Pshuflw_xmm_xmmm128_imm8", Code.Pshuflw_xmm_xmmm128_imm8 },
				{ "VEX_Vpshuflw_xmm_xmmm128_imm8", Code.VEX_Vpshuflw_xmm_xmmm128_imm8 },
				{ "VEX_Vpshuflw_ymm_ymmm256_imm8", Code.VEX_Vpshuflw_ymm_ymmm256_imm8 },
				{ "EVEX_Vpshuflw_xmm_k1z_xmmm128_imm8", Code.EVEX_Vpshuflw_xmm_k1z_xmmm128_imm8 },
				{ "EVEX_Vpshuflw_ymm_k1z_ymmm256_imm8", Code.EVEX_Vpshuflw_ymm_k1z_ymmm256_imm8 },
				{ "EVEX_Vpshuflw_zmm_k1z_zmmm512_imm8", Code.EVEX_Vpshuflw_zmm_k1z_zmmm512_imm8 },
				{ "Psrlw_mm_imm8", Code.Psrlw_mm_imm8 },
				{ "Psrlw_xmm_imm8", Code.Psrlw_xmm_imm8 },
				{ "VEX_Vpsrlw_xmm_xmm_imm8", Code.VEX_Vpsrlw_xmm_xmm_imm8 },
				{ "VEX_Vpsrlw_ymm_ymm_imm8", Code.VEX_Vpsrlw_ymm_ymm_imm8 },
				{ "EVEX_Vpsrlw_xmm_k1z_xmmm128_imm8", Code.EVEX_Vpsrlw_xmm_k1z_xmmm128_imm8 },
				{ "EVEX_Vpsrlw_ymm_k1z_ymmm256_imm8", Code.EVEX_Vpsrlw_ymm_k1z_ymmm256_imm8 },
				{ "EVEX_Vpsrlw_zmm_k1z_zmmm512_imm8", Code.EVEX_Vpsrlw_zmm_k1z_zmmm512_imm8 },
				{ "Psraw_mm_imm8", Code.Psraw_mm_imm8 },
				{ "Psraw_xmm_imm8", Code.Psraw_xmm_imm8 },
				{ "VEX_Vpsraw_xmm_xmm_imm8", Code.VEX_Vpsraw_xmm_xmm_imm8 },
				{ "VEX_Vpsraw_ymm_ymm_imm8", Code.VEX_Vpsraw_ymm_ymm_imm8 },
				{ "EVEX_Vpsraw_xmm_k1z_xmmm128_imm8", Code.EVEX_Vpsraw_xmm_k1z_xmmm128_imm8 },
				{ "EVEX_Vpsraw_ymm_k1z_ymmm256_imm8", Code.EVEX_Vpsraw_ymm_k1z_ymmm256_imm8 },
				{ "EVEX_Vpsraw_zmm_k1z_zmmm512_imm8", Code.EVEX_Vpsraw_zmm_k1z_zmmm512_imm8 },
				{ "Psllw_mm_imm8", Code.Psllw_mm_imm8 },
				{ "Psllw_xmm_imm8", Code.Psllw_xmm_imm8 },
				{ "VEX_Vpsllw_xmm_xmm_imm8", Code.VEX_Vpsllw_xmm_xmm_imm8 },
				{ "VEX_Vpsllw_ymm_ymm_imm8", Code.VEX_Vpsllw_ymm_ymm_imm8 },
				{ "EVEX_Vpsllw_xmm_k1z_xmmm128_imm8", Code.EVEX_Vpsllw_xmm_k1z_xmmm128_imm8 },
				{ "EVEX_Vpsllw_ymm_k1z_ymmm256_imm8", Code.EVEX_Vpsllw_ymm_k1z_ymmm256_imm8 },
				{ "EVEX_Vpsllw_zmm_k1z_zmmm512_imm8", Code.EVEX_Vpsllw_zmm_k1z_zmmm512_imm8 },
				{ "EVEX_Vprord_xmm_k1z_xmmm128b32_imm8", Code.EVEX_Vprord_xmm_k1z_xmmm128b32_imm8 },
				{ "EVEX_Vprord_ymm_k1z_ymmm256b32_imm8", Code.EVEX_Vprord_ymm_k1z_ymmm256b32_imm8 },
				{ "EVEX_Vprord_zmm_k1z_zmmm512b32_imm8", Code.EVEX_Vprord_zmm_k1z_zmmm512b32_imm8 },
				{ "EVEX_Vprorq_xmm_k1z_xmmm128b64_imm8", Code.EVEX_Vprorq_xmm_k1z_xmmm128b64_imm8 },
				{ "EVEX_Vprorq_ymm_k1z_ymmm256b64_imm8", Code.EVEX_Vprorq_ymm_k1z_ymmm256b64_imm8 },
				{ "EVEX_Vprorq_zmm_k1z_zmmm512b64_imm8", Code.EVEX_Vprorq_zmm_k1z_zmmm512b64_imm8 },
				{ "EVEX_Vprold_xmm_k1z_xmmm128b32_imm8", Code.EVEX_Vprold_xmm_k1z_xmmm128b32_imm8 },
				{ "EVEX_Vprold_ymm_k1z_ymmm256b32_imm8", Code.EVEX_Vprold_ymm_k1z_ymmm256b32_imm8 },
				{ "EVEX_Vprold_zmm_k1z_zmmm512b32_imm8", Code.EVEX_Vprold_zmm_k1z_zmmm512b32_imm8 },
				{ "EVEX_Vprolq_xmm_k1z_xmmm128b64_imm8", Code.EVEX_Vprolq_xmm_k1z_xmmm128b64_imm8 },
				{ "EVEX_Vprolq_ymm_k1z_ymmm256b64_imm8", Code.EVEX_Vprolq_ymm_k1z_ymmm256b64_imm8 },
				{ "EVEX_Vprolq_zmm_k1z_zmmm512b64_imm8", Code.EVEX_Vprolq_zmm_k1z_zmmm512b64_imm8 },
				{ "Psrld_mm_imm8", Code.Psrld_mm_imm8 },
				{ "Psrld_xmm_imm8", Code.Psrld_xmm_imm8 },
				{ "VEX_Vpsrld_xmm_xmm_imm8", Code.VEX_Vpsrld_xmm_xmm_imm8 },
				{ "VEX_Vpsrld_ymm_ymm_imm8", Code.VEX_Vpsrld_ymm_ymm_imm8 },
				{ "EVEX_Vpsrld_xmm_k1z_xmmm128b32_imm8", Code.EVEX_Vpsrld_xmm_k1z_xmmm128b32_imm8 },
				{ "EVEX_Vpsrld_ymm_k1z_ymmm256b32_imm8", Code.EVEX_Vpsrld_ymm_k1z_ymmm256b32_imm8 },
				{ "EVEX_Vpsrld_zmm_k1z_zmmm512b32_imm8", Code.EVEX_Vpsrld_zmm_k1z_zmmm512b32_imm8 },
				{ "Psrad_mm_imm8", Code.Psrad_mm_imm8 },
				{ "Psrad_xmm_imm8", Code.Psrad_xmm_imm8 },
				{ "VEX_Vpsrad_xmm_xmm_imm8", Code.VEX_Vpsrad_xmm_xmm_imm8 },
				{ "VEX_Vpsrad_ymm_ymm_imm8", Code.VEX_Vpsrad_ymm_ymm_imm8 },
				{ "EVEX_Vpsrad_xmm_k1z_xmmm128b32_imm8", Code.EVEX_Vpsrad_xmm_k1z_xmmm128b32_imm8 },
				{ "EVEX_Vpsrad_ymm_k1z_ymmm256b32_imm8", Code.EVEX_Vpsrad_ymm_k1z_ymmm256b32_imm8 },
				{ "EVEX_Vpsrad_zmm_k1z_zmmm512b32_imm8", Code.EVEX_Vpsrad_zmm_k1z_zmmm512b32_imm8 },
				{ "EVEX_Vpsraq_xmm_k1z_xmmm128b64_imm8", Code.EVEX_Vpsraq_xmm_k1z_xmmm128b64_imm8 },
				{ "EVEX_Vpsraq_ymm_k1z_ymmm256b64_imm8", Code.EVEX_Vpsraq_ymm_k1z_ymmm256b64_imm8 },
				{ "EVEX_Vpsraq_zmm_k1z_zmmm512b64_imm8", Code.EVEX_Vpsraq_zmm_k1z_zmmm512b64_imm8 },
				{ "Pslld_mm_imm8", Code.Pslld_mm_imm8 },
				{ "Pslld_xmm_imm8", Code.Pslld_xmm_imm8 },
				{ "VEX_Vpslld_xmm_xmm_imm8", Code.VEX_Vpslld_xmm_xmm_imm8 },
				{ "VEX_Vpslld_ymm_ymm_imm8", Code.VEX_Vpslld_ymm_ymm_imm8 },
				{ "EVEX_Vpslld_xmm_k1z_xmmm128b32_imm8", Code.EVEX_Vpslld_xmm_k1z_xmmm128b32_imm8 },
				{ "EVEX_Vpslld_ymm_k1z_ymmm256b32_imm8", Code.EVEX_Vpslld_ymm_k1z_ymmm256b32_imm8 },
				{ "EVEX_Vpslld_zmm_k1z_zmmm512b32_imm8", Code.EVEX_Vpslld_zmm_k1z_zmmm512b32_imm8 },
				{ "Psrlq_mm_imm8", Code.Psrlq_mm_imm8 },
				{ "Psrlq_xmm_imm8", Code.Psrlq_xmm_imm8 },
				{ "VEX_Vpsrlq_xmm_xmm_imm8", Code.VEX_Vpsrlq_xmm_xmm_imm8 },
				{ "VEX_Vpsrlq_ymm_ymm_imm8", Code.VEX_Vpsrlq_ymm_ymm_imm8 },
				{ "EVEX_Vpsrlq_xmm_k1z_xmmm128b64_imm8", Code.EVEX_Vpsrlq_xmm_k1z_xmmm128b64_imm8 },
				{ "EVEX_Vpsrlq_ymm_k1z_ymmm256b64_imm8", Code.EVEX_Vpsrlq_ymm_k1z_ymmm256b64_imm8 },
				{ "EVEX_Vpsrlq_zmm_k1z_zmmm512b64_imm8", Code.EVEX_Vpsrlq_zmm_k1z_zmmm512b64_imm8 },
				{ "Psrldq_xmm_imm8", Code.Psrldq_xmm_imm8 },
				{ "VEX_Vpsrldq_xmm_xmm_imm8", Code.VEX_Vpsrldq_xmm_xmm_imm8 },
				{ "VEX_Vpsrldq_ymm_ymm_imm8", Code.VEX_Vpsrldq_ymm_ymm_imm8 },
				{ "EVEX_Vpsrldq_xmm_xmmm128_imm8", Code.EVEX_Vpsrldq_xmm_xmmm128_imm8 },
				{ "EVEX_Vpsrldq_ymm_ymmm256_imm8", Code.EVEX_Vpsrldq_ymm_ymmm256_imm8 },
				{ "EVEX_Vpsrldq_zmm_zmmm512_imm8", Code.EVEX_Vpsrldq_zmm_zmmm512_imm8 },
				{ "Psllq_mm_imm8", Code.Psllq_mm_imm8 },
				{ "Psllq_xmm_imm8", Code.Psllq_xmm_imm8 },
				{ "VEX_Vpsllq_xmm_xmm_imm8", Code.VEX_Vpsllq_xmm_xmm_imm8 },
				{ "VEX_Vpsllq_ymm_ymm_imm8", Code.VEX_Vpsllq_ymm_ymm_imm8 },
				{ "EVEX_Vpsllq_xmm_k1z_xmmm128b64_imm8", Code.EVEX_Vpsllq_xmm_k1z_xmmm128b64_imm8 },
				{ "EVEX_Vpsllq_ymm_k1z_ymmm256b64_imm8", Code.EVEX_Vpsllq_ymm_k1z_ymmm256b64_imm8 },
				{ "EVEX_Vpsllq_zmm_k1z_zmmm512b64_imm8", Code.EVEX_Vpsllq_zmm_k1z_zmmm512b64_imm8 },
				{ "Pslldq_xmm_imm8", Code.Pslldq_xmm_imm8 },
				{ "VEX_Vpslldq_xmm_xmm_imm8", Code.VEX_Vpslldq_xmm_xmm_imm8 },
				{ "VEX_Vpslldq_ymm_ymm_imm8", Code.VEX_Vpslldq_ymm_ymm_imm8 },
				{ "EVEX_Vpslldq_xmm_xmmm128_imm8", Code.EVEX_Vpslldq_xmm_xmmm128_imm8 },
				{ "EVEX_Vpslldq_ymm_ymmm256_imm8", Code.EVEX_Vpslldq_ymm_ymmm256_imm8 },
				{ "EVEX_Vpslldq_zmm_zmmm512_imm8", Code.EVEX_Vpslldq_zmm_zmmm512_imm8 },
				{ "Pcmpeqb_mm_mmm64", Code.Pcmpeqb_mm_mmm64 },
				{ "Pcmpeqb_xmm_xmmm128", Code.Pcmpeqb_xmm_xmmm128 },
				{ "VEX_Vpcmpeqb_xmm_xmm_xmmm128", Code.VEX_Vpcmpeqb_xmm_xmm_xmmm128 },
				{ "VEX_Vpcmpeqb_ymm_ymm_ymmm256", Code.VEX_Vpcmpeqb_ymm_ymm_ymmm256 },
				{ "EVEX_Vpcmpeqb_kr_k1_xmm_xmmm128", Code.EVEX_Vpcmpeqb_kr_k1_xmm_xmmm128 },
				{ "EVEX_Vpcmpeqb_kr_k1_ymm_ymmm256", Code.EVEX_Vpcmpeqb_kr_k1_ymm_ymmm256 },
				{ "EVEX_Vpcmpeqb_kr_k1_zmm_zmmm512", Code.EVEX_Vpcmpeqb_kr_k1_zmm_zmmm512 },
				{ "Pcmpeqw_mm_mmm64", Code.Pcmpeqw_mm_mmm64 },
				{ "Pcmpeqw_xmm_xmmm128", Code.Pcmpeqw_xmm_xmmm128 },
				{ "VEX_Vpcmpeqw_xmm_xmm_xmmm128", Code.VEX_Vpcmpeqw_xmm_xmm_xmmm128 },
				{ "VEX_Vpcmpeqw_ymm_ymm_ymmm256", Code.VEX_Vpcmpeqw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpcmpeqw_kr_k1_xmm_xmmm128", Code.EVEX_Vpcmpeqw_kr_k1_xmm_xmmm128 },
				{ "EVEX_Vpcmpeqw_kr_k1_ymm_ymmm256", Code.EVEX_Vpcmpeqw_kr_k1_ymm_ymmm256 },
				{ "EVEX_Vpcmpeqw_kr_k1_zmm_zmmm512", Code.EVEX_Vpcmpeqw_kr_k1_zmm_zmmm512 },
				{ "Pcmpeqd_mm_mmm64", Code.Pcmpeqd_mm_mmm64 },
				{ "Pcmpeqd_xmm_xmmm128", Code.Pcmpeqd_xmm_xmmm128 },
				{ "VEX_Vpcmpeqd_xmm_xmm_xmmm128", Code.VEX_Vpcmpeqd_xmm_xmm_xmmm128 },
				{ "VEX_Vpcmpeqd_ymm_ymm_ymmm256", Code.VEX_Vpcmpeqd_ymm_ymm_ymmm256 },
				{ "EVEX_Vpcmpeqd_kr_k1_xmm_xmmm128b32", Code.EVEX_Vpcmpeqd_kr_k1_xmm_xmmm128b32 },
				{ "EVEX_Vpcmpeqd_kr_k1_ymm_ymmm256b32", Code.EVEX_Vpcmpeqd_kr_k1_ymm_ymmm256b32 },
				{ "EVEX_Vpcmpeqd_kr_k1_zmm_zmmm512b32", Code.EVEX_Vpcmpeqd_kr_k1_zmm_zmmm512b32 },
				{ "Emms", Code.Emms },
				{ "VEX_Vzeroupper", Code.VEX_Vzeroupper },
				{ "VEX_Vzeroall", Code.VEX_Vzeroall },
				{ "Vmread_rm32_r32", Code.Vmread_rm32_r32 },
				{ "Vmread_rm64_r64", Code.Vmread_rm64_r64 },
				{ "EVEX_Vcvttps2udq_xmm_k1z_xmmm128b32", Code.EVEX_Vcvttps2udq_xmm_k1z_xmmm128b32 },
				{ "EVEX_Vcvttps2udq_ymm_k1z_ymmm256b32", Code.EVEX_Vcvttps2udq_ymm_k1z_ymmm256b32 },
				{ "EVEX_Vcvttps2udq_zmm_k1z_zmmm512b32_sae", Code.EVEX_Vcvttps2udq_zmm_k1z_zmmm512b32_sae },
				{ "EVEX_Vcvttpd2udq_xmm_k1z_xmmm128b64", Code.EVEX_Vcvttpd2udq_xmm_k1z_xmmm128b64 },
				{ "EVEX_Vcvttpd2udq_xmm_k1z_ymmm256b64", Code.EVEX_Vcvttpd2udq_xmm_k1z_ymmm256b64 },
				{ "EVEX_Vcvttpd2udq_ymm_k1z_zmmm512b64_sae", Code.EVEX_Vcvttpd2udq_ymm_k1z_zmmm512b64_sae },
				{ "Extrq_xmm_imm8_imm8", Code.Extrq_xmm_imm8_imm8 },
				{ "EVEX_Vcvttps2uqq_xmm_k1z_xmmm64b32", Code.EVEX_Vcvttps2uqq_xmm_k1z_xmmm64b32 },
				{ "EVEX_Vcvttps2uqq_ymm_k1z_xmmm128b32", Code.EVEX_Vcvttps2uqq_ymm_k1z_xmmm128b32 },
				{ "EVEX_Vcvttps2uqq_zmm_k1z_ymmm256b32_sae", Code.EVEX_Vcvttps2uqq_zmm_k1z_ymmm256b32_sae },
				{ "EVEX_Vcvttpd2uqq_xmm_k1z_xmmm128b64", Code.EVEX_Vcvttpd2uqq_xmm_k1z_xmmm128b64 },
				{ "EVEX_Vcvttpd2uqq_ymm_k1z_ymmm256b64", Code.EVEX_Vcvttpd2uqq_ymm_k1z_ymmm256b64 },
				{ "EVEX_Vcvttpd2uqq_zmm_k1z_zmmm512b64_sae", Code.EVEX_Vcvttpd2uqq_zmm_k1z_zmmm512b64_sae },
				{ "EVEX_Vcvttss2usi_r32_xmmm32_sae", Code.EVEX_Vcvttss2usi_r32_xmmm32_sae },
				{ "EVEX_Vcvttss2usi_r64_xmmm32_sae", Code.EVEX_Vcvttss2usi_r64_xmmm32_sae },
				{ "Insertq_xmm_xmm_imm8_imm8", Code.Insertq_xmm_xmm_imm8_imm8 },
				{ "EVEX_Vcvttsd2usi_r32_xmmm64_sae", Code.EVEX_Vcvttsd2usi_r32_xmmm64_sae },
				{ "EVEX_Vcvttsd2usi_r64_xmmm64_sae", Code.EVEX_Vcvttsd2usi_r64_xmmm64_sae },
				{ "Vmwrite_r32_rm32", Code.Vmwrite_r32_rm32 },
				{ "Vmwrite_r64_rm64", Code.Vmwrite_r64_rm64 },
				{ "EVEX_Vcvtps2udq_xmm_k1z_xmmm128b32", Code.EVEX_Vcvtps2udq_xmm_k1z_xmmm128b32 },
				{ "EVEX_Vcvtps2udq_ymm_k1z_ymmm256b32", Code.EVEX_Vcvtps2udq_ymm_k1z_ymmm256b32 },
				{ "EVEX_Vcvtps2udq_zmm_k1z_zmmm512b32_er", Code.EVEX_Vcvtps2udq_zmm_k1z_zmmm512b32_er },
				{ "EVEX_Vcvtpd2udq_xmm_k1z_xmmm128b64", Code.EVEX_Vcvtpd2udq_xmm_k1z_xmmm128b64 },
				{ "EVEX_Vcvtpd2udq_xmm_k1z_ymmm256b64", Code.EVEX_Vcvtpd2udq_xmm_k1z_ymmm256b64 },
				{ "EVEX_Vcvtpd2udq_ymm_k1z_zmmm512b64_er", Code.EVEX_Vcvtpd2udq_ymm_k1z_zmmm512b64_er },
				{ "Extrq_xmm_xmm", Code.Extrq_xmm_xmm },
				{ "EVEX_Vcvtps2uqq_xmm_k1z_xmmm64b32", Code.EVEX_Vcvtps2uqq_xmm_k1z_xmmm64b32 },
				{ "EVEX_Vcvtps2uqq_ymm_k1z_xmmm128b32", Code.EVEX_Vcvtps2uqq_ymm_k1z_xmmm128b32 },
				{ "EVEX_Vcvtps2uqq_zmm_k1z_ymmm256b32_er", Code.EVEX_Vcvtps2uqq_zmm_k1z_ymmm256b32_er },
				{ "EVEX_Vcvtpd2uqq_xmm_k1z_xmmm128b64", Code.EVEX_Vcvtpd2uqq_xmm_k1z_xmmm128b64 },
				{ "EVEX_Vcvtpd2uqq_ymm_k1z_ymmm256b64", Code.EVEX_Vcvtpd2uqq_ymm_k1z_ymmm256b64 },
				{ "EVEX_Vcvtpd2uqq_zmm_k1z_zmmm512b64_er", Code.EVEX_Vcvtpd2uqq_zmm_k1z_zmmm512b64_er },
				{ "EVEX_Vcvtss2usi_r32_xmmm32_er", Code.EVEX_Vcvtss2usi_r32_xmmm32_er },
				{ "EVEX_Vcvtss2usi_r64_xmmm32_er", Code.EVEX_Vcvtss2usi_r64_xmmm32_er },
				{ "Insertq_xmm_xmm", Code.Insertq_xmm_xmm },
				{ "EVEX_Vcvtsd2usi_r32_xmmm64_er", Code.EVEX_Vcvtsd2usi_r32_xmmm64_er },
				{ "EVEX_Vcvtsd2usi_r64_xmmm64_er", Code.EVEX_Vcvtsd2usi_r64_xmmm64_er },
				{ "EVEX_Vcvttps2qq_xmm_k1z_xmmm64b32", Code.EVEX_Vcvttps2qq_xmm_k1z_xmmm64b32 },
				{ "EVEX_Vcvttps2qq_ymm_k1z_xmmm128b32", Code.EVEX_Vcvttps2qq_ymm_k1z_xmmm128b32 },
				{ "EVEX_Vcvttps2qq_zmm_k1z_ymmm256b32_sae", Code.EVEX_Vcvttps2qq_zmm_k1z_ymmm256b32_sae },
				{ "EVEX_Vcvttpd2qq_xmm_k1z_xmmm128b64", Code.EVEX_Vcvttpd2qq_xmm_k1z_xmmm128b64 },
				{ "EVEX_Vcvttpd2qq_ymm_k1z_ymmm256b64", Code.EVEX_Vcvttpd2qq_ymm_k1z_ymmm256b64 },
				{ "EVEX_Vcvttpd2qq_zmm_k1z_zmmm512b64_sae", Code.EVEX_Vcvttpd2qq_zmm_k1z_zmmm512b64_sae },
				{ "EVEX_Vcvtudq2pd_xmm_k1z_xmmm64b32", Code.EVEX_Vcvtudq2pd_xmm_k1z_xmmm64b32 },
				{ "EVEX_Vcvtudq2pd_ymm_k1z_xmmm128b32", Code.EVEX_Vcvtudq2pd_ymm_k1z_xmmm128b32 },
				{ "EVEX_Vcvtudq2pd_zmm_k1z_ymmm256b32_er", Code.EVEX_Vcvtudq2pd_zmm_k1z_ymmm256b32_er },
				{ "EVEX_Vcvtuqq2pd_xmm_k1z_xmmm128b64", Code.EVEX_Vcvtuqq2pd_xmm_k1z_xmmm128b64 },
				{ "EVEX_Vcvtuqq2pd_ymm_k1z_ymmm256b64", Code.EVEX_Vcvtuqq2pd_ymm_k1z_ymmm256b64 },
				{ "EVEX_Vcvtuqq2pd_zmm_k1z_zmmm512b64_er", Code.EVEX_Vcvtuqq2pd_zmm_k1z_zmmm512b64_er },
				{ "EVEX_Vcvtudq2ps_xmm_k1z_xmmm128b32", Code.EVEX_Vcvtudq2ps_xmm_k1z_xmmm128b32 },
				{ "EVEX_Vcvtudq2ps_ymm_k1z_ymmm256b32", Code.EVEX_Vcvtudq2ps_ymm_k1z_ymmm256b32 },
				{ "EVEX_Vcvtudq2ps_zmm_k1z_zmmm512b32_er", Code.EVEX_Vcvtudq2ps_zmm_k1z_zmmm512b32_er },
				{ "EVEX_Vcvtuqq2ps_xmm_k1z_xmmm128b64", Code.EVEX_Vcvtuqq2ps_xmm_k1z_xmmm128b64 },
				{ "EVEX_Vcvtuqq2ps_xmm_k1z_ymmm256b64", Code.EVEX_Vcvtuqq2ps_xmm_k1z_ymmm256b64 },
				{ "EVEX_Vcvtuqq2ps_ymm_k1z_zmmm512b64_er", Code.EVEX_Vcvtuqq2ps_ymm_k1z_zmmm512b64_er },
				{ "EVEX_Vcvtps2qq_xmm_k1z_xmmm64b32", Code.EVEX_Vcvtps2qq_xmm_k1z_xmmm64b32 },
				{ "EVEX_Vcvtps2qq_ymm_k1z_xmmm128b32", Code.EVEX_Vcvtps2qq_ymm_k1z_xmmm128b32 },
				{ "EVEX_Vcvtps2qq_zmm_k1z_ymmm256b32_er", Code.EVEX_Vcvtps2qq_zmm_k1z_ymmm256b32_er },
				{ "EVEX_Vcvtpd2qq_xmm_k1z_xmmm128b64", Code.EVEX_Vcvtpd2qq_xmm_k1z_xmmm128b64 },
				{ "EVEX_Vcvtpd2qq_ymm_k1z_ymmm256b64", Code.EVEX_Vcvtpd2qq_ymm_k1z_ymmm256b64 },
				{ "EVEX_Vcvtpd2qq_zmm_k1z_zmmm512b64_er", Code.EVEX_Vcvtpd2qq_zmm_k1z_zmmm512b64_er },
				{ "EVEX_Vcvtusi2ss_xmm_xmm_rm32_er", Code.EVEX_Vcvtusi2ss_xmm_xmm_rm32_er },
				{ "EVEX_Vcvtusi2ss_xmm_xmm_rm64_er", Code.EVEX_Vcvtusi2ss_xmm_xmm_rm64_er },
				{ "EVEX_Vcvtusi2sd_xmm_xmm_rm32_er", Code.EVEX_Vcvtusi2sd_xmm_xmm_rm32_er },
				{ "EVEX_Vcvtusi2sd_xmm_xmm_rm64_er", Code.EVEX_Vcvtusi2sd_xmm_xmm_rm64_er },
				{ "Haddpd_xmm_xmmm128", Code.Haddpd_xmm_xmmm128 },
				{ "VEX_Vhaddpd_xmm_xmm_xmmm128", Code.VEX_Vhaddpd_xmm_xmm_xmmm128 },
				{ "VEX_Vhaddpd_ymm_ymm_ymmm256", Code.VEX_Vhaddpd_ymm_ymm_ymmm256 },
				{ "Haddps_xmm_xmmm128", Code.Haddps_xmm_xmmm128 },
				{ "VEX_Vhaddps_xmm_xmm_xmmm128", Code.VEX_Vhaddps_xmm_xmm_xmmm128 },
				{ "VEX_Vhaddps_ymm_ymm_ymmm256", Code.VEX_Vhaddps_ymm_ymm_ymmm256 },
				{ "Hsubpd_xmm_xmmm128", Code.Hsubpd_xmm_xmmm128 },
				{ "VEX_Vhsubpd_xmm_xmm_xmmm128", Code.VEX_Vhsubpd_xmm_xmm_xmmm128 },
				{ "VEX_Vhsubpd_ymm_ymm_ymmm256", Code.VEX_Vhsubpd_ymm_ymm_ymmm256 },
				{ "Hsubps_xmm_xmmm128", Code.Hsubps_xmm_xmmm128 },
				{ "VEX_Vhsubps_xmm_xmm_xmmm128", Code.VEX_Vhsubps_xmm_xmm_xmmm128 },
				{ "VEX_Vhsubps_ymm_ymm_ymmm256", Code.VEX_Vhsubps_ymm_ymm_ymmm256 },
				{ "Movd_rm32_mm", Code.Movd_rm32_mm },
				{ "Movq_rm64_mm", Code.Movq_rm64_mm },
				{ "Movd_rm32_xmm", Code.Movd_rm32_xmm },
				{ "Movq_rm64_xmm", Code.Movq_rm64_xmm },
				{ "VEX_Vmovd_rm32_xmm", Code.VEX_Vmovd_rm32_xmm },
				{ "VEX_Vmovq_rm64_xmm", Code.VEX_Vmovq_rm64_xmm },
				{ "EVEX_Vmovd_rm32_xmm", Code.EVEX_Vmovd_rm32_xmm },
				{ "EVEX_Vmovq_rm64_xmm", Code.EVEX_Vmovq_rm64_xmm },
				{ "Movq_xmm_xmmm64", Code.Movq_xmm_xmmm64 },
				{ "VEX_Vmovq_xmm_xmmm64", Code.VEX_Vmovq_xmm_xmmm64 },
				{ "EVEX_Vmovq_xmm_xmmm64", Code.EVEX_Vmovq_xmm_xmmm64 },
				{ "Movq_mmm64_mm", Code.Movq_mmm64_mm },
				{ "Movdqa_xmmm128_xmm", Code.Movdqa_xmmm128_xmm },
				{ "VEX_Vmovdqa_xmmm128_xmm", Code.VEX_Vmovdqa_xmmm128_xmm },
				{ "VEX_Vmovdqa_ymmm256_ymm", Code.VEX_Vmovdqa_ymmm256_ymm },
				{ "EVEX_Vmovdqa32_xmmm128_k1z_xmm", Code.EVEX_Vmovdqa32_xmmm128_k1z_xmm },
				{ "EVEX_Vmovdqa32_ymmm256_k1z_ymm", Code.EVEX_Vmovdqa32_ymmm256_k1z_ymm },
				{ "EVEX_Vmovdqa32_zmmm512_k1z_zmm", Code.EVEX_Vmovdqa32_zmmm512_k1z_zmm },
				{ "EVEX_Vmovdqa64_xmmm128_k1z_xmm", Code.EVEX_Vmovdqa64_xmmm128_k1z_xmm },
				{ "EVEX_Vmovdqa64_ymmm256_k1z_ymm", Code.EVEX_Vmovdqa64_ymmm256_k1z_ymm },
				{ "EVEX_Vmovdqa64_zmmm512_k1z_zmm", Code.EVEX_Vmovdqa64_zmmm512_k1z_zmm },
				{ "Movdqu_xmmm128_xmm", Code.Movdqu_xmmm128_xmm },
				{ "VEX_Vmovdqu_xmmm128_xmm", Code.VEX_Vmovdqu_xmmm128_xmm },
				{ "VEX_Vmovdqu_ymmm256_ymm", Code.VEX_Vmovdqu_ymmm256_ymm },
				{ "EVEX_Vmovdqu32_xmmm128_k1z_xmm", Code.EVEX_Vmovdqu32_xmmm128_k1z_xmm },
				{ "EVEX_Vmovdqu32_ymmm256_k1z_ymm", Code.EVEX_Vmovdqu32_ymmm256_k1z_ymm },
				{ "EVEX_Vmovdqu32_zmmm512_k1z_zmm", Code.EVEX_Vmovdqu32_zmmm512_k1z_zmm },
				{ "EVEX_Vmovdqu64_xmmm128_k1z_xmm", Code.EVEX_Vmovdqu64_xmmm128_k1z_xmm },
				{ "EVEX_Vmovdqu64_ymmm256_k1z_ymm", Code.EVEX_Vmovdqu64_ymmm256_k1z_ymm },
				{ "EVEX_Vmovdqu64_zmmm512_k1z_zmm", Code.EVEX_Vmovdqu64_zmmm512_k1z_zmm },
				{ "EVEX_Vmovdqu8_xmmm128_k1z_xmm", Code.EVEX_Vmovdqu8_xmmm128_k1z_xmm },
				{ "EVEX_Vmovdqu8_ymmm256_k1z_ymm", Code.EVEX_Vmovdqu8_ymmm256_k1z_ymm },
				{ "EVEX_Vmovdqu8_zmmm512_k1z_zmm", Code.EVEX_Vmovdqu8_zmmm512_k1z_zmm },
				{ "EVEX_Vmovdqu16_xmmm128_k1z_xmm", Code.EVEX_Vmovdqu16_xmmm128_k1z_xmm },
				{ "EVEX_Vmovdqu16_ymmm256_k1z_ymm", Code.EVEX_Vmovdqu16_ymmm256_k1z_ymm },
				{ "EVEX_Vmovdqu16_zmmm512_k1z_zmm", Code.EVEX_Vmovdqu16_zmmm512_k1z_zmm },
				{ "Jo_rel16", Code.Jo_rel16 },
				{ "Jo_rel32_32", Code.Jo_rel32_32 },
				{ "Jo_rel32_64", Code.Jo_rel32_64 },
				{ "Jno_rel16", Code.Jno_rel16 },
				{ "Jno_rel32_32", Code.Jno_rel32_32 },
				{ "Jno_rel32_64", Code.Jno_rel32_64 },
				{ "Jb_rel16", Code.Jb_rel16 },
				{ "Jb_rel32_32", Code.Jb_rel32_32 },
				{ "Jb_rel32_64", Code.Jb_rel32_64 },
				{ "Jae_rel16", Code.Jae_rel16 },
				{ "Jae_rel32_32", Code.Jae_rel32_32 },
				{ "Jae_rel32_64", Code.Jae_rel32_64 },
				{ "Je_rel16", Code.Je_rel16 },
				{ "Je_rel32_32", Code.Je_rel32_32 },
				{ "Je_rel32_64", Code.Je_rel32_64 },
				{ "Jne_rel16", Code.Jne_rel16 },
				{ "Jne_rel32_32", Code.Jne_rel32_32 },
				{ "Jne_rel32_64", Code.Jne_rel32_64 },
				{ "Jbe_rel16", Code.Jbe_rel16 },
				{ "Jbe_rel32_32", Code.Jbe_rel32_32 },
				{ "Jbe_rel32_64", Code.Jbe_rel32_64 },
				{ "Ja_rel16", Code.Ja_rel16 },
				{ "Ja_rel32_32", Code.Ja_rel32_32 },
				{ "Ja_rel32_64", Code.Ja_rel32_64 },
				{ "Js_rel16", Code.Js_rel16 },
				{ "Js_rel32_32", Code.Js_rel32_32 },
				{ "Js_rel32_64", Code.Js_rel32_64 },
				{ "Jns_rel16", Code.Jns_rel16 },
				{ "Jns_rel32_32", Code.Jns_rel32_32 },
				{ "Jns_rel32_64", Code.Jns_rel32_64 },
				{ "Jp_rel16", Code.Jp_rel16 },
				{ "Jp_rel32_32", Code.Jp_rel32_32 },
				{ "Jp_rel32_64", Code.Jp_rel32_64 },
				{ "Jnp_rel16", Code.Jnp_rel16 },
				{ "Jnp_rel32_32", Code.Jnp_rel32_32 },
				{ "Jnp_rel32_64", Code.Jnp_rel32_64 },
				{ "Jl_rel16", Code.Jl_rel16 },
				{ "Jl_rel32_32", Code.Jl_rel32_32 },
				{ "Jl_rel32_64", Code.Jl_rel32_64 },
				{ "Jge_rel16", Code.Jge_rel16 },
				{ "Jge_rel32_32", Code.Jge_rel32_32 },
				{ "Jge_rel32_64", Code.Jge_rel32_64 },
				{ "Jle_rel16", Code.Jle_rel16 },
				{ "Jle_rel32_32", Code.Jle_rel32_32 },
				{ "Jle_rel32_64", Code.Jle_rel32_64 },
				{ "Jg_rel16", Code.Jg_rel16 },
				{ "Jg_rel32_32", Code.Jg_rel32_32 },
				{ "Jg_rel32_64", Code.Jg_rel32_64 },
				{ "Seto_rm8", Code.Seto_rm8 },
				{ "Setno_rm8", Code.Setno_rm8 },
				{ "Setb_rm8", Code.Setb_rm8 },
				{ "Setae_rm8", Code.Setae_rm8 },
				{ "Sete_rm8", Code.Sete_rm8 },
				{ "Setne_rm8", Code.Setne_rm8 },
				{ "Setbe_rm8", Code.Setbe_rm8 },
				{ "Seta_rm8", Code.Seta_rm8 },
				{ "Sets_rm8", Code.Sets_rm8 },
				{ "Setns_rm8", Code.Setns_rm8 },
				{ "Setp_rm8", Code.Setp_rm8 },
				{ "Setnp_rm8", Code.Setnp_rm8 },
				{ "Setl_rm8", Code.Setl_rm8 },
				{ "Setge_rm8", Code.Setge_rm8 },
				{ "Setle_rm8", Code.Setle_rm8 },
				{ "Setg_rm8", Code.Setg_rm8 },
				{ "VEX_Kmovw_kr_km16", Code.VEX_Kmovw_kr_km16 },
				{ "VEX_Kmovq_kr_km64", Code.VEX_Kmovq_kr_km64 },
				{ "VEX_Kmovb_kr_km8", Code.VEX_Kmovb_kr_km8 },
				{ "VEX_Kmovd_kr_km32", Code.VEX_Kmovd_kr_km32 },
				{ "VEX_Kmovw_m16_kr", Code.VEX_Kmovw_m16_kr },
				{ "VEX_Kmovq_m64_kr", Code.VEX_Kmovq_m64_kr },
				{ "VEX_Kmovb_m8_kr", Code.VEX_Kmovb_m8_kr },
				{ "VEX_Kmovd_m32_kr", Code.VEX_Kmovd_m32_kr },
				{ "VEX_Kmovw_kr_r32", Code.VEX_Kmovw_kr_r32 },
				{ "VEX_Kmovb_kr_r32", Code.VEX_Kmovb_kr_r32 },
				{ "VEX_Kmovd_kr_r32", Code.VEX_Kmovd_kr_r32 },
				{ "VEX_Kmovq_kr_r64", Code.VEX_Kmovq_kr_r64 },
				{ "VEX_Kmovw_r32_kr", Code.VEX_Kmovw_r32_kr },
				{ "VEX_Kmovb_r32_kr", Code.VEX_Kmovb_r32_kr },
				{ "VEX_Kmovd_r32_kr", Code.VEX_Kmovd_r32_kr },
				{ "VEX_Kmovq_r64_kr", Code.VEX_Kmovq_r64_kr },
				{ "VEX_Kortestw_kr_kr", Code.VEX_Kortestw_kr_kr },
				{ "VEX_Kortestq_kr_kr", Code.VEX_Kortestq_kr_kr },
				{ "VEX_Kortestb_kr_kr", Code.VEX_Kortestb_kr_kr },
				{ "VEX_Kortestd_kr_kr", Code.VEX_Kortestd_kr_kr },
				{ "VEX_Ktestw_kr_kr", Code.VEX_Ktestw_kr_kr },
				{ "VEX_Ktestq_kr_kr", Code.VEX_Ktestq_kr_kr },
				{ "VEX_Ktestb_kr_kr", Code.VEX_Ktestb_kr_kr },
				{ "VEX_Ktestd_kr_kr", Code.VEX_Ktestd_kr_kr },
				{ "Pushw_FS", Code.Pushw_FS },
				{ "Pushd_FS", Code.Pushd_FS },
				{ "Pushq_FS", Code.Pushq_FS },
				{ "Popw_FS", Code.Popw_FS },
				{ "Popd_FS", Code.Popd_FS },
				{ "Popq_FS", Code.Popq_FS },
				{ "Cpuid", Code.Cpuid },
				{ "Bt_rm16_r16", Code.Bt_rm16_r16 },
				{ "Bt_rm32_r32", Code.Bt_rm32_r32 },
				{ "Bt_rm64_r64", Code.Bt_rm64_r64 },
				{ "Shld_rm16_r16_imm8", Code.Shld_rm16_r16_imm8 },
				{ "Shld_rm32_r32_imm8", Code.Shld_rm32_r32_imm8 },
				{ "Shld_rm64_r64_imm8", Code.Shld_rm64_r64_imm8 },
				{ "Shld_rm16_r16_CL", Code.Shld_rm16_r16_CL },
				{ "Shld_rm32_r32_CL", Code.Shld_rm32_r32_CL },
				{ "Shld_rm64_r64_CL", Code.Shld_rm64_r64_CL },
				{ "Montmul_16", Code.Montmul_16 },
				{ "Montmul_32", Code.Montmul_32 },
				{ "Montmul_64", Code.Montmul_64 },
				{ "Xsha1_16", Code.Xsha1_16 },
				{ "Xsha1_32", Code.Xsha1_32 },
				{ "Xsha1_64", Code.Xsha1_64 },
				{ "Xsha256_16", Code.Xsha256_16 },
				{ "Xsha256_32", Code.Xsha256_32 },
				{ "Xsha256_64", Code.Xsha256_64 },
				{ "Xbts_r16_rm16", Code.Xbts_r16_rm16 },
				{ "Xbts_r32_rm32", Code.Xbts_r32_rm32 },
				{ "Xstore_16", Code.Xstore_16 },
				{ "Xstore_32", Code.Xstore_32 },
				{ "Xstore_64", Code.Xstore_64 },
				{ "Xcryptecb_16", Code.Xcryptecb_16 },
				{ "Xcryptecb_32", Code.Xcryptecb_32 },
				{ "Xcryptecb_64", Code.Xcryptecb_64 },
				{ "Xcryptcbc_16", Code.Xcryptcbc_16 },
				{ "Xcryptcbc_32", Code.Xcryptcbc_32 },
				{ "Xcryptcbc_64", Code.Xcryptcbc_64 },
				{ "Xcryptctr_16", Code.Xcryptctr_16 },
				{ "Xcryptctr_32", Code.Xcryptctr_32 },
				{ "Xcryptctr_64", Code.Xcryptctr_64 },
				{ "Xcryptcfb_16", Code.Xcryptcfb_16 },
				{ "Xcryptcfb_32", Code.Xcryptcfb_32 },
				{ "Xcryptcfb_64", Code.Xcryptcfb_64 },
				{ "Xcryptofb_16", Code.Xcryptofb_16 },
				{ "Xcryptofb_32", Code.Xcryptofb_32 },
				{ "Xcryptofb_64", Code.Xcryptofb_64 },
				{ "Ibts_rm16_r16", Code.Ibts_rm16_r16 },
				{ "Ibts_rm32_r32", Code.Ibts_rm32_r32 },
				{ "Cmpxchg486_rm8_r8", Code.Cmpxchg486_rm8_r8 },
				{ "Cmpxchg486_rm16_r16", Code.Cmpxchg486_rm16_r16 },
				{ "Cmpxchg486_rm32_r32", Code.Cmpxchg486_rm32_r32 },
				{ "Pushw_GS", Code.Pushw_GS },
				{ "Pushd_GS", Code.Pushd_GS },
				{ "Pushq_GS", Code.Pushq_GS },
				{ "Popw_GS", Code.Popw_GS },
				{ "Popd_GS", Code.Popd_GS },
				{ "Popq_GS", Code.Popq_GS },
				{ "Rsm", Code.Rsm },
				{ "Bts_rm16_r16", Code.Bts_rm16_r16 },
				{ "Bts_rm32_r32", Code.Bts_rm32_r32 },
				{ "Bts_rm64_r64", Code.Bts_rm64_r64 },
				{ "Shrd_rm16_r16_imm8", Code.Shrd_rm16_r16_imm8 },
				{ "Shrd_rm32_r32_imm8", Code.Shrd_rm32_r32_imm8 },
				{ "Shrd_rm64_r64_imm8", Code.Shrd_rm64_r64_imm8 },
				{ "Shrd_rm16_r16_CL", Code.Shrd_rm16_r16_CL },
				{ "Shrd_rm32_r32_CL", Code.Shrd_rm32_r32_CL },
				{ "Shrd_rm64_r64_CL", Code.Shrd_rm64_r64_CL },
				{ "Fxsave_m512byte", Code.Fxsave_m512byte },
				{ "Fxsave64_m512byte", Code.Fxsave64_m512byte },
				{ "Rdfsbase_r32", Code.Rdfsbase_r32 },
				{ "Rdfsbase_r64", Code.Rdfsbase_r64 },
				{ "Fxrstor_m512byte", Code.Fxrstor_m512byte },
				{ "Fxrstor64_m512byte", Code.Fxrstor64_m512byte },
				{ "Rdgsbase_r32", Code.Rdgsbase_r32 },
				{ "Rdgsbase_r64", Code.Rdgsbase_r64 },
				{ "Ldmxcsr_m32", Code.Ldmxcsr_m32 },
				{ "Wrfsbase_r32", Code.Wrfsbase_r32 },
				{ "Wrfsbase_r64", Code.Wrfsbase_r64 },
				{ "VEX_Vldmxcsr_m32", Code.VEX_Vldmxcsr_m32 },
				{ "Stmxcsr_m32", Code.Stmxcsr_m32 },
				{ "Wrgsbase_r32", Code.Wrgsbase_r32 },
				{ "Wrgsbase_r64", Code.Wrgsbase_r64 },
				{ "VEX_Vstmxcsr_m32", Code.VEX_Vstmxcsr_m32 },
				{ "Xsave_mem", Code.Xsave_mem },
				{ "Xsave64_mem", Code.Xsave64_mem },
				{ "Ptwrite_rm32", Code.Ptwrite_rm32 },
				{ "Ptwrite_rm64", Code.Ptwrite_rm64 },
				{ "Xrstor_mem", Code.Xrstor_mem },
				{ "Xrstor64_mem", Code.Xrstor64_mem },
				{ "Incsspd_r32", Code.Incsspd_r32 },
				{ "Incsspq_r64", Code.Incsspq_r64 },
				{ "Xsaveopt_mem", Code.Xsaveopt_mem },
				{ "Xsaveopt64_mem", Code.Xsaveopt64_mem },
				{ "Clwb_m8", Code.Clwb_m8 },
				{ "Tpause_r32", Code.Tpause_r32 },
				{ "Tpause_r64", Code.Tpause_r64 },
				{ "Clrssbsy_m64", Code.Clrssbsy_m64 },
				{ "Umonitor_r16", Code.Umonitor_r16 },
				{ "Umonitor_r32", Code.Umonitor_r32 },
				{ "Umonitor_r64", Code.Umonitor_r64 },
				{ "Umwait_r32", Code.Umwait_r32 },
				{ "Umwait_r64", Code.Umwait_r64 },
				{ "Clflush_m8", Code.Clflush_m8 },
				{ "Clflushopt_m8", Code.Clflushopt_m8 },
				{ "Lfence", Code.Lfence },
				{ "Lfence_E9", Code.Lfence_E9 },
				{ "Lfence_EA", Code.Lfence_EA },
				{ "Lfence_EB", Code.Lfence_EB },
				{ "Lfence_EC", Code.Lfence_EC },
				{ "Lfence_ED", Code.Lfence_ED },
				{ "Lfence_EE", Code.Lfence_EE },
				{ "Lfence_EF", Code.Lfence_EF },
				{ "Mfence", Code.Mfence },
				{ "Mfence_F1", Code.Mfence_F1 },
				{ "Mfence_F2", Code.Mfence_F2 },
				{ "Mfence_F3", Code.Mfence_F3 },
				{ "Mfence_F4", Code.Mfence_F4 },
				{ "Mfence_F5", Code.Mfence_F5 },
				{ "Mfence_F6", Code.Mfence_F6 },
				{ "Mfence_F7", Code.Mfence_F7 },
				{ "Sfence", Code.Sfence },
				{ "Sfence_F9", Code.Sfence_F9 },
				{ "Sfence_FA", Code.Sfence_FA },
				{ "Sfence_FB", Code.Sfence_FB },
				{ "Sfence_FC", Code.Sfence_FC },
				{ "Sfence_FD", Code.Sfence_FD },
				{ "Sfence_FE", Code.Sfence_FE },
				{ "Sfence_FF", Code.Sfence_FF },
				{ "Pcommit", Code.Pcommit },
				{ "Imul_r16_rm16", Code.Imul_r16_rm16 },
				{ "Imul_r32_rm32", Code.Imul_r32_rm32 },
				{ "Imul_r64_rm64", Code.Imul_r64_rm64 },
				{ "Cmpxchg_rm8_r8", Code.Cmpxchg_rm8_r8 },
				{ "Cmpxchg_rm16_r16", Code.Cmpxchg_rm16_r16 },
				{ "Cmpxchg_rm32_r32", Code.Cmpxchg_rm32_r32 },
				{ "Cmpxchg_rm64_r64", Code.Cmpxchg_rm64_r64 },
				{ "Lss_r16_m1616", Code.Lss_r16_m1616 },
				{ "Lss_r32_m1632", Code.Lss_r32_m1632 },
				{ "Lss_r64_m1664", Code.Lss_r64_m1664 },
				{ "Btr_rm16_r16", Code.Btr_rm16_r16 },
				{ "Btr_rm32_r32", Code.Btr_rm32_r32 },
				{ "Btr_rm64_r64", Code.Btr_rm64_r64 },
				{ "Lfs_r16_m1616", Code.Lfs_r16_m1616 },
				{ "Lfs_r32_m1632", Code.Lfs_r32_m1632 },
				{ "Lfs_r64_m1664", Code.Lfs_r64_m1664 },
				{ "Lgs_r16_m1616", Code.Lgs_r16_m1616 },
				{ "Lgs_r32_m1632", Code.Lgs_r32_m1632 },
				{ "Lgs_r64_m1664", Code.Lgs_r64_m1664 },
				{ "Movzx_r16_rm8", Code.Movzx_r16_rm8 },
				{ "Movzx_r32_rm8", Code.Movzx_r32_rm8 },
				{ "Movzx_r64_rm8", Code.Movzx_r64_rm8 },
				{ "Movzx_r16_rm16", Code.Movzx_r16_rm16 },
				{ "Movzx_r32_rm16", Code.Movzx_r32_rm16 },
				{ "Movzx_r64_rm16", Code.Movzx_r64_rm16 },
				{ "Jmpe_disp16", Code.Jmpe_disp16 },
				{ "Jmpe_disp32", Code.Jmpe_disp32 },
				{ "Popcnt_r16_rm16", Code.Popcnt_r16_rm16 },
				{ "Popcnt_r32_rm32", Code.Popcnt_r32_rm32 },
				{ "Popcnt_r64_rm64", Code.Popcnt_r64_rm64 },
				{ "Ud1_r16_rm16", Code.Ud1_r16_rm16 },
				{ "Ud1_r32_rm32", Code.Ud1_r32_rm32 },
				{ "Ud1_r64_rm64", Code.Ud1_r64_rm64 },
				{ "Bt_rm16_imm8", Code.Bt_rm16_imm8 },
				{ "Bt_rm32_imm8", Code.Bt_rm32_imm8 },
				{ "Bt_rm64_imm8", Code.Bt_rm64_imm8 },
				{ "Bts_rm16_imm8", Code.Bts_rm16_imm8 },
				{ "Bts_rm32_imm8", Code.Bts_rm32_imm8 },
				{ "Bts_rm64_imm8", Code.Bts_rm64_imm8 },
				{ "Btr_rm16_imm8", Code.Btr_rm16_imm8 },
				{ "Btr_rm32_imm8", Code.Btr_rm32_imm8 },
				{ "Btr_rm64_imm8", Code.Btr_rm64_imm8 },
				{ "Btc_rm16_imm8", Code.Btc_rm16_imm8 },
				{ "Btc_rm32_imm8", Code.Btc_rm32_imm8 },
				{ "Btc_rm64_imm8", Code.Btc_rm64_imm8 },
				{ "Btc_rm16_r16", Code.Btc_rm16_r16 },
				{ "Btc_rm32_r32", Code.Btc_rm32_r32 },
				{ "Btc_rm64_r64", Code.Btc_rm64_r64 },
				{ "Bsf_r16_rm16", Code.Bsf_r16_rm16 },
				{ "Bsf_r32_rm32", Code.Bsf_r32_rm32 },
				{ "Bsf_r64_rm64", Code.Bsf_r64_rm64 },
				{ "Tzcnt_r16_rm16", Code.Tzcnt_r16_rm16 },
				{ "Tzcnt_r32_rm32", Code.Tzcnt_r32_rm32 },
				{ "Tzcnt_r64_rm64", Code.Tzcnt_r64_rm64 },
				{ "Bsr_r16_rm16", Code.Bsr_r16_rm16 },
				{ "Bsr_r32_rm32", Code.Bsr_r32_rm32 },
				{ "Bsr_r64_rm64", Code.Bsr_r64_rm64 },
				{ "Lzcnt_r16_rm16", Code.Lzcnt_r16_rm16 },
				{ "Lzcnt_r32_rm32", Code.Lzcnt_r32_rm32 },
				{ "Lzcnt_r64_rm64", Code.Lzcnt_r64_rm64 },
				{ "Movsx_r16_rm8", Code.Movsx_r16_rm8 },
				{ "Movsx_r32_rm8", Code.Movsx_r32_rm8 },
				{ "Movsx_r64_rm8", Code.Movsx_r64_rm8 },
				{ "Movsx_r16_rm16", Code.Movsx_r16_rm16 },
				{ "Movsx_r32_rm16", Code.Movsx_r32_rm16 },
				{ "Movsx_r64_rm16", Code.Movsx_r64_rm16 },
				{ "Xadd_rm8_r8", Code.Xadd_rm8_r8 },
				{ "Xadd_rm16_r16", Code.Xadd_rm16_r16 },
				{ "Xadd_rm32_r32", Code.Xadd_rm32_r32 },
				{ "Xadd_rm64_r64", Code.Xadd_rm64_r64 },
				{ "Cmpps_xmm_xmmm128_imm8", Code.Cmpps_xmm_xmmm128_imm8 },
				{ "VEX_Vcmpps_xmm_xmm_xmmm128_imm8", Code.VEX_Vcmpps_xmm_xmm_xmmm128_imm8 },
				{ "VEX_Vcmpps_ymm_ymm_ymmm256_imm8", Code.VEX_Vcmpps_ymm_ymm_ymmm256_imm8 },
				{ "EVEX_Vcmpps_kr_k1_xmm_xmmm128b32_imm8", Code.EVEX_Vcmpps_kr_k1_xmm_xmmm128b32_imm8 },
				{ "EVEX_Vcmpps_kr_k1_ymm_ymmm256b32_imm8", Code.EVEX_Vcmpps_kr_k1_ymm_ymmm256b32_imm8 },
				{ "EVEX_Vcmpps_kr_k1_zmm_zmmm512b32_imm8_sae", Code.EVEX_Vcmpps_kr_k1_zmm_zmmm512b32_imm8_sae },
				{ "Cmppd_xmm_xmmm128_imm8", Code.Cmppd_xmm_xmmm128_imm8 },
				{ "VEX_Vcmppd_xmm_xmm_xmmm128_imm8", Code.VEX_Vcmppd_xmm_xmm_xmmm128_imm8 },
				{ "VEX_Vcmppd_ymm_ymm_ymmm256_imm8", Code.VEX_Vcmppd_ymm_ymm_ymmm256_imm8 },
				{ "EVEX_Vcmppd_kr_k1_xmm_xmmm128b64_imm8", Code.EVEX_Vcmppd_kr_k1_xmm_xmmm128b64_imm8 },
				{ "EVEX_Vcmppd_kr_k1_ymm_ymmm256b64_imm8", Code.EVEX_Vcmppd_kr_k1_ymm_ymmm256b64_imm8 },
				{ "EVEX_Vcmppd_kr_k1_zmm_zmmm512b64_imm8_sae", Code.EVEX_Vcmppd_kr_k1_zmm_zmmm512b64_imm8_sae },
				{ "Cmpss_xmm_xmmm32_imm8", Code.Cmpss_xmm_xmmm32_imm8 },
				{ "VEX_Vcmpss_xmm_xmm_xmmm32_imm8", Code.VEX_Vcmpss_xmm_xmm_xmmm32_imm8 },
				{ "EVEX_Vcmpss_kr_k1_xmm_xmmm32_imm8_sae", Code.EVEX_Vcmpss_kr_k1_xmm_xmmm32_imm8_sae },
				{ "Cmpsd_xmm_xmmm64_imm8", Code.Cmpsd_xmm_xmmm64_imm8 },
				{ "VEX_Vcmpsd_xmm_xmm_xmmm64_imm8", Code.VEX_Vcmpsd_xmm_xmm_xmmm64_imm8 },
				{ "EVEX_Vcmpsd_kr_k1_xmm_xmmm64_imm8_sae", Code.EVEX_Vcmpsd_kr_k1_xmm_xmmm64_imm8_sae },
				{ "Movnti_m32_r32", Code.Movnti_m32_r32 },
				{ "Movnti_m64_r64", Code.Movnti_m64_r64 },
				{ "Pinsrw_mm_r32m16_imm8", Code.Pinsrw_mm_r32m16_imm8 },
				{ "Pinsrw_mm_r64m16_imm8", Code.Pinsrw_mm_r64m16_imm8 },
				{ "Pinsrw_xmm_r32m16_imm8", Code.Pinsrw_xmm_r32m16_imm8 },
				{ "Pinsrw_xmm_r64m16_imm8", Code.Pinsrw_xmm_r64m16_imm8 },
				{ "VEX_Vpinsrw_xmm_xmm_r32m16_imm8", Code.VEX_Vpinsrw_xmm_xmm_r32m16_imm8 },
				{ "VEX_Vpinsrw_xmm_xmm_r64m16_imm8", Code.VEX_Vpinsrw_xmm_xmm_r64m16_imm8 },
				{ "EVEX_Vpinsrw_xmm_xmm_r32m16_imm8", Code.EVEX_Vpinsrw_xmm_xmm_r32m16_imm8 },
				{ "EVEX_Vpinsrw_xmm_xmm_r64m16_imm8", Code.EVEX_Vpinsrw_xmm_xmm_r64m16_imm8 },
				{ "Pextrw_r32_mm_imm8", Code.Pextrw_r32_mm_imm8 },
				{ "Pextrw_r64_mm_imm8", Code.Pextrw_r64_mm_imm8 },
				{ "Pextrw_r32_xmm_imm8", Code.Pextrw_r32_xmm_imm8 },
				{ "Pextrw_r64_xmm_imm8", Code.Pextrw_r64_xmm_imm8 },
				{ "VEX_Vpextrw_r32_xmm_imm8", Code.VEX_Vpextrw_r32_xmm_imm8 },
				{ "VEX_Vpextrw_r64_xmm_imm8", Code.VEX_Vpextrw_r64_xmm_imm8 },
				{ "EVEX_Vpextrw_r32_xmm_imm8", Code.EVEX_Vpextrw_r32_xmm_imm8 },
				{ "EVEX_Vpextrw_r64_xmm_imm8", Code.EVEX_Vpextrw_r64_xmm_imm8 },
				{ "Shufps_xmm_xmmm128_imm8", Code.Shufps_xmm_xmmm128_imm8 },
				{ "VEX_Vshufps_xmm_xmm_xmmm128_imm8", Code.VEX_Vshufps_xmm_xmm_xmmm128_imm8 },
				{ "VEX_Vshufps_ymm_ymm_ymmm256_imm8", Code.VEX_Vshufps_ymm_ymm_ymmm256_imm8 },
				{ "EVEX_Vshufps_xmm_k1z_xmm_xmmm128b32_imm8", Code.EVEX_Vshufps_xmm_k1z_xmm_xmmm128b32_imm8 },
				{ "EVEX_Vshufps_ymm_k1z_ymm_ymmm256b32_imm8", Code.EVEX_Vshufps_ymm_k1z_ymm_ymmm256b32_imm8 },
				{ "EVEX_Vshufps_zmm_k1z_zmm_zmmm512b32_imm8", Code.EVEX_Vshufps_zmm_k1z_zmm_zmmm512b32_imm8 },
				{ "Shufpd_xmm_xmmm128_imm8", Code.Shufpd_xmm_xmmm128_imm8 },
				{ "VEX_Vshufpd_xmm_xmm_xmmm128_imm8", Code.VEX_Vshufpd_xmm_xmm_xmmm128_imm8 },
				{ "VEX_Vshufpd_ymm_ymm_ymmm256_imm8", Code.VEX_Vshufpd_ymm_ymm_ymmm256_imm8 },
				{ "EVEX_Vshufpd_xmm_k1z_xmm_xmmm128b64_imm8", Code.EVEX_Vshufpd_xmm_k1z_xmm_xmmm128b64_imm8 },
				{ "EVEX_Vshufpd_ymm_k1z_ymm_ymmm256b64_imm8", Code.EVEX_Vshufpd_ymm_k1z_ymm_ymmm256b64_imm8 },
				{ "EVEX_Vshufpd_zmm_k1z_zmm_zmmm512b64_imm8", Code.EVEX_Vshufpd_zmm_k1z_zmm_zmmm512b64_imm8 },
				{ "Cmpxchg8b_m64", Code.Cmpxchg8b_m64 },
				{ "Cmpxchg16b_m128", Code.Cmpxchg16b_m128 },
				{ "Xrstors_mem", Code.Xrstors_mem },
				{ "Xrstors64_mem", Code.Xrstors64_mem },
				{ "Xsavec_mem", Code.Xsavec_mem },
				{ "Xsavec64_mem", Code.Xsavec64_mem },
				{ "Xsaves_mem", Code.Xsaves_mem },
				{ "Xsaves64_mem", Code.Xsaves64_mem },
				{ "Vmptrld_m64", Code.Vmptrld_m64 },
				{ "Vmclear_m64", Code.Vmclear_m64 },
				{ "Vmxon_m64", Code.Vmxon_m64 },
				{ "Rdrand_r16", Code.Rdrand_r16 },
				{ "Rdrand_r32", Code.Rdrand_r32 },
				{ "Rdrand_r64", Code.Rdrand_r64 },
				{ "Vmptrst_m64", Code.Vmptrst_m64 },
				{ "Rdseed_r16", Code.Rdseed_r16 },
				{ "Rdseed_r32", Code.Rdseed_r32 },
				{ "Rdseed_r64", Code.Rdseed_r64 },
				{ "Rdpid_r32", Code.Rdpid_r32 },
				{ "Rdpid_r64", Code.Rdpid_r64 },
				{ "Bswap_r16", Code.Bswap_r16 },
				{ "Bswap_r32", Code.Bswap_r32 },
				{ "Bswap_r64", Code.Bswap_r64 },
				{ "Addsubpd_xmm_xmmm128", Code.Addsubpd_xmm_xmmm128 },
				{ "VEX_Vaddsubpd_xmm_xmm_xmmm128", Code.VEX_Vaddsubpd_xmm_xmm_xmmm128 },
				{ "VEX_Vaddsubpd_ymm_ymm_ymmm256", Code.VEX_Vaddsubpd_ymm_ymm_ymmm256 },
				{ "Addsubps_xmm_xmmm128", Code.Addsubps_xmm_xmmm128 },
				{ "VEX_Vaddsubps_xmm_xmm_xmmm128", Code.VEX_Vaddsubps_xmm_xmm_xmmm128 },
				{ "VEX_Vaddsubps_ymm_ymm_ymmm256", Code.VEX_Vaddsubps_ymm_ymm_ymmm256 },
				{ "Psrlw_mm_mmm64", Code.Psrlw_mm_mmm64 },
				{ "Psrlw_xmm_xmmm128", Code.Psrlw_xmm_xmmm128 },
				{ "VEX_Vpsrlw_xmm_xmm_xmmm128", Code.VEX_Vpsrlw_xmm_xmm_xmmm128 },
				{ "VEX_Vpsrlw_ymm_ymm_xmmm128", Code.VEX_Vpsrlw_ymm_ymm_xmmm128 },
				{ "EVEX_Vpsrlw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpsrlw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpsrlw_ymm_k1z_ymm_xmmm128", Code.EVEX_Vpsrlw_ymm_k1z_ymm_xmmm128 },
				{ "EVEX_Vpsrlw_zmm_k1z_zmm_xmmm128", Code.EVEX_Vpsrlw_zmm_k1z_zmm_xmmm128 },
				{ "Psrld_mm_mmm64", Code.Psrld_mm_mmm64 },
				{ "Psrld_xmm_xmmm128", Code.Psrld_xmm_xmmm128 },
				{ "VEX_Vpsrld_xmm_xmm_xmmm128", Code.VEX_Vpsrld_xmm_xmm_xmmm128 },
				{ "VEX_Vpsrld_ymm_ymm_xmmm128", Code.VEX_Vpsrld_ymm_ymm_xmmm128 },
				{ "EVEX_Vpsrld_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpsrld_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpsrld_ymm_k1z_ymm_xmmm128", Code.EVEX_Vpsrld_ymm_k1z_ymm_xmmm128 },
				{ "EVEX_Vpsrld_zmm_k1z_zmm_xmmm128", Code.EVEX_Vpsrld_zmm_k1z_zmm_xmmm128 },
				{ "Psrlq_mm_mmm64", Code.Psrlq_mm_mmm64 },
				{ "Psrlq_xmm_xmmm128", Code.Psrlq_xmm_xmmm128 },
				{ "VEX_Vpsrlq_xmm_xmm_xmmm128", Code.VEX_Vpsrlq_xmm_xmm_xmmm128 },
				{ "VEX_Vpsrlq_ymm_ymm_xmmm128", Code.VEX_Vpsrlq_ymm_ymm_xmmm128 },
				{ "EVEX_Vpsrlq_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpsrlq_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpsrlq_ymm_k1z_ymm_xmmm128", Code.EVEX_Vpsrlq_ymm_k1z_ymm_xmmm128 },
				{ "EVEX_Vpsrlq_zmm_k1z_zmm_xmmm128", Code.EVEX_Vpsrlq_zmm_k1z_zmm_xmmm128 },
				{ "Paddq_mm_mmm64", Code.Paddq_mm_mmm64 },
				{ "Paddq_xmm_xmmm128", Code.Paddq_xmm_xmmm128 },
				{ "VEX_Vpaddq_xmm_xmm_xmmm128", Code.VEX_Vpaddq_xmm_xmm_xmmm128 },
				{ "VEX_Vpaddq_ymm_ymm_ymmm256", Code.VEX_Vpaddq_ymm_ymm_ymmm256 },
				{ "EVEX_Vpaddq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpaddq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpaddq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpaddq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpaddq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpaddq_zmm_k1z_zmm_zmmm512b64 },
				{ "Pmullw_mm_mmm64", Code.Pmullw_mm_mmm64 },
				{ "Pmullw_xmm_xmmm128", Code.Pmullw_xmm_xmmm128 },
				{ "VEX_Vpmullw_xmm_xmm_xmmm128", Code.VEX_Vpmullw_xmm_xmm_xmmm128 },
				{ "VEX_Vpmullw_ymm_ymm_ymmm256", Code.VEX_Vpmullw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpmullw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpmullw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpmullw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpmullw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpmullw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpmullw_zmm_k1z_zmm_zmmm512 },
				{ "Movq_xmmm64_xmm", Code.Movq_xmmm64_xmm },
				{ "VEX_Vmovq_xmmm64_xmm", Code.VEX_Vmovq_xmmm64_xmm },
				{ "EVEX_Vmovq_xmmm64_xmm", Code.EVEX_Vmovq_xmmm64_xmm },
				{ "Movq2dq_xmm_mm", Code.Movq2dq_xmm_mm },
				{ "Movdq2q_mm_xmm", Code.Movdq2q_mm_xmm },
				{ "Pmovmskb_r32_mm", Code.Pmovmskb_r32_mm },
				{ "Pmovmskb_r64_mm", Code.Pmovmskb_r64_mm },
				{ "Pmovmskb_r32_xmm", Code.Pmovmskb_r32_xmm },
				{ "Pmovmskb_r64_xmm", Code.Pmovmskb_r64_xmm },
				{ "VEX_Vpmovmskb_r32_xmm", Code.VEX_Vpmovmskb_r32_xmm },
				{ "VEX_Vpmovmskb_r64_xmm", Code.VEX_Vpmovmskb_r64_xmm },
				{ "VEX_Vpmovmskb_r32_ymm", Code.VEX_Vpmovmskb_r32_ymm },
				{ "VEX_Vpmovmskb_r64_ymm", Code.VEX_Vpmovmskb_r64_ymm },
				{ "Psubusb_mm_mmm64", Code.Psubusb_mm_mmm64 },
				{ "Psubusb_xmm_xmmm128", Code.Psubusb_xmm_xmmm128 },
				{ "VEX_Vpsubusb_xmm_xmm_xmmm128", Code.VEX_Vpsubusb_xmm_xmm_xmmm128 },
				{ "VEX_Vpsubusb_ymm_ymm_ymmm256", Code.VEX_Vpsubusb_ymm_ymm_ymmm256 },
				{ "EVEX_Vpsubusb_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpsubusb_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpsubusb_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpsubusb_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpsubusb_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpsubusb_zmm_k1z_zmm_zmmm512 },
				{ "Psubusw_mm_mmm64", Code.Psubusw_mm_mmm64 },
				{ "Psubusw_xmm_xmmm128", Code.Psubusw_xmm_xmmm128 },
				{ "VEX_Vpsubusw_xmm_xmm_xmmm128", Code.VEX_Vpsubusw_xmm_xmm_xmmm128 },
				{ "VEX_Vpsubusw_ymm_ymm_ymmm256", Code.VEX_Vpsubusw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpsubusw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpsubusw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpsubusw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpsubusw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpsubusw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpsubusw_zmm_k1z_zmm_zmmm512 },
				{ "Pminub_mm_mmm64", Code.Pminub_mm_mmm64 },
				{ "Pminub_xmm_xmmm128", Code.Pminub_xmm_xmmm128 },
				{ "VEX_Vpminub_xmm_xmm_xmmm128", Code.VEX_Vpminub_xmm_xmm_xmmm128 },
				{ "VEX_Vpminub_ymm_ymm_ymmm256", Code.VEX_Vpminub_ymm_ymm_ymmm256 },
				{ "EVEX_Vpminub_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpminub_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpminub_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpminub_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpminub_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpminub_zmm_k1z_zmm_zmmm512 },
				{ "Pand_mm_mmm64", Code.Pand_mm_mmm64 },
				{ "Pand_xmm_xmmm128", Code.Pand_xmm_xmmm128 },
				{ "VEX_Vpand_xmm_xmm_xmmm128", Code.VEX_Vpand_xmm_xmm_xmmm128 },
				{ "VEX_Vpand_ymm_ymm_ymmm256", Code.VEX_Vpand_ymm_ymm_ymmm256 },
				{ "EVEX_Vpandd_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpandd_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpandd_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpandd_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpandd_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpandd_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpandq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpandq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpandq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpandq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpandq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpandq_zmm_k1z_zmm_zmmm512b64 },
				{ "Paddusb_mm_mmm64", Code.Paddusb_mm_mmm64 },
				{ "Paddusb_xmm_xmmm128", Code.Paddusb_xmm_xmmm128 },
				{ "VEX_Vpaddusb_xmm_xmm_xmmm128", Code.VEX_Vpaddusb_xmm_xmm_xmmm128 },
				{ "VEX_Vpaddusb_ymm_ymm_ymmm256", Code.VEX_Vpaddusb_ymm_ymm_ymmm256 },
				{ "EVEX_Vpaddusb_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpaddusb_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpaddusb_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpaddusb_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpaddusb_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpaddusb_zmm_k1z_zmm_zmmm512 },
				{ "Paddusw_mm_mmm64", Code.Paddusw_mm_mmm64 },
				{ "Paddusw_xmm_xmmm128", Code.Paddusw_xmm_xmmm128 },
				{ "VEX_Vpaddusw_xmm_xmm_xmmm128", Code.VEX_Vpaddusw_xmm_xmm_xmmm128 },
				{ "VEX_Vpaddusw_ymm_ymm_ymmm256", Code.VEX_Vpaddusw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpaddusw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpaddusw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpaddusw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpaddusw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpaddusw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpaddusw_zmm_k1z_zmm_zmmm512 },
				{ "Pmaxub_mm_mmm64", Code.Pmaxub_mm_mmm64 },
				{ "Pmaxub_xmm_xmmm128", Code.Pmaxub_xmm_xmmm128 },
				{ "VEX_Vpmaxub_xmm_xmm_xmmm128", Code.VEX_Vpmaxub_xmm_xmm_xmmm128 },
				{ "VEX_Vpmaxub_ymm_ymm_ymmm256", Code.VEX_Vpmaxub_ymm_ymm_ymmm256 },
				{ "EVEX_Vpmaxub_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpmaxub_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpmaxub_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpmaxub_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpmaxub_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpmaxub_zmm_k1z_zmm_zmmm512 },
				{ "Pandn_mm_mmm64", Code.Pandn_mm_mmm64 },
				{ "Pandn_xmm_xmmm128", Code.Pandn_xmm_xmmm128 },
				{ "VEX_Vpandn_xmm_xmm_xmmm128", Code.VEX_Vpandn_xmm_xmm_xmmm128 },
				{ "VEX_Vpandn_ymm_ymm_ymmm256", Code.VEX_Vpandn_ymm_ymm_ymmm256 },
				{ "EVEX_Vpandnd_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpandnd_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpandnd_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpandnd_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpandnd_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpandnd_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpandnq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpandnq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpandnq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpandnq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpandnq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpandnq_zmm_k1z_zmm_zmmm512b64 },
				{ "Pavgb_mm_mmm64", Code.Pavgb_mm_mmm64 },
				{ "Pavgb_xmm_xmmm128", Code.Pavgb_xmm_xmmm128 },
				{ "VEX_Vpavgb_xmm_xmm_xmmm128", Code.VEX_Vpavgb_xmm_xmm_xmmm128 },
				{ "VEX_Vpavgb_ymm_ymm_ymmm256", Code.VEX_Vpavgb_ymm_ymm_ymmm256 },
				{ "EVEX_Vpavgb_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpavgb_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpavgb_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpavgb_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpavgb_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpavgb_zmm_k1z_zmm_zmmm512 },
				{ "Psraw_mm_mmm64", Code.Psraw_mm_mmm64 },
				{ "Psraw_xmm_xmmm128", Code.Psraw_xmm_xmmm128 },
				{ "VEX_Vpsraw_xmm_xmm_xmmm128", Code.VEX_Vpsraw_xmm_xmm_xmmm128 },
				{ "VEX_Vpsraw_ymm_ymm_xmmm128", Code.VEX_Vpsraw_ymm_ymm_xmmm128 },
				{ "EVEX_Vpsraw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpsraw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpsraw_ymm_k1z_ymm_xmmm128", Code.EVEX_Vpsraw_ymm_k1z_ymm_xmmm128 },
				{ "EVEX_Vpsraw_zmm_k1z_zmm_xmmm128", Code.EVEX_Vpsraw_zmm_k1z_zmm_xmmm128 },
				{ "Psrad_mm_mmm64", Code.Psrad_mm_mmm64 },
				{ "Psrad_xmm_xmmm128", Code.Psrad_xmm_xmmm128 },
				{ "VEX_Vpsrad_xmm_xmm_xmmm128", Code.VEX_Vpsrad_xmm_xmm_xmmm128 },
				{ "VEX_Vpsrad_ymm_ymm_xmmm128", Code.VEX_Vpsrad_ymm_ymm_xmmm128 },
				{ "EVEX_Vpsrad_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpsrad_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpsrad_ymm_k1z_ymm_xmmm128", Code.EVEX_Vpsrad_ymm_k1z_ymm_xmmm128 },
				{ "EVEX_Vpsrad_zmm_k1z_zmm_xmmm128", Code.EVEX_Vpsrad_zmm_k1z_zmm_xmmm128 },
				{ "EVEX_Vpsraq_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpsraq_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpsraq_ymm_k1z_ymm_xmmm128", Code.EVEX_Vpsraq_ymm_k1z_ymm_xmmm128 },
				{ "EVEX_Vpsraq_zmm_k1z_zmm_xmmm128", Code.EVEX_Vpsraq_zmm_k1z_zmm_xmmm128 },
				{ "Pavgw_mm_mmm64", Code.Pavgw_mm_mmm64 },
				{ "Pavgw_xmm_xmmm128", Code.Pavgw_xmm_xmmm128 },
				{ "VEX_Vpavgw_xmm_xmm_xmmm128", Code.VEX_Vpavgw_xmm_xmm_xmmm128 },
				{ "VEX_Vpavgw_ymm_ymm_ymmm256", Code.VEX_Vpavgw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpavgw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpavgw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpavgw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpavgw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpavgw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpavgw_zmm_k1z_zmm_zmmm512 },
				{ "Pmulhuw_mm_mmm64", Code.Pmulhuw_mm_mmm64 },
				{ "Pmulhuw_xmm_xmmm128", Code.Pmulhuw_xmm_xmmm128 },
				{ "VEX_Vpmulhuw_xmm_xmm_xmmm128", Code.VEX_Vpmulhuw_xmm_xmm_xmmm128 },
				{ "VEX_Vpmulhuw_ymm_ymm_ymmm256", Code.VEX_Vpmulhuw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpmulhuw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpmulhuw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpmulhuw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpmulhuw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpmulhuw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpmulhuw_zmm_k1z_zmm_zmmm512 },
				{ "Pmulhw_mm_mmm64", Code.Pmulhw_mm_mmm64 },
				{ "Pmulhw_xmm_xmmm128", Code.Pmulhw_xmm_xmmm128 },
				{ "VEX_Vpmulhw_xmm_xmm_xmmm128", Code.VEX_Vpmulhw_xmm_xmm_xmmm128 },
				{ "VEX_Vpmulhw_ymm_ymm_ymmm256", Code.VEX_Vpmulhw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpmulhw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpmulhw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpmulhw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpmulhw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpmulhw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpmulhw_zmm_k1z_zmm_zmmm512 },
				{ "Cvttpd2dq_xmm_xmmm128", Code.Cvttpd2dq_xmm_xmmm128 },
				{ "VEX_Vcvttpd2dq_xmm_xmmm128", Code.VEX_Vcvttpd2dq_xmm_xmmm128 },
				{ "VEX_Vcvttpd2dq_xmm_ymmm256", Code.VEX_Vcvttpd2dq_xmm_ymmm256 },
				{ "EVEX_Vcvttpd2dq_xmm_k1z_xmmm128b64", Code.EVEX_Vcvttpd2dq_xmm_k1z_xmmm128b64 },
				{ "EVEX_Vcvttpd2dq_xmm_k1z_ymmm256b64", Code.EVEX_Vcvttpd2dq_xmm_k1z_ymmm256b64 },
				{ "EVEX_Vcvttpd2dq_ymm_k1z_zmmm512b64_sae", Code.EVEX_Vcvttpd2dq_ymm_k1z_zmmm512b64_sae },
				{ "Cvtdq2pd_xmm_xmmm64", Code.Cvtdq2pd_xmm_xmmm64 },
				{ "VEX_Vcvtdq2pd_xmm_xmmm64", Code.VEX_Vcvtdq2pd_xmm_xmmm64 },
				{ "VEX_Vcvtdq2pd_ymm_xmmm128", Code.VEX_Vcvtdq2pd_ymm_xmmm128 },
				{ "EVEX_Vcvtdq2pd_xmm_k1z_xmmm64b32", Code.EVEX_Vcvtdq2pd_xmm_k1z_xmmm64b32 },
				{ "EVEX_Vcvtdq2pd_ymm_k1z_xmmm128b32", Code.EVEX_Vcvtdq2pd_ymm_k1z_xmmm128b32 },
				{ "EVEX_Vcvtdq2pd_zmm_k1z_ymmm256b32_er", Code.EVEX_Vcvtdq2pd_zmm_k1z_ymmm256b32_er },
				{ "EVEX_Vcvtqq2pd_xmm_k1z_xmmm128b64", Code.EVEX_Vcvtqq2pd_xmm_k1z_xmmm128b64 },
				{ "EVEX_Vcvtqq2pd_ymm_k1z_ymmm256b64", Code.EVEX_Vcvtqq2pd_ymm_k1z_ymmm256b64 },
				{ "EVEX_Vcvtqq2pd_zmm_k1z_zmmm512b64_er", Code.EVEX_Vcvtqq2pd_zmm_k1z_zmmm512b64_er },
				{ "Cvtpd2dq_xmm_xmmm128", Code.Cvtpd2dq_xmm_xmmm128 },
				{ "VEX_Vcvtpd2dq_xmm_xmmm128", Code.VEX_Vcvtpd2dq_xmm_xmmm128 },
				{ "VEX_Vcvtpd2dq_xmm_ymmm256", Code.VEX_Vcvtpd2dq_xmm_ymmm256 },
				{ "EVEX_Vcvtpd2dq_xmm_k1z_xmmm128b64", Code.EVEX_Vcvtpd2dq_xmm_k1z_xmmm128b64 },
				{ "EVEX_Vcvtpd2dq_xmm_k1z_ymmm256b64", Code.EVEX_Vcvtpd2dq_xmm_k1z_ymmm256b64 },
				{ "EVEX_Vcvtpd2dq_ymm_k1z_zmmm512b64_er", Code.EVEX_Vcvtpd2dq_ymm_k1z_zmmm512b64_er },
				{ "Movntq_m64_mm", Code.Movntq_m64_mm },
				{ "Movntdq_m128_xmm", Code.Movntdq_m128_xmm },
				{ "VEX_Vmovntdq_m128_xmm", Code.VEX_Vmovntdq_m128_xmm },
				{ "VEX_Vmovntdq_m256_ymm", Code.VEX_Vmovntdq_m256_ymm },
				{ "EVEX_Vmovntdq_m128_xmm", Code.EVEX_Vmovntdq_m128_xmm },
				{ "EVEX_Vmovntdq_m256_ymm", Code.EVEX_Vmovntdq_m256_ymm },
				{ "EVEX_Vmovntdq_m512_zmm", Code.EVEX_Vmovntdq_m512_zmm },
				{ "Psubsb_mm_mmm64", Code.Psubsb_mm_mmm64 },
				{ "Psubsb_xmm_xmmm128", Code.Psubsb_xmm_xmmm128 },
				{ "VEX_Vpsubsb_xmm_xmm_xmmm128", Code.VEX_Vpsubsb_xmm_xmm_xmmm128 },
				{ "VEX_Vpsubsb_ymm_ymm_ymmm256", Code.VEX_Vpsubsb_ymm_ymm_ymmm256 },
				{ "EVEX_Vpsubsb_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpsubsb_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpsubsb_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpsubsb_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpsubsb_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpsubsb_zmm_k1z_zmm_zmmm512 },
				{ "Psubsw_mm_mmm64", Code.Psubsw_mm_mmm64 },
				{ "Psubsw_xmm_xmmm128", Code.Psubsw_xmm_xmmm128 },
				{ "VEX_Vpsubsw_xmm_xmm_xmmm128", Code.VEX_Vpsubsw_xmm_xmm_xmmm128 },
				{ "VEX_Vpsubsw_ymm_ymm_ymmm256", Code.VEX_Vpsubsw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpsubsw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpsubsw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpsubsw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpsubsw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpsubsw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpsubsw_zmm_k1z_zmm_zmmm512 },
				{ "Pminsw_mm_mmm64", Code.Pminsw_mm_mmm64 },
				{ "Pminsw_xmm_xmmm128", Code.Pminsw_xmm_xmmm128 },
				{ "VEX_Vpminsw_xmm_xmm_xmmm128", Code.VEX_Vpminsw_xmm_xmm_xmmm128 },
				{ "VEX_Vpminsw_ymm_ymm_ymmm256", Code.VEX_Vpminsw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpminsw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpminsw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpminsw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpminsw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpminsw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpminsw_zmm_k1z_zmm_zmmm512 },
				{ "Por_mm_mmm64", Code.Por_mm_mmm64 },
				{ "Por_xmm_xmmm128", Code.Por_xmm_xmmm128 },
				{ "VEX_Vpor_xmm_xmm_xmmm128", Code.VEX_Vpor_xmm_xmm_xmmm128 },
				{ "VEX_Vpor_ymm_ymm_ymmm256", Code.VEX_Vpor_ymm_ymm_ymmm256 },
				{ "EVEX_Vpord_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpord_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpord_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpord_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpord_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpord_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vporq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vporq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vporq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vporq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vporq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vporq_zmm_k1z_zmm_zmmm512b64 },
				{ "Paddsb_mm_mmm64", Code.Paddsb_mm_mmm64 },
				{ "Paddsb_xmm_xmmm128", Code.Paddsb_xmm_xmmm128 },
				{ "VEX_Vpaddsb_xmm_xmm_xmmm128", Code.VEX_Vpaddsb_xmm_xmm_xmmm128 },
				{ "VEX_Vpaddsb_ymm_ymm_ymmm256", Code.VEX_Vpaddsb_ymm_ymm_ymmm256 },
				{ "EVEX_Vpaddsb_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpaddsb_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpaddsb_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpaddsb_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpaddsb_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpaddsb_zmm_k1z_zmm_zmmm512 },
				{ "Paddsw_mm_mmm64", Code.Paddsw_mm_mmm64 },
				{ "Paddsw_xmm_xmmm128", Code.Paddsw_xmm_xmmm128 },
				{ "VEX_Vpaddsw_xmm_xmm_xmmm128", Code.VEX_Vpaddsw_xmm_xmm_xmmm128 },
				{ "VEX_Vpaddsw_ymm_ymm_ymmm256", Code.VEX_Vpaddsw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpaddsw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpaddsw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpaddsw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpaddsw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpaddsw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpaddsw_zmm_k1z_zmm_zmmm512 },
				{ "Pmaxsw_mm_mmm64", Code.Pmaxsw_mm_mmm64 },
				{ "Pmaxsw_xmm_xmmm128", Code.Pmaxsw_xmm_xmmm128 },
				{ "VEX_Vpmaxsw_xmm_xmm_xmmm128", Code.VEX_Vpmaxsw_xmm_xmm_xmmm128 },
				{ "VEX_Vpmaxsw_ymm_ymm_ymmm256", Code.VEX_Vpmaxsw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpmaxsw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpmaxsw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpmaxsw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpmaxsw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpmaxsw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpmaxsw_zmm_k1z_zmm_zmmm512 },
				{ "Pxor_mm_mmm64", Code.Pxor_mm_mmm64 },
				{ "Pxor_xmm_xmmm128", Code.Pxor_xmm_xmmm128 },
				{ "VEX_Vpxor_xmm_xmm_xmmm128", Code.VEX_Vpxor_xmm_xmm_xmmm128 },
				{ "VEX_Vpxor_ymm_ymm_ymmm256", Code.VEX_Vpxor_ymm_ymm_ymmm256 },
				{ "EVEX_Vpxord_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpxord_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpxord_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpxord_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpxord_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpxord_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpxorq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpxorq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpxorq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpxorq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpxorq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpxorq_zmm_k1z_zmm_zmmm512b64 },
				{ "Lddqu_xmm_m128", Code.Lddqu_xmm_m128 },
				{ "VEX_Vlddqu_xmm_m128", Code.VEX_Vlddqu_xmm_m128 },
				{ "VEX_Vlddqu_ymm_m256", Code.VEX_Vlddqu_ymm_m256 },
				{ "Psllw_mm_mmm64", Code.Psllw_mm_mmm64 },
				{ "Psllw_xmm_xmmm128", Code.Psllw_xmm_xmmm128 },
				{ "VEX_Vpsllw_xmm_xmm_xmmm128", Code.VEX_Vpsllw_xmm_xmm_xmmm128 },
				{ "VEX_Vpsllw_ymm_ymm_xmmm128", Code.VEX_Vpsllw_ymm_ymm_xmmm128 },
				{ "EVEX_Vpsllw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpsllw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpsllw_ymm_k1z_ymm_xmmm128", Code.EVEX_Vpsllw_ymm_k1z_ymm_xmmm128 },
				{ "EVEX_Vpsllw_zmm_k1z_zmm_xmmm128", Code.EVEX_Vpsllw_zmm_k1z_zmm_xmmm128 },
				{ "Pslld_mm_mmm64", Code.Pslld_mm_mmm64 },
				{ "Pslld_xmm_xmmm128", Code.Pslld_xmm_xmmm128 },
				{ "VEX_Vpslld_xmm_xmm_xmmm128", Code.VEX_Vpslld_xmm_xmm_xmmm128 },
				{ "VEX_Vpslld_ymm_ymm_xmmm128", Code.VEX_Vpslld_ymm_ymm_xmmm128 },
				{ "EVEX_Vpslld_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpslld_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpslld_ymm_k1z_ymm_xmmm128", Code.EVEX_Vpslld_ymm_k1z_ymm_xmmm128 },
				{ "EVEX_Vpslld_zmm_k1z_zmm_xmmm128", Code.EVEX_Vpslld_zmm_k1z_zmm_xmmm128 },
				{ "Psllq_mm_mmm64", Code.Psllq_mm_mmm64 },
				{ "Psllq_xmm_xmmm128", Code.Psllq_xmm_xmmm128 },
				{ "VEX_Vpsllq_xmm_xmm_xmmm128", Code.VEX_Vpsllq_xmm_xmm_xmmm128 },
				{ "VEX_Vpsllq_ymm_ymm_xmmm128", Code.VEX_Vpsllq_ymm_ymm_xmmm128 },
				{ "EVEX_Vpsllq_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpsllq_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpsllq_ymm_k1z_ymm_xmmm128", Code.EVEX_Vpsllq_ymm_k1z_ymm_xmmm128 },
				{ "EVEX_Vpsllq_zmm_k1z_zmm_xmmm128", Code.EVEX_Vpsllq_zmm_k1z_zmm_xmmm128 },
				{ "Pmuludq_mm_mmm64", Code.Pmuludq_mm_mmm64 },
				{ "Pmuludq_xmm_xmmm128", Code.Pmuludq_xmm_xmmm128 },
				{ "VEX_Vpmuludq_xmm_xmm_xmmm128", Code.VEX_Vpmuludq_xmm_xmm_xmmm128 },
				{ "VEX_Vpmuludq_ymm_ymm_ymmm256", Code.VEX_Vpmuludq_ymm_ymm_ymmm256 },
				{ "EVEX_Vpmuludq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpmuludq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpmuludq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpmuludq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpmuludq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpmuludq_zmm_k1z_zmm_zmmm512b64 },
				{ "Pmaddwd_mm_mmm64", Code.Pmaddwd_mm_mmm64 },
				{ "Pmaddwd_xmm_xmmm128", Code.Pmaddwd_xmm_xmmm128 },
				{ "VEX_Vpmaddwd_xmm_xmm_xmmm128", Code.VEX_Vpmaddwd_xmm_xmm_xmmm128 },
				{ "VEX_Vpmaddwd_ymm_ymm_ymmm256", Code.VEX_Vpmaddwd_ymm_ymm_ymmm256 },
				{ "EVEX_Vpmaddwd_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpmaddwd_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpmaddwd_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpmaddwd_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpmaddwd_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpmaddwd_zmm_k1z_zmm_zmmm512 },
				{ "Psadbw_mm_mmm64", Code.Psadbw_mm_mmm64 },
				{ "Psadbw_xmm_xmmm128", Code.Psadbw_xmm_xmmm128 },
				{ "VEX_Vpsadbw_xmm_xmm_xmmm128", Code.VEX_Vpsadbw_xmm_xmm_xmmm128 },
				{ "VEX_Vpsadbw_ymm_ymm_ymmm256", Code.VEX_Vpsadbw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpsadbw_xmm_xmm_xmmm128", Code.EVEX_Vpsadbw_xmm_xmm_xmmm128 },
				{ "EVEX_Vpsadbw_ymm_ymm_ymmm256", Code.EVEX_Vpsadbw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpsadbw_zmm_zmm_zmmm512", Code.EVEX_Vpsadbw_zmm_zmm_zmmm512 },
				{ "Maskmovq_rDI_mm_mm", Code.Maskmovq_rDI_mm_mm },
				{ "Maskmovdqu_rDI_xmm_xmm", Code.Maskmovdqu_rDI_xmm_xmm },
				{ "VEX_Vmaskmovdqu_rDI_xmm_xmm", Code.VEX_Vmaskmovdqu_rDI_xmm_xmm },
				{ "Psubb_mm_mmm64", Code.Psubb_mm_mmm64 },
				{ "Psubb_xmm_xmmm128", Code.Psubb_xmm_xmmm128 },
				{ "VEX_Vpsubb_xmm_xmm_xmmm128", Code.VEX_Vpsubb_xmm_xmm_xmmm128 },
				{ "VEX_Vpsubb_ymm_ymm_ymmm256", Code.VEX_Vpsubb_ymm_ymm_ymmm256 },
				{ "EVEX_Vpsubb_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpsubb_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpsubb_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpsubb_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpsubb_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpsubb_zmm_k1z_zmm_zmmm512 },
				{ "Psubw_mm_mmm64", Code.Psubw_mm_mmm64 },
				{ "Psubw_xmm_xmmm128", Code.Psubw_xmm_xmmm128 },
				{ "VEX_Vpsubw_xmm_xmm_xmmm128", Code.VEX_Vpsubw_xmm_xmm_xmmm128 },
				{ "VEX_Vpsubw_ymm_ymm_ymmm256", Code.VEX_Vpsubw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpsubw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpsubw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpsubw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpsubw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpsubw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpsubw_zmm_k1z_zmm_zmmm512 },
				{ "Psubd_mm_mmm64", Code.Psubd_mm_mmm64 },
				{ "Psubd_xmm_xmmm128", Code.Psubd_xmm_xmmm128 },
				{ "VEX_Vpsubd_xmm_xmm_xmmm128", Code.VEX_Vpsubd_xmm_xmm_xmmm128 },
				{ "VEX_Vpsubd_ymm_ymm_ymmm256", Code.VEX_Vpsubd_ymm_ymm_ymmm256 },
				{ "EVEX_Vpsubd_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpsubd_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpsubd_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpsubd_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpsubd_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpsubd_zmm_k1z_zmm_zmmm512b32 },
				{ "Psubq_mm_mmm64", Code.Psubq_mm_mmm64 },
				{ "Psubq_xmm_xmmm128", Code.Psubq_xmm_xmmm128 },
				{ "VEX_Vpsubq_xmm_xmm_xmmm128", Code.VEX_Vpsubq_xmm_xmm_xmmm128 },
				{ "VEX_Vpsubq_ymm_ymm_ymmm256", Code.VEX_Vpsubq_ymm_ymm_ymmm256 },
				{ "EVEX_Vpsubq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpsubq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpsubq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpsubq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpsubq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpsubq_zmm_k1z_zmm_zmmm512b64 },
				{ "Paddb_mm_mmm64", Code.Paddb_mm_mmm64 },
				{ "Paddb_xmm_xmmm128", Code.Paddb_xmm_xmmm128 },
				{ "VEX_Vpaddb_xmm_xmm_xmmm128", Code.VEX_Vpaddb_xmm_xmm_xmmm128 },
				{ "VEX_Vpaddb_ymm_ymm_ymmm256", Code.VEX_Vpaddb_ymm_ymm_ymmm256 },
				{ "EVEX_Vpaddb_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpaddb_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpaddb_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpaddb_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpaddb_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpaddb_zmm_k1z_zmm_zmmm512 },
				{ "Paddw_mm_mmm64", Code.Paddw_mm_mmm64 },
				{ "Paddw_xmm_xmmm128", Code.Paddw_xmm_xmmm128 },
				{ "VEX_Vpaddw_xmm_xmm_xmmm128", Code.VEX_Vpaddw_xmm_xmm_xmmm128 },
				{ "VEX_Vpaddw_ymm_ymm_ymmm256", Code.VEX_Vpaddw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpaddw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpaddw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpaddw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpaddw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpaddw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpaddw_zmm_k1z_zmm_zmmm512 },
				{ "Paddd_mm_mmm64", Code.Paddd_mm_mmm64 },
				{ "Paddd_xmm_xmmm128", Code.Paddd_xmm_xmmm128 },
				{ "VEX_Vpaddd_xmm_xmm_xmmm128", Code.VEX_Vpaddd_xmm_xmm_xmmm128 },
				{ "VEX_Vpaddd_ymm_ymm_ymmm256", Code.VEX_Vpaddd_ymm_ymm_ymmm256 },
				{ "EVEX_Vpaddd_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpaddd_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpaddd_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpaddd_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpaddd_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpaddd_zmm_k1z_zmm_zmmm512b32 },
				{ "Ud0_r16_rm16", Code.Ud0_r16_rm16 },
				{ "Ud0_r32_rm32", Code.Ud0_r32_rm32 },
				{ "Ud0_r64_rm64", Code.Ud0_r64_rm64 },
				{ "Pshufb_mm_mmm64", Code.Pshufb_mm_mmm64 },
				{ "Pshufb_xmm_xmmm128", Code.Pshufb_xmm_xmmm128 },
				{ "VEX_Vpshufb_xmm_xmm_xmmm128", Code.VEX_Vpshufb_xmm_xmm_xmmm128 },
				{ "VEX_Vpshufb_ymm_ymm_ymmm256", Code.VEX_Vpshufb_ymm_ymm_ymmm256 },
				{ "EVEX_Vpshufb_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpshufb_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpshufb_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpshufb_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpshufb_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpshufb_zmm_k1z_zmm_zmmm512 },
				{ "Phaddw_mm_mmm64", Code.Phaddw_mm_mmm64 },
				{ "Phaddw_xmm_xmmm128", Code.Phaddw_xmm_xmmm128 },
				{ "VEX_Vphaddw_xmm_xmm_xmmm128", Code.VEX_Vphaddw_xmm_xmm_xmmm128 },
				{ "VEX_Vphaddw_ymm_ymm_ymmm256", Code.VEX_Vphaddw_ymm_ymm_ymmm256 },
				{ "Phaddd_mm_mmm64", Code.Phaddd_mm_mmm64 },
				{ "Phaddd_xmm_xmmm128", Code.Phaddd_xmm_xmmm128 },
				{ "VEX_Vphaddd_xmm_xmm_xmmm128", Code.VEX_Vphaddd_xmm_xmm_xmmm128 },
				{ "VEX_Vphaddd_ymm_ymm_ymmm256", Code.VEX_Vphaddd_ymm_ymm_ymmm256 },
				{ "Phaddsw_mm_mmm64", Code.Phaddsw_mm_mmm64 },
				{ "Phaddsw_xmm_xmmm128", Code.Phaddsw_xmm_xmmm128 },
				{ "VEX_Vphaddsw_xmm_xmm_xmmm128", Code.VEX_Vphaddsw_xmm_xmm_xmmm128 },
				{ "VEX_Vphaddsw_ymm_ymm_ymmm256", Code.VEX_Vphaddsw_ymm_ymm_ymmm256 },
				{ "Pmaddubsw_mm_mmm64", Code.Pmaddubsw_mm_mmm64 },
				{ "Pmaddubsw_xmm_xmmm128", Code.Pmaddubsw_xmm_xmmm128 },
				{ "VEX_Vpmaddubsw_xmm_xmm_xmmm128", Code.VEX_Vpmaddubsw_xmm_xmm_xmmm128 },
				{ "VEX_Vpmaddubsw_ymm_ymm_ymmm256", Code.VEX_Vpmaddubsw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpmaddubsw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpmaddubsw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpmaddubsw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpmaddubsw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpmaddubsw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpmaddubsw_zmm_k1z_zmm_zmmm512 },
				{ "Phsubw_mm_mmm64", Code.Phsubw_mm_mmm64 },
				{ "Phsubw_xmm_xmmm128", Code.Phsubw_xmm_xmmm128 },
				{ "VEX_Vphsubw_xmm_xmm_xmmm128", Code.VEX_Vphsubw_xmm_xmm_xmmm128 },
				{ "VEX_Vphsubw_ymm_ymm_ymmm256", Code.VEX_Vphsubw_ymm_ymm_ymmm256 },
				{ "Phsubd_mm_mmm64", Code.Phsubd_mm_mmm64 },
				{ "Phsubd_xmm_xmmm128", Code.Phsubd_xmm_xmmm128 },
				{ "VEX_Vphsubd_xmm_xmm_xmmm128", Code.VEX_Vphsubd_xmm_xmm_xmmm128 },
				{ "VEX_Vphsubd_ymm_ymm_ymmm256", Code.VEX_Vphsubd_ymm_ymm_ymmm256 },
				{ "Phsubsw_mm_mmm64", Code.Phsubsw_mm_mmm64 },
				{ "Phsubsw_xmm_xmmm128", Code.Phsubsw_xmm_xmmm128 },
				{ "VEX_Vphsubsw_xmm_xmm_xmmm128", Code.VEX_Vphsubsw_xmm_xmm_xmmm128 },
				{ "VEX_Vphsubsw_ymm_ymm_ymmm256", Code.VEX_Vphsubsw_ymm_ymm_ymmm256 },
				{ "Psignb_mm_mmm64", Code.Psignb_mm_mmm64 },
				{ "Psignb_xmm_xmmm128", Code.Psignb_xmm_xmmm128 },
				{ "VEX_Vpsignb_xmm_xmm_xmmm128", Code.VEX_Vpsignb_xmm_xmm_xmmm128 },
				{ "VEX_Vpsignb_ymm_ymm_ymmm256", Code.VEX_Vpsignb_ymm_ymm_ymmm256 },
				{ "Psignw_mm_mmm64", Code.Psignw_mm_mmm64 },
				{ "Psignw_xmm_xmmm128", Code.Psignw_xmm_xmmm128 },
				{ "VEX_Vpsignw_xmm_xmm_xmmm128", Code.VEX_Vpsignw_xmm_xmm_xmmm128 },
				{ "VEX_Vpsignw_ymm_ymm_ymmm256", Code.VEX_Vpsignw_ymm_ymm_ymmm256 },
				{ "Psignd_mm_mmm64", Code.Psignd_mm_mmm64 },
				{ "Psignd_xmm_xmmm128", Code.Psignd_xmm_xmmm128 },
				{ "VEX_Vpsignd_xmm_xmm_xmmm128", Code.VEX_Vpsignd_xmm_xmm_xmmm128 },
				{ "VEX_Vpsignd_ymm_ymm_ymmm256", Code.VEX_Vpsignd_ymm_ymm_ymmm256 },
				{ "Pmulhrsw_mm_mmm64", Code.Pmulhrsw_mm_mmm64 },
				{ "Pmulhrsw_xmm_xmmm128", Code.Pmulhrsw_xmm_xmmm128 },
				{ "VEX_Vpmulhrsw_xmm_xmm_xmmm128", Code.VEX_Vpmulhrsw_xmm_xmm_xmmm128 },
				{ "VEX_Vpmulhrsw_ymm_ymm_ymmm256", Code.VEX_Vpmulhrsw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpmulhrsw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpmulhrsw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpmulhrsw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpmulhrsw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpmulhrsw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpmulhrsw_zmm_k1z_zmm_zmmm512 },
				{ "VEX_Vpermilps_xmm_xmm_xmmm128", Code.VEX_Vpermilps_xmm_xmm_xmmm128 },
				{ "VEX_Vpermilps_ymm_ymm_ymmm256", Code.VEX_Vpermilps_ymm_ymm_ymmm256 },
				{ "EVEX_Vpermilps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpermilps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpermilps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpermilps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpermilps_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpermilps_zmm_k1z_zmm_zmmm512b32 },
				{ "VEX_Vpermilpd_xmm_xmm_xmmm128", Code.VEX_Vpermilpd_xmm_xmm_xmmm128 },
				{ "VEX_Vpermilpd_ymm_ymm_ymmm256", Code.VEX_Vpermilpd_ymm_ymm_ymmm256 },
				{ "EVEX_Vpermilpd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpermilpd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpermilpd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpermilpd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpermilpd_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpermilpd_zmm_k1z_zmm_zmmm512b64 },
				{ "VEX_Vtestps_xmm_xmmm128", Code.VEX_Vtestps_xmm_xmmm128 },
				{ "VEX_Vtestps_ymm_ymmm256", Code.VEX_Vtestps_ymm_ymmm256 },
				{ "VEX_Vtestpd_xmm_xmmm128", Code.VEX_Vtestpd_xmm_xmmm128 },
				{ "VEX_Vtestpd_ymm_ymmm256", Code.VEX_Vtestpd_ymm_ymmm256 },
				{ "Pblendvb_xmm_xmmm128", Code.Pblendvb_xmm_xmmm128 },
				{ "EVEX_Vpsrlvw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpsrlvw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpsrlvw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpsrlvw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpsrlvw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpsrlvw_zmm_k1z_zmm_zmmm512 },
				{ "EVEX_Vpmovuswb_xmmm64_k1z_xmm", Code.EVEX_Vpmovuswb_xmmm64_k1z_xmm },
				{ "EVEX_Vpmovuswb_xmmm128_k1z_ymm", Code.EVEX_Vpmovuswb_xmmm128_k1z_ymm },
				{ "EVEX_Vpmovuswb_ymmm256_k1z_zmm", Code.EVEX_Vpmovuswb_ymmm256_k1z_zmm },
				{ "EVEX_Vpsravw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpsravw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpsravw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpsravw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpsravw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpsravw_zmm_k1z_zmm_zmmm512 },
				{ "EVEX_Vpmovusdb_xmmm32_k1z_xmm", Code.EVEX_Vpmovusdb_xmmm32_k1z_xmm },
				{ "EVEX_Vpmovusdb_xmmm64_k1z_ymm", Code.EVEX_Vpmovusdb_xmmm64_k1z_ymm },
				{ "EVEX_Vpmovusdb_xmmm128_k1z_zmm", Code.EVEX_Vpmovusdb_xmmm128_k1z_zmm },
				{ "EVEX_Vpsllvw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpsllvw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpsllvw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpsllvw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpsllvw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpsllvw_zmm_k1z_zmm_zmmm512 },
				{ "EVEX_Vpmovusqb_xmmm16_k1z_xmm", Code.EVEX_Vpmovusqb_xmmm16_k1z_xmm },
				{ "EVEX_Vpmovusqb_xmmm32_k1z_ymm", Code.EVEX_Vpmovusqb_xmmm32_k1z_ymm },
				{ "EVEX_Vpmovusqb_xmmm64_k1z_zmm", Code.EVEX_Vpmovusqb_xmmm64_k1z_zmm },
				{ "VEX_Vcvtph2ps_xmm_xmmm64", Code.VEX_Vcvtph2ps_xmm_xmmm64 },
				{ "VEX_Vcvtph2ps_ymm_xmmm128", Code.VEX_Vcvtph2ps_ymm_xmmm128 },
				{ "EVEX_Vcvtph2ps_xmm_k1z_xmmm64", Code.EVEX_Vcvtph2ps_xmm_k1z_xmmm64 },
				{ "EVEX_Vcvtph2ps_ymm_k1z_xmmm128", Code.EVEX_Vcvtph2ps_ymm_k1z_xmmm128 },
				{ "EVEX_Vcvtph2ps_zmm_k1z_ymmm256_sae", Code.EVEX_Vcvtph2ps_zmm_k1z_ymmm256_sae },
				{ "EVEX_Vpmovusdw_xmmm64_k1z_xmm", Code.EVEX_Vpmovusdw_xmmm64_k1z_xmm },
				{ "EVEX_Vpmovusdw_xmmm128_k1z_ymm", Code.EVEX_Vpmovusdw_xmmm128_k1z_ymm },
				{ "EVEX_Vpmovusdw_ymmm256_k1z_zmm", Code.EVEX_Vpmovusdw_ymmm256_k1z_zmm },
				{ "Blendvps_xmm_xmmm128", Code.Blendvps_xmm_xmmm128 },
				{ "EVEX_Vprorvd_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vprorvd_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vprorvd_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vprorvd_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vprorvd_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vprorvd_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vprorvq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vprorvq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vprorvq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vprorvq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vprorvq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vprorvq_zmm_k1z_zmm_zmmm512b64 },
				{ "EVEX_Vpmovusqw_xmmm32_k1z_xmm", Code.EVEX_Vpmovusqw_xmmm32_k1z_xmm },
				{ "EVEX_Vpmovusqw_xmmm64_k1z_ymm", Code.EVEX_Vpmovusqw_xmmm64_k1z_ymm },
				{ "EVEX_Vpmovusqw_xmmm128_k1z_zmm", Code.EVEX_Vpmovusqw_xmmm128_k1z_zmm },
				{ "Blendvpd_xmm_xmmm128", Code.Blendvpd_xmm_xmmm128 },
				{ "EVEX_Vprolvd_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vprolvd_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vprolvd_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vprolvd_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vprolvd_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vprolvd_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vprolvq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vprolvq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vprolvq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vprolvq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vprolvq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vprolvq_zmm_k1z_zmm_zmmm512b64 },
				{ "EVEX_Vpmovusqd_xmmm64_k1z_xmm", Code.EVEX_Vpmovusqd_xmmm64_k1z_xmm },
				{ "EVEX_Vpmovusqd_xmmm128_k1z_ymm", Code.EVEX_Vpmovusqd_xmmm128_k1z_ymm },
				{ "EVEX_Vpmovusqd_ymmm256_k1z_zmm", Code.EVEX_Vpmovusqd_ymmm256_k1z_zmm },
				{ "VEX_Vpermps_ymm_ymm_ymmm256", Code.VEX_Vpermps_ymm_ymm_ymmm256 },
				{ "EVEX_Vpermps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpermps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpermps_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpermps_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpermpd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpermpd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpermpd_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpermpd_zmm_k1z_zmm_zmmm512b64 },
				{ "Ptest_xmm_xmmm128", Code.Ptest_xmm_xmmm128 },
				{ "VEX_Vptest_xmm_xmmm128", Code.VEX_Vptest_xmm_xmmm128 },
				{ "VEX_Vptest_ymm_ymmm256", Code.VEX_Vptest_ymm_ymmm256 },
				{ "VEX_Vbroadcastss_xmm_m32", Code.VEX_Vbroadcastss_xmm_m32 },
				{ "VEX_Vbroadcastss_ymm_m32", Code.VEX_Vbroadcastss_ymm_m32 },
				{ "EVEX_Vbroadcastss_xmm_k1z_xmmm32", Code.EVEX_Vbroadcastss_xmm_k1z_xmmm32 },
				{ "EVEX_Vbroadcastss_ymm_k1z_xmmm32", Code.EVEX_Vbroadcastss_ymm_k1z_xmmm32 },
				{ "EVEX_Vbroadcastss_zmm_k1z_xmmm32", Code.EVEX_Vbroadcastss_zmm_k1z_xmmm32 },
				{ "VEX_Vbroadcastsd_ymm_m64", Code.VEX_Vbroadcastsd_ymm_m64 },
				{ "EVEX_Vbroadcastf32x2_ymm_k1z_xmmm64", Code.EVEX_Vbroadcastf32x2_ymm_k1z_xmmm64 },
				{ "EVEX_Vbroadcastf32x2_zmm_k1z_xmmm64", Code.EVEX_Vbroadcastf32x2_zmm_k1z_xmmm64 },
				{ "EVEX_Vbroadcastsd_ymm_k1z_xmmm64", Code.EVEX_Vbroadcastsd_ymm_k1z_xmmm64 },
				{ "EVEX_Vbroadcastsd_zmm_k1z_xmmm64", Code.EVEX_Vbroadcastsd_zmm_k1z_xmmm64 },
				{ "VEX_Vbroadcastf128_ymm_m128", Code.VEX_Vbroadcastf128_ymm_m128 },
				{ "EVEX_Vbroadcastf32x4_ymm_k1z_m128", Code.EVEX_Vbroadcastf32x4_ymm_k1z_m128 },
				{ "EVEX_Vbroadcastf32x4_zmm_k1z_m128", Code.EVEX_Vbroadcastf32x4_zmm_k1z_m128 },
				{ "EVEX_Vbroadcastf64x2_ymm_k1z_m128", Code.EVEX_Vbroadcastf64x2_ymm_k1z_m128 },
				{ "EVEX_Vbroadcastf64x2_zmm_k1z_m128", Code.EVEX_Vbroadcastf64x2_zmm_k1z_m128 },
				{ "EVEX_Vbroadcastf32x8_zmm_k1z_m256", Code.EVEX_Vbroadcastf32x8_zmm_k1z_m256 },
				{ "EVEX_Vbroadcastf64x4_zmm_k1z_m256", Code.EVEX_Vbroadcastf64x4_zmm_k1z_m256 },
				{ "Pabsb_mm_mmm64", Code.Pabsb_mm_mmm64 },
				{ "Pabsb_xmm_xmmm128", Code.Pabsb_xmm_xmmm128 },
				{ "VEX_Vpabsb_xmm_xmmm128", Code.VEX_Vpabsb_xmm_xmmm128 },
				{ "VEX_Vpabsb_ymm_ymmm256", Code.VEX_Vpabsb_ymm_ymmm256 },
				{ "EVEX_Vpabsb_xmm_k1z_xmmm128", Code.EVEX_Vpabsb_xmm_k1z_xmmm128 },
				{ "EVEX_Vpabsb_ymm_k1z_ymmm256", Code.EVEX_Vpabsb_ymm_k1z_ymmm256 },
				{ "EVEX_Vpabsb_zmm_k1z_zmmm512", Code.EVEX_Vpabsb_zmm_k1z_zmmm512 },
				{ "Pabsw_mm_mmm64", Code.Pabsw_mm_mmm64 },
				{ "Pabsw_xmm_xmmm128", Code.Pabsw_xmm_xmmm128 },
				{ "VEX_Vpabsw_xmm_xmmm128", Code.VEX_Vpabsw_xmm_xmmm128 },
				{ "VEX_Vpabsw_ymm_ymmm256", Code.VEX_Vpabsw_ymm_ymmm256 },
				{ "EVEX_Vpabsw_xmm_k1z_xmmm128", Code.EVEX_Vpabsw_xmm_k1z_xmmm128 },
				{ "EVEX_Vpabsw_ymm_k1z_ymmm256", Code.EVEX_Vpabsw_ymm_k1z_ymmm256 },
				{ "EVEX_Vpabsw_zmm_k1z_zmmm512", Code.EVEX_Vpabsw_zmm_k1z_zmmm512 },
				{ "Pabsd_mm_mmm64", Code.Pabsd_mm_mmm64 },
				{ "Pabsd_xmm_xmmm128", Code.Pabsd_xmm_xmmm128 },
				{ "VEX_Vpabsd_xmm_xmmm128", Code.VEX_Vpabsd_xmm_xmmm128 },
				{ "VEX_Vpabsd_ymm_ymmm256", Code.VEX_Vpabsd_ymm_ymmm256 },
				{ "EVEX_Vpabsd_xmm_k1z_xmmm128b32", Code.EVEX_Vpabsd_xmm_k1z_xmmm128b32 },
				{ "EVEX_Vpabsd_ymm_k1z_ymmm256b32", Code.EVEX_Vpabsd_ymm_k1z_ymmm256b32 },
				{ "EVEX_Vpabsd_zmm_k1z_zmmm512b32", Code.EVEX_Vpabsd_zmm_k1z_zmmm512b32 },
				{ "EVEX_Vpabsq_xmm_k1z_xmmm128b64", Code.EVEX_Vpabsq_xmm_k1z_xmmm128b64 },
				{ "EVEX_Vpabsq_ymm_k1z_ymmm256b64", Code.EVEX_Vpabsq_ymm_k1z_ymmm256b64 },
				{ "EVEX_Vpabsq_zmm_k1z_zmmm512b64", Code.EVEX_Vpabsq_zmm_k1z_zmmm512b64 },
				{ "Pmovsxbw_xmm_xmmm64", Code.Pmovsxbw_xmm_xmmm64 },
				{ "VEX_Vpmovsxbw_xmm_xmmm64", Code.VEX_Vpmovsxbw_xmm_xmmm64 },
				{ "VEX_Vpmovsxbw_ymm_xmmm128", Code.VEX_Vpmovsxbw_ymm_xmmm128 },
				{ "EVEX_Vpmovsxbw_xmm_k1z_xmmm64", Code.EVEX_Vpmovsxbw_xmm_k1z_xmmm64 },
				{ "EVEX_Vpmovsxbw_ymm_k1z_xmmm128", Code.EVEX_Vpmovsxbw_ymm_k1z_xmmm128 },
				{ "EVEX_Vpmovsxbw_zmm_k1z_ymmm256", Code.EVEX_Vpmovsxbw_zmm_k1z_ymmm256 },
				{ "EVEX_Vpmovswb_xmmm64_k1z_xmm", Code.EVEX_Vpmovswb_xmmm64_k1z_xmm },
				{ "EVEX_Vpmovswb_xmmm128_k1z_ymm", Code.EVEX_Vpmovswb_xmmm128_k1z_ymm },
				{ "EVEX_Vpmovswb_ymmm256_k1z_zmm", Code.EVEX_Vpmovswb_ymmm256_k1z_zmm },
				{ "Pmovsxbd_xmm_xmmm32", Code.Pmovsxbd_xmm_xmmm32 },
				{ "VEX_Vpmovsxbd_xmm_xmmm32", Code.VEX_Vpmovsxbd_xmm_xmmm32 },
				{ "VEX_Vpmovsxbd_ymm_xmmm64", Code.VEX_Vpmovsxbd_ymm_xmmm64 },
				{ "EVEX_Vpmovsxbd_xmm_k1z_xmmm32", Code.EVEX_Vpmovsxbd_xmm_k1z_xmmm32 },
				{ "EVEX_Vpmovsxbd_ymm_k1z_xmmm64", Code.EVEX_Vpmovsxbd_ymm_k1z_xmmm64 },
				{ "EVEX_Vpmovsxbd_zmm_k1z_xmmm128", Code.EVEX_Vpmovsxbd_zmm_k1z_xmmm128 },
				{ "EVEX_Vpmovsdb_xmmm32_k1z_xmm", Code.EVEX_Vpmovsdb_xmmm32_k1z_xmm },
				{ "EVEX_Vpmovsdb_xmmm64_k1z_ymm", Code.EVEX_Vpmovsdb_xmmm64_k1z_ymm },
				{ "EVEX_Vpmovsdb_xmmm128_k1z_zmm", Code.EVEX_Vpmovsdb_xmmm128_k1z_zmm },
				{ "Pmovsxbq_xmm_xmmm16", Code.Pmovsxbq_xmm_xmmm16 },
				{ "VEX_Vpmovsxbq_xmm_xmmm16", Code.VEX_Vpmovsxbq_xmm_xmmm16 },
				{ "VEX_Vpmovsxbq_ymm_xmmm32", Code.VEX_Vpmovsxbq_ymm_xmmm32 },
				{ "EVEX_Vpmovsxbq_xmm_k1z_xmmm16", Code.EVEX_Vpmovsxbq_xmm_k1z_xmmm16 },
				{ "EVEX_Vpmovsxbq_ymm_k1z_xmmm32", Code.EVEX_Vpmovsxbq_ymm_k1z_xmmm32 },
				{ "EVEX_Vpmovsxbq_zmm_k1z_xmmm64", Code.EVEX_Vpmovsxbq_zmm_k1z_xmmm64 },
				{ "EVEX_Vpmovsqb_xmmm16_k1z_xmm", Code.EVEX_Vpmovsqb_xmmm16_k1z_xmm },
				{ "EVEX_Vpmovsqb_xmmm32_k1z_ymm", Code.EVEX_Vpmovsqb_xmmm32_k1z_ymm },
				{ "EVEX_Vpmovsqb_xmmm64_k1z_zmm", Code.EVEX_Vpmovsqb_xmmm64_k1z_zmm },
				{ "Pmovsxwd_xmm_xmmm64", Code.Pmovsxwd_xmm_xmmm64 },
				{ "VEX_Vpmovsxwd_xmm_xmmm64", Code.VEX_Vpmovsxwd_xmm_xmmm64 },
				{ "VEX_Vpmovsxwd_ymm_xmmm128", Code.VEX_Vpmovsxwd_ymm_xmmm128 },
				{ "EVEX_Vpmovsxwd_xmm_k1z_xmmm64", Code.EVEX_Vpmovsxwd_xmm_k1z_xmmm64 },
				{ "EVEX_Vpmovsxwd_ymm_k1z_xmmm128", Code.EVEX_Vpmovsxwd_ymm_k1z_xmmm128 },
				{ "EVEX_Vpmovsxwd_zmm_k1z_ymmm256", Code.EVEX_Vpmovsxwd_zmm_k1z_ymmm256 },
				{ "EVEX_Vpmovsdw_xmmm64_k1z_xmm", Code.EVEX_Vpmovsdw_xmmm64_k1z_xmm },
				{ "EVEX_Vpmovsdw_xmmm128_k1z_ymm", Code.EVEX_Vpmovsdw_xmmm128_k1z_ymm },
				{ "EVEX_Vpmovsdw_ymmm256_k1z_zmm", Code.EVEX_Vpmovsdw_ymmm256_k1z_zmm },
				{ "Pmovsxwq_xmm_xmmm32", Code.Pmovsxwq_xmm_xmmm32 },
				{ "VEX_Vpmovsxwq_xmm_xmmm32", Code.VEX_Vpmovsxwq_xmm_xmmm32 },
				{ "VEX_Vpmovsxwq_ymm_xmmm64", Code.VEX_Vpmovsxwq_ymm_xmmm64 },
				{ "EVEX_Vpmovsxwq_xmm_k1z_xmmm32", Code.EVEX_Vpmovsxwq_xmm_k1z_xmmm32 },
				{ "EVEX_Vpmovsxwq_ymm_k1z_xmmm64", Code.EVEX_Vpmovsxwq_ymm_k1z_xmmm64 },
				{ "EVEX_Vpmovsxwq_zmm_k1z_xmmm128", Code.EVEX_Vpmovsxwq_zmm_k1z_xmmm128 },
				{ "EVEX_Vpmovsqw_xmmm32_k1z_xmm", Code.EVEX_Vpmovsqw_xmmm32_k1z_xmm },
				{ "EVEX_Vpmovsqw_xmmm64_k1z_ymm", Code.EVEX_Vpmovsqw_xmmm64_k1z_ymm },
				{ "EVEX_Vpmovsqw_xmmm128_k1z_zmm", Code.EVEX_Vpmovsqw_xmmm128_k1z_zmm },
				{ "Pmovsxdq_xmm_xmmm64", Code.Pmovsxdq_xmm_xmmm64 },
				{ "VEX_Vpmovsxdq_xmm_xmmm64", Code.VEX_Vpmovsxdq_xmm_xmmm64 },
				{ "VEX_Vpmovsxdq_ymm_xmmm128", Code.VEX_Vpmovsxdq_ymm_xmmm128 },
				{ "EVEX_Vpmovsxdq_xmm_k1z_xmmm64", Code.EVEX_Vpmovsxdq_xmm_k1z_xmmm64 },
				{ "EVEX_Vpmovsxdq_ymm_k1z_xmmm128", Code.EVEX_Vpmovsxdq_ymm_k1z_xmmm128 },
				{ "EVEX_Vpmovsxdq_zmm_k1z_ymmm256", Code.EVEX_Vpmovsxdq_zmm_k1z_ymmm256 },
				{ "EVEX_Vpmovsqd_xmmm64_k1z_xmm", Code.EVEX_Vpmovsqd_xmmm64_k1z_xmm },
				{ "EVEX_Vpmovsqd_xmmm128_k1z_ymm", Code.EVEX_Vpmovsqd_xmmm128_k1z_ymm },
				{ "EVEX_Vpmovsqd_ymmm256_k1z_zmm", Code.EVEX_Vpmovsqd_ymmm256_k1z_zmm },
				{ "EVEX_Vptestmb_kr_k1_xmm_xmmm128", Code.EVEX_Vptestmb_kr_k1_xmm_xmmm128 },
				{ "EVEX_Vptestmb_kr_k1_ymm_ymmm256", Code.EVEX_Vptestmb_kr_k1_ymm_ymmm256 },
				{ "EVEX_Vptestmb_kr_k1_zmm_zmmm512", Code.EVEX_Vptestmb_kr_k1_zmm_zmmm512 },
				{ "EVEX_Vptestmw_kr_k1_xmm_xmmm128", Code.EVEX_Vptestmw_kr_k1_xmm_xmmm128 },
				{ "EVEX_Vptestmw_kr_k1_ymm_ymmm256", Code.EVEX_Vptestmw_kr_k1_ymm_ymmm256 },
				{ "EVEX_Vptestmw_kr_k1_zmm_zmmm512", Code.EVEX_Vptestmw_kr_k1_zmm_zmmm512 },
				{ "EVEX_Vptestnmb_kr_k1_xmm_xmmm128", Code.EVEX_Vptestnmb_kr_k1_xmm_xmmm128 },
				{ "EVEX_Vptestnmb_kr_k1_ymm_ymmm256", Code.EVEX_Vptestnmb_kr_k1_ymm_ymmm256 },
				{ "EVEX_Vptestnmb_kr_k1_zmm_zmmm512", Code.EVEX_Vptestnmb_kr_k1_zmm_zmmm512 },
				{ "EVEX_Vptestnmw_kr_k1_xmm_xmmm128", Code.EVEX_Vptestnmw_kr_k1_xmm_xmmm128 },
				{ "EVEX_Vptestnmw_kr_k1_ymm_ymmm256", Code.EVEX_Vptestnmw_kr_k1_ymm_ymmm256 },
				{ "EVEX_Vptestnmw_kr_k1_zmm_zmmm512", Code.EVEX_Vptestnmw_kr_k1_zmm_zmmm512 },
				{ "EVEX_Vptestmd_kr_k1_xmm_xmmm128b32", Code.EVEX_Vptestmd_kr_k1_xmm_xmmm128b32 },
				{ "EVEX_Vptestmd_kr_k1_ymm_ymmm256b32", Code.EVEX_Vptestmd_kr_k1_ymm_ymmm256b32 },
				{ "EVEX_Vptestmd_kr_k1_zmm_zmmm512b32", Code.EVEX_Vptestmd_kr_k1_zmm_zmmm512b32 },
				{ "EVEX_Vptestmq_kr_k1_xmm_xmmm128b64", Code.EVEX_Vptestmq_kr_k1_xmm_xmmm128b64 },
				{ "EVEX_Vptestmq_kr_k1_ymm_ymmm256b64", Code.EVEX_Vptestmq_kr_k1_ymm_ymmm256b64 },
				{ "EVEX_Vptestmq_kr_k1_zmm_zmmm512b64", Code.EVEX_Vptestmq_kr_k1_zmm_zmmm512b64 },
				{ "EVEX_Vptestnmd_kr_k1_xmm_xmmm128b32", Code.EVEX_Vptestnmd_kr_k1_xmm_xmmm128b32 },
				{ "EVEX_Vptestnmd_kr_k1_ymm_ymmm256b32", Code.EVEX_Vptestnmd_kr_k1_ymm_ymmm256b32 },
				{ "EVEX_Vptestnmd_kr_k1_zmm_zmmm512b32", Code.EVEX_Vptestnmd_kr_k1_zmm_zmmm512b32 },
				{ "EVEX_Vptestnmq_kr_k1_xmm_xmmm128b64", Code.EVEX_Vptestnmq_kr_k1_xmm_xmmm128b64 },
				{ "EVEX_Vptestnmq_kr_k1_ymm_ymmm256b64", Code.EVEX_Vptestnmq_kr_k1_ymm_ymmm256b64 },
				{ "EVEX_Vptestnmq_kr_k1_zmm_zmmm512b64", Code.EVEX_Vptestnmq_kr_k1_zmm_zmmm512b64 },
				{ "Pmuldq_xmm_xmmm128", Code.Pmuldq_xmm_xmmm128 },
				{ "VEX_Vpmuldq_xmm_xmm_xmmm128", Code.VEX_Vpmuldq_xmm_xmm_xmmm128 },
				{ "VEX_Vpmuldq_ymm_ymm_ymmm256", Code.VEX_Vpmuldq_ymm_ymm_ymmm256 },
				{ "EVEX_Vpmuldq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpmuldq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpmuldq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpmuldq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpmuldq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpmuldq_zmm_k1z_zmm_zmmm512b64 },
				{ "EVEX_Vpmovm2b_xmm_kr", Code.EVEX_Vpmovm2b_xmm_kr },
				{ "EVEX_Vpmovm2b_ymm_kr", Code.EVEX_Vpmovm2b_ymm_kr },
				{ "EVEX_Vpmovm2b_zmm_kr", Code.EVEX_Vpmovm2b_zmm_kr },
				{ "EVEX_Vpmovm2w_xmm_kr", Code.EVEX_Vpmovm2w_xmm_kr },
				{ "EVEX_Vpmovm2w_ymm_kr", Code.EVEX_Vpmovm2w_ymm_kr },
				{ "EVEX_Vpmovm2w_zmm_kr", Code.EVEX_Vpmovm2w_zmm_kr },
				{ "Pcmpeqq_xmm_xmmm128", Code.Pcmpeqq_xmm_xmmm128 },
				{ "VEX_Vpcmpeqq_xmm_xmm_xmmm128", Code.VEX_Vpcmpeqq_xmm_xmm_xmmm128 },
				{ "VEX_Vpcmpeqq_ymm_ymm_ymmm256", Code.VEX_Vpcmpeqq_ymm_ymm_ymmm256 },
				{ "EVEX_Vpcmpeqq_kr_k1_xmm_xmmm128b64", Code.EVEX_Vpcmpeqq_kr_k1_xmm_xmmm128b64 },
				{ "EVEX_Vpcmpeqq_kr_k1_ymm_ymmm256b64", Code.EVEX_Vpcmpeqq_kr_k1_ymm_ymmm256b64 },
				{ "EVEX_Vpcmpeqq_kr_k1_zmm_zmmm512b64", Code.EVEX_Vpcmpeqq_kr_k1_zmm_zmmm512b64 },
				{ "EVEX_Vpmovb2m_kr_xmm", Code.EVEX_Vpmovb2m_kr_xmm },
				{ "EVEX_Vpmovb2m_kr_ymm", Code.EVEX_Vpmovb2m_kr_ymm },
				{ "EVEX_Vpmovb2m_kr_zmm", Code.EVEX_Vpmovb2m_kr_zmm },
				{ "EVEX_Vpmovw2m_kr_xmm", Code.EVEX_Vpmovw2m_kr_xmm },
				{ "EVEX_Vpmovw2m_kr_ymm", Code.EVEX_Vpmovw2m_kr_ymm },
				{ "EVEX_Vpmovw2m_kr_zmm", Code.EVEX_Vpmovw2m_kr_zmm },
				{ "Movntdqa_xmm_m128", Code.Movntdqa_xmm_m128 },
				{ "VEX_Vmovntdqa_xmm_m128", Code.VEX_Vmovntdqa_xmm_m128 },
				{ "VEX_Vmovntdqa_ymm_m256", Code.VEX_Vmovntdqa_ymm_m256 },
				{ "EVEX_Vmovntdqa_xmm_m128", Code.EVEX_Vmovntdqa_xmm_m128 },
				{ "EVEX_Vmovntdqa_ymm_m256", Code.EVEX_Vmovntdqa_ymm_m256 },
				{ "EVEX_Vmovntdqa_zmm_m512", Code.EVEX_Vmovntdqa_zmm_m512 },
				{ "EVEX_Vpbroadcastmb2q_xmm_kr", Code.EVEX_Vpbroadcastmb2q_xmm_kr },
				{ "EVEX_Vpbroadcastmb2q_ymm_kr", Code.EVEX_Vpbroadcastmb2q_ymm_kr },
				{ "EVEX_Vpbroadcastmb2q_zmm_kr", Code.EVEX_Vpbroadcastmb2q_zmm_kr },
				{ "Packusdw_xmm_xmmm128", Code.Packusdw_xmm_xmmm128 },
				{ "VEX_Vpackusdw_xmm_xmm_xmmm128", Code.VEX_Vpackusdw_xmm_xmm_xmmm128 },
				{ "VEX_Vpackusdw_ymm_ymm_ymmm256", Code.VEX_Vpackusdw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpackusdw_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpackusdw_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpackusdw_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpackusdw_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpackusdw_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpackusdw_zmm_k1z_zmm_zmmm512b32 },
				{ "VEX_Vmaskmovps_xmm_xmm_m128", Code.VEX_Vmaskmovps_xmm_xmm_m128 },
				{ "VEX_Vmaskmovps_ymm_ymm_m256", Code.VEX_Vmaskmovps_ymm_ymm_m256 },
				{ "EVEX_Vscalefps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vscalefps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vscalefps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vscalefps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vscalefps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vscalefps_zmm_k1z_zmm_zmmm512b32_er },
				{ "EVEX_Vscalefpd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vscalefpd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vscalefpd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vscalefpd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vscalefpd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vscalefpd_zmm_k1z_zmm_zmmm512b64_er },
				{ "VEX_Vmaskmovpd_xmm_xmm_m128", Code.VEX_Vmaskmovpd_xmm_xmm_m128 },
				{ "VEX_Vmaskmovpd_ymm_ymm_m256", Code.VEX_Vmaskmovpd_ymm_ymm_m256 },
				{ "EVEX_Vscalefss_xmm_k1z_xmm_xmmm32_er", Code.EVEX_Vscalefss_xmm_k1z_xmm_xmmm32_er },
				{ "EVEX_Vscalefsd_xmm_k1z_xmm_xmmm64_er", Code.EVEX_Vscalefsd_xmm_k1z_xmm_xmmm64_er },
				{ "VEX_Vmaskmovps_m128_xmm_xmm", Code.VEX_Vmaskmovps_m128_xmm_xmm },
				{ "VEX_Vmaskmovps_m256_ymm_ymm", Code.VEX_Vmaskmovps_m256_ymm_ymm },
				{ "VEX_Vmaskmovpd_m128_xmm_xmm", Code.VEX_Vmaskmovpd_m128_xmm_xmm },
				{ "VEX_Vmaskmovpd_m256_ymm_ymm", Code.VEX_Vmaskmovpd_m256_ymm_ymm },
				{ "Pmovzxbw_xmm_xmmm64", Code.Pmovzxbw_xmm_xmmm64 },
				{ "VEX_Vpmovzxbw_xmm_xmmm64", Code.VEX_Vpmovzxbw_xmm_xmmm64 },
				{ "VEX_Vpmovzxbw_ymm_xmmm128", Code.VEX_Vpmovzxbw_ymm_xmmm128 },
				{ "EVEX_Vpmovzxbw_xmm_k1z_xmmm64", Code.EVEX_Vpmovzxbw_xmm_k1z_xmmm64 },
				{ "EVEX_Vpmovzxbw_ymm_k1z_xmmm128", Code.EVEX_Vpmovzxbw_ymm_k1z_xmmm128 },
				{ "EVEX_Vpmovzxbw_zmm_k1z_ymmm256", Code.EVEX_Vpmovzxbw_zmm_k1z_ymmm256 },
				{ "EVEX_Vpmovwb_xmmm64_k1z_xmm", Code.EVEX_Vpmovwb_xmmm64_k1z_xmm },
				{ "EVEX_Vpmovwb_xmmm128_k1z_ymm", Code.EVEX_Vpmovwb_xmmm128_k1z_ymm },
				{ "EVEX_Vpmovwb_ymmm256_k1z_zmm", Code.EVEX_Vpmovwb_ymmm256_k1z_zmm },
				{ "Pmovzxbd_xmm_xmmm32", Code.Pmovzxbd_xmm_xmmm32 },
				{ "VEX_Vpmovzxbd_xmm_xmmm32", Code.VEX_Vpmovzxbd_xmm_xmmm32 },
				{ "VEX_Vpmovzxbd_ymm_xmmm64", Code.VEX_Vpmovzxbd_ymm_xmmm64 },
				{ "EVEX_Vpmovzxbd_xmm_k1z_xmmm32", Code.EVEX_Vpmovzxbd_xmm_k1z_xmmm32 },
				{ "EVEX_Vpmovzxbd_ymm_k1z_xmmm64", Code.EVEX_Vpmovzxbd_ymm_k1z_xmmm64 },
				{ "EVEX_Vpmovzxbd_zmm_k1z_xmmm128", Code.EVEX_Vpmovzxbd_zmm_k1z_xmmm128 },
				{ "EVEX_Vpmovdb_xmmm32_k1z_xmm", Code.EVEX_Vpmovdb_xmmm32_k1z_xmm },
				{ "EVEX_Vpmovdb_xmmm64_k1z_ymm", Code.EVEX_Vpmovdb_xmmm64_k1z_ymm },
				{ "EVEX_Vpmovdb_xmmm128_k1z_zmm", Code.EVEX_Vpmovdb_xmmm128_k1z_zmm },
				{ "Pmovzxbq_xmm_xmmm16", Code.Pmovzxbq_xmm_xmmm16 },
				{ "VEX_Vpmovzxbq_xmm_xmmm16", Code.VEX_Vpmovzxbq_xmm_xmmm16 },
				{ "VEX_Vpmovzxbq_ymm_xmmm32", Code.VEX_Vpmovzxbq_ymm_xmmm32 },
				{ "EVEX_Vpmovzxbq_xmm_k1z_xmmm16", Code.EVEX_Vpmovzxbq_xmm_k1z_xmmm16 },
				{ "EVEX_Vpmovzxbq_ymm_k1z_xmmm32", Code.EVEX_Vpmovzxbq_ymm_k1z_xmmm32 },
				{ "EVEX_Vpmovzxbq_zmm_k1z_xmmm64", Code.EVEX_Vpmovzxbq_zmm_k1z_xmmm64 },
				{ "EVEX_Vpmovqb_xmmm16_k1z_xmm", Code.EVEX_Vpmovqb_xmmm16_k1z_xmm },
				{ "EVEX_Vpmovqb_xmmm32_k1z_ymm", Code.EVEX_Vpmovqb_xmmm32_k1z_ymm },
				{ "EVEX_Vpmovqb_xmmm64_k1z_zmm", Code.EVEX_Vpmovqb_xmmm64_k1z_zmm },
				{ "Pmovzxwd_xmm_xmmm64", Code.Pmovzxwd_xmm_xmmm64 },
				{ "VEX_Vpmovzxwd_xmm_xmmm64", Code.VEX_Vpmovzxwd_xmm_xmmm64 },
				{ "VEX_Vpmovzxwd_ymm_xmmm128", Code.VEX_Vpmovzxwd_ymm_xmmm128 },
				{ "EVEX_Vpmovzxwd_xmm_k1z_xmmm64", Code.EVEX_Vpmovzxwd_xmm_k1z_xmmm64 },
				{ "EVEX_Vpmovzxwd_ymm_k1z_xmmm128", Code.EVEX_Vpmovzxwd_ymm_k1z_xmmm128 },
				{ "EVEX_Vpmovzxwd_zmm_k1z_ymmm256", Code.EVEX_Vpmovzxwd_zmm_k1z_ymmm256 },
				{ "EVEX_Vpmovdw_xmmm64_k1z_xmm", Code.EVEX_Vpmovdw_xmmm64_k1z_xmm },
				{ "EVEX_Vpmovdw_xmmm128_k1z_ymm", Code.EVEX_Vpmovdw_xmmm128_k1z_ymm },
				{ "EVEX_Vpmovdw_ymmm256_k1z_zmm", Code.EVEX_Vpmovdw_ymmm256_k1z_zmm },
				{ "Pmovzxwq_xmm_xmmm32", Code.Pmovzxwq_xmm_xmmm32 },
				{ "VEX_Vpmovzxwq_xmm_xmmm32", Code.VEX_Vpmovzxwq_xmm_xmmm32 },
				{ "VEX_Vpmovzxwq_ymm_xmmm64", Code.VEX_Vpmovzxwq_ymm_xmmm64 },
				{ "EVEX_Vpmovzxwq_xmm_k1z_xmmm32", Code.EVEX_Vpmovzxwq_xmm_k1z_xmmm32 },
				{ "EVEX_Vpmovzxwq_ymm_k1z_xmmm64", Code.EVEX_Vpmovzxwq_ymm_k1z_xmmm64 },
				{ "EVEX_Vpmovzxwq_zmm_k1z_xmmm128", Code.EVEX_Vpmovzxwq_zmm_k1z_xmmm128 },
				{ "EVEX_Vpmovqw_xmmm32_k1z_xmm", Code.EVEX_Vpmovqw_xmmm32_k1z_xmm },
				{ "EVEX_Vpmovqw_xmmm64_k1z_ymm", Code.EVEX_Vpmovqw_xmmm64_k1z_ymm },
				{ "EVEX_Vpmovqw_xmmm128_k1z_zmm", Code.EVEX_Vpmovqw_xmmm128_k1z_zmm },
				{ "Pmovzxdq_xmm_xmmm64", Code.Pmovzxdq_xmm_xmmm64 },
				{ "VEX_Vpmovzxdq_xmm_xmmm64", Code.VEX_Vpmovzxdq_xmm_xmmm64 },
				{ "VEX_Vpmovzxdq_ymm_xmmm128", Code.VEX_Vpmovzxdq_ymm_xmmm128 },
				{ "EVEX_Vpmovzxdq_xmm_k1z_xmmm64", Code.EVEX_Vpmovzxdq_xmm_k1z_xmmm64 },
				{ "EVEX_Vpmovzxdq_ymm_k1z_xmmm128", Code.EVEX_Vpmovzxdq_ymm_k1z_xmmm128 },
				{ "EVEX_Vpmovzxdq_zmm_k1z_ymmm256", Code.EVEX_Vpmovzxdq_zmm_k1z_ymmm256 },
				{ "EVEX_Vpmovqd_xmmm64_k1z_xmm", Code.EVEX_Vpmovqd_xmmm64_k1z_xmm },
				{ "EVEX_Vpmovqd_xmmm128_k1z_ymm", Code.EVEX_Vpmovqd_xmmm128_k1z_ymm },
				{ "EVEX_Vpmovqd_ymmm256_k1z_zmm", Code.EVEX_Vpmovqd_ymmm256_k1z_zmm },
				{ "VEX_Vpermd_ymm_ymm_ymmm256", Code.VEX_Vpermd_ymm_ymm_ymmm256 },
				{ "EVEX_Vpermd_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpermd_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpermd_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpermd_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpermq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpermq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpermq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpermq_zmm_k1z_zmm_zmmm512b64 },
				{ "Pcmpgtq_xmm_xmmm128", Code.Pcmpgtq_xmm_xmmm128 },
				{ "VEX_Vpcmpgtq_xmm_xmm_xmmm128", Code.VEX_Vpcmpgtq_xmm_xmm_xmmm128 },
				{ "VEX_Vpcmpgtq_ymm_ymm_ymmm256", Code.VEX_Vpcmpgtq_ymm_ymm_ymmm256 },
				{ "EVEX_Vpcmpgtq_kr_k1_xmm_xmmm128b64", Code.EVEX_Vpcmpgtq_kr_k1_xmm_xmmm128b64 },
				{ "EVEX_Vpcmpgtq_kr_k1_ymm_ymmm256b64", Code.EVEX_Vpcmpgtq_kr_k1_ymm_ymmm256b64 },
				{ "EVEX_Vpcmpgtq_kr_k1_zmm_zmmm512b64", Code.EVEX_Vpcmpgtq_kr_k1_zmm_zmmm512b64 },
				{ "Pminsb_xmm_xmmm128", Code.Pminsb_xmm_xmmm128 },
				{ "VEX_Vpminsb_xmm_xmm_xmmm128", Code.VEX_Vpminsb_xmm_xmm_xmmm128 },
				{ "VEX_Vpminsb_ymm_ymm_ymmm256", Code.VEX_Vpminsb_ymm_ymm_ymmm256 },
				{ "EVEX_Vpminsb_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpminsb_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpminsb_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpminsb_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpminsb_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpminsb_zmm_k1z_zmm_zmmm512 },
				{ "EVEX_Vpmovm2d_xmm_kr", Code.EVEX_Vpmovm2d_xmm_kr },
				{ "EVEX_Vpmovm2d_ymm_kr", Code.EVEX_Vpmovm2d_ymm_kr },
				{ "EVEX_Vpmovm2d_zmm_kr", Code.EVEX_Vpmovm2d_zmm_kr },
				{ "EVEX_Vpmovm2q_xmm_kr", Code.EVEX_Vpmovm2q_xmm_kr },
				{ "EVEX_Vpmovm2q_ymm_kr", Code.EVEX_Vpmovm2q_ymm_kr },
				{ "EVEX_Vpmovm2q_zmm_kr", Code.EVEX_Vpmovm2q_zmm_kr },
				{ "Pminsd_xmm_xmmm128", Code.Pminsd_xmm_xmmm128 },
				{ "VEX_Vpminsd_xmm_xmm_xmmm128", Code.VEX_Vpminsd_xmm_xmm_xmmm128 },
				{ "VEX_Vpminsd_ymm_ymm_ymmm256", Code.VEX_Vpminsd_ymm_ymm_ymmm256 },
				{ "EVEX_Vpminsd_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpminsd_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpminsd_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpminsd_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpminsd_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpminsd_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpminsq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpminsq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpminsq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpminsq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpminsq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpminsq_zmm_k1z_zmm_zmmm512b64 },
				{ "EVEX_Vpmovd2m_kr_xmm", Code.EVEX_Vpmovd2m_kr_xmm },
				{ "EVEX_Vpmovd2m_kr_ymm", Code.EVEX_Vpmovd2m_kr_ymm },
				{ "EVEX_Vpmovd2m_kr_zmm", Code.EVEX_Vpmovd2m_kr_zmm },
				{ "EVEX_Vpmovq2m_kr_xmm", Code.EVEX_Vpmovq2m_kr_xmm },
				{ "EVEX_Vpmovq2m_kr_ymm", Code.EVEX_Vpmovq2m_kr_ymm },
				{ "EVEX_Vpmovq2m_kr_zmm", Code.EVEX_Vpmovq2m_kr_zmm },
				{ "Pminuw_xmm_xmmm128", Code.Pminuw_xmm_xmmm128 },
				{ "VEX_Vpminuw_xmm_xmm_xmmm128", Code.VEX_Vpminuw_xmm_xmm_xmmm128 },
				{ "VEX_Vpminuw_ymm_ymm_ymmm256", Code.VEX_Vpminuw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpminuw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpminuw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpminuw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpminuw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpminuw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpminuw_zmm_k1z_zmm_zmmm512 },
				{ "EVEX_Vpbroadcastmw2d_xmm_kr", Code.EVEX_Vpbroadcastmw2d_xmm_kr },
				{ "EVEX_Vpbroadcastmw2d_ymm_kr", Code.EVEX_Vpbroadcastmw2d_ymm_kr },
				{ "EVEX_Vpbroadcastmw2d_zmm_kr", Code.EVEX_Vpbroadcastmw2d_zmm_kr },
				{ "Pminud_xmm_xmmm128", Code.Pminud_xmm_xmmm128 },
				{ "VEX_Vpminud_xmm_xmm_xmmm128", Code.VEX_Vpminud_xmm_xmm_xmmm128 },
				{ "VEX_Vpminud_ymm_ymm_ymmm256", Code.VEX_Vpminud_ymm_ymm_ymmm256 },
				{ "EVEX_Vpminud_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpminud_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpminud_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpminud_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpminud_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpminud_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpminuq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpminuq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpminuq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpminuq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpminuq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpminuq_zmm_k1z_zmm_zmmm512b64 },
				{ "Pmaxsb_xmm_xmmm128", Code.Pmaxsb_xmm_xmmm128 },
				{ "VEX_Vpmaxsb_xmm_xmm_xmmm128", Code.VEX_Vpmaxsb_xmm_xmm_xmmm128 },
				{ "VEX_Vpmaxsb_ymm_ymm_ymmm256", Code.VEX_Vpmaxsb_ymm_ymm_ymmm256 },
				{ "EVEX_Vpmaxsb_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpmaxsb_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpmaxsb_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpmaxsb_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpmaxsb_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpmaxsb_zmm_k1z_zmm_zmmm512 },
				{ "Pmaxsd_xmm_xmmm128", Code.Pmaxsd_xmm_xmmm128 },
				{ "VEX_Vpmaxsd_xmm_xmm_xmmm128", Code.VEX_Vpmaxsd_xmm_xmm_xmmm128 },
				{ "VEX_Vpmaxsd_ymm_ymm_ymmm256", Code.VEX_Vpmaxsd_ymm_ymm_ymmm256 },
				{ "EVEX_Vpmaxsd_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpmaxsd_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpmaxsd_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpmaxsd_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpmaxsd_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpmaxsd_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpmaxsq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpmaxsq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpmaxsq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpmaxsq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpmaxsq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpmaxsq_zmm_k1z_zmm_zmmm512b64 },
				{ "Pmaxuw_xmm_xmmm128", Code.Pmaxuw_xmm_xmmm128 },
				{ "VEX_Vpmaxuw_xmm_xmm_xmmm128", Code.VEX_Vpmaxuw_xmm_xmm_xmmm128 },
				{ "VEX_Vpmaxuw_ymm_ymm_ymmm256", Code.VEX_Vpmaxuw_ymm_ymm_ymmm256 },
				{ "EVEX_Vpmaxuw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpmaxuw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpmaxuw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpmaxuw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpmaxuw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpmaxuw_zmm_k1z_zmm_zmmm512 },
				{ "Pmaxud_xmm_xmmm128", Code.Pmaxud_xmm_xmmm128 },
				{ "VEX_Vpmaxud_xmm_xmm_xmmm128", Code.VEX_Vpmaxud_xmm_xmm_xmmm128 },
				{ "VEX_Vpmaxud_ymm_ymm_ymmm256", Code.VEX_Vpmaxud_ymm_ymm_ymmm256 },
				{ "EVEX_Vpmaxud_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpmaxud_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpmaxud_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpmaxud_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpmaxud_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpmaxud_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpmaxuq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpmaxuq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpmaxuq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpmaxuq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpmaxuq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpmaxuq_zmm_k1z_zmm_zmmm512b64 },
				{ "Pmulld_xmm_xmmm128", Code.Pmulld_xmm_xmmm128 },
				{ "VEX_Vpmulld_xmm_xmm_xmmm128", Code.VEX_Vpmulld_xmm_xmm_xmmm128 },
				{ "VEX_Vpmulld_ymm_ymm_ymmm256", Code.VEX_Vpmulld_ymm_ymm_ymmm256 },
				{ "EVEX_Vpmulld_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpmulld_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpmulld_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpmulld_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpmulld_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpmulld_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpmullq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpmullq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpmullq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpmullq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpmullq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpmullq_zmm_k1z_zmm_zmmm512b64 },
				{ "Phminposuw_xmm_xmmm128", Code.Phminposuw_xmm_xmmm128 },
				{ "VEX_Vphminposuw_xmm_xmmm128", Code.VEX_Vphminposuw_xmm_xmmm128 },
				{ "EVEX_Vgetexpps_xmm_k1z_xmmm128b32", Code.EVEX_Vgetexpps_xmm_k1z_xmmm128b32 },
				{ "EVEX_Vgetexpps_ymm_k1z_ymmm256b32", Code.EVEX_Vgetexpps_ymm_k1z_ymmm256b32 },
				{ "EVEX_Vgetexpps_zmm_k1z_zmmm512b32_sae", Code.EVEX_Vgetexpps_zmm_k1z_zmmm512b32_sae },
				{ "EVEX_Vgetexppd_xmm_k1z_xmmm128b64", Code.EVEX_Vgetexppd_xmm_k1z_xmmm128b64 },
				{ "EVEX_Vgetexppd_ymm_k1z_ymmm256b64", Code.EVEX_Vgetexppd_ymm_k1z_ymmm256b64 },
				{ "EVEX_Vgetexppd_zmm_k1z_zmmm512b64_sae", Code.EVEX_Vgetexppd_zmm_k1z_zmmm512b64_sae },
				{ "EVEX_Vgetexpss_xmm_k1z_xmm_xmmm32_sae", Code.EVEX_Vgetexpss_xmm_k1z_xmm_xmmm32_sae },
				{ "EVEX_Vgetexpsd_xmm_k1z_xmm_xmmm64_sae", Code.EVEX_Vgetexpsd_xmm_k1z_xmm_xmmm64_sae },
				{ "EVEX_Vplzcntd_xmm_k1z_xmmm128b32", Code.EVEX_Vplzcntd_xmm_k1z_xmmm128b32 },
				{ "EVEX_Vplzcntd_ymm_k1z_ymmm256b32", Code.EVEX_Vplzcntd_ymm_k1z_ymmm256b32 },
				{ "EVEX_Vplzcntd_zmm_k1z_zmmm512b32", Code.EVEX_Vplzcntd_zmm_k1z_zmmm512b32 },
				{ "EVEX_Vplzcntq_xmm_k1z_xmmm128b64", Code.EVEX_Vplzcntq_xmm_k1z_xmmm128b64 },
				{ "EVEX_Vplzcntq_ymm_k1z_ymmm256b64", Code.EVEX_Vplzcntq_ymm_k1z_ymmm256b64 },
				{ "EVEX_Vplzcntq_zmm_k1z_zmmm512b64", Code.EVEX_Vplzcntq_zmm_k1z_zmmm512b64 },
				{ "VEX_Vpsrlvd_xmm_xmm_xmmm128", Code.VEX_Vpsrlvd_xmm_xmm_xmmm128 },
				{ "VEX_Vpsrlvd_ymm_ymm_ymmm256", Code.VEX_Vpsrlvd_ymm_ymm_ymmm256 },
				{ "VEX_Vpsrlvq_xmm_xmm_xmmm128", Code.VEX_Vpsrlvq_xmm_xmm_xmmm128 },
				{ "VEX_Vpsrlvq_ymm_ymm_ymmm256", Code.VEX_Vpsrlvq_ymm_ymm_ymmm256 },
				{ "EVEX_Vpsrlvd_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpsrlvd_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpsrlvd_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpsrlvd_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpsrlvd_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpsrlvd_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpsrlvq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpsrlvq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpsrlvq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpsrlvq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpsrlvq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpsrlvq_zmm_k1z_zmm_zmmm512b64 },
				{ "VEX_Vpsravd_xmm_xmm_xmmm128", Code.VEX_Vpsravd_xmm_xmm_xmmm128 },
				{ "VEX_Vpsravd_ymm_ymm_ymmm256", Code.VEX_Vpsravd_ymm_ymm_ymmm256 },
				{ "EVEX_Vpsravd_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpsravd_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpsravd_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpsravd_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpsravd_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpsravd_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpsravq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpsravq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpsravq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpsravq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpsravq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpsravq_zmm_k1z_zmm_zmmm512b64 },
				{ "VEX_Vpsllvd_xmm_xmm_xmmm128", Code.VEX_Vpsllvd_xmm_xmm_xmmm128 },
				{ "VEX_Vpsllvd_ymm_ymm_ymmm256", Code.VEX_Vpsllvd_ymm_ymm_ymmm256 },
				{ "VEX_Vpsllvq_xmm_xmm_xmmm128", Code.VEX_Vpsllvq_xmm_xmm_xmmm128 },
				{ "VEX_Vpsllvq_ymm_ymm_ymmm256", Code.VEX_Vpsllvq_ymm_ymm_ymmm256 },
				{ "EVEX_Vpsllvd_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpsllvd_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpsllvd_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpsllvd_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpsllvd_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpsllvd_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpsllvq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpsllvq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpsllvq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpsllvq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpsllvq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpsllvq_zmm_k1z_zmm_zmmm512b64 },
				{ "EVEX_Vrcp14ps_xmm_k1z_xmmm128b32", Code.EVEX_Vrcp14ps_xmm_k1z_xmmm128b32 },
				{ "EVEX_Vrcp14ps_ymm_k1z_ymmm256b32", Code.EVEX_Vrcp14ps_ymm_k1z_ymmm256b32 },
				{ "EVEX_Vrcp14ps_zmm_k1z_zmmm512b32", Code.EVEX_Vrcp14ps_zmm_k1z_zmmm512b32 },
				{ "EVEX_Vrcp14pd_xmm_k1z_xmmm128b64", Code.EVEX_Vrcp14pd_xmm_k1z_xmmm128b64 },
				{ "EVEX_Vrcp14pd_ymm_k1z_ymmm256b64", Code.EVEX_Vrcp14pd_ymm_k1z_ymmm256b64 },
				{ "EVEX_Vrcp14pd_zmm_k1z_zmmm512b64", Code.EVEX_Vrcp14pd_zmm_k1z_zmmm512b64 },
				{ "EVEX_Vrcp14ss_xmm_k1z_xmm_xmmm32", Code.EVEX_Vrcp14ss_xmm_k1z_xmm_xmmm32 },
				{ "EVEX_Vrcp14sd_xmm_k1z_xmm_xmmm64", Code.EVEX_Vrcp14sd_xmm_k1z_xmm_xmmm64 },
				{ "EVEX_Vrsqrt14ps_xmm_k1z_xmmm128b32", Code.EVEX_Vrsqrt14ps_xmm_k1z_xmmm128b32 },
				{ "EVEX_Vrsqrt14ps_ymm_k1z_ymmm256b32", Code.EVEX_Vrsqrt14ps_ymm_k1z_ymmm256b32 },
				{ "EVEX_Vrsqrt14ps_zmm_k1z_zmmm512b32", Code.EVEX_Vrsqrt14ps_zmm_k1z_zmmm512b32 },
				{ "EVEX_Vrsqrt14pd_xmm_k1z_xmmm128b64", Code.EVEX_Vrsqrt14pd_xmm_k1z_xmmm128b64 },
				{ "EVEX_Vrsqrt14pd_ymm_k1z_ymmm256b64", Code.EVEX_Vrsqrt14pd_ymm_k1z_ymmm256b64 },
				{ "EVEX_Vrsqrt14pd_zmm_k1z_zmmm512b64", Code.EVEX_Vrsqrt14pd_zmm_k1z_zmmm512b64 },
				{ "EVEX_Vrsqrt14ss_xmm_k1z_xmm_xmmm32", Code.EVEX_Vrsqrt14ss_xmm_k1z_xmm_xmmm32 },
				{ "EVEX_Vrsqrt14sd_xmm_k1z_xmm_xmmm64", Code.EVEX_Vrsqrt14sd_xmm_k1z_xmm_xmmm64 },
				{ "EVEX_Vpdpbusd_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpdpbusd_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpdpbusd_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpdpbusd_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpdpbusd_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpdpbusd_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpdpbusds_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpdpbusds_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpdpbusds_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpdpbusds_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpdpbusds_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpdpbusds_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpdpwssd_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpdpwssd_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpdpwssd_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpdpwssd_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpdpwssd_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpdpwssd_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vdpbf16ps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vdpbf16ps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vdpbf16ps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vdpbf16ps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vdpbf16ps_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vdpbf16ps_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vp4dpwssd_zmm_k1z_zmmp3_m128", Code.EVEX_Vp4dpwssd_zmm_k1z_zmmp3_m128 },
				{ "EVEX_Vpdpwssds_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpdpwssds_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpdpwssds_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpdpwssds_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpdpwssds_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpdpwssds_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vp4dpwssds_zmm_k1z_zmmp3_m128", Code.EVEX_Vp4dpwssds_zmm_k1z_zmmp3_m128 },
				{ "EVEX_Vpopcntb_xmm_k1z_xmmm128", Code.EVEX_Vpopcntb_xmm_k1z_xmmm128 },
				{ "EVEX_Vpopcntb_ymm_k1z_ymmm256", Code.EVEX_Vpopcntb_ymm_k1z_ymmm256 },
				{ "EVEX_Vpopcntb_zmm_k1z_zmmm512", Code.EVEX_Vpopcntb_zmm_k1z_zmmm512 },
				{ "EVEX_Vpopcntw_xmm_k1z_xmmm128", Code.EVEX_Vpopcntw_xmm_k1z_xmmm128 },
				{ "EVEX_Vpopcntw_ymm_k1z_ymmm256", Code.EVEX_Vpopcntw_ymm_k1z_ymmm256 },
				{ "EVEX_Vpopcntw_zmm_k1z_zmmm512", Code.EVEX_Vpopcntw_zmm_k1z_zmmm512 },
				{ "EVEX_Vpopcntd_xmm_k1z_xmmm128b32", Code.EVEX_Vpopcntd_xmm_k1z_xmmm128b32 },
				{ "EVEX_Vpopcntd_ymm_k1z_ymmm256b32", Code.EVEX_Vpopcntd_ymm_k1z_ymmm256b32 },
				{ "EVEX_Vpopcntd_zmm_k1z_zmmm512b32", Code.EVEX_Vpopcntd_zmm_k1z_zmmm512b32 },
				{ "EVEX_Vpopcntq_xmm_k1z_xmmm128b64", Code.EVEX_Vpopcntq_xmm_k1z_xmmm128b64 },
				{ "EVEX_Vpopcntq_ymm_k1z_ymmm256b64", Code.EVEX_Vpopcntq_ymm_k1z_ymmm256b64 },
				{ "EVEX_Vpopcntq_zmm_k1z_zmmm512b64", Code.EVEX_Vpopcntq_zmm_k1z_zmmm512b64 },
				{ "VEX_Vpbroadcastd_xmm_xmmm32", Code.VEX_Vpbroadcastd_xmm_xmmm32 },
				{ "VEX_Vpbroadcastd_ymm_xmmm32", Code.VEX_Vpbroadcastd_ymm_xmmm32 },
				{ "EVEX_Vpbroadcastd_xmm_k1z_xmmm32", Code.EVEX_Vpbroadcastd_xmm_k1z_xmmm32 },
				{ "EVEX_Vpbroadcastd_ymm_k1z_xmmm32", Code.EVEX_Vpbroadcastd_ymm_k1z_xmmm32 },
				{ "EVEX_Vpbroadcastd_zmm_k1z_xmmm32", Code.EVEX_Vpbroadcastd_zmm_k1z_xmmm32 },
				{ "VEX_Vpbroadcastq_xmm_xmmm64", Code.VEX_Vpbroadcastq_xmm_xmmm64 },
				{ "VEX_Vpbroadcastq_ymm_xmmm64", Code.VEX_Vpbroadcastq_ymm_xmmm64 },
				{ "EVEX_Vbroadcasti32x2_xmm_k1z_xmmm64", Code.EVEX_Vbroadcasti32x2_xmm_k1z_xmmm64 },
				{ "EVEX_Vbroadcasti32x2_ymm_k1z_xmmm64", Code.EVEX_Vbroadcasti32x2_ymm_k1z_xmmm64 },
				{ "EVEX_Vbroadcasti32x2_zmm_k1z_xmmm64", Code.EVEX_Vbroadcasti32x2_zmm_k1z_xmmm64 },
				{ "EVEX_Vpbroadcastq_xmm_k1z_xmmm64", Code.EVEX_Vpbroadcastq_xmm_k1z_xmmm64 },
				{ "EVEX_Vpbroadcastq_ymm_k1z_xmmm64", Code.EVEX_Vpbroadcastq_ymm_k1z_xmmm64 },
				{ "EVEX_Vpbroadcastq_zmm_k1z_xmmm64", Code.EVEX_Vpbroadcastq_zmm_k1z_xmmm64 },
				{ "VEX_Vbroadcasti128_ymm_m128", Code.VEX_Vbroadcasti128_ymm_m128 },
				{ "EVEX_Vbroadcasti32x4_ymm_k1z_m128", Code.EVEX_Vbroadcasti32x4_ymm_k1z_m128 },
				{ "EVEX_Vbroadcasti32x4_zmm_k1z_m128", Code.EVEX_Vbroadcasti32x4_zmm_k1z_m128 },
				{ "EVEX_Vbroadcasti64x2_ymm_k1z_m128", Code.EVEX_Vbroadcasti64x2_ymm_k1z_m128 },
				{ "EVEX_Vbroadcasti64x2_zmm_k1z_m128", Code.EVEX_Vbroadcasti64x2_zmm_k1z_m128 },
				{ "EVEX_Vbroadcasti32x8_zmm_k1z_m256", Code.EVEX_Vbroadcasti32x8_zmm_k1z_m256 },
				{ "EVEX_Vbroadcasti64x4_zmm_k1z_m256", Code.EVEX_Vbroadcasti64x4_zmm_k1z_m256 },
				{ "EVEX_Vpexpandb_xmm_k1z_xmmm128", Code.EVEX_Vpexpandb_xmm_k1z_xmmm128 },
				{ "EVEX_Vpexpandb_ymm_k1z_ymmm256", Code.EVEX_Vpexpandb_ymm_k1z_ymmm256 },
				{ "EVEX_Vpexpandb_zmm_k1z_zmmm512", Code.EVEX_Vpexpandb_zmm_k1z_zmmm512 },
				{ "EVEX_Vpexpandw_xmm_k1z_xmmm128", Code.EVEX_Vpexpandw_xmm_k1z_xmmm128 },
				{ "EVEX_Vpexpandw_ymm_k1z_ymmm256", Code.EVEX_Vpexpandw_ymm_k1z_ymmm256 },
				{ "EVEX_Vpexpandw_zmm_k1z_zmmm512", Code.EVEX_Vpexpandw_zmm_k1z_zmmm512 },
				{ "EVEX_Vpcompressb_xmmm128_k1z_xmm", Code.EVEX_Vpcompressb_xmmm128_k1z_xmm },
				{ "EVEX_Vpcompressb_ymmm256_k1z_ymm", Code.EVEX_Vpcompressb_ymmm256_k1z_ymm },
				{ "EVEX_Vpcompressb_zmmm512_k1z_zmm", Code.EVEX_Vpcompressb_zmmm512_k1z_zmm },
				{ "EVEX_Vpcompressw_xmmm128_k1z_xmm", Code.EVEX_Vpcompressw_xmmm128_k1z_xmm },
				{ "EVEX_Vpcompressw_ymmm256_k1z_ymm", Code.EVEX_Vpcompressw_ymmm256_k1z_ymm },
				{ "EVEX_Vpcompressw_zmmm512_k1z_zmm", Code.EVEX_Vpcompressw_zmmm512_k1z_zmm },
				{ "EVEX_Vpblendmd_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpblendmd_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpblendmd_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpblendmd_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpblendmd_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpblendmd_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpblendmq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpblendmq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpblendmq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpblendmq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpblendmq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpblendmq_zmm_k1z_zmm_zmmm512b64 },
				{ "EVEX_Vblendmps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vblendmps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vblendmps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vblendmps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vblendmps_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vblendmps_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vblendmpd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vblendmpd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vblendmpd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vblendmpd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vblendmpd_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vblendmpd_zmm_k1z_zmm_zmmm512b64 },
				{ "EVEX_Vpblendmb_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpblendmb_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpblendmb_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpblendmb_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpblendmb_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpblendmb_zmm_k1z_zmm_zmmm512 },
				{ "EVEX_Vpblendmw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpblendmw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpblendmw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpblendmw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpblendmw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpblendmw_zmm_k1z_zmm_zmmm512 },
				{ "EVEX_Vp2intersectd_kp1_xmm_xmmm128b32", Code.EVEX_Vp2intersectd_kp1_xmm_xmmm128b32 },
				{ "EVEX_Vp2intersectd_kp1_ymm_ymmm256b32", Code.EVEX_Vp2intersectd_kp1_ymm_ymmm256b32 },
				{ "EVEX_Vp2intersectd_kp1_zmm_zmmm512b32", Code.EVEX_Vp2intersectd_kp1_zmm_zmmm512b32 },
				{ "EVEX_Vp2intersectq_kp1_xmm_xmmm128b64", Code.EVEX_Vp2intersectq_kp1_xmm_xmmm128b64 },
				{ "EVEX_Vp2intersectq_kp1_ymm_ymmm256b64", Code.EVEX_Vp2intersectq_kp1_ymm_ymmm256b64 },
				{ "EVEX_Vp2intersectq_kp1_zmm_zmmm512b64", Code.EVEX_Vp2intersectq_kp1_zmm_zmmm512b64 },
				{ "EVEX_Vpshldvw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpshldvw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpshldvw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpshldvw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpshldvw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpshldvw_zmm_k1z_zmm_zmmm512 },
				{ "EVEX_Vpshldvd_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpshldvd_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpshldvd_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpshldvd_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpshldvd_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpshldvd_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpshldvq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpshldvq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpshldvq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpshldvq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpshldvq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpshldvq_zmm_k1z_zmm_zmmm512b64 },
				{ "EVEX_Vpshrdvw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpshrdvw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpshrdvw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpshrdvw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpshrdvw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpshrdvw_zmm_k1z_zmm_zmmm512 },
				{ "EVEX_Vcvtneps2bf16_xmm_k1z_xmmm128b32", Code.EVEX_Vcvtneps2bf16_xmm_k1z_xmmm128b32 },
				{ "EVEX_Vcvtneps2bf16_xmm_k1z_ymmm256b32", Code.EVEX_Vcvtneps2bf16_xmm_k1z_ymmm256b32 },
				{ "EVEX_Vcvtneps2bf16_ymm_k1z_zmmm512b32", Code.EVEX_Vcvtneps2bf16_ymm_k1z_zmmm512b32 },
				{ "EVEX_Vcvtne2ps2bf16_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vcvtne2ps2bf16_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vcvtne2ps2bf16_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vcvtne2ps2bf16_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vcvtne2ps2bf16_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vcvtne2ps2bf16_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpshrdvd_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpshrdvd_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpshrdvd_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpshrdvd_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpshrdvd_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpshrdvd_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpshrdvq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpshrdvq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpshrdvq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpshrdvq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpshrdvq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpshrdvq_zmm_k1z_zmm_zmmm512b64 },
				{ "EVEX_Vpermi2b_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpermi2b_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpermi2b_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpermi2b_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpermi2b_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpermi2b_zmm_k1z_zmm_zmmm512 },
				{ "EVEX_Vpermi2w_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpermi2w_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpermi2w_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpermi2w_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpermi2w_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpermi2w_zmm_k1z_zmm_zmmm512 },
				{ "EVEX_Vpermi2d_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpermi2d_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpermi2d_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpermi2d_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpermi2d_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpermi2d_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpermi2q_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpermi2q_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpermi2q_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpermi2q_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpermi2q_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpermi2q_zmm_k1z_zmm_zmmm512b64 },
				{ "EVEX_Vpermi2ps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpermi2ps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpermi2ps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpermi2ps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpermi2ps_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpermi2ps_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpermi2pd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpermi2pd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpermi2pd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpermi2pd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpermi2pd_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpermi2pd_zmm_k1z_zmm_zmmm512b64 },
				{ "VEX_Vpbroadcastb_xmm_xmmm8", Code.VEX_Vpbroadcastb_xmm_xmmm8 },
				{ "VEX_Vpbroadcastb_ymm_xmmm8", Code.VEX_Vpbroadcastb_ymm_xmmm8 },
				{ "EVEX_Vpbroadcastb_xmm_k1z_xmmm8", Code.EVEX_Vpbroadcastb_xmm_k1z_xmmm8 },
				{ "EVEX_Vpbroadcastb_ymm_k1z_xmmm8", Code.EVEX_Vpbroadcastb_ymm_k1z_xmmm8 },
				{ "EVEX_Vpbroadcastb_zmm_k1z_xmmm8", Code.EVEX_Vpbroadcastb_zmm_k1z_xmmm8 },
				{ "VEX_Vpbroadcastw_xmm_xmmm16", Code.VEX_Vpbroadcastw_xmm_xmmm16 },
				{ "VEX_Vpbroadcastw_ymm_xmmm16", Code.VEX_Vpbroadcastw_ymm_xmmm16 },
				{ "EVEX_Vpbroadcastw_xmm_k1z_xmmm16", Code.EVEX_Vpbroadcastw_xmm_k1z_xmmm16 },
				{ "EVEX_Vpbroadcastw_ymm_k1z_xmmm16", Code.EVEX_Vpbroadcastw_ymm_k1z_xmmm16 },
				{ "EVEX_Vpbroadcastw_zmm_k1z_xmmm16", Code.EVEX_Vpbroadcastw_zmm_k1z_xmmm16 },
				{ "EVEX_Vpbroadcastb_xmm_k1z_r32", Code.EVEX_Vpbroadcastb_xmm_k1z_r32 },
				{ "EVEX_Vpbroadcastb_ymm_k1z_r32", Code.EVEX_Vpbroadcastb_ymm_k1z_r32 },
				{ "EVEX_Vpbroadcastb_zmm_k1z_r32", Code.EVEX_Vpbroadcastb_zmm_k1z_r32 },
				{ "EVEX_Vpbroadcastw_xmm_k1z_r32", Code.EVEX_Vpbroadcastw_xmm_k1z_r32 },
				{ "EVEX_Vpbroadcastw_ymm_k1z_r32", Code.EVEX_Vpbroadcastw_ymm_k1z_r32 },
				{ "EVEX_Vpbroadcastw_zmm_k1z_r32", Code.EVEX_Vpbroadcastw_zmm_k1z_r32 },
				{ "EVEX_Vpbroadcastd_xmm_k1z_r32", Code.EVEX_Vpbroadcastd_xmm_k1z_r32 },
				{ "EVEX_Vpbroadcastd_ymm_k1z_r32", Code.EVEX_Vpbroadcastd_ymm_k1z_r32 },
				{ "EVEX_Vpbroadcastd_zmm_k1z_r32", Code.EVEX_Vpbroadcastd_zmm_k1z_r32 },
				{ "EVEX_Vpbroadcastq_xmm_k1z_r64", Code.EVEX_Vpbroadcastq_xmm_k1z_r64 },
				{ "EVEX_Vpbroadcastq_ymm_k1z_r64", Code.EVEX_Vpbroadcastq_ymm_k1z_r64 },
				{ "EVEX_Vpbroadcastq_zmm_k1z_r64", Code.EVEX_Vpbroadcastq_zmm_k1z_r64 },
				{ "EVEX_Vpermt2b_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpermt2b_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpermt2b_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpermt2b_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpermt2b_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpermt2b_zmm_k1z_zmm_zmmm512 },
				{ "EVEX_Vpermt2w_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpermt2w_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpermt2w_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpermt2w_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpermt2w_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpermt2w_zmm_k1z_zmm_zmmm512 },
				{ "EVEX_Vpermt2d_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpermt2d_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpermt2d_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpermt2d_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpermt2d_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpermt2d_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpermt2q_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpermt2q_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpermt2q_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpermt2q_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpermt2q_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpermt2q_zmm_k1z_zmm_zmmm512b64 },
				{ "EVEX_Vpermt2ps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vpermt2ps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vpermt2ps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vpermt2ps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vpermt2ps_zmm_k1z_zmm_zmmm512b32", Code.EVEX_Vpermt2ps_zmm_k1z_zmm_zmmm512b32 },
				{ "EVEX_Vpermt2pd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpermt2pd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpermt2pd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpermt2pd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpermt2pd_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpermt2pd_zmm_k1z_zmm_zmmm512b64 },
				{ "Invept_r32_m128", Code.Invept_r32_m128 },
				{ "Invept_r64_m128", Code.Invept_r64_m128 },
				{ "Invvpid_r32_m128", Code.Invvpid_r32_m128 },
				{ "Invvpid_r64_m128", Code.Invvpid_r64_m128 },
				{ "Invpcid_r32_m128", Code.Invpcid_r32_m128 },
				{ "Invpcid_r64_m128", Code.Invpcid_r64_m128 },
				{ "EVEX_Vpmultishiftqb_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpmultishiftqb_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpmultishiftqb_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpmultishiftqb_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpmultishiftqb_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpmultishiftqb_zmm_k1z_zmm_zmmm512b64 },
				{ "EVEX_Vexpandps_xmm_k1z_xmmm128", Code.EVEX_Vexpandps_xmm_k1z_xmmm128 },
				{ "EVEX_Vexpandps_ymm_k1z_ymmm256", Code.EVEX_Vexpandps_ymm_k1z_ymmm256 },
				{ "EVEX_Vexpandps_zmm_k1z_zmmm512", Code.EVEX_Vexpandps_zmm_k1z_zmmm512 },
				{ "EVEX_Vexpandpd_xmm_k1z_xmmm128", Code.EVEX_Vexpandpd_xmm_k1z_xmmm128 },
				{ "EVEX_Vexpandpd_ymm_k1z_ymmm256", Code.EVEX_Vexpandpd_ymm_k1z_ymmm256 },
				{ "EVEX_Vexpandpd_zmm_k1z_zmmm512", Code.EVEX_Vexpandpd_zmm_k1z_zmmm512 },
				{ "EVEX_Vpexpandd_xmm_k1z_xmmm128", Code.EVEX_Vpexpandd_xmm_k1z_xmmm128 },
				{ "EVEX_Vpexpandd_ymm_k1z_ymmm256", Code.EVEX_Vpexpandd_ymm_k1z_ymmm256 },
				{ "EVEX_Vpexpandd_zmm_k1z_zmmm512", Code.EVEX_Vpexpandd_zmm_k1z_zmmm512 },
				{ "EVEX_Vpexpandq_xmm_k1z_xmmm128", Code.EVEX_Vpexpandq_xmm_k1z_xmmm128 },
				{ "EVEX_Vpexpandq_ymm_k1z_ymmm256", Code.EVEX_Vpexpandq_ymm_k1z_ymmm256 },
				{ "EVEX_Vpexpandq_zmm_k1z_zmmm512", Code.EVEX_Vpexpandq_zmm_k1z_zmmm512 },
				{ "EVEX_Vcompressps_xmmm128_k1z_xmm", Code.EVEX_Vcompressps_xmmm128_k1z_xmm },
				{ "EVEX_Vcompressps_ymmm256_k1z_ymm", Code.EVEX_Vcompressps_ymmm256_k1z_ymm },
				{ "EVEX_Vcompressps_zmmm512_k1z_zmm", Code.EVEX_Vcompressps_zmmm512_k1z_zmm },
				{ "EVEX_Vcompresspd_xmmm128_k1z_xmm", Code.EVEX_Vcompresspd_xmmm128_k1z_xmm },
				{ "EVEX_Vcompresspd_ymmm256_k1z_ymm", Code.EVEX_Vcompresspd_ymmm256_k1z_ymm },
				{ "EVEX_Vcompresspd_zmmm512_k1z_zmm", Code.EVEX_Vcompresspd_zmmm512_k1z_zmm },
				{ "EVEX_Vpcompressd_xmmm128_k1z_xmm", Code.EVEX_Vpcompressd_xmmm128_k1z_xmm },
				{ "EVEX_Vpcompressd_ymmm256_k1z_ymm", Code.EVEX_Vpcompressd_ymmm256_k1z_ymm },
				{ "EVEX_Vpcompressd_zmmm512_k1z_zmm", Code.EVEX_Vpcompressd_zmmm512_k1z_zmm },
				{ "EVEX_Vpcompressq_xmmm128_k1z_xmm", Code.EVEX_Vpcompressq_xmmm128_k1z_xmm },
				{ "EVEX_Vpcompressq_ymmm256_k1z_ymm", Code.EVEX_Vpcompressq_ymmm256_k1z_ymm },
				{ "EVEX_Vpcompressq_zmmm512_k1z_zmm", Code.EVEX_Vpcompressq_zmmm512_k1z_zmm },
				{ "VEX_Vpmaskmovd_xmm_xmm_m128", Code.VEX_Vpmaskmovd_xmm_xmm_m128 },
				{ "VEX_Vpmaskmovd_ymm_ymm_m256", Code.VEX_Vpmaskmovd_ymm_ymm_m256 },
				{ "VEX_Vpmaskmovq_xmm_xmm_m128", Code.VEX_Vpmaskmovq_xmm_xmm_m128 },
				{ "VEX_Vpmaskmovq_ymm_ymm_m256", Code.VEX_Vpmaskmovq_ymm_ymm_m256 },
				{ "EVEX_Vpermb_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpermb_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpermb_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpermb_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpermb_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpermb_zmm_k1z_zmm_zmmm512 },
				{ "EVEX_Vpermw_xmm_k1z_xmm_xmmm128", Code.EVEX_Vpermw_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vpermw_ymm_k1z_ymm_ymmm256", Code.EVEX_Vpermw_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vpermw_zmm_k1z_zmm_zmmm512", Code.EVEX_Vpermw_zmm_k1z_zmm_zmmm512 },
				{ "VEX_Vpmaskmovd_m128_xmm_xmm", Code.VEX_Vpmaskmovd_m128_xmm_xmm },
				{ "VEX_Vpmaskmovd_m256_ymm_ymm", Code.VEX_Vpmaskmovd_m256_ymm_ymm },
				{ "VEX_Vpmaskmovq_m128_xmm_xmm", Code.VEX_Vpmaskmovq_m128_xmm_xmm },
				{ "VEX_Vpmaskmovq_m256_ymm_ymm", Code.VEX_Vpmaskmovq_m256_ymm_ymm },
				{ "EVEX_Vpshufbitqmb_kr_k1_xmm_xmmm128", Code.EVEX_Vpshufbitqmb_kr_k1_xmm_xmmm128 },
				{ "EVEX_Vpshufbitqmb_kr_k1_ymm_ymmm256", Code.EVEX_Vpshufbitqmb_kr_k1_ymm_ymmm256 },
				{ "EVEX_Vpshufbitqmb_kr_k1_zmm_zmmm512", Code.EVEX_Vpshufbitqmb_kr_k1_zmm_zmmm512 },
				{ "VEX_Vpgatherdd_xmm_vm32x_xmm", Code.VEX_Vpgatherdd_xmm_vm32x_xmm },
				{ "VEX_Vpgatherdd_ymm_vm32y_ymm", Code.VEX_Vpgatherdd_ymm_vm32y_ymm },
				{ "VEX_Vpgatherdq_xmm_vm32x_xmm", Code.VEX_Vpgatherdq_xmm_vm32x_xmm },
				{ "VEX_Vpgatherdq_ymm_vm32x_ymm", Code.VEX_Vpgatherdq_ymm_vm32x_ymm },
				{ "EVEX_Vpgatherdd_xmm_k1_vm32x", Code.EVEX_Vpgatherdd_xmm_k1_vm32x },
				{ "EVEX_Vpgatherdd_ymm_k1_vm32y", Code.EVEX_Vpgatherdd_ymm_k1_vm32y },
				{ "EVEX_Vpgatherdd_zmm_k1_vm32z", Code.EVEX_Vpgatherdd_zmm_k1_vm32z },
				{ "EVEX_Vpgatherdq_xmm_k1_vm32x", Code.EVEX_Vpgatherdq_xmm_k1_vm32x },
				{ "EVEX_Vpgatherdq_ymm_k1_vm32x", Code.EVEX_Vpgatherdq_ymm_k1_vm32x },
				{ "EVEX_Vpgatherdq_zmm_k1_vm32y", Code.EVEX_Vpgatherdq_zmm_k1_vm32y },
				{ "VEX_Vpgatherqd_xmm_vm64x_xmm", Code.VEX_Vpgatherqd_xmm_vm64x_xmm },
				{ "VEX_Vpgatherqd_xmm_vm64y_xmm", Code.VEX_Vpgatherqd_xmm_vm64y_xmm },
				{ "VEX_Vpgatherqq_xmm_vm64x_xmm", Code.VEX_Vpgatherqq_xmm_vm64x_xmm },
				{ "VEX_Vpgatherqq_ymm_vm64y_ymm", Code.VEX_Vpgatherqq_ymm_vm64y_ymm },
				{ "EVEX_Vpgatherqd_xmm_k1_vm64x", Code.EVEX_Vpgatherqd_xmm_k1_vm64x },
				{ "EVEX_Vpgatherqd_xmm_k1_vm64y", Code.EVEX_Vpgatherqd_xmm_k1_vm64y },
				{ "EVEX_Vpgatherqd_ymm_k1_vm64z", Code.EVEX_Vpgatherqd_ymm_k1_vm64z },
				{ "EVEX_Vpgatherqq_xmm_k1_vm64x", Code.EVEX_Vpgatherqq_xmm_k1_vm64x },
				{ "EVEX_Vpgatherqq_ymm_k1_vm64y", Code.EVEX_Vpgatherqq_ymm_k1_vm64y },
				{ "EVEX_Vpgatherqq_zmm_k1_vm64z", Code.EVEX_Vpgatherqq_zmm_k1_vm64z },
				{ "VEX_Vgatherdps_xmm_vm32x_xmm", Code.VEX_Vgatherdps_xmm_vm32x_xmm },
				{ "VEX_Vgatherdps_ymm_vm32y_ymm", Code.VEX_Vgatherdps_ymm_vm32y_ymm },
				{ "VEX_Vgatherdpd_xmm_vm32x_xmm", Code.VEX_Vgatherdpd_xmm_vm32x_xmm },
				{ "VEX_Vgatherdpd_ymm_vm32x_ymm", Code.VEX_Vgatherdpd_ymm_vm32x_ymm },
				{ "EVEX_Vgatherdps_xmm_k1_vm32x", Code.EVEX_Vgatherdps_xmm_k1_vm32x },
				{ "EVEX_Vgatherdps_ymm_k1_vm32y", Code.EVEX_Vgatherdps_ymm_k1_vm32y },
				{ "EVEX_Vgatherdps_zmm_k1_vm32z", Code.EVEX_Vgatherdps_zmm_k1_vm32z },
				{ "EVEX_Vgatherdpd_xmm_k1_vm32x", Code.EVEX_Vgatherdpd_xmm_k1_vm32x },
				{ "EVEX_Vgatherdpd_ymm_k1_vm32x", Code.EVEX_Vgatherdpd_ymm_k1_vm32x },
				{ "EVEX_Vgatherdpd_zmm_k1_vm32y", Code.EVEX_Vgatherdpd_zmm_k1_vm32y },
				{ "VEX_Vgatherqps_xmm_vm64x_xmm", Code.VEX_Vgatherqps_xmm_vm64x_xmm },
				{ "VEX_Vgatherqps_xmm_vm64y_xmm", Code.VEX_Vgatherqps_xmm_vm64y_xmm },
				{ "VEX_Vgatherqpd_xmm_vm64x_xmm", Code.VEX_Vgatherqpd_xmm_vm64x_xmm },
				{ "VEX_Vgatherqpd_ymm_vm64y_ymm", Code.VEX_Vgatherqpd_ymm_vm64y_ymm },
				{ "EVEX_Vgatherqps_xmm_k1_vm64x", Code.EVEX_Vgatherqps_xmm_k1_vm64x },
				{ "EVEX_Vgatherqps_xmm_k1_vm64y", Code.EVEX_Vgatherqps_xmm_k1_vm64y },
				{ "EVEX_Vgatherqps_ymm_k1_vm64z", Code.EVEX_Vgatherqps_ymm_k1_vm64z },
				{ "EVEX_Vgatherqpd_xmm_k1_vm64x", Code.EVEX_Vgatherqpd_xmm_k1_vm64x },
				{ "EVEX_Vgatherqpd_ymm_k1_vm64y", Code.EVEX_Vgatherqpd_ymm_k1_vm64y },
				{ "EVEX_Vgatherqpd_zmm_k1_vm64z", Code.EVEX_Vgatherqpd_zmm_k1_vm64z },
				{ "VEX_Vfmaddsub132ps_xmm_xmm_xmmm128", Code.VEX_Vfmaddsub132ps_xmm_xmm_xmmm128 },
				{ "VEX_Vfmaddsub132ps_ymm_ymm_ymmm256", Code.VEX_Vfmaddsub132ps_ymm_ymm_ymmm256 },
				{ "VEX_Vfmaddsub132pd_xmm_xmm_xmmm128", Code.VEX_Vfmaddsub132pd_xmm_xmm_xmmm128 },
				{ "VEX_Vfmaddsub132pd_ymm_ymm_ymmm256", Code.VEX_Vfmaddsub132pd_ymm_ymm_ymmm256 },
				{ "EVEX_Vfmaddsub132ps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vfmaddsub132ps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vfmaddsub132ps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vfmaddsub132ps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vfmaddsub132ps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vfmaddsub132ps_zmm_k1z_zmm_zmmm512b32_er },
				{ "EVEX_Vfmaddsub132pd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vfmaddsub132pd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vfmaddsub132pd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vfmaddsub132pd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vfmaddsub132pd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vfmaddsub132pd_zmm_k1z_zmm_zmmm512b64_er },
				{ "VEX_Vfmsubadd132ps_xmm_xmm_xmmm128", Code.VEX_Vfmsubadd132ps_xmm_xmm_xmmm128 },
				{ "VEX_Vfmsubadd132ps_ymm_ymm_ymmm256", Code.VEX_Vfmsubadd132ps_ymm_ymm_ymmm256 },
				{ "VEX_Vfmsubadd132pd_xmm_xmm_xmmm128", Code.VEX_Vfmsubadd132pd_xmm_xmm_xmmm128 },
				{ "VEX_Vfmsubadd132pd_ymm_ymm_ymmm256", Code.VEX_Vfmsubadd132pd_ymm_ymm_ymmm256 },
				{ "EVEX_Vfmsubadd132ps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vfmsubadd132ps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vfmsubadd132ps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vfmsubadd132ps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vfmsubadd132ps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vfmsubadd132ps_zmm_k1z_zmm_zmmm512b32_er },
				{ "EVEX_Vfmsubadd132pd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vfmsubadd132pd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vfmsubadd132pd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vfmsubadd132pd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vfmsubadd132pd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vfmsubadd132pd_zmm_k1z_zmm_zmmm512b64_er },
				{ "VEX_Vfmadd132ps_xmm_xmm_xmmm128", Code.VEX_Vfmadd132ps_xmm_xmm_xmmm128 },
				{ "VEX_Vfmadd132ps_ymm_ymm_ymmm256", Code.VEX_Vfmadd132ps_ymm_ymm_ymmm256 },
				{ "VEX_Vfmadd132pd_xmm_xmm_xmmm128", Code.VEX_Vfmadd132pd_xmm_xmm_xmmm128 },
				{ "VEX_Vfmadd132pd_ymm_ymm_ymmm256", Code.VEX_Vfmadd132pd_ymm_ymm_ymmm256 },
				{ "EVEX_Vfmadd132ps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vfmadd132ps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vfmadd132ps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vfmadd132ps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vfmadd132ps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vfmadd132ps_zmm_k1z_zmm_zmmm512b32_er },
				{ "EVEX_Vfmadd132pd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vfmadd132pd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vfmadd132pd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vfmadd132pd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vfmadd132pd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vfmadd132pd_zmm_k1z_zmm_zmmm512b64_er },
				{ "VEX_Vfmadd132ss_xmm_xmm_xmmm32", Code.VEX_Vfmadd132ss_xmm_xmm_xmmm32 },
				{ "VEX_Vfmadd132sd_xmm_xmm_xmmm64", Code.VEX_Vfmadd132sd_xmm_xmm_xmmm64 },
				{ "EVEX_Vfmadd132ss_xmm_k1z_xmm_xmmm32_er", Code.EVEX_Vfmadd132ss_xmm_k1z_xmm_xmmm32_er },
				{ "EVEX_Vfmadd132sd_xmm_k1z_xmm_xmmm64_er", Code.EVEX_Vfmadd132sd_xmm_k1z_xmm_xmmm64_er },
				{ "VEX_Vfmsub132ps_xmm_xmm_xmmm128", Code.VEX_Vfmsub132ps_xmm_xmm_xmmm128 },
				{ "VEX_Vfmsub132ps_ymm_ymm_ymmm256", Code.VEX_Vfmsub132ps_ymm_ymm_ymmm256 },
				{ "VEX_Vfmsub132pd_xmm_xmm_xmmm128", Code.VEX_Vfmsub132pd_xmm_xmm_xmmm128 },
				{ "VEX_Vfmsub132pd_ymm_ymm_ymmm256", Code.VEX_Vfmsub132pd_ymm_ymm_ymmm256 },
				{ "EVEX_Vfmsub132ps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vfmsub132ps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vfmsub132ps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vfmsub132ps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vfmsub132ps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vfmsub132ps_zmm_k1z_zmm_zmmm512b32_er },
				{ "EVEX_Vfmsub132pd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vfmsub132pd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vfmsub132pd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vfmsub132pd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vfmsub132pd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vfmsub132pd_zmm_k1z_zmm_zmmm512b64_er },
				{ "EVEX_V4fmaddps_zmm_k1z_zmmp3_m128", Code.EVEX_V4fmaddps_zmm_k1z_zmmp3_m128 },
				{ "VEX_Vfmsub132ss_xmm_xmm_xmmm32", Code.VEX_Vfmsub132ss_xmm_xmm_xmmm32 },
				{ "VEX_Vfmsub132sd_xmm_xmm_xmmm64", Code.VEX_Vfmsub132sd_xmm_xmm_xmmm64 },
				{ "EVEX_Vfmsub132ss_xmm_k1z_xmm_xmmm32_er", Code.EVEX_Vfmsub132ss_xmm_k1z_xmm_xmmm32_er },
				{ "EVEX_Vfmsub132sd_xmm_k1z_xmm_xmmm64_er", Code.EVEX_Vfmsub132sd_xmm_k1z_xmm_xmmm64_er },
				{ "EVEX_V4fmaddss_xmm_k1z_xmmp3_m128", Code.EVEX_V4fmaddss_xmm_k1z_xmmp3_m128 },
				{ "VEX_Vfnmadd132ps_xmm_xmm_xmmm128", Code.VEX_Vfnmadd132ps_xmm_xmm_xmmm128 },
				{ "VEX_Vfnmadd132ps_ymm_ymm_ymmm256", Code.VEX_Vfnmadd132ps_ymm_ymm_ymmm256 },
				{ "VEX_Vfnmadd132pd_xmm_xmm_xmmm128", Code.VEX_Vfnmadd132pd_xmm_xmm_xmmm128 },
				{ "VEX_Vfnmadd132pd_ymm_ymm_ymmm256", Code.VEX_Vfnmadd132pd_ymm_ymm_ymmm256 },
				{ "EVEX_Vfnmadd132ps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vfnmadd132ps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vfnmadd132ps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vfnmadd132ps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vfnmadd132ps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vfnmadd132ps_zmm_k1z_zmm_zmmm512b32_er },
				{ "EVEX_Vfnmadd132pd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vfnmadd132pd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vfnmadd132pd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vfnmadd132pd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vfnmadd132pd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vfnmadd132pd_zmm_k1z_zmm_zmmm512b64_er },
				{ "VEX_Vfnmadd132ss_xmm_xmm_xmmm32", Code.VEX_Vfnmadd132ss_xmm_xmm_xmmm32 },
				{ "VEX_Vfnmadd132sd_xmm_xmm_xmmm64", Code.VEX_Vfnmadd132sd_xmm_xmm_xmmm64 },
				{ "EVEX_Vfnmadd132ss_xmm_k1z_xmm_xmmm32_er", Code.EVEX_Vfnmadd132ss_xmm_k1z_xmm_xmmm32_er },
				{ "EVEX_Vfnmadd132sd_xmm_k1z_xmm_xmmm64_er", Code.EVEX_Vfnmadd132sd_xmm_k1z_xmm_xmmm64_er },
				{ "VEX_Vfnmsub132ps_xmm_xmm_xmmm128", Code.VEX_Vfnmsub132ps_xmm_xmm_xmmm128 },
				{ "VEX_Vfnmsub132ps_ymm_ymm_ymmm256", Code.VEX_Vfnmsub132ps_ymm_ymm_ymmm256 },
				{ "VEX_Vfnmsub132pd_xmm_xmm_xmmm128", Code.VEX_Vfnmsub132pd_xmm_xmm_xmmm128 },
				{ "VEX_Vfnmsub132pd_ymm_ymm_ymmm256", Code.VEX_Vfnmsub132pd_ymm_ymm_ymmm256 },
				{ "EVEX_Vfnmsub132ps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vfnmsub132ps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vfnmsub132ps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vfnmsub132ps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vfnmsub132ps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vfnmsub132ps_zmm_k1z_zmm_zmmm512b32_er },
				{ "EVEX_Vfnmsub132pd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vfnmsub132pd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vfnmsub132pd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vfnmsub132pd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vfnmsub132pd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vfnmsub132pd_zmm_k1z_zmm_zmmm512b64_er },
				{ "VEX_Vfnmsub132ss_xmm_xmm_xmmm32", Code.VEX_Vfnmsub132ss_xmm_xmm_xmmm32 },
				{ "VEX_Vfnmsub132sd_xmm_xmm_xmmm64", Code.VEX_Vfnmsub132sd_xmm_xmm_xmmm64 },
				{ "EVEX_Vfnmsub132ss_xmm_k1z_xmm_xmmm32_er", Code.EVEX_Vfnmsub132ss_xmm_k1z_xmm_xmmm32_er },
				{ "EVEX_Vfnmsub132sd_xmm_k1z_xmm_xmmm64_er", Code.EVEX_Vfnmsub132sd_xmm_k1z_xmm_xmmm64_er },
				{ "EVEX_Vpscatterdd_vm32x_k1_xmm", Code.EVEX_Vpscatterdd_vm32x_k1_xmm },
				{ "EVEX_Vpscatterdd_vm32y_k1_ymm", Code.EVEX_Vpscatterdd_vm32y_k1_ymm },
				{ "EVEX_Vpscatterdd_vm32z_k1_zmm", Code.EVEX_Vpscatterdd_vm32z_k1_zmm },
				{ "EVEX_Vpscatterdq_vm32x_k1_xmm", Code.EVEX_Vpscatterdq_vm32x_k1_xmm },
				{ "EVEX_Vpscatterdq_vm32x_k1_ymm", Code.EVEX_Vpscatterdq_vm32x_k1_ymm },
				{ "EVEX_Vpscatterdq_vm32y_k1_zmm", Code.EVEX_Vpscatterdq_vm32y_k1_zmm },
				{ "EVEX_Vpscatterqd_vm64x_k1_xmm", Code.EVEX_Vpscatterqd_vm64x_k1_xmm },
				{ "EVEX_Vpscatterqd_vm64y_k1_xmm", Code.EVEX_Vpscatterqd_vm64y_k1_xmm },
				{ "EVEX_Vpscatterqd_vm64z_k1_ymm", Code.EVEX_Vpscatterqd_vm64z_k1_ymm },
				{ "EVEX_Vpscatterqq_vm64x_k1_xmm", Code.EVEX_Vpscatterqq_vm64x_k1_xmm },
				{ "EVEX_Vpscatterqq_vm64y_k1_ymm", Code.EVEX_Vpscatterqq_vm64y_k1_ymm },
				{ "EVEX_Vpscatterqq_vm64z_k1_zmm", Code.EVEX_Vpscatterqq_vm64z_k1_zmm },
				{ "EVEX_Vscatterdps_vm32x_k1_xmm", Code.EVEX_Vscatterdps_vm32x_k1_xmm },
				{ "EVEX_Vscatterdps_vm32y_k1_ymm", Code.EVEX_Vscatterdps_vm32y_k1_ymm },
				{ "EVEX_Vscatterdps_vm32z_k1_zmm", Code.EVEX_Vscatterdps_vm32z_k1_zmm },
				{ "EVEX_Vscatterdpd_vm32x_k1_xmm", Code.EVEX_Vscatterdpd_vm32x_k1_xmm },
				{ "EVEX_Vscatterdpd_vm32x_k1_ymm", Code.EVEX_Vscatterdpd_vm32x_k1_ymm },
				{ "EVEX_Vscatterdpd_vm32y_k1_zmm", Code.EVEX_Vscatterdpd_vm32y_k1_zmm },
				{ "EVEX_Vscatterqps_vm64x_k1_xmm", Code.EVEX_Vscatterqps_vm64x_k1_xmm },
				{ "EVEX_Vscatterqps_vm64y_k1_xmm", Code.EVEX_Vscatterqps_vm64y_k1_xmm },
				{ "EVEX_Vscatterqps_vm64z_k1_ymm", Code.EVEX_Vscatterqps_vm64z_k1_ymm },
				{ "EVEX_Vscatterqpd_vm64x_k1_xmm", Code.EVEX_Vscatterqpd_vm64x_k1_xmm },
				{ "EVEX_Vscatterqpd_vm64y_k1_ymm", Code.EVEX_Vscatterqpd_vm64y_k1_ymm },
				{ "EVEX_Vscatterqpd_vm64z_k1_zmm", Code.EVEX_Vscatterqpd_vm64z_k1_zmm },
				{ "VEX_Vfmaddsub213ps_xmm_xmm_xmmm128", Code.VEX_Vfmaddsub213ps_xmm_xmm_xmmm128 },
				{ "VEX_Vfmaddsub213ps_ymm_ymm_ymmm256", Code.VEX_Vfmaddsub213ps_ymm_ymm_ymmm256 },
				{ "VEX_Vfmaddsub213pd_xmm_xmm_xmmm128", Code.VEX_Vfmaddsub213pd_xmm_xmm_xmmm128 },
				{ "VEX_Vfmaddsub213pd_ymm_ymm_ymmm256", Code.VEX_Vfmaddsub213pd_ymm_ymm_ymmm256 },
				{ "EVEX_Vfmaddsub213ps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vfmaddsub213ps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vfmaddsub213ps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vfmaddsub213ps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vfmaddsub213ps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vfmaddsub213ps_zmm_k1z_zmm_zmmm512b32_er },
				{ "EVEX_Vfmaddsub213pd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vfmaddsub213pd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vfmaddsub213pd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vfmaddsub213pd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vfmaddsub213pd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vfmaddsub213pd_zmm_k1z_zmm_zmmm512b64_er },
				{ "VEX_Vfmsubadd213ps_xmm_xmm_xmmm128", Code.VEX_Vfmsubadd213ps_xmm_xmm_xmmm128 },
				{ "VEX_Vfmsubadd213ps_ymm_ymm_ymmm256", Code.VEX_Vfmsubadd213ps_ymm_ymm_ymmm256 },
				{ "VEX_Vfmsubadd213pd_xmm_xmm_xmmm128", Code.VEX_Vfmsubadd213pd_xmm_xmm_xmmm128 },
				{ "VEX_Vfmsubadd213pd_ymm_ymm_ymmm256", Code.VEX_Vfmsubadd213pd_ymm_ymm_ymmm256 },
				{ "EVEX_Vfmsubadd213ps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vfmsubadd213ps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vfmsubadd213ps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vfmsubadd213ps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vfmsubadd213ps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vfmsubadd213ps_zmm_k1z_zmm_zmmm512b32_er },
				{ "EVEX_Vfmsubadd213pd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vfmsubadd213pd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vfmsubadd213pd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vfmsubadd213pd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vfmsubadd213pd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vfmsubadd213pd_zmm_k1z_zmm_zmmm512b64_er },
				{ "VEX_Vfmadd213ps_xmm_xmm_xmmm128", Code.VEX_Vfmadd213ps_xmm_xmm_xmmm128 },
				{ "VEX_Vfmadd213ps_ymm_ymm_ymmm256", Code.VEX_Vfmadd213ps_ymm_ymm_ymmm256 },
				{ "VEX_Vfmadd213pd_xmm_xmm_xmmm128", Code.VEX_Vfmadd213pd_xmm_xmm_xmmm128 },
				{ "VEX_Vfmadd213pd_ymm_ymm_ymmm256", Code.VEX_Vfmadd213pd_ymm_ymm_ymmm256 },
				{ "EVEX_Vfmadd213ps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vfmadd213ps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vfmadd213ps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vfmadd213ps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vfmadd213ps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vfmadd213ps_zmm_k1z_zmm_zmmm512b32_er },
				{ "EVEX_Vfmadd213pd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vfmadd213pd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vfmadd213pd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vfmadd213pd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vfmadd213pd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vfmadd213pd_zmm_k1z_zmm_zmmm512b64_er },
				{ "VEX_Vfmadd213ss_xmm_xmm_xmmm32", Code.VEX_Vfmadd213ss_xmm_xmm_xmmm32 },
				{ "VEX_Vfmadd213sd_xmm_xmm_xmmm64", Code.VEX_Vfmadd213sd_xmm_xmm_xmmm64 },
				{ "EVEX_Vfmadd213ss_xmm_k1z_xmm_xmmm32_er", Code.EVEX_Vfmadd213ss_xmm_k1z_xmm_xmmm32_er },
				{ "EVEX_Vfmadd213sd_xmm_k1z_xmm_xmmm64_er", Code.EVEX_Vfmadd213sd_xmm_k1z_xmm_xmmm64_er },
				{ "VEX_Vfmsub213ps_xmm_xmm_xmmm128", Code.VEX_Vfmsub213ps_xmm_xmm_xmmm128 },
				{ "VEX_Vfmsub213ps_ymm_ymm_ymmm256", Code.VEX_Vfmsub213ps_ymm_ymm_ymmm256 },
				{ "VEX_Vfmsub213pd_xmm_xmm_xmmm128", Code.VEX_Vfmsub213pd_xmm_xmm_xmmm128 },
				{ "VEX_Vfmsub213pd_ymm_ymm_ymmm256", Code.VEX_Vfmsub213pd_ymm_ymm_ymmm256 },
				{ "EVEX_Vfmsub213ps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vfmsub213ps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vfmsub213ps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vfmsub213ps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vfmsub213ps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vfmsub213ps_zmm_k1z_zmm_zmmm512b32_er },
				{ "EVEX_Vfmsub213pd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vfmsub213pd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vfmsub213pd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vfmsub213pd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vfmsub213pd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vfmsub213pd_zmm_k1z_zmm_zmmm512b64_er },
				{ "EVEX_V4fnmaddps_zmm_k1z_zmmp3_m128", Code.EVEX_V4fnmaddps_zmm_k1z_zmmp3_m128 },
				{ "VEX_Vfmsub213ss_xmm_xmm_xmmm32", Code.VEX_Vfmsub213ss_xmm_xmm_xmmm32 },
				{ "VEX_Vfmsub213sd_xmm_xmm_xmmm64", Code.VEX_Vfmsub213sd_xmm_xmm_xmmm64 },
				{ "EVEX_Vfmsub213ss_xmm_k1z_xmm_xmmm32_er", Code.EVEX_Vfmsub213ss_xmm_k1z_xmm_xmmm32_er },
				{ "EVEX_Vfmsub213sd_xmm_k1z_xmm_xmmm64_er", Code.EVEX_Vfmsub213sd_xmm_k1z_xmm_xmmm64_er },
				{ "EVEX_V4fnmaddss_xmm_k1z_xmmp3_m128", Code.EVEX_V4fnmaddss_xmm_k1z_xmmp3_m128 },
				{ "VEX_Vfnmadd213ps_xmm_xmm_xmmm128", Code.VEX_Vfnmadd213ps_xmm_xmm_xmmm128 },
				{ "VEX_Vfnmadd213ps_ymm_ymm_ymmm256", Code.VEX_Vfnmadd213ps_ymm_ymm_ymmm256 },
				{ "VEX_Vfnmadd213pd_xmm_xmm_xmmm128", Code.VEX_Vfnmadd213pd_xmm_xmm_xmmm128 },
				{ "VEX_Vfnmadd213pd_ymm_ymm_ymmm256", Code.VEX_Vfnmadd213pd_ymm_ymm_ymmm256 },
				{ "EVEX_Vfnmadd213ps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vfnmadd213ps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vfnmadd213ps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vfnmadd213ps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vfnmadd213ps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vfnmadd213ps_zmm_k1z_zmm_zmmm512b32_er },
				{ "EVEX_Vfnmadd213pd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vfnmadd213pd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vfnmadd213pd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vfnmadd213pd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vfnmadd213pd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vfnmadd213pd_zmm_k1z_zmm_zmmm512b64_er },
				{ "VEX_Vfnmadd213ss_xmm_xmm_xmmm32", Code.VEX_Vfnmadd213ss_xmm_xmm_xmmm32 },
				{ "VEX_Vfnmadd213sd_xmm_xmm_xmmm64", Code.VEX_Vfnmadd213sd_xmm_xmm_xmmm64 },
				{ "EVEX_Vfnmadd213ss_xmm_k1z_xmm_xmmm32_er", Code.EVEX_Vfnmadd213ss_xmm_k1z_xmm_xmmm32_er },
				{ "EVEX_Vfnmadd213sd_xmm_k1z_xmm_xmmm64_er", Code.EVEX_Vfnmadd213sd_xmm_k1z_xmm_xmmm64_er },
				{ "VEX_Vfnmsub213ps_xmm_xmm_xmmm128", Code.VEX_Vfnmsub213ps_xmm_xmm_xmmm128 },
				{ "VEX_Vfnmsub213ps_ymm_ymm_ymmm256", Code.VEX_Vfnmsub213ps_ymm_ymm_ymmm256 },
				{ "VEX_Vfnmsub213pd_xmm_xmm_xmmm128", Code.VEX_Vfnmsub213pd_xmm_xmm_xmmm128 },
				{ "VEX_Vfnmsub213pd_ymm_ymm_ymmm256", Code.VEX_Vfnmsub213pd_ymm_ymm_ymmm256 },
				{ "EVEX_Vfnmsub213ps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vfnmsub213ps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vfnmsub213ps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vfnmsub213ps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vfnmsub213ps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vfnmsub213ps_zmm_k1z_zmm_zmmm512b32_er },
				{ "EVEX_Vfnmsub213pd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vfnmsub213pd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vfnmsub213pd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vfnmsub213pd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vfnmsub213pd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vfnmsub213pd_zmm_k1z_zmm_zmmm512b64_er },
				{ "VEX_Vfnmsub213ss_xmm_xmm_xmmm32", Code.VEX_Vfnmsub213ss_xmm_xmm_xmmm32 },
				{ "VEX_Vfnmsub213sd_xmm_xmm_xmmm64", Code.VEX_Vfnmsub213sd_xmm_xmm_xmmm64 },
				{ "EVEX_Vfnmsub213ss_xmm_k1z_xmm_xmmm32_er", Code.EVEX_Vfnmsub213ss_xmm_k1z_xmm_xmmm32_er },
				{ "EVEX_Vfnmsub213sd_xmm_k1z_xmm_xmmm64_er", Code.EVEX_Vfnmsub213sd_xmm_k1z_xmm_xmmm64_er },
				{ "EVEX_Vpmadd52luq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpmadd52luq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpmadd52luq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpmadd52luq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpmadd52luq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpmadd52luq_zmm_k1z_zmm_zmmm512b64 },
				{ "EVEX_Vpmadd52huq_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vpmadd52huq_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vpmadd52huq_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vpmadd52huq_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vpmadd52huq_zmm_k1z_zmm_zmmm512b64", Code.EVEX_Vpmadd52huq_zmm_k1z_zmm_zmmm512b64 },
				{ "VEX_Vfmaddsub231ps_xmm_xmm_xmmm128", Code.VEX_Vfmaddsub231ps_xmm_xmm_xmmm128 },
				{ "VEX_Vfmaddsub231ps_ymm_ymm_ymmm256", Code.VEX_Vfmaddsub231ps_ymm_ymm_ymmm256 },
				{ "VEX_Vfmaddsub231pd_xmm_xmm_xmmm128", Code.VEX_Vfmaddsub231pd_xmm_xmm_xmmm128 },
				{ "VEX_Vfmaddsub231pd_ymm_ymm_ymmm256", Code.VEX_Vfmaddsub231pd_ymm_ymm_ymmm256 },
				{ "EVEX_Vfmaddsub231ps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vfmaddsub231ps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vfmaddsub231ps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vfmaddsub231ps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vfmaddsub231ps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vfmaddsub231ps_zmm_k1z_zmm_zmmm512b32_er },
				{ "EVEX_Vfmaddsub231pd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vfmaddsub231pd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vfmaddsub231pd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vfmaddsub231pd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vfmaddsub231pd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vfmaddsub231pd_zmm_k1z_zmm_zmmm512b64_er },
				{ "VEX_Vfmsubadd231ps_xmm_xmm_xmmm128", Code.VEX_Vfmsubadd231ps_xmm_xmm_xmmm128 },
				{ "VEX_Vfmsubadd231ps_ymm_ymm_ymmm256", Code.VEX_Vfmsubadd231ps_ymm_ymm_ymmm256 },
				{ "VEX_Vfmsubadd231pd_xmm_xmm_xmmm128", Code.VEX_Vfmsubadd231pd_xmm_xmm_xmmm128 },
				{ "VEX_Vfmsubadd231pd_ymm_ymm_ymmm256", Code.VEX_Vfmsubadd231pd_ymm_ymm_ymmm256 },
				{ "EVEX_Vfmsubadd231ps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vfmsubadd231ps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vfmsubadd231ps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vfmsubadd231ps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vfmsubadd231ps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vfmsubadd231ps_zmm_k1z_zmm_zmmm512b32_er },
				{ "EVEX_Vfmsubadd231pd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vfmsubadd231pd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vfmsubadd231pd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vfmsubadd231pd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vfmsubadd231pd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vfmsubadd231pd_zmm_k1z_zmm_zmmm512b64_er },
				{ "VEX_Vfmadd231ps_xmm_xmm_xmmm128", Code.VEX_Vfmadd231ps_xmm_xmm_xmmm128 },
				{ "VEX_Vfmadd231ps_ymm_ymm_ymmm256", Code.VEX_Vfmadd231ps_ymm_ymm_ymmm256 },
				{ "VEX_Vfmadd231pd_xmm_xmm_xmmm128", Code.VEX_Vfmadd231pd_xmm_xmm_xmmm128 },
				{ "VEX_Vfmadd231pd_ymm_ymm_ymmm256", Code.VEX_Vfmadd231pd_ymm_ymm_ymmm256 },
				{ "EVEX_Vfmadd231ps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vfmadd231ps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vfmadd231ps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vfmadd231ps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vfmadd231ps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vfmadd231ps_zmm_k1z_zmm_zmmm512b32_er },
				{ "EVEX_Vfmadd231pd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vfmadd231pd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vfmadd231pd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vfmadd231pd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vfmadd231pd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vfmadd231pd_zmm_k1z_zmm_zmmm512b64_er },
				{ "VEX_Vfmadd231ss_xmm_xmm_xmmm32", Code.VEX_Vfmadd231ss_xmm_xmm_xmmm32 },
				{ "VEX_Vfmadd231sd_xmm_xmm_xmmm64", Code.VEX_Vfmadd231sd_xmm_xmm_xmmm64 },
				{ "EVEX_Vfmadd231ss_xmm_k1z_xmm_xmmm32_er", Code.EVEX_Vfmadd231ss_xmm_k1z_xmm_xmmm32_er },
				{ "EVEX_Vfmadd231sd_xmm_k1z_xmm_xmmm64_er", Code.EVEX_Vfmadd231sd_xmm_k1z_xmm_xmmm64_er },
				{ "VEX_Vfmsub231ps_xmm_xmm_xmmm128", Code.VEX_Vfmsub231ps_xmm_xmm_xmmm128 },
				{ "VEX_Vfmsub231ps_ymm_ymm_ymmm256", Code.VEX_Vfmsub231ps_ymm_ymm_ymmm256 },
				{ "VEX_Vfmsub231pd_xmm_xmm_xmmm128", Code.VEX_Vfmsub231pd_xmm_xmm_xmmm128 },
				{ "VEX_Vfmsub231pd_ymm_ymm_ymmm256", Code.VEX_Vfmsub231pd_ymm_ymm_ymmm256 },
				{ "EVEX_Vfmsub231ps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vfmsub231ps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vfmsub231ps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vfmsub231ps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vfmsub231ps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vfmsub231ps_zmm_k1z_zmm_zmmm512b32_er },
				{ "EVEX_Vfmsub231pd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vfmsub231pd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vfmsub231pd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vfmsub231pd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vfmsub231pd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vfmsub231pd_zmm_k1z_zmm_zmmm512b64_er },
				{ "VEX_Vfmsub231ss_xmm_xmm_xmmm32", Code.VEX_Vfmsub231ss_xmm_xmm_xmmm32 },
				{ "VEX_Vfmsub231sd_xmm_xmm_xmmm64", Code.VEX_Vfmsub231sd_xmm_xmm_xmmm64 },
				{ "EVEX_Vfmsub231ss_xmm_k1z_xmm_xmmm32_er", Code.EVEX_Vfmsub231ss_xmm_k1z_xmm_xmmm32_er },
				{ "EVEX_Vfmsub231sd_xmm_k1z_xmm_xmmm64_er", Code.EVEX_Vfmsub231sd_xmm_k1z_xmm_xmmm64_er },
				{ "VEX_Vfnmadd231ps_xmm_xmm_xmmm128", Code.VEX_Vfnmadd231ps_xmm_xmm_xmmm128 },
				{ "VEX_Vfnmadd231ps_ymm_ymm_ymmm256", Code.VEX_Vfnmadd231ps_ymm_ymm_ymmm256 },
				{ "VEX_Vfnmadd231pd_xmm_xmm_xmmm128", Code.VEX_Vfnmadd231pd_xmm_xmm_xmmm128 },
				{ "VEX_Vfnmadd231pd_ymm_ymm_ymmm256", Code.VEX_Vfnmadd231pd_ymm_ymm_ymmm256 },
				{ "EVEX_Vfnmadd231ps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vfnmadd231ps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vfnmadd231ps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vfnmadd231ps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vfnmadd231ps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vfnmadd231ps_zmm_k1z_zmm_zmmm512b32_er },
				{ "EVEX_Vfnmadd231pd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vfnmadd231pd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vfnmadd231pd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vfnmadd231pd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vfnmadd231pd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vfnmadd231pd_zmm_k1z_zmm_zmmm512b64_er },
				{ "VEX_Vfnmadd231ss_xmm_xmm_xmmm32", Code.VEX_Vfnmadd231ss_xmm_xmm_xmmm32 },
				{ "VEX_Vfnmadd231sd_xmm_xmm_xmmm64", Code.VEX_Vfnmadd231sd_xmm_xmm_xmmm64 },
				{ "EVEX_Vfnmadd231ss_xmm_k1z_xmm_xmmm32_er", Code.EVEX_Vfnmadd231ss_xmm_k1z_xmm_xmmm32_er },
				{ "EVEX_Vfnmadd231sd_xmm_k1z_xmm_xmmm64_er", Code.EVEX_Vfnmadd231sd_xmm_k1z_xmm_xmmm64_er },
				{ "VEX_Vfnmsub231ps_xmm_xmm_xmmm128", Code.VEX_Vfnmsub231ps_xmm_xmm_xmmm128 },
				{ "VEX_Vfnmsub231ps_ymm_ymm_ymmm256", Code.VEX_Vfnmsub231ps_ymm_ymm_ymmm256 },
				{ "VEX_Vfnmsub231pd_xmm_xmm_xmmm128", Code.VEX_Vfnmsub231pd_xmm_xmm_xmmm128 },
				{ "VEX_Vfnmsub231pd_ymm_ymm_ymmm256", Code.VEX_Vfnmsub231pd_ymm_ymm_ymmm256 },
				{ "EVEX_Vfnmsub231ps_xmm_k1z_xmm_xmmm128b32", Code.EVEX_Vfnmsub231ps_xmm_k1z_xmm_xmmm128b32 },
				{ "EVEX_Vfnmsub231ps_ymm_k1z_ymm_ymmm256b32", Code.EVEX_Vfnmsub231ps_ymm_k1z_ymm_ymmm256b32 },
				{ "EVEX_Vfnmsub231ps_zmm_k1z_zmm_zmmm512b32_er", Code.EVEX_Vfnmsub231ps_zmm_k1z_zmm_zmmm512b32_er },
				{ "EVEX_Vfnmsub231pd_xmm_k1z_xmm_xmmm128b64", Code.EVEX_Vfnmsub231pd_xmm_k1z_xmm_xmmm128b64 },
				{ "EVEX_Vfnmsub231pd_ymm_k1z_ymm_ymmm256b64", Code.EVEX_Vfnmsub231pd_ymm_k1z_ymm_ymmm256b64 },
				{ "EVEX_Vfnmsub231pd_zmm_k1z_zmm_zmmm512b64_er", Code.EVEX_Vfnmsub231pd_zmm_k1z_zmm_zmmm512b64_er },
				{ "VEX_Vfnmsub231ss_xmm_xmm_xmmm32", Code.VEX_Vfnmsub231ss_xmm_xmm_xmmm32 },
				{ "VEX_Vfnmsub231sd_xmm_xmm_xmmm64", Code.VEX_Vfnmsub231sd_xmm_xmm_xmmm64 },
				{ "EVEX_Vfnmsub231ss_xmm_k1z_xmm_xmmm32_er", Code.EVEX_Vfnmsub231ss_xmm_k1z_xmm_xmmm32_er },
				{ "EVEX_Vfnmsub231sd_xmm_k1z_xmm_xmmm64_er", Code.EVEX_Vfnmsub231sd_xmm_k1z_xmm_xmmm64_er },
				{ "EVEX_Vpconflictd_xmm_k1z_xmmm128b32", Code.EVEX_Vpconflictd_xmm_k1z_xmmm128b32 },
				{ "EVEX_Vpconflictd_ymm_k1z_ymmm256b32", Code.EVEX_Vpconflictd_ymm_k1z_ymmm256b32 },
				{ "EVEX_Vpconflictd_zmm_k1z_zmmm512b32", Code.EVEX_Vpconflictd_zmm_k1z_zmmm512b32 },
				{ "EVEX_Vpconflictq_xmm_k1z_xmmm128b64", Code.EVEX_Vpconflictq_xmm_k1z_xmmm128b64 },
				{ "EVEX_Vpconflictq_ymm_k1z_ymmm256b64", Code.EVEX_Vpconflictq_ymm_k1z_ymmm256b64 },
				{ "EVEX_Vpconflictq_zmm_k1z_zmmm512b64", Code.EVEX_Vpconflictq_zmm_k1z_zmmm512b64 },
				{ "EVEX_Vgatherpf0dps_vm32z_k1", Code.EVEX_Vgatherpf0dps_vm32z_k1 },
				{ "EVEX_Vgatherpf0dpd_vm32y_k1", Code.EVEX_Vgatherpf0dpd_vm32y_k1 },
				{ "EVEX_Vgatherpf1dps_vm32z_k1", Code.EVEX_Vgatherpf1dps_vm32z_k1 },
				{ "EVEX_Vgatherpf1dpd_vm32y_k1", Code.EVEX_Vgatherpf1dpd_vm32y_k1 },
				{ "EVEX_Vscatterpf0dps_vm32z_k1", Code.EVEX_Vscatterpf0dps_vm32z_k1 },
				{ "EVEX_Vscatterpf0dpd_vm32y_k1", Code.EVEX_Vscatterpf0dpd_vm32y_k1 },
				{ "EVEX_Vscatterpf1dps_vm32z_k1", Code.EVEX_Vscatterpf1dps_vm32z_k1 },
				{ "EVEX_Vscatterpf1dpd_vm32y_k1", Code.EVEX_Vscatterpf1dpd_vm32y_k1 },
				{ "EVEX_Vgatherpf0qps_vm64z_k1", Code.EVEX_Vgatherpf0qps_vm64z_k1 },
				{ "EVEX_Vgatherpf0qpd_vm64z_k1", Code.EVEX_Vgatherpf0qpd_vm64z_k1 },
				{ "EVEX_Vgatherpf1qps_vm64z_k1", Code.EVEX_Vgatherpf1qps_vm64z_k1 },
				{ "EVEX_Vgatherpf1qpd_vm64z_k1", Code.EVEX_Vgatherpf1qpd_vm64z_k1 },
				{ "EVEX_Vscatterpf0qps_vm64z_k1", Code.EVEX_Vscatterpf0qps_vm64z_k1 },
				{ "EVEX_Vscatterpf0qpd_vm64z_k1", Code.EVEX_Vscatterpf0qpd_vm64z_k1 },
				{ "EVEX_Vscatterpf1qps_vm64z_k1", Code.EVEX_Vscatterpf1qps_vm64z_k1 },
				{ "EVEX_Vscatterpf1qpd_vm64z_k1", Code.EVEX_Vscatterpf1qpd_vm64z_k1 },
				{ "Sha1nexte_xmm_xmmm128", Code.Sha1nexte_xmm_xmmm128 },
				{ "EVEX_Vexp2ps_zmm_k1z_zmmm512b32_sae", Code.EVEX_Vexp2ps_zmm_k1z_zmmm512b32_sae },
				{ "EVEX_Vexp2pd_zmm_k1z_zmmm512b64_sae", Code.EVEX_Vexp2pd_zmm_k1z_zmmm512b64_sae },
				{ "Sha1msg1_xmm_xmmm128", Code.Sha1msg1_xmm_xmmm128 },
				{ "Sha1msg2_xmm_xmmm128", Code.Sha1msg2_xmm_xmmm128 },
				{ "EVEX_Vrcp28ps_zmm_k1z_zmmm512b32_sae", Code.EVEX_Vrcp28ps_zmm_k1z_zmmm512b32_sae },
				{ "EVEX_Vrcp28pd_zmm_k1z_zmmm512b64_sae", Code.EVEX_Vrcp28pd_zmm_k1z_zmmm512b64_sae },
				{ "Sha256rnds2_xmm_xmmm128", Code.Sha256rnds2_xmm_xmmm128 },
				{ "EVEX_Vrcp28ss_xmm_k1z_xmm_xmmm32_sae", Code.EVEX_Vrcp28ss_xmm_k1z_xmm_xmmm32_sae },
				{ "EVEX_Vrcp28sd_xmm_k1z_xmm_xmmm64_sae", Code.EVEX_Vrcp28sd_xmm_k1z_xmm_xmmm64_sae },
				{ "Sha256msg1_xmm_xmmm128", Code.Sha256msg1_xmm_xmmm128 },
				{ "EVEX_Vrsqrt28ps_zmm_k1z_zmmm512b32_sae", Code.EVEX_Vrsqrt28ps_zmm_k1z_zmmm512b32_sae },
				{ "EVEX_Vrsqrt28pd_zmm_k1z_zmmm512b64_sae", Code.EVEX_Vrsqrt28pd_zmm_k1z_zmmm512b64_sae },
				{ "Sha256msg2_xmm_xmmm128", Code.Sha256msg2_xmm_xmmm128 },
				{ "EVEX_Vrsqrt28ss_xmm_k1z_xmm_xmmm32_sae", Code.EVEX_Vrsqrt28ss_xmm_k1z_xmm_xmmm32_sae },
				{ "EVEX_Vrsqrt28sd_xmm_k1z_xmm_xmmm64_sae", Code.EVEX_Vrsqrt28sd_xmm_k1z_xmm_xmmm64_sae },
				{ "Gf2p8mulb_xmm_xmmm128", Code.Gf2p8mulb_xmm_xmmm128 },
				{ "VEX_Vgf2p8mulb_xmm_xmm_xmmm128", Code.VEX_Vgf2p8mulb_xmm_xmm_xmmm128 },
				{ "VEX_Vgf2p8mulb_ymm_ymm_ymmm256", Code.VEX_Vgf2p8mulb_ymm_ymm_ymmm256 },
				{ "EVEX_Vgf2p8mulb_xmm_k1z_xmm_xmmm128", Code.EVEX_Vgf2p8mulb_xmm_k1z_xmm_xmmm128 },
				{ "EVEX_Vgf2p8mulb_ymm_k1z_ymm_ymmm256", Code.EVEX_Vgf2p8mulb_ymm_k1z_ymm_ymmm256 },
				{ "EVEX_Vgf2p8mulb_zmm_k1z_zmm_zmmm512", Code.EVEX_Vgf2p8mulb_zmm_k1z_zmm_zmmm512 },
				{ "Aesimc_xmm_xmmm128", Code.Aesimc_xmm_xmmm128 },
				{ "VEX_Vaesimc_xmm_xmmm128", Code.VEX_Vaesimc_xmm_xmmm128 },
				{ "Aesenc_xmm_xmmm128", Code.Aesenc_xmm_xmmm128 },
				{ "VEX_Vaesenc_xmm_xmm_xmmm128", Code.VEX_Vaesenc_xmm_xmm_xmmm128 },
				{ "VEX_Vaesenc_ymm_ymm_ymmm256", Code.VEX_Vaesenc_ymm_ymm_ymmm256 },
				{ "EVEX_Vaesenc_xmm_xmm_xmmm128", Code.EVEX_Vaesenc_xmm_xmm_xmmm128 },
				{ "EVEX_Vaesenc_ymm_ymm_ymmm256", Code.EVEX_Vaesenc_ymm_ymm_ymmm256 },
				{ "EVEX_Vaesenc_zmm_zmm_zmmm512", Code.EVEX_Vaesenc_zmm_zmm_zmmm512 },
				{ "Aesenclast_xmm_xmmm128", Code.Aesenclast_xmm_xmmm128 },
				{ "VEX_Vaesenclast_xmm_xmm_xmmm128", Code.VEX_Vaesenclast_xmm_xmm_xmmm128 },
				{ "VEX_Vaesenclast_ymm_ymm_ymmm256", Code.VEX_Vaesenclast_ymm_ymm_ymmm256 },
				{ "EVEX_Vaesenclast_xmm_xmm_xmmm128", Code.EVEX_Vaesenclast_xmm_xmm_xmmm128 },
				{ "EVEX_Vaesenclast_ymm_ymm_ymmm256", Code.EVEX_Vaesenclast_ymm_ymm_ymmm256 },
				{ "EVEX_Vaesenclast_zmm_zmm_zmmm512", Code.EVEX_Vaesenclast_zmm_zmm_zmmm512 },
				{ "Aesdec_xmm_xmmm128", Code.Aesdec_xmm_xmmm128 },
				{ "VEX_Vaesdec_xmm_xmm_xmmm128", Code.VEX_Vaesdec_xmm_xmm_xmmm128 },
				{ "VEX_Vaesdec_ymm_ymm_ymmm256", Code.VEX_Vaesdec_ymm_ymm_ymmm256 },
				{ "EVEX_Vaesdec_xmm_xmm_xmmm128", Code.EVEX_Vaesdec_xmm_xmm_xmmm128 },
				{ "EVEX_Vaesdec_ymm_ymm_ymmm256", Code.EVEX_Vaesdec_ymm_ymm_ymmm256 },
				{ "EVEX_Vaesdec_zmm_zmm_zmmm512", Code.EVEX_Vaesdec_zmm_zmm_zmmm512 },
				{ "Aesdeclast_xmm_xmmm128", Code.Aesdeclast_xmm_xmmm128 },
				{ "VEX_Vaesdeclast_xmm_xmm_xmmm128", Code.VEX_Vaesdeclast_xmm_xmm_xmmm128 },
				{ "VEX_Vaesdeclast_ymm_ymm_ymmm256", Code.VEX_Vaesdeclast_ymm_ymm_ymmm256 },
				{ "EVEX_Vaesdeclast_xmm_xmm_xmmm128", Code.EVEX_Vaesdeclast_xmm_xmm_xmmm128 },
				{ "EVEX_Vaesdeclast_ymm_ymm_ymmm256", Code.EVEX_Vaesdeclast_ymm_ymm_ymmm256 },
				{ "EVEX_Vaesdeclast_zmm_zmm_zmmm512", Code.EVEX_Vaesdeclast_zmm_zmm_zmmm512 },
				{ "Movbe_r16_m16", Code.Movbe_r16_m16 },
				{ "Movbe_r32_m32", Code.Movbe_r32_m32 },
				{ "Movbe_r64_m64", Code.Movbe_r64_m64 },
				{ "Crc32_r32_rm8", Code.Crc32_r32_rm8 },
				{ "Crc32_r64_rm8", Code.Crc32_r64_rm8 },
				{ "Movbe_m16_r16", Code.Movbe_m16_r16 },
				{ "Movbe_m32_r32", Code.Movbe_m32_r32 },
				{ "Movbe_m64_r64", Code.Movbe_m64_r64 },
				{ "Crc32_r32_rm16", Code.Crc32_r32_rm16 },
				{ "Crc32_r32_rm32", Code.Crc32_r32_rm32 },
				{ "Crc32_r64_rm64", Code.Crc32_r64_rm64 },
				{ "VEX_Andn_r32_r32_rm32", Code.VEX_Andn_r32_r32_rm32 },
				{ "VEX_Andn_r64_r64_rm64", Code.VEX_Andn_r64_r64_rm64 },
				{ "VEX_Blsr_r32_rm32", Code.VEX_Blsr_r32_rm32 },
				{ "VEX_Blsr_r64_rm64", Code.VEX_Blsr_r64_rm64 },
				{ "VEX_Blsmsk_r32_rm32", Code.VEX_Blsmsk_r32_rm32 },
				{ "VEX_Blsmsk_r64_rm64", Code.VEX_Blsmsk_r64_rm64 },
				{ "VEX_Blsi_r32_rm32", Code.VEX_Blsi_r32_rm32 },
				{ "VEX_Blsi_r64_rm64", Code.VEX_Blsi_r64_rm64 },
				{ "VEX_Bzhi_r32_rm32_r32", Code.VEX_Bzhi_r32_rm32_r32 },
				{ "VEX_Bzhi_r64_rm64_r64", Code.VEX_Bzhi_r64_rm64_r64 },
				{ "Wrussd_m32_r32", Code.Wrussd_m32_r32 },
				{ "Wrussq_m64_r64", Code.Wrussq_m64_r64 },
				{ "VEX_Pext_r32_r32_rm32", Code.VEX_Pext_r32_r32_rm32 },
				{ "VEX_Pext_r64_r64_rm64", Code.VEX_Pext_r64_r64_rm64 },
				{ "VEX_Pdep_r32_r32_rm32", Code.VEX_Pdep_r32_r32_rm32 },
				{ "VEX_Pdep_r64_r64_rm64", Code.VEX_Pdep_r64_r64_rm64 },
				{ "Wrssd_m32_r32", Code.Wrssd_m32_r32 },
				{ "Wrssq_m64_r64", Code.Wrssq_m64_r64 },
				{ "Adcx_r32_rm32", Code.Adcx_r32_rm32 },
				{ "Adcx_r64_rm64", Code.Adcx_r64_rm64 },
				{ "Adox_r32_rm32", Code.Adox_r32_rm32 },
				{ "Adox_r64_rm64", Code.Adox_r64_rm64 },
				{ "VEX_Mulx_r32_r32_rm32", Code.VEX_Mulx_r32_r32_rm32 },
				{ "VEX_Mulx_r64_r64_rm64", Code.VEX_Mulx_r64_r64_rm64 },
				{ "VEX_Bextr_r32_rm32_r32", Code.VEX_Bextr_r32_rm32_r32 },
				{ "VEX_Bextr_r64_rm64_r64", Code.VEX_Bextr_r64_rm64_r64 },
				{ "VEX_Shlx_r32_rm32_r32", Code.VEX_Shlx_r32_rm32_r32 },
				{ "VEX_Shlx_r64_rm64_r64", Code.VEX_Shlx_r64_rm64_r64 },
				{ "VEX_Sarx_r32_rm32_r32", Code.VEX_Sarx_r32_rm32_r32 },
				{ "VEX_Sarx_r64_rm64_r64", Code.VEX_Sarx_r64_rm64_r64 },
				{ "VEX_Shrx_r32_rm32_r32", Code.VEX_Shrx_r32_rm32_r32 },
				{ "VEX_Shrx_r64_rm64_r64", Code.VEX_Shrx_r64_rm64_r64 },
				{ "Movdir64b_r16_m512", Code.Movdir64b_r16_m512 },
				{ "Movdir64b_r32_m512", Code.Movdir64b_r32_m512 },
				{ "Movdir64b_r64_m512", Code.Movdir64b_r64_m512 },
				{ "Enqcmds_r16_m512", Code.Enqcmds_r16_m512 },
				{ "Enqcmds_r32_m512", Code.Enqcmds_r32_m512 },
				{ "Enqcmds_r64_m512", Code.Enqcmds_r64_m512 },
				{ "Enqcmd_r16_m512", Code.Enqcmd_r16_m512 },
				{ "Enqcmd_r32_m512", Code.Enqcmd_r32_m512 },
				{ "Enqcmd_r64_m512", Code.Enqcmd_r64_m512 },
				{ "Movdiri_m32_r32", Code.Movdiri_m32_r32 },
				{ "Movdiri_m64_r64", Code.Movdiri_m64_r64 },
				{ "VEX_Vpermq_ymm_ymmm256_imm8", Code.VEX_Vpermq_ymm_ymmm256_imm8 },
				{ "EVEX_Vpermq_ymm_k1z_ymmm256b64_imm8", Code.EVEX_Vpermq_ymm_k1z_ymmm256b64_imm8 },
				{ "EVEX_Vpermq_zmm_k1z_zmmm512b64_imm8", Code.EVEX_Vpermq_zmm_k1z_zmmm512b64_imm8 },
				{ "VEX_Vpermpd_ymm_ymmm256_imm8", Code.VEX_Vpermpd_ymm_ymmm256_imm8 },
				{ "EVEX_Vpermpd_ymm_k1z_ymmm256b64_imm8", Code.EVEX_Vpermpd_ymm_k1z_ymmm256b64_imm8 },
				{ "EVEX_Vpermpd_zmm_k1z_zmmm512b64_imm8", Code.EVEX_Vpermpd_zmm_k1z_zmmm512b64_imm8 },
				{ "VEX_Vpblendd_xmm_xmm_xmmm128_imm8", Code.VEX_Vpblendd_xmm_xmm_xmmm128_imm8 },
				{ "VEX_Vpblendd_ymm_ymm_ymmm256_imm8", Code.VEX_Vpblendd_ymm_ymm_ymmm256_imm8 },
				{ "EVEX_Valignd_xmm_k1z_xmm_xmmm128b32_imm8", Code.EVEX_Valignd_xmm_k1z_xmm_xmmm128b32_imm8 },
				{ "EVEX_Valignd_ymm_k1z_ymm_ymmm256b32_imm8", Code.EVEX_Valignd_ymm_k1z_ymm_ymmm256b32_imm8 },
				{ "EVEX_Valignd_zmm_k1z_zmm_zmmm512b32_imm8", Code.EVEX_Valignd_zmm_k1z_zmm_zmmm512b32_imm8 },
				{ "EVEX_Valignq_xmm_k1z_xmm_xmmm128b64_imm8", Code.EVEX_Valignq_xmm_k1z_xmm_xmmm128b64_imm8 },
				{ "EVEX_Valignq_ymm_k1z_ymm_ymmm256b64_imm8", Code.EVEX_Valignq_ymm_k1z_ymm_ymmm256b64_imm8 },
				{ "EVEX_Valignq_zmm_k1z_zmm_zmmm512b64_imm8", Code.EVEX_Valignq_zmm_k1z_zmm_zmmm512b64_imm8 },
				{ "VEX_Vpermilps_xmm_xmmm128_imm8", Code.VEX_Vpermilps_xmm_xmmm128_imm8 },
				{ "VEX_Vpermilps_ymm_ymmm256_imm8", Code.VEX_Vpermilps_ymm_ymmm256_imm8 },
				{ "EVEX_Vpermilps_xmm_k1z_xmmm128b32_imm8", Code.EVEX_Vpermilps_xmm_k1z_xmmm128b32_imm8 },
				{ "EVEX_Vpermilps_ymm_k1z_ymmm256b32_imm8", Code.EVEX_Vpermilps_ymm_k1z_ymmm256b32_imm8 },
				{ "EVEX_Vpermilps_zmm_k1z_zmmm512b32_imm8", Code.EVEX_Vpermilps_zmm_k1z_zmmm512b32_imm8 },
				{ "VEX_Vpermilpd_xmm_xmmm128_imm8", Code.VEX_Vpermilpd_xmm_xmmm128_imm8 },
				{ "VEX_Vpermilpd_ymm_ymmm256_imm8", Code.VEX_Vpermilpd_ymm_ymmm256_imm8 },
				{ "EVEX_Vpermilpd_xmm_k1z_xmmm128b64_imm8", Code.EVEX_Vpermilpd_xmm_k1z_xmmm128b64_imm8 },
				{ "EVEX_Vpermilpd_ymm_k1z_ymmm256b64_imm8", Code.EVEX_Vpermilpd_ymm_k1z_ymmm256b64_imm8 },
				{ "EVEX_Vpermilpd_zmm_k1z_zmmm512b64_imm8", Code.EVEX_Vpermilpd_zmm_k1z_zmmm512b64_imm8 },
				{ "VEX_Vperm2f128_ymm_ymm_ymmm256_imm8", Code.VEX_Vperm2f128_ymm_ymm_ymmm256_imm8 },
				{ "Roundps_xmm_xmmm128_imm8", Code.Roundps_xmm_xmmm128_imm8 },
				{ "VEX_Vroundps_xmm_xmmm128_imm8", Code.VEX_Vroundps_xmm_xmmm128_imm8 },
				{ "VEX_Vroundps_ymm_ymmm256_imm8", Code.VEX_Vroundps_ymm_ymmm256_imm8 },
				{ "EVEX_Vrndscaleps_xmm_k1z_xmmm128b32_imm8", Code.EVEX_Vrndscaleps_xmm_k1z_xmmm128b32_imm8 },
				{ "EVEX_Vrndscaleps_ymm_k1z_ymmm256b32_imm8", Code.EVEX_Vrndscaleps_ymm_k1z_ymmm256b32_imm8 },
				{ "EVEX_Vrndscaleps_zmm_k1z_zmmm512b32_imm8_sae", Code.EVEX_Vrndscaleps_zmm_k1z_zmmm512b32_imm8_sae },
				{ "Roundpd_xmm_xmmm128_imm8", Code.Roundpd_xmm_xmmm128_imm8 },
				{ "VEX_Vroundpd_xmm_xmmm128_imm8", Code.VEX_Vroundpd_xmm_xmmm128_imm8 },
				{ "VEX_Vroundpd_ymm_ymmm256_imm8", Code.VEX_Vroundpd_ymm_ymmm256_imm8 },
				{ "EVEX_Vrndscalepd_xmm_k1z_xmmm128b64_imm8", Code.EVEX_Vrndscalepd_xmm_k1z_xmmm128b64_imm8 },
				{ "EVEX_Vrndscalepd_ymm_k1z_ymmm256b64_imm8", Code.EVEX_Vrndscalepd_ymm_k1z_ymmm256b64_imm8 },
				{ "EVEX_Vrndscalepd_zmm_k1z_zmmm512b64_imm8_sae", Code.EVEX_Vrndscalepd_zmm_k1z_zmmm512b64_imm8_sae },
				{ "Roundss_xmm_xmmm32_imm8", Code.Roundss_xmm_xmmm32_imm8 },
				{ "VEX_Vroundss_xmm_xmm_xmmm32_imm8", Code.VEX_Vroundss_xmm_xmm_xmmm32_imm8 },
				{ "EVEX_Vrndscaless_xmm_k1z_xmm_xmmm32_imm8_sae", Code.EVEX_Vrndscaless_xmm_k1z_xmm_xmmm32_imm8_sae },
				{ "Roundsd_xmm_xmmm64_imm8", Code.Roundsd_xmm_xmmm64_imm8 },
				{ "VEX_Vroundsd_xmm_xmm_xmmm64_imm8", Code.VEX_Vroundsd_xmm_xmm_xmmm64_imm8 },
				{ "EVEX_Vrndscalesd_xmm_k1z_xmm_xmmm64_imm8_sae", Code.EVEX_Vrndscalesd_xmm_k1z_xmm_xmmm64_imm8_sae },
				{ "Blendps_xmm_xmmm128_imm8", Code.Blendps_xmm_xmmm128_imm8 },
				{ "VEX_Vblendps_xmm_xmm_xmmm128_imm8", Code.VEX_Vblendps_xmm_xmm_xmmm128_imm8 },
				{ "VEX_Vblendps_ymm_ymm_ymmm256_imm8", Code.VEX_Vblendps_ymm_ymm_ymmm256_imm8 },
				{ "Blendpd_xmm_xmmm128_imm8", Code.Blendpd_xmm_xmmm128_imm8 },
				{ "VEX_Vblendpd_xmm_xmm_xmmm128_imm8", Code.VEX_Vblendpd_xmm_xmm_xmmm128_imm8 },
				{ "VEX_Vblendpd_ymm_ymm_ymmm256_imm8", Code.VEX_Vblendpd_ymm_ymm_ymmm256_imm8 },
				{ "Pblendw_xmm_xmmm128_imm8", Code.Pblendw_xmm_xmmm128_imm8 },
				{ "VEX_Vpblendw_xmm_xmm_xmmm128_imm8", Code.VEX_Vpblendw_xmm_xmm_xmmm128_imm8 },
				{ "VEX_Vpblendw_ymm_ymm_ymmm256_imm8", Code.VEX_Vpblendw_ymm_ymm_ymmm256_imm8 },
				{ "Palignr_mm_mmm64_imm8", Code.Palignr_mm_mmm64_imm8 },
				{ "Palignr_xmm_xmmm128_imm8", Code.Palignr_xmm_xmmm128_imm8 },
				{ "VEX_Vpalignr_xmm_xmm_xmmm128_imm8", Code.VEX_Vpalignr_xmm_xmm_xmmm128_imm8 },
				{ "VEX_Vpalignr_ymm_ymm_ymmm256_imm8", Code.VEX_Vpalignr_ymm_ymm_ymmm256_imm8 },
				{ "EVEX_Vpalignr_xmm_k1z_xmm_xmmm128_imm8", Code.EVEX_Vpalignr_xmm_k1z_xmm_xmmm128_imm8 },
				{ "EVEX_Vpalignr_ymm_k1z_ymm_ymmm256_imm8", Code.EVEX_Vpalignr_ymm_k1z_ymm_ymmm256_imm8 },
				{ "EVEX_Vpalignr_zmm_k1z_zmm_zmmm512_imm8", Code.EVEX_Vpalignr_zmm_k1z_zmm_zmmm512_imm8 },
				{ "Pextrb_r32m8_xmm_imm8", Code.Pextrb_r32m8_xmm_imm8 },
				{ "Pextrb_r64m8_xmm_imm8", Code.Pextrb_r64m8_xmm_imm8 },
				{ "VEX_Vpextrb_r32m8_xmm_imm8", Code.VEX_Vpextrb_r32m8_xmm_imm8 },
				{ "VEX_Vpextrb_r64m8_xmm_imm8", Code.VEX_Vpextrb_r64m8_xmm_imm8 },
				{ "EVEX_Vpextrb_r32m8_xmm_imm8", Code.EVEX_Vpextrb_r32m8_xmm_imm8 },
				{ "EVEX_Vpextrb_r64m8_xmm_imm8", Code.EVEX_Vpextrb_r64m8_xmm_imm8 },
				{ "Pextrw_r32m16_xmm_imm8", Code.Pextrw_r32m16_xmm_imm8 },
				{ "Pextrw_r64m16_xmm_imm8", Code.Pextrw_r64m16_xmm_imm8 },
				{ "VEX_Vpextrw_r32m16_xmm_imm8", Code.VEX_Vpextrw_r32m16_xmm_imm8 },
				{ "VEX_Vpextrw_r64m16_xmm_imm8", Code.VEX_Vpextrw_r64m16_xmm_imm8 },
				{ "EVEX_Vpextrw_r32m16_xmm_imm8", Code.EVEX_Vpextrw_r32m16_xmm_imm8 },
				{ "EVEX_Vpextrw_r64m16_xmm_imm8", Code.EVEX_Vpextrw_r64m16_xmm_imm8 },
				{ "Pextrd_rm32_xmm_imm8", Code.Pextrd_rm32_xmm_imm8 },
				{ "Pextrq_rm64_xmm_imm8", Code.Pextrq_rm64_xmm_imm8 },
				{ "VEX_Vpextrd_rm32_xmm_imm8", Code.VEX_Vpextrd_rm32_xmm_imm8 },
				{ "VEX_Vpextrq_rm64_xmm_imm8", Code.VEX_Vpextrq_rm64_xmm_imm8 },
				{ "EVEX_Vpextrd_rm32_xmm_imm8", Code.EVEX_Vpextrd_rm32_xmm_imm8 },
				{ "EVEX_Vpextrq_rm64_xmm_imm8", Code.EVEX_Vpextrq_rm64_xmm_imm8 },
				{ "Extractps_rm32_xmm_imm8", Code.Extractps_rm32_xmm_imm8 },
				{ "Extractps_r64m32_xmm_imm8", Code.Extractps_r64m32_xmm_imm8 },
				{ "VEX_Vextractps_rm32_xmm_imm8", Code.VEX_Vextractps_rm32_xmm_imm8 },
				{ "VEX_Vextractps_r64m32_xmm_imm8", Code.VEX_Vextractps_r64m32_xmm_imm8 },
				{ "EVEX_Vextractps_rm32_xmm_imm8", Code.EVEX_Vextractps_rm32_xmm_imm8 },
				{ "EVEX_Vextractps_r64m32_xmm_imm8", Code.EVEX_Vextractps_r64m32_xmm_imm8 },
				{ "VEX_Vinsertf128_ymm_ymm_xmmm128_imm8", Code.VEX_Vinsertf128_ymm_ymm_xmmm128_imm8 },
				{ "EVEX_Vinsertf32x4_ymm_k1z_ymm_xmmm128_imm8", Code.EVEX_Vinsertf32x4_ymm_k1z_ymm_xmmm128_imm8 },
				{ "EVEX_Vinsertf32x4_zmm_k1z_zmm_xmmm128_imm8", Code.EVEX_Vinsertf32x4_zmm_k1z_zmm_xmmm128_imm8 },
				{ "EVEX_Vinsertf64x2_ymm_k1z_ymm_xmmm128_imm8", Code.EVEX_Vinsertf64x2_ymm_k1z_ymm_xmmm128_imm8 },
				{ "EVEX_Vinsertf64x2_zmm_k1z_zmm_xmmm128_imm8", Code.EVEX_Vinsertf64x2_zmm_k1z_zmm_xmmm128_imm8 },
				{ "VEX_Vextractf128_xmmm128_ymm_imm8", Code.VEX_Vextractf128_xmmm128_ymm_imm8 },
				{ "EVEX_Vextractf32x4_xmmm128_k1z_ymm_imm8", Code.EVEX_Vextractf32x4_xmmm128_k1z_ymm_imm8 },
				{ "EVEX_Vextractf32x4_xmmm128_k1z_zmm_imm8", Code.EVEX_Vextractf32x4_xmmm128_k1z_zmm_imm8 },
				{ "EVEX_Vextractf64x2_xmmm128_k1z_ymm_imm8", Code.EVEX_Vextractf64x2_xmmm128_k1z_ymm_imm8 },
				{ "EVEX_Vextractf64x2_xmmm128_k1z_zmm_imm8", Code.EVEX_Vextractf64x2_xmmm128_k1z_zmm_imm8 },
				{ "EVEX_Vinsertf32x8_zmm_k1z_zmm_ymmm256_imm8", Code.EVEX_Vinsertf32x8_zmm_k1z_zmm_ymmm256_imm8 },
				{ "EVEX_Vinsertf64x4_zmm_k1z_zmm_ymmm256_imm8", Code.EVEX_Vinsertf64x4_zmm_k1z_zmm_ymmm256_imm8 },
				{ "EVEX_Vextractf32x8_ymmm256_k1z_zmm_imm8", Code.EVEX_Vextractf32x8_ymmm256_k1z_zmm_imm8 },
				{ "EVEX_Vextractf64x4_ymmm256_k1z_zmm_imm8", Code.EVEX_Vextractf64x4_ymmm256_k1z_zmm_imm8 },
				{ "VEX_Vcvtps2ph_xmmm64_xmm_imm8", Code.VEX_Vcvtps2ph_xmmm64_xmm_imm8 },
				{ "VEX_Vcvtps2ph_xmmm128_ymm_imm8", Code.VEX_Vcvtps2ph_xmmm128_ymm_imm8 },
				{ "EVEX_Vcvtps2ph_xmmm64_k1z_xmm_imm8", Code.EVEX_Vcvtps2ph_xmmm64_k1z_xmm_imm8 },
				{ "EVEX_Vcvtps2ph_xmmm128_k1z_ymm_imm8", Code.EVEX_Vcvtps2ph_xmmm128_k1z_ymm_imm8 },
				{ "EVEX_Vcvtps2ph_ymmm256_k1z_zmm_imm8_sae", Code.EVEX_Vcvtps2ph_ymmm256_k1z_zmm_imm8_sae },
				{ "EVEX_Vpcmpud_kr_k1_xmm_xmmm128b32_imm8", Code.EVEX_Vpcmpud_kr_k1_xmm_xmmm128b32_imm8 },
				{ "EVEX_Vpcmpud_kr_k1_ymm_ymmm256b32_imm8", Code.EVEX_Vpcmpud_kr_k1_ymm_ymmm256b32_imm8 },
				{ "EVEX_Vpcmpud_kr_k1_zmm_zmmm512b32_imm8", Code.EVEX_Vpcmpud_kr_k1_zmm_zmmm512b32_imm8 },
				{ "EVEX_Vpcmpuq_kr_k1_xmm_xmmm128b64_imm8", Code.EVEX_Vpcmpuq_kr_k1_xmm_xmmm128b64_imm8 },
				{ "EVEX_Vpcmpuq_kr_k1_ymm_ymmm256b64_imm8", Code.EVEX_Vpcmpuq_kr_k1_ymm_ymmm256b64_imm8 },
				{ "EVEX_Vpcmpuq_kr_k1_zmm_zmmm512b64_imm8", Code.EVEX_Vpcmpuq_kr_k1_zmm_zmmm512b64_imm8 },
				{ "EVEX_Vpcmpd_kr_k1_xmm_xmmm128b32_imm8", Code.EVEX_Vpcmpd_kr_k1_xmm_xmmm128b32_imm8 },
				{ "EVEX_Vpcmpd_kr_k1_ymm_ymmm256b32_imm8", Code.EVEX_Vpcmpd_kr_k1_ymm_ymmm256b32_imm8 },
				{ "EVEX_Vpcmpd_kr_k1_zmm_zmmm512b32_imm8", Code.EVEX_Vpcmpd_kr_k1_zmm_zmmm512b32_imm8 },
				{ "EVEX_Vpcmpq_kr_k1_xmm_xmmm128b64_imm8", Code.EVEX_Vpcmpq_kr_k1_xmm_xmmm128b64_imm8 },
				{ "EVEX_Vpcmpq_kr_k1_ymm_ymmm256b64_imm8", Code.EVEX_Vpcmpq_kr_k1_ymm_ymmm256b64_imm8 },
				{ "EVEX_Vpcmpq_kr_k1_zmm_zmmm512b64_imm8", Code.EVEX_Vpcmpq_kr_k1_zmm_zmmm512b64_imm8 },
				{ "Pinsrb_xmm_r32m8_imm8", Code.Pinsrb_xmm_r32m8_imm8 },
				{ "Pinsrb_xmm_r64m8_imm8", Code.Pinsrb_xmm_r64m8_imm8 },
				{ "VEX_Vpinsrb_xmm_xmm_r32m8_imm8", Code.VEX_Vpinsrb_xmm_xmm_r32m8_imm8 },
				{ "VEX_Vpinsrb_xmm_xmm_r64m8_imm8", Code.VEX_Vpinsrb_xmm_xmm_r64m8_imm8 },
				{ "EVEX_Vpinsrb_xmm_xmm_r32m8_imm8", Code.EVEX_Vpinsrb_xmm_xmm_r32m8_imm8 },
				{ "EVEX_Vpinsrb_xmm_xmm_r64m8_imm8", Code.EVEX_Vpinsrb_xmm_xmm_r64m8_imm8 },
				{ "Insertps_xmm_xmmm32_imm8", Code.Insertps_xmm_xmmm32_imm8 },
				{ "VEX_Vinsertps_xmm_xmm_xmmm32_imm8", Code.VEX_Vinsertps_xmm_xmm_xmmm32_imm8 },
				{ "EVEX_Vinsertps_xmm_xmm_xmmm32_imm8", Code.EVEX_Vinsertps_xmm_xmm_xmmm32_imm8 },
				{ "Pinsrd_xmm_rm32_imm8", Code.Pinsrd_xmm_rm32_imm8 },
				{ "Pinsrq_xmm_rm64_imm8", Code.Pinsrq_xmm_rm64_imm8 },
				{ "VEX_Vpinsrd_xmm_xmm_rm32_imm8", Code.VEX_Vpinsrd_xmm_xmm_rm32_imm8 },
				{ "VEX_Vpinsrq_xmm_xmm_rm64_imm8", Code.VEX_Vpinsrq_xmm_xmm_rm64_imm8 },
				{ "EVEX_Vpinsrd_xmm_xmm_rm32_imm8", Code.EVEX_Vpinsrd_xmm_xmm_rm32_imm8 },
				{ "EVEX_Vpinsrq_xmm_xmm_rm64_imm8", Code.EVEX_Vpinsrq_xmm_xmm_rm64_imm8 },
				{ "EVEX_Vshuff32x4_ymm_k1z_ymm_ymmm256b32_imm8", Code.EVEX_Vshuff32x4_ymm_k1z_ymm_ymmm256b32_imm8 },
				{ "EVEX_Vshuff32x4_zmm_k1z_zmm_zmmm512b32_imm8", Code.EVEX_Vshuff32x4_zmm_k1z_zmm_zmmm512b32_imm8 },
				{ "EVEX_Vshuff64x2_ymm_k1z_ymm_ymmm256b64_imm8", Code.EVEX_Vshuff64x2_ymm_k1z_ymm_ymmm256b64_imm8 },
				{ "EVEX_Vshuff64x2_zmm_k1z_zmm_zmmm512b64_imm8", Code.EVEX_Vshuff64x2_zmm_k1z_zmm_zmmm512b64_imm8 },
				{ "EVEX_Vpternlogd_xmm_k1z_xmm_xmmm128b32_imm8", Code.EVEX_Vpternlogd_xmm_k1z_xmm_xmmm128b32_imm8 },
				{ "EVEX_Vpternlogd_ymm_k1z_ymm_ymmm256b32_imm8", Code.EVEX_Vpternlogd_ymm_k1z_ymm_ymmm256b32_imm8 },
				{ "EVEX_Vpternlogd_zmm_k1z_zmm_zmmm512b32_imm8", Code.EVEX_Vpternlogd_zmm_k1z_zmm_zmmm512b32_imm8 },
				{ "EVEX_Vpternlogq_xmm_k1z_xmm_xmmm128b64_imm8", Code.EVEX_Vpternlogq_xmm_k1z_xmm_xmmm128b64_imm8 },
				{ "EVEX_Vpternlogq_ymm_k1z_ymm_ymmm256b64_imm8", Code.EVEX_Vpternlogq_ymm_k1z_ymm_ymmm256b64_imm8 },
				{ "EVEX_Vpternlogq_zmm_k1z_zmm_zmmm512b64_imm8", Code.EVEX_Vpternlogq_zmm_k1z_zmm_zmmm512b64_imm8 },
				{ "EVEX_Vgetmantps_xmm_k1z_xmmm128b32_imm8", Code.EVEX_Vgetmantps_xmm_k1z_xmmm128b32_imm8 },
				{ "EVEX_Vgetmantps_ymm_k1z_ymmm256b32_imm8", Code.EVEX_Vgetmantps_ymm_k1z_ymmm256b32_imm8 },
				{ "EVEX_Vgetmantps_zmm_k1z_zmmm512b32_imm8_sae", Code.EVEX_Vgetmantps_zmm_k1z_zmmm512b32_imm8_sae },
				{ "EVEX_Vgetmantpd_xmm_k1z_xmmm128b64_imm8", Code.EVEX_Vgetmantpd_xmm_k1z_xmmm128b64_imm8 },
				{ "EVEX_Vgetmantpd_ymm_k1z_ymmm256b64_imm8", Code.EVEX_Vgetmantpd_ymm_k1z_ymmm256b64_imm8 },
				{ "EVEX_Vgetmantpd_zmm_k1z_zmmm512b64_imm8_sae", Code.EVEX_Vgetmantpd_zmm_k1z_zmmm512b64_imm8_sae },
				{ "EVEX_Vgetmantss_xmm_k1z_xmm_xmmm32_imm8_sae", Code.EVEX_Vgetmantss_xmm_k1z_xmm_xmmm32_imm8_sae },
				{ "EVEX_Vgetmantsd_xmm_k1z_xmm_xmmm64_imm8_sae", Code.EVEX_Vgetmantsd_xmm_k1z_xmm_xmmm64_imm8_sae },
				{ "VEX_Kshiftrb_kr_kr_imm8", Code.VEX_Kshiftrb_kr_kr_imm8 },
				{ "VEX_Kshiftrw_kr_kr_imm8", Code.VEX_Kshiftrw_kr_kr_imm8 },
				{ "VEX_Kshiftrd_kr_kr_imm8", Code.VEX_Kshiftrd_kr_kr_imm8 },
				{ "VEX_Kshiftrq_kr_kr_imm8", Code.VEX_Kshiftrq_kr_kr_imm8 },
				{ "VEX_Kshiftlb_kr_kr_imm8", Code.VEX_Kshiftlb_kr_kr_imm8 },
				{ "VEX_Kshiftlw_kr_kr_imm8", Code.VEX_Kshiftlw_kr_kr_imm8 },
				{ "VEX_Kshiftld_kr_kr_imm8", Code.VEX_Kshiftld_kr_kr_imm8 },
				{ "VEX_Kshiftlq_kr_kr_imm8", Code.VEX_Kshiftlq_kr_kr_imm8 },
				{ "VEX_Vinserti128_ymm_ymm_xmmm128_imm8", Code.VEX_Vinserti128_ymm_ymm_xmmm128_imm8 },
				{ "EVEX_Vinserti32x4_ymm_k1z_ymm_xmmm128_imm8", Code.EVEX_Vinserti32x4_ymm_k1z_ymm_xmmm128_imm8 },
				{ "EVEX_Vinserti32x4_zmm_k1z_zmm_xmmm128_imm8", Code.EVEX_Vinserti32x4_zmm_k1z_zmm_xmmm128_imm8 },
				{ "EVEX_Vinserti64x2_ymm_k1z_ymm_xmmm128_imm8", Code.EVEX_Vinserti64x2_ymm_k1z_ymm_xmmm128_imm8 },
				{ "EVEX_Vinserti64x2_zmm_k1z_zmm_xmmm128_imm8", Code.EVEX_Vinserti64x2_zmm_k1z_zmm_xmmm128_imm8 },
				{ "VEX_Vextracti128_xmmm128_ymm_imm8", Code.VEX_Vextracti128_xmmm128_ymm_imm8 },
				{ "EVEX_Vextracti32x4_xmmm128_k1z_ymm_imm8", Code.EVEX_Vextracti32x4_xmmm128_k1z_ymm_imm8 },
				{ "EVEX_Vextracti32x4_xmmm128_k1z_zmm_imm8", Code.EVEX_Vextracti32x4_xmmm128_k1z_zmm_imm8 },
				{ "EVEX_Vextracti64x2_xmmm128_k1z_ymm_imm8", Code.EVEX_Vextracti64x2_xmmm128_k1z_ymm_imm8 },
				{ "EVEX_Vextracti64x2_xmmm128_k1z_zmm_imm8", Code.EVEX_Vextracti64x2_xmmm128_k1z_zmm_imm8 },
				{ "EVEX_Vinserti32x8_zmm_k1z_zmm_ymmm256_imm8", Code.EVEX_Vinserti32x8_zmm_k1z_zmm_ymmm256_imm8 },
				{ "EVEX_Vinserti64x4_zmm_k1z_zmm_ymmm256_imm8", Code.EVEX_Vinserti64x4_zmm_k1z_zmm_ymmm256_imm8 },
				{ "EVEX_Vextracti32x8_ymmm256_k1z_zmm_imm8", Code.EVEX_Vextracti32x8_ymmm256_k1z_zmm_imm8 },
				{ "EVEX_Vextracti64x4_ymmm256_k1z_zmm_imm8", Code.EVEX_Vextracti64x4_ymmm256_k1z_zmm_imm8 },
				{ "EVEX_Vpcmpub_kr_k1_xmm_xmmm128_imm8", Code.EVEX_Vpcmpub_kr_k1_xmm_xmmm128_imm8 },
				{ "EVEX_Vpcmpub_kr_k1_ymm_ymmm256_imm8", Code.EVEX_Vpcmpub_kr_k1_ymm_ymmm256_imm8 },
				{ "EVEX_Vpcmpub_kr_k1_zmm_zmmm512_imm8", Code.EVEX_Vpcmpub_kr_k1_zmm_zmmm512_imm8 },
				{ "EVEX_Vpcmpuw_kr_k1_xmm_xmmm128_imm8", Code.EVEX_Vpcmpuw_kr_k1_xmm_xmmm128_imm8 },
				{ "EVEX_Vpcmpuw_kr_k1_ymm_ymmm256_imm8", Code.EVEX_Vpcmpuw_kr_k1_ymm_ymmm256_imm8 },
				{ "EVEX_Vpcmpuw_kr_k1_zmm_zmmm512_imm8", Code.EVEX_Vpcmpuw_kr_k1_zmm_zmmm512_imm8 },
				{ "EVEX_Vpcmpb_kr_k1_xmm_xmmm128_imm8", Code.EVEX_Vpcmpb_kr_k1_xmm_xmmm128_imm8 },
				{ "EVEX_Vpcmpb_kr_k1_ymm_ymmm256_imm8", Code.EVEX_Vpcmpb_kr_k1_ymm_ymmm256_imm8 },
				{ "EVEX_Vpcmpb_kr_k1_zmm_zmmm512_imm8", Code.EVEX_Vpcmpb_kr_k1_zmm_zmmm512_imm8 },
				{ "EVEX_Vpcmpw_kr_k1_xmm_xmmm128_imm8", Code.EVEX_Vpcmpw_kr_k1_xmm_xmmm128_imm8 },
				{ "EVEX_Vpcmpw_kr_k1_ymm_ymmm256_imm8", Code.EVEX_Vpcmpw_kr_k1_ymm_ymmm256_imm8 },
				{ "EVEX_Vpcmpw_kr_k1_zmm_zmmm512_imm8", Code.EVEX_Vpcmpw_kr_k1_zmm_zmmm512_imm8 },
				{ "Dpps_xmm_xmmm128_imm8", Code.Dpps_xmm_xmmm128_imm8 },
				{ "VEX_Vdpps_xmm_xmm_xmmm128_imm8", Code.VEX_Vdpps_xmm_xmm_xmmm128_imm8 },
				{ "VEX_Vdpps_ymm_ymm_ymmm256_imm8", Code.VEX_Vdpps_ymm_ymm_ymmm256_imm8 },
				{ "Dppd_xmm_xmmm128_imm8", Code.Dppd_xmm_xmmm128_imm8 },
				{ "VEX_Vdppd_xmm_xmm_xmmm128_imm8", Code.VEX_Vdppd_xmm_xmm_xmmm128_imm8 },
				{ "Mpsadbw_xmm_xmmm128_imm8", Code.Mpsadbw_xmm_xmmm128_imm8 },
				{ "VEX_Vmpsadbw_xmm_xmm_xmmm128_imm8", Code.VEX_Vmpsadbw_xmm_xmm_xmmm128_imm8 },
				{ "VEX_Vmpsadbw_ymm_ymm_ymmm256_imm8", Code.VEX_Vmpsadbw_ymm_ymm_ymmm256_imm8 },
				{ "EVEX_Vdbpsadbw_xmm_k1z_xmm_xmmm128_imm8", Code.EVEX_Vdbpsadbw_xmm_k1z_xmm_xmmm128_imm8 },
				{ "EVEX_Vdbpsadbw_ymm_k1z_ymm_ymmm256_imm8", Code.EVEX_Vdbpsadbw_ymm_k1z_ymm_ymmm256_imm8 },
				{ "EVEX_Vdbpsadbw_zmm_k1z_zmm_zmmm512_imm8", Code.EVEX_Vdbpsadbw_zmm_k1z_zmm_zmmm512_imm8 },
				{ "EVEX_Vshufi32x4_ymm_k1z_ymm_ymmm256b32_imm8", Code.EVEX_Vshufi32x4_ymm_k1z_ymm_ymmm256b32_imm8 },
				{ "EVEX_Vshufi32x4_zmm_k1z_zmm_zmmm512b32_imm8", Code.EVEX_Vshufi32x4_zmm_k1z_zmm_zmmm512b32_imm8 },
				{ "EVEX_Vshufi64x2_ymm_k1z_ymm_ymmm256b64_imm8", Code.EVEX_Vshufi64x2_ymm_k1z_ymm_ymmm256b64_imm8 },
				{ "EVEX_Vshufi64x2_zmm_k1z_zmm_zmmm512b64_imm8", Code.EVEX_Vshufi64x2_zmm_k1z_zmm_zmmm512b64_imm8 },
				{ "Pclmulqdq_xmm_xmmm128_imm8", Code.Pclmulqdq_xmm_xmmm128_imm8 },
				{ "VEX_Vpclmulqdq_xmm_xmm_xmmm128_imm8", Code.VEX_Vpclmulqdq_xmm_xmm_xmmm128_imm8 },
				{ "VEX_Vpclmulqdq_ymm_ymm_ymmm256_imm8", Code.VEX_Vpclmulqdq_ymm_ymm_ymmm256_imm8 },
				{ "EVEX_Vpclmulqdq_xmm_xmm_xmmm128_imm8", Code.EVEX_Vpclmulqdq_xmm_xmm_xmmm128_imm8 },
				{ "EVEX_Vpclmulqdq_ymm_ymm_ymmm256_imm8", Code.EVEX_Vpclmulqdq_ymm_ymm_ymmm256_imm8 },
				{ "EVEX_Vpclmulqdq_zmm_zmm_zmmm512_imm8", Code.EVEX_Vpclmulqdq_zmm_zmm_zmmm512_imm8 },
				{ "VEX_Vperm2i128_ymm_ymm_ymmm256_imm8", Code.VEX_Vperm2i128_ymm_ymm_ymmm256_imm8 },
				{ "VEX_Vpermil2ps_xmm_xmm_xmmm128_xmm_imm2", Code.VEX_Vpermil2ps_xmm_xmm_xmmm128_xmm_imm2 },
				{ "VEX_Vpermil2ps_ymm_ymm_ymmm256_ymm_imm2", Code.VEX_Vpermil2ps_ymm_ymm_ymmm256_ymm_imm2 },
				{ "VEX_Vpermil2ps_xmm_xmm_xmm_xmmm128_imm2", Code.VEX_Vpermil2ps_xmm_xmm_xmm_xmmm128_imm2 },
				{ "VEX_Vpermil2ps_ymm_ymm_ymm_ymmm256_imm2", Code.VEX_Vpermil2ps_ymm_ymm_ymm_ymmm256_imm2 },
				{ "VEX_Vpermil2pd_xmm_xmm_xmmm128_xmm_imm2", Code.VEX_Vpermil2pd_xmm_xmm_xmmm128_xmm_imm2 },
				{ "VEX_Vpermil2pd_ymm_ymm_ymmm256_ymm_imm2", Code.VEX_Vpermil2pd_ymm_ymm_ymmm256_ymm_imm2 },
				{ "VEX_Vpermil2pd_xmm_xmm_xmm_xmmm128_imm2", Code.VEX_Vpermil2pd_xmm_xmm_xmm_xmmm128_imm2 },
				{ "VEX_Vpermil2pd_ymm_ymm_ymm_ymmm256_imm2", Code.VEX_Vpermil2pd_ymm_ymm_ymm_ymmm256_imm2 },
				{ "VEX_Vblendvps_xmm_xmm_xmmm128_xmm", Code.VEX_Vblendvps_xmm_xmm_xmmm128_xmm },
				{ "VEX_Vblendvps_ymm_ymm_ymmm256_ymm", Code.VEX_Vblendvps_ymm_ymm_ymmm256_ymm },
				{ "VEX_Vblendvpd_xmm_xmm_xmmm128_xmm", Code.VEX_Vblendvpd_xmm_xmm_xmmm128_xmm },
				{ "VEX_Vblendvpd_ymm_ymm_ymmm256_ymm", Code.VEX_Vblendvpd_ymm_ymm_ymmm256_ymm },
				{ "VEX_Vpblendvb_xmm_xmm_xmmm128_xmm", Code.VEX_Vpblendvb_xmm_xmm_xmmm128_xmm },
				{ "VEX_Vpblendvb_ymm_ymm_ymmm256_ymm", Code.VEX_Vpblendvb_ymm_ymm_ymmm256_ymm },
				{ "EVEX_Vrangeps_xmm_k1z_xmm_xmmm128b32_imm8", Code.EVEX_Vrangeps_xmm_k1z_xmm_xmmm128b32_imm8 },
				{ "EVEX_Vrangeps_ymm_k1z_ymm_ymmm256b32_imm8", Code.EVEX_Vrangeps_ymm_k1z_ymm_ymmm256b32_imm8 },
				{ "EVEX_Vrangeps_zmm_k1z_zmm_zmmm512b32_imm8_sae", Code.EVEX_Vrangeps_zmm_k1z_zmm_zmmm512b32_imm8_sae },
				{ "EVEX_Vrangepd_xmm_k1z_xmm_xmmm128b64_imm8", Code.EVEX_Vrangepd_xmm_k1z_xmm_xmmm128b64_imm8 },
				{ "EVEX_Vrangepd_ymm_k1z_ymm_ymmm256b64_imm8", Code.EVEX_Vrangepd_ymm_k1z_ymm_ymmm256b64_imm8 },
				{ "EVEX_Vrangepd_zmm_k1z_zmm_zmmm512b64_imm8_sae", Code.EVEX_Vrangepd_zmm_k1z_zmm_zmmm512b64_imm8_sae },
				{ "EVEX_Vrangess_xmm_k1z_xmm_xmmm32_imm8_sae", Code.EVEX_Vrangess_xmm_k1z_xmm_xmmm32_imm8_sae },
				{ "EVEX_Vrangesd_xmm_k1z_xmm_xmmm64_imm8_sae", Code.EVEX_Vrangesd_xmm_k1z_xmm_xmmm64_imm8_sae },
				{ "EVEX_Vfixupimmps_xmm_k1z_xmm_xmmm128b32_imm8", Code.EVEX_Vfixupimmps_xmm_k1z_xmm_xmmm128b32_imm8 },
				{ "EVEX_Vfixupimmps_ymm_k1z_ymm_ymmm256b32_imm8", Code.EVEX_Vfixupimmps_ymm_k1z_ymm_ymmm256b32_imm8 },
				{ "EVEX_Vfixupimmps_zmm_k1z_zmm_zmmm512b32_imm8_sae", Code.EVEX_Vfixupimmps_zmm_k1z_zmm_zmmm512b32_imm8_sae },
				{ "EVEX_Vfixupimmpd_xmm_k1z_xmm_xmmm128b64_imm8", Code.EVEX_Vfixupimmpd_xmm_k1z_xmm_xmmm128b64_imm8 },
				{ "EVEX_Vfixupimmpd_ymm_k1z_ymm_ymmm256b64_imm8", Code.EVEX_Vfixupimmpd_ymm_k1z_ymm_ymmm256b64_imm8 },
				{ "EVEX_Vfixupimmpd_zmm_k1z_zmm_zmmm512b64_imm8_sae", Code.EVEX_Vfixupimmpd_zmm_k1z_zmm_zmmm512b64_imm8_sae },
				{ "EVEX_Vfixupimmss_xmm_k1z_xmm_xmmm32_imm8_sae", Code.EVEX_Vfixupimmss_xmm_k1z_xmm_xmmm32_imm8_sae },
				{ "EVEX_Vfixupimmsd_xmm_k1z_xmm_xmmm64_imm8_sae", Code.EVEX_Vfixupimmsd_xmm_k1z_xmm_xmmm64_imm8_sae },
				{ "EVEX_Vreduceps_xmm_k1z_xmmm128b32_imm8", Code.EVEX_Vreduceps_xmm_k1z_xmmm128b32_imm8 },
				{ "EVEX_Vreduceps_ymm_k1z_ymmm256b32_imm8", Code.EVEX_Vreduceps_ymm_k1z_ymmm256b32_imm8 },
				{ "EVEX_Vreduceps_zmm_k1z_zmmm512b32_imm8_sae", Code.EVEX_Vreduceps_zmm_k1z_zmmm512b32_imm8_sae },
				{ "EVEX_Vreducepd_xmm_k1z_xmmm128b64_imm8", Code.EVEX_Vreducepd_xmm_k1z_xmmm128b64_imm8 },
				{ "EVEX_Vreducepd_ymm_k1z_ymmm256b64_imm8", Code.EVEX_Vreducepd_ymm_k1z_ymmm256b64_imm8 },
				{ "EVEX_Vreducepd_zmm_k1z_zmmm512b64_imm8_sae", Code.EVEX_Vreducepd_zmm_k1z_zmmm512b64_imm8_sae },
				{ "EVEX_Vreducess_xmm_k1z_xmm_xmmm32_imm8_sae", Code.EVEX_Vreducess_xmm_k1z_xmm_xmmm32_imm8_sae },
				{ "EVEX_Vreducesd_xmm_k1z_xmm_xmmm64_imm8_sae", Code.EVEX_Vreducesd_xmm_k1z_xmm_xmmm64_imm8_sae },
				{ "VEX_Vfmaddsubps_xmm_xmm_xmmm128_xmm", Code.VEX_Vfmaddsubps_xmm_xmm_xmmm128_xmm },
				{ "VEX_Vfmaddsubps_ymm_ymm_ymmm256_ymm", Code.VEX_Vfmaddsubps_ymm_ymm_ymmm256_ymm },
				{ "VEX_Vfmaddsubps_xmm_xmm_xmm_xmmm128", Code.VEX_Vfmaddsubps_xmm_xmm_xmm_xmmm128 },
				{ "VEX_Vfmaddsubps_ymm_ymm_ymm_ymmm256", Code.VEX_Vfmaddsubps_ymm_ymm_ymm_ymmm256 },
				{ "VEX_Vfmaddsubpd_xmm_xmm_xmmm128_xmm", Code.VEX_Vfmaddsubpd_xmm_xmm_xmmm128_xmm },
				{ "VEX_Vfmaddsubpd_ymm_ymm_ymmm256_ymm", Code.VEX_Vfmaddsubpd_ymm_ymm_ymmm256_ymm },
				{ "VEX_Vfmaddsubpd_xmm_xmm_xmm_xmmm128", Code.VEX_Vfmaddsubpd_xmm_xmm_xmm_xmmm128 },
				{ "VEX_Vfmaddsubpd_ymm_ymm_ymm_ymmm256", Code.VEX_Vfmaddsubpd_ymm_ymm_ymm_ymmm256 },
				{ "VEX_Vfmsubaddps_xmm_xmm_xmmm128_xmm", Code.VEX_Vfmsubaddps_xmm_xmm_xmmm128_xmm },
				{ "VEX_Vfmsubaddps_ymm_ymm_ymmm256_ymm", Code.VEX_Vfmsubaddps_ymm_ymm_ymmm256_ymm },
				{ "VEX_Vfmsubaddps_xmm_xmm_xmm_xmmm128", Code.VEX_Vfmsubaddps_xmm_xmm_xmm_xmmm128 },
				{ "VEX_Vfmsubaddps_ymm_ymm_ymm_ymmm256", Code.VEX_Vfmsubaddps_ymm_ymm_ymm_ymmm256 },
				{ "VEX_Vfmsubaddpd_xmm_xmm_xmmm128_xmm", Code.VEX_Vfmsubaddpd_xmm_xmm_xmmm128_xmm },
				{ "VEX_Vfmsubaddpd_ymm_ymm_ymmm256_ymm", Code.VEX_Vfmsubaddpd_ymm_ymm_ymmm256_ymm },
				{ "VEX_Vfmsubaddpd_xmm_xmm_xmm_xmmm128", Code.VEX_Vfmsubaddpd_xmm_xmm_xmm_xmmm128 },
				{ "VEX_Vfmsubaddpd_ymm_ymm_ymm_ymmm256", Code.VEX_Vfmsubaddpd_ymm_ymm_ymm_ymmm256 },
				{ "Pcmpestrm_xmm_xmmm128_imm8", Code.Pcmpestrm_xmm_xmmm128_imm8 },
				{ "Pcmpestrm64_xmm_xmmm128_imm8", Code.Pcmpestrm64_xmm_xmmm128_imm8 },
				{ "VEX_Vpcmpestrm_xmm_xmmm128_imm8", Code.VEX_Vpcmpestrm_xmm_xmmm128_imm8 },
				{ "VEX_Vpcmpestrm64_xmm_xmmm128_imm8", Code.VEX_Vpcmpestrm64_xmm_xmmm128_imm8 },
				{ "Pcmpestri_xmm_xmmm128_imm8", Code.Pcmpestri_xmm_xmmm128_imm8 },
				{ "Pcmpestri64_xmm_xmmm128_imm8", Code.Pcmpestri64_xmm_xmmm128_imm8 },
				{ "VEX_Vpcmpestri_xmm_xmmm128_imm8", Code.VEX_Vpcmpestri_xmm_xmmm128_imm8 },
				{ "VEX_Vpcmpestri64_xmm_xmmm128_imm8", Code.VEX_Vpcmpestri64_xmm_xmmm128_imm8 },
				{ "Pcmpistrm_xmm_xmmm128_imm8", Code.Pcmpistrm_xmm_xmmm128_imm8 },
				{ "VEX_Vpcmpistrm_xmm_xmmm128_imm8", Code.VEX_Vpcmpistrm_xmm_xmmm128_imm8 },
				{ "Pcmpistri_xmm_xmmm128_imm8", Code.Pcmpistri_xmm_xmmm128_imm8 },
				{ "VEX_Vpcmpistri_xmm_xmmm128_imm8", Code.VEX_Vpcmpistri_xmm_xmmm128_imm8 },
				{ "EVEX_Vfpclassps_kr_k1_xmmm128b32_imm8", Code.EVEX_Vfpclassps_kr_k1_xmmm128b32_imm8 },
				{ "EVEX_Vfpclassps_kr_k1_ymmm256b32_imm8", Code.EVEX_Vfpclassps_kr_k1_ymmm256b32_imm8 },
				{ "EVEX_Vfpclassps_kr_k1_zmmm512b32_imm8", Code.EVEX_Vfpclassps_kr_k1_zmmm512b32_imm8 },
				{ "EVEX_Vfpclasspd_kr_k1_xmmm128b64_imm8", Code.EVEX_Vfpclasspd_kr_k1_xmmm128b64_imm8 },
				{ "EVEX_Vfpclasspd_kr_k1_ymmm256b64_imm8", Code.EVEX_Vfpclasspd_kr_k1_ymmm256b64_imm8 },
				{ "EVEX_Vfpclasspd_kr_k1_zmmm512b64_imm8", Code.EVEX_Vfpclasspd_kr_k1_zmmm512b64_imm8 },
				{ "EVEX_Vfpclassss_kr_k1_xmmm32_imm8", Code.EVEX_Vfpclassss_kr_k1_xmmm32_imm8 },
				{ "EVEX_Vfpclasssd_kr_k1_xmmm64_imm8", Code.EVEX_Vfpclasssd_kr_k1_xmmm64_imm8 },
				{ "VEX_Vfmaddps_xmm_xmm_xmmm128_xmm", Code.VEX_Vfmaddps_xmm_xmm_xmmm128_xmm },
				{ "VEX_Vfmaddps_ymm_ymm_ymmm256_ymm", Code.VEX_Vfmaddps_ymm_ymm_ymmm256_ymm },
				{ "VEX_Vfmaddps_xmm_xmm_xmm_xmmm128", Code.VEX_Vfmaddps_xmm_xmm_xmm_xmmm128 },
				{ "VEX_Vfmaddps_ymm_ymm_ymm_ymmm256", Code.VEX_Vfmaddps_ymm_ymm_ymm_ymmm256 },
				{ "VEX_Vfmaddpd_xmm_xmm_xmmm128_xmm", Code.VEX_Vfmaddpd_xmm_xmm_xmmm128_xmm },
				{ "VEX_Vfmaddpd_ymm_ymm_ymmm256_ymm", Code.VEX_Vfmaddpd_ymm_ymm_ymmm256_ymm },
				{ "VEX_Vfmaddpd_xmm_xmm_xmm_xmmm128", Code.VEX_Vfmaddpd_xmm_xmm_xmm_xmmm128 },
				{ "VEX_Vfmaddpd_ymm_ymm_ymm_ymmm256", Code.VEX_Vfmaddpd_ymm_ymm_ymm_ymmm256 },
				{ "VEX_Vfmaddss_xmm_xmm_xmmm32_xmm", Code.VEX_Vfmaddss_xmm_xmm_xmmm32_xmm },
				{ "VEX_Vfmaddss_xmm_xmm_xmm_xmmm32", Code.VEX_Vfmaddss_xmm_xmm_xmm_xmmm32 },
				{ "VEX_Vfmaddsd_xmm_xmm_xmmm64_xmm", Code.VEX_Vfmaddsd_xmm_xmm_xmmm64_xmm },
				{ "VEX_Vfmaddsd_xmm_xmm_xmm_xmmm64", Code.VEX_Vfmaddsd_xmm_xmm_xmm_xmmm64 },
				{ "VEX_Vfmsubps_xmm_xmm_xmmm128_xmm", Code.VEX_Vfmsubps_xmm_xmm_xmmm128_xmm },
				{ "VEX_Vfmsubps_ymm_ymm_ymmm256_ymm", Code.VEX_Vfmsubps_ymm_ymm_ymmm256_ymm },
				{ "VEX_Vfmsubps_xmm_xmm_xmm_xmmm128", Code.VEX_Vfmsubps_xmm_xmm_xmm_xmmm128 },
				{ "VEX_Vfmsubps_ymm_ymm_ymm_ymmm256", Code.VEX_Vfmsubps_ymm_ymm_ymm_ymmm256 },
				{ "VEX_Vfmsubpd_xmm_xmm_xmmm128_xmm", Code.VEX_Vfmsubpd_xmm_xmm_xmmm128_xmm },
				{ "VEX_Vfmsubpd_ymm_ymm_ymmm256_ymm", Code.VEX_Vfmsubpd_ymm_ymm_ymmm256_ymm },
				{ "VEX_Vfmsubpd_xmm_xmm_xmm_xmmm128", Code.VEX_Vfmsubpd_xmm_xmm_xmm_xmmm128 },
				{ "VEX_Vfmsubpd_ymm_ymm_ymm_ymmm256", Code.VEX_Vfmsubpd_ymm_ymm_ymm_ymmm256 },
				{ "VEX_Vfmsubss_xmm_xmm_xmmm32_xmm", Code.VEX_Vfmsubss_xmm_xmm_xmmm32_xmm },
				{ "VEX_Vfmsubss_xmm_xmm_xmm_xmmm32", Code.VEX_Vfmsubss_xmm_xmm_xmm_xmmm32 },
				{ "VEX_Vfmsubsd_xmm_xmm_xmmm64_xmm", Code.VEX_Vfmsubsd_xmm_xmm_xmmm64_xmm },
				{ "VEX_Vfmsubsd_xmm_xmm_xmm_xmmm64", Code.VEX_Vfmsubsd_xmm_xmm_xmm_xmmm64 },
				{ "EVEX_Vpshldw_xmm_k1z_xmm_xmmm128_imm8", Code.EVEX_Vpshldw_xmm_k1z_xmm_xmmm128_imm8 },
				{ "EVEX_Vpshldw_ymm_k1z_ymm_ymmm256_imm8", Code.EVEX_Vpshldw_ymm_k1z_ymm_ymmm256_imm8 },
				{ "EVEX_Vpshldw_zmm_k1z_zmm_zmmm512_imm8", Code.EVEX_Vpshldw_zmm_k1z_zmm_zmmm512_imm8 },
				{ "EVEX_Vpshldd_xmm_k1z_xmm_xmmm128b32_imm8", Code.EVEX_Vpshldd_xmm_k1z_xmm_xmmm128b32_imm8 },
				{ "EVEX_Vpshldd_ymm_k1z_ymm_ymmm256b32_imm8", Code.EVEX_Vpshldd_ymm_k1z_ymm_ymmm256b32_imm8 },
				{ "EVEX_Vpshldd_zmm_k1z_zmm_zmmm512b32_imm8", Code.EVEX_Vpshldd_zmm_k1z_zmm_zmmm512b32_imm8 },
				{ "EVEX_Vpshldq_xmm_k1z_xmm_xmmm128b64_imm8", Code.EVEX_Vpshldq_xmm_k1z_xmm_xmmm128b64_imm8 },
				{ "EVEX_Vpshldq_ymm_k1z_ymm_ymmm256b64_imm8", Code.EVEX_Vpshldq_ymm_k1z_ymm_ymmm256b64_imm8 },
				{ "EVEX_Vpshldq_zmm_k1z_zmm_zmmm512b64_imm8", Code.EVEX_Vpshldq_zmm_k1z_zmm_zmmm512b64_imm8 },
				{ "EVEX_Vpshrdw_xmm_k1z_xmm_xmmm128_imm8", Code.EVEX_Vpshrdw_xmm_k1z_xmm_xmmm128_imm8 },
				{ "EVEX_Vpshrdw_ymm_k1z_ymm_ymmm256_imm8", Code.EVEX_Vpshrdw_ymm_k1z_ymm_ymmm256_imm8 },
				{ "EVEX_Vpshrdw_zmm_k1z_zmm_zmmm512_imm8", Code.EVEX_Vpshrdw_zmm_k1z_zmm_zmmm512_imm8 },
				{ "EVEX_Vpshrdd_xmm_k1z_xmm_xmmm128b32_imm8", Code.EVEX_Vpshrdd_xmm_k1z_xmm_xmmm128b32_imm8 },
				{ "EVEX_Vpshrdd_ymm_k1z_ymm_ymmm256b32_imm8", Code.EVEX_Vpshrdd_ymm_k1z_ymm_ymmm256b32_imm8 },
				{ "EVEX_Vpshrdd_zmm_k1z_zmm_zmmm512b32_imm8", Code.EVEX_Vpshrdd_zmm_k1z_zmm_zmmm512b32_imm8 },
				{ "EVEX_Vpshrdq_xmm_k1z_xmm_xmmm128b64_imm8", Code.EVEX_Vpshrdq_xmm_k1z_xmm_xmmm128b64_imm8 },
				{ "EVEX_Vpshrdq_ymm_k1z_ymm_ymmm256b64_imm8", Code.EVEX_Vpshrdq_ymm_k1z_ymm_ymmm256b64_imm8 },
				{ "EVEX_Vpshrdq_zmm_k1z_zmm_zmmm512b64_imm8", Code.EVEX_Vpshrdq_zmm_k1z_zmm_zmmm512b64_imm8 },
				{ "VEX_Vfnmaddps_xmm_xmm_xmmm128_xmm", Code.VEX_Vfnmaddps_xmm_xmm_xmmm128_xmm },
				{ "VEX_Vfnmaddps_ymm_ymm_ymmm256_ymm", Code.VEX_Vfnmaddps_ymm_ymm_ymmm256_ymm },
				{ "VEX_Vfnmaddps_xmm_xmm_xmm_xmmm128", Code.VEX_Vfnmaddps_xmm_xmm_xmm_xmmm128 },
				{ "VEX_Vfnmaddps_ymm_ymm_ymm_ymmm256", Code.VEX_Vfnmaddps_ymm_ymm_ymm_ymmm256 },
				{ "VEX_Vfnmaddpd_xmm_xmm_xmmm128_xmm", Code.VEX_Vfnmaddpd_xmm_xmm_xmmm128_xmm },
				{ "VEX_Vfnmaddpd_ymm_ymm_ymmm256_ymm", Code.VEX_Vfnmaddpd_ymm_ymm_ymmm256_ymm },
				{ "VEX_Vfnmaddpd_xmm_xmm_xmm_xmmm128", Code.VEX_Vfnmaddpd_xmm_xmm_xmm_xmmm128 },
				{ "VEX_Vfnmaddpd_ymm_ymm_ymm_ymmm256", Code.VEX_Vfnmaddpd_ymm_ymm_ymm_ymmm256 },
				{ "VEX_Vfnmaddss_xmm_xmm_xmmm32_xmm", Code.VEX_Vfnmaddss_xmm_xmm_xmmm32_xmm },
				{ "VEX_Vfnmaddss_xmm_xmm_xmm_xmmm32", Code.VEX_Vfnmaddss_xmm_xmm_xmm_xmmm32 },
				{ "VEX_Vfnmaddsd_xmm_xmm_xmmm64_xmm", Code.VEX_Vfnmaddsd_xmm_xmm_xmmm64_xmm },
				{ "VEX_Vfnmaddsd_xmm_xmm_xmm_xmmm64", Code.VEX_Vfnmaddsd_xmm_xmm_xmm_xmmm64 },
				{ "VEX_Vfnmsubps_xmm_xmm_xmmm128_xmm", Code.VEX_Vfnmsubps_xmm_xmm_xmmm128_xmm },
				{ "VEX_Vfnmsubps_ymm_ymm_ymmm256_ymm", Code.VEX_Vfnmsubps_ymm_ymm_ymmm256_ymm },
				{ "VEX_Vfnmsubps_xmm_xmm_xmm_xmmm128", Code.VEX_Vfnmsubps_xmm_xmm_xmm_xmmm128 },
				{ "VEX_Vfnmsubps_ymm_ymm_ymm_ymmm256", Code.VEX_Vfnmsubps_ymm_ymm_ymm_ymmm256 },
				{ "VEX_Vfnmsubpd_xmm_xmm_xmmm128_xmm", Code.VEX_Vfnmsubpd_xmm_xmm_xmmm128_xmm },
				{ "VEX_Vfnmsubpd_ymm_ymm_ymmm256_ymm", Code.VEX_Vfnmsubpd_ymm_ymm_ymmm256_ymm },
				{ "VEX_Vfnmsubpd_xmm_xmm_xmm_xmmm128", Code.VEX_Vfnmsubpd_xmm_xmm_xmm_xmmm128 },
				{ "VEX_Vfnmsubpd_ymm_ymm_ymm_ymmm256", Code.VEX_Vfnmsubpd_ymm_ymm_ymm_ymmm256 },
				{ "VEX_Vfnmsubss_xmm_xmm_xmmm32_xmm", Code.VEX_Vfnmsubss_xmm_xmm_xmmm32_xmm },
				{ "VEX_Vfnmsubss_xmm_xmm_xmm_xmmm32", Code.VEX_Vfnmsubss_xmm_xmm_xmm_xmmm32 },
				{ "VEX_Vfnmsubsd_xmm_xmm_xmmm64_xmm", Code.VEX_Vfnmsubsd_xmm_xmm_xmmm64_xmm },
				{ "VEX_Vfnmsubsd_xmm_xmm_xmm_xmmm64", Code.VEX_Vfnmsubsd_xmm_xmm_xmm_xmmm64 },
				{ "Sha1rnds4_xmm_xmmm128_imm8", Code.Sha1rnds4_xmm_xmmm128_imm8 },
				{ "Gf2p8affineqb_xmm_xmmm128_imm8", Code.Gf2p8affineqb_xmm_xmmm128_imm8 },
				{ "VEX_Vgf2p8affineqb_xmm_xmm_xmmm128_imm8", Code.VEX_Vgf2p8affineqb_xmm_xmm_xmmm128_imm8 },
				{ "VEX_Vgf2p8affineqb_ymm_ymm_ymmm256_imm8", Code.VEX_Vgf2p8affineqb_ymm_ymm_ymmm256_imm8 },
				{ "EVEX_Vgf2p8affineqb_xmm_k1z_xmm_xmmm128b64_imm8", Code.EVEX_Vgf2p8affineqb_xmm_k1z_xmm_xmmm128b64_imm8 },
				{ "EVEX_Vgf2p8affineqb_ymm_k1z_ymm_ymmm256b64_imm8", Code.EVEX_Vgf2p8affineqb_ymm_k1z_ymm_ymmm256b64_imm8 },
				{ "EVEX_Vgf2p8affineqb_zmm_k1z_zmm_zmmm512b64_imm8", Code.EVEX_Vgf2p8affineqb_zmm_k1z_zmm_zmmm512b64_imm8 },
				{ "Gf2p8affineinvqb_xmm_xmmm128_imm8", Code.Gf2p8affineinvqb_xmm_xmmm128_imm8 },
				{ "VEX_Vgf2p8affineinvqb_xmm_xmm_xmmm128_imm8", Code.VEX_Vgf2p8affineinvqb_xmm_xmm_xmmm128_imm8 },
				{ "VEX_Vgf2p8affineinvqb_ymm_ymm_ymmm256_imm8", Code.VEX_Vgf2p8affineinvqb_ymm_ymm_ymmm256_imm8 },
				{ "EVEX_Vgf2p8affineinvqb_xmm_k1z_xmm_xmmm128b64_imm8", Code.EVEX_Vgf2p8affineinvqb_xmm_k1z_xmm_xmmm128b64_imm8 },
				{ "EVEX_Vgf2p8affineinvqb_ymm_k1z_ymm_ymmm256b64_imm8", Code.EVEX_Vgf2p8affineinvqb_ymm_k1z_ymm_ymmm256b64_imm8 },
				{ "EVEX_Vgf2p8affineinvqb_zmm_k1z_zmm_zmmm512b64_imm8", Code.EVEX_Vgf2p8affineinvqb_zmm_k1z_zmm_zmmm512b64_imm8 },
				{ "Aeskeygenassist_xmm_xmmm128_imm8", Code.Aeskeygenassist_xmm_xmmm128_imm8 },
				{ "VEX_Vaeskeygenassist_xmm_xmmm128_imm8", Code.VEX_Vaeskeygenassist_xmm_xmmm128_imm8 },
				{ "VEX_Rorx_r32_rm32_imm8", Code.VEX_Rorx_r32_rm32_imm8 },
				{ "VEX_Rorx_r64_rm64_imm8", Code.VEX_Rorx_r64_rm64_imm8 },
				{ "XOP_Vpmacssww_xmm_xmm_xmmm128_xmm", Code.XOP_Vpmacssww_xmm_xmm_xmmm128_xmm },
				{ "XOP_Vpmacsswd_xmm_xmm_xmmm128_xmm", Code.XOP_Vpmacsswd_xmm_xmm_xmmm128_xmm },
				{ "XOP_Vpmacssdql_xmm_xmm_xmmm128_xmm", Code.XOP_Vpmacssdql_xmm_xmm_xmmm128_xmm },
				{ "XOP_Vpmacssdd_xmm_xmm_xmmm128_xmm", Code.XOP_Vpmacssdd_xmm_xmm_xmmm128_xmm },
				{ "XOP_Vpmacssdqh_xmm_xmm_xmmm128_xmm", Code.XOP_Vpmacssdqh_xmm_xmm_xmmm128_xmm },
				{ "XOP_Vpmacsww_xmm_xmm_xmmm128_xmm", Code.XOP_Vpmacsww_xmm_xmm_xmmm128_xmm },
				{ "XOP_Vpmacswd_xmm_xmm_xmmm128_xmm", Code.XOP_Vpmacswd_xmm_xmm_xmmm128_xmm },
				{ "XOP_Vpmacsdql_xmm_xmm_xmmm128_xmm", Code.XOP_Vpmacsdql_xmm_xmm_xmmm128_xmm },
				{ "XOP_Vpmacsdd_xmm_xmm_xmmm128_xmm", Code.XOP_Vpmacsdd_xmm_xmm_xmmm128_xmm },
				{ "XOP_Vpmacsdqh_xmm_xmm_xmmm128_xmm", Code.XOP_Vpmacsdqh_xmm_xmm_xmmm128_xmm },
				{ "XOP_Vpcmov_xmm_xmm_xmmm128_xmm", Code.XOP_Vpcmov_xmm_xmm_xmmm128_xmm },
				{ "XOP_Vpcmov_ymm_ymm_ymmm256_ymm", Code.XOP_Vpcmov_ymm_ymm_ymmm256_ymm },
				{ "XOP_Vpcmov_xmm_xmm_xmm_xmmm128", Code.XOP_Vpcmov_xmm_xmm_xmm_xmmm128 },
				{ "XOP_Vpcmov_ymm_ymm_ymm_ymmm256", Code.XOP_Vpcmov_ymm_ymm_ymm_ymmm256 },
				{ "XOP_Vpperm_xmm_xmm_xmmm128_xmm", Code.XOP_Vpperm_xmm_xmm_xmmm128_xmm },
				{ "XOP_Vpperm_xmm_xmm_xmm_xmmm128", Code.XOP_Vpperm_xmm_xmm_xmm_xmmm128 },
				{ "XOP_Vpmadcsswd_xmm_xmm_xmmm128_xmm", Code.XOP_Vpmadcsswd_xmm_xmm_xmmm128_xmm },
				{ "XOP_Vpmadcswd_xmm_xmm_xmmm128_xmm", Code.XOP_Vpmadcswd_xmm_xmm_xmmm128_xmm },
				{ "XOP_Vprotb_xmm_xmmm128_imm8", Code.XOP_Vprotb_xmm_xmmm128_imm8 },
				{ "XOP_Vprotw_xmm_xmmm128_imm8", Code.XOP_Vprotw_xmm_xmmm128_imm8 },
				{ "XOP_Vprotd_xmm_xmmm128_imm8", Code.XOP_Vprotd_xmm_xmmm128_imm8 },
				{ "XOP_Vprotq_xmm_xmmm128_imm8", Code.XOP_Vprotq_xmm_xmmm128_imm8 },
				{ "XOP_Vpcomb_xmm_xmm_xmmm128_imm8", Code.XOP_Vpcomb_xmm_xmm_xmmm128_imm8 },
				{ "XOP_Vpcomw_xmm_xmm_xmmm128_imm8", Code.XOP_Vpcomw_xmm_xmm_xmmm128_imm8 },
				{ "XOP_Vpcomd_xmm_xmm_xmmm128_imm8", Code.XOP_Vpcomd_xmm_xmm_xmmm128_imm8 },
				{ "XOP_Vpcomq_xmm_xmm_xmmm128_imm8", Code.XOP_Vpcomq_xmm_xmm_xmmm128_imm8 },
				{ "XOP_Vpcomub_xmm_xmm_xmmm128_imm8", Code.XOP_Vpcomub_xmm_xmm_xmmm128_imm8 },
				{ "XOP_Vpcomuw_xmm_xmm_xmmm128_imm8", Code.XOP_Vpcomuw_xmm_xmm_xmmm128_imm8 },
				{ "XOP_Vpcomud_xmm_xmm_xmmm128_imm8", Code.XOP_Vpcomud_xmm_xmm_xmmm128_imm8 },
				{ "XOP_Vpcomuq_xmm_xmm_xmmm128_imm8", Code.XOP_Vpcomuq_xmm_xmm_xmmm128_imm8 },
				{ "XOP_Blcfill_r32_rm32", Code.XOP_Blcfill_r32_rm32 },
				{ "XOP_Blcfill_r64_rm64", Code.XOP_Blcfill_r64_rm64 },
				{ "XOP_Blsfill_r32_rm32", Code.XOP_Blsfill_r32_rm32 },
				{ "XOP_Blsfill_r64_rm64", Code.XOP_Blsfill_r64_rm64 },
				{ "XOP_Blcs_r32_rm32", Code.XOP_Blcs_r32_rm32 },
				{ "XOP_Blcs_r64_rm64", Code.XOP_Blcs_r64_rm64 },
				{ "XOP_Tzmsk_r32_rm32", Code.XOP_Tzmsk_r32_rm32 },
				{ "XOP_Tzmsk_r64_rm64", Code.XOP_Tzmsk_r64_rm64 },
				{ "XOP_Blcic_r32_rm32", Code.XOP_Blcic_r32_rm32 },
				{ "XOP_Blcic_r64_rm64", Code.XOP_Blcic_r64_rm64 },
				{ "XOP_Blsic_r32_rm32", Code.XOP_Blsic_r32_rm32 },
				{ "XOP_Blsic_r64_rm64", Code.XOP_Blsic_r64_rm64 },
				{ "XOP_T1mskc_r32_rm32", Code.XOP_T1mskc_r32_rm32 },
				{ "XOP_T1mskc_r64_rm64", Code.XOP_T1mskc_r64_rm64 },
				{ "XOP_Blcmsk_r32_rm32", Code.XOP_Blcmsk_r32_rm32 },
				{ "XOP_Blcmsk_r64_rm64", Code.XOP_Blcmsk_r64_rm64 },
				{ "XOP_Blci_r32_rm32", Code.XOP_Blci_r32_rm32 },
				{ "XOP_Blci_r64_rm64", Code.XOP_Blci_r64_rm64 },
				{ "XOP_Llwpcb_r32", Code.XOP_Llwpcb_r32 },
				{ "XOP_Llwpcb_r64", Code.XOP_Llwpcb_r64 },
				{ "XOP_Slwpcb_r32", Code.XOP_Slwpcb_r32 },
				{ "XOP_Slwpcb_r64", Code.XOP_Slwpcb_r64 },
				{ "XOP_Vfrczps_xmm_xmmm128", Code.XOP_Vfrczps_xmm_xmmm128 },
				{ "XOP_Vfrczps_ymm_ymmm256", Code.XOP_Vfrczps_ymm_ymmm256 },
				{ "XOP_Vfrczpd_xmm_xmmm128", Code.XOP_Vfrczpd_xmm_xmmm128 },
				{ "XOP_Vfrczpd_ymm_ymmm256", Code.XOP_Vfrczpd_ymm_ymmm256 },
				{ "XOP_Vfrczss_xmm_xmmm32", Code.XOP_Vfrczss_xmm_xmmm32 },
				{ "XOP_Vfrczsd_xmm_xmmm64", Code.XOP_Vfrczsd_xmm_xmmm64 },
				{ "XOP_Vprotb_xmm_xmmm128_xmm", Code.XOP_Vprotb_xmm_xmmm128_xmm },
				{ "XOP_Vprotb_xmm_xmm_xmmm128", Code.XOP_Vprotb_xmm_xmm_xmmm128 },
				{ "XOP_Vprotw_xmm_xmmm128_xmm", Code.XOP_Vprotw_xmm_xmmm128_xmm },
				{ "XOP_Vprotw_xmm_xmm_xmmm128", Code.XOP_Vprotw_xmm_xmm_xmmm128 },
				{ "XOP_Vprotd_xmm_xmmm128_xmm", Code.XOP_Vprotd_xmm_xmmm128_xmm },
				{ "XOP_Vprotd_xmm_xmm_xmmm128", Code.XOP_Vprotd_xmm_xmm_xmmm128 },
				{ "XOP_Vprotq_xmm_xmmm128_xmm", Code.XOP_Vprotq_xmm_xmmm128_xmm },
				{ "XOP_Vprotq_xmm_xmm_xmmm128", Code.XOP_Vprotq_xmm_xmm_xmmm128 },
				{ "XOP_Vpshlb_xmm_xmmm128_xmm", Code.XOP_Vpshlb_xmm_xmmm128_xmm },
				{ "XOP_Vpshlb_xmm_xmm_xmmm128", Code.XOP_Vpshlb_xmm_xmm_xmmm128 },
				{ "XOP_Vpshlw_xmm_xmmm128_xmm", Code.XOP_Vpshlw_xmm_xmmm128_xmm },
				{ "XOP_Vpshlw_xmm_xmm_xmmm128", Code.XOP_Vpshlw_xmm_xmm_xmmm128 },
				{ "XOP_Vpshld_xmm_xmmm128_xmm", Code.XOP_Vpshld_xmm_xmmm128_xmm },
				{ "XOP_Vpshld_xmm_xmm_xmmm128", Code.XOP_Vpshld_xmm_xmm_xmmm128 },
				{ "XOP_Vpshlq_xmm_xmmm128_xmm", Code.XOP_Vpshlq_xmm_xmmm128_xmm },
				{ "XOP_Vpshlq_xmm_xmm_xmmm128", Code.XOP_Vpshlq_xmm_xmm_xmmm128 },
				{ "XOP_Vpshab_xmm_xmmm128_xmm", Code.XOP_Vpshab_xmm_xmmm128_xmm },
				{ "XOP_Vpshab_xmm_xmm_xmmm128", Code.XOP_Vpshab_xmm_xmm_xmmm128 },
				{ "XOP_Vpshaw_xmm_xmmm128_xmm", Code.XOP_Vpshaw_xmm_xmmm128_xmm },
				{ "XOP_Vpshaw_xmm_xmm_xmmm128", Code.XOP_Vpshaw_xmm_xmm_xmmm128 },
				{ "XOP_Vpshad_xmm_xmmm128_xmm", Code.XOP_Vpshad_xmm_xmmm128_xmm },
				{ "XOP_Vpshad_xmm_xmm_xmmm128", Code.XOP_Vpshad_xmm_xmm_xmmm128 },
				{ "XOP_Vpshaq_xmm_xmmm128_xmm", Code.XOP_Vpshaq_xmm_xmmm128_xmm },
				{ "XOP_Vpshaq_xmm_xmm_xmmm128", Code.XOP_Vpshaq_xmm_xmm_xmmm128 },
				{ "XOP_Vphaddbw_xmm_xmmm128", Code.XOP_Vphaddbw_xmm_xmmm128 },
				{ "XOP_Vphaddbd_xmm_xmmm128", Code.XOP_Vphaddbd_xmm_xmmm128 },
				{ "XOP_Vphaddbq_xmm_xmmm128", Code.XOP_Vphaddbq_xmm_xmmm128 },
				{ "XOP_Vphaddwd_xmm_xmmm128", Code.XOP_Vphaddwd_xmm_xmmm128 },
				{ "XOP_Vphaddwq_xmm_xmmm128", Code.XOP_Vphaddwq_xmm_xmmm128 },
				{ "XOP_Vphadddq_xmm_xmmm128", Code.XOP_Vphadddq_xmm_xmmm128 },
				{ "XOP_Vphaddubw_xmm_xmmm128", Code.XOP_Vphaddubw_xmm_xmmm128 },
				{ "XOP_Vphaddubd_xmm_xmmm128", Code.XOP_Vphaddubd_xmm_xmmm128 },
				{ "XOP_Vphaddubq_xmm_xmmm128", Code.XOP_Vphaddubq_xmm_xmmm128 },
				{ "XOP_Vphadduwd_xmm_xmmm128", Code.XOP_Vphadduwd_xmm_xmmm128 },
				{ "XOP_Vphadduwq_xmm_xmmm128", Code.XOP_Vphadduwq_xmm_xmmm128 },
				{ "XOP_Vphaddudq_xmm_xmmm128", Code.XOP_Vphaddudq_xmm_xmmm128 },
				{ "XOP_Vphsubbw_xmm_xmmm128", Code.XOP_Vphsubbw_xmm_xmmm128 },
				{ "XOP_Vphsubwd_xmm_xmmm128", Code.XOP_Vphsubwd_xmm_xmmm128 },
				{ "XOP_Vphsubdq_xmm_xmmm128", Code.XOP_Vphsubdq_xmm_xmmm128 },
				{ "XOP_Bextr_r32_rm32_imm32", Code.XOP_Bextr_r32_rm32_imm32 },
				{ "XOP_Bextr_r64_rm64_imm32", Code.XOP_Bextr_r64_rm64_imm32 },
				{ "XOP_Lwpins_r32_rm32_imm32", Code.XOP_Lwpins_r32_rm32_imm32 },
				{ "XOP_Lwpins_r64_rm32_imm32", Code.XOP_Lwpins_r64_rm32_imm32 },
				{ "XOP_Lwpval_r32_rm32_imm32", Code.XOP_Lwpval_r32_rm32_imm32 },
				{ "XOP_Lwpval_r64_rm32_imm32", Code.XOP_Lwpval_r64_rm32_imm32 },
				{ "D3NOW_Pi2fw_mm_mmm64", Code.D3NOW_Pi2fw_mm_mmm64 },
				{ "D3NOW_Pi2fd_mm_mmm64", Code.D3NOW_Pi2fd_mm_mmm64 },
				{ "D3NOW_Pf2iw_mm_mmm64", Code.D3NOW_Pf2iw_mm_mmm64 },
				{ "D3NOW_Pf2id_mm_mmm64", Code.D3NOW_Pf2id_mm_mmm64 },
				{ "D3NOW_Pfrcpv_mm_mmm64", Code.D3NOW_Pfrcpv_mm_mmm64 },
				{ "D3NOW_Pfrsqrtv_mm_mmm64", Code.D3NOW_Pfrsqrtv_mm_mmm64 },
				{ "D3NOW_Pfnacc_mm_mmm64", Code.D3NOW_Pfnacc_mm_mmm64 },
				{ "D3NOW_Pfpnacc_mm_mmm64", Code.D3NOW_Pfpnacc_mm_mmm64 },
				{ "D3NOW_Pfcmpge_mm_mmm64", Code.D3NOW_Pfcmpge_mm_mmm64 },
				{ "D3NOW_Pfmin_mm_mmm64", Code.D3NOW_Pfmin_mm_mmm64 },
				{ "D3NOW_Pfrcp_mm_mmm64", Code.D3NOW_Pfrcp_mm_mmm64 },
				{ "D3NOW_Pfrsqrt_mm_mmm64", Code.D3NOW_Pfrsqrt_mm_mmm64 },
				{ "D3NOW_Pfsub_mm_mmm64", Code.D3NOW_Pfsub_mm_mmm64 },
				{ "D3NOW_Pfadd_mm_mmm64", Code.D3NOW_Pfadd_mm_mmm64 },
				{ "D3NOW_Pfcmpgt_mm_mmm64", Code.D3NOW_Pfcmpgt_mm_mmm64 },
				{ "D3NOW_Pfmax_mm_mmm64", Code.D3NOW_Pfmax_mm_mmm64 },
				{ "D3NOW_Pfrcpit1_mm_mmm64", Code.D3NOW_Pfrcpit1_mm_mmm64 },
				{ "D3NOW_Pfrsqit1_mm_mmm64", Code.D3NOW_Pfrsqit1_mm_mmm64 },
				{ "D3NOW_Pfsubr_mm_mmm64", Code.D3NOW_Pfsubr_mm_mmm64 },
				{ "D3NOW_Pfacc_mm_mmm64", Code.D3NOW_Pfacc_mm_mmm64 },
				{ "D3NOW_Pfcmpeq_mm_mmm64", Code.D3NOW_Pfcmpeq_mm_mmm64 },
				{ "D3NOW_Pfmul_mm_mmm64", Code.D3NOW_Pfmul_mm_mmm64 },
				{ "D3NOW_Pfrcpit2_mm_mmm64", Code.D3NOW_Pfrcpit2_mm_mmm64 },
				{ "D3NOW_Pmulhrw_mm_mmm64", Code.D3NOW_Pmulhrw_mm_mmm64 },
				{ "D3NOW_Pswapd_mm_mmm64", Code.D3NOW_Pswapd_mm_mmm64 },
				{ "D3NOW_Pavgusb_mm_mmm64", Code.D3NOW_Pavgusb_mm_mmm64 },
				{ "Rmpadjust", Code.Rmpadjust },
				{ "Rmpupdate", Code.Rmpupdate },
				{ "Psmash", Code.Psmash },
				{ "Pvalidatew", Code.Pvalidatew },
				{ "Pvalidated", Code.Pvalidated },
				{ "Pvalidateq", Code.Pvalidateq },
				{ "Serialize", Code.Serialize },
				{ "Xsusldtrk", Code.Xsusldtrk },
				{ "Xresldtrk", Code.Xresldtrk },
				{ "Invlpgbw", Code.Invlpgbw },
				{ "Invlpgbd", Code.Invlpgbd },
				{ "Invlpgbq", Code.Invlpgbq },
				{ "Tlbsync", Code.Tlbsync },
				{ "Prefetchreserved3_m8", Code.Prefetchreserved3_m8 },
				{ "Prefetchreserved4_m8", Code.Prefetchreserved4_m8 },
				{ "Prefetchreserved5_m8", Code.Prefetchreserved5_m8 },
				{ "Prefetchreserved6_m8", Code.Prefetchreserved6_m8 },
				{ "Prefetchreserved7_m8", Code.Prefetchreserved7_m8 },
				{ "Ud0", Code.Ud0 },
				{ "Vmgexit", Code.Vmgexit },
				{ "Getsecq", Code.Getsecq },
				{ "VEX_Ldtilecfg_m512", Code.VEX_Ldtilecfg_m512 },
				{ "VEX_Tilerelease", Code.VEX_Tilerelease },
				{ "VEX_Sttilecfg_m512", Code.VEX_Sttilecfg_m512 },
				{ "VEX_Tilezero_tmm", Code.VEX_Tilezero_tmm },
				{ "VEX_Tileloaddt1_tmm_sibmem", Code.VEX_Tileloaddt1_tmm_sibmem },
				{ "VEX_Tilestored_sibmem_tmm", Code.VEX_Tilestored_sibmem_tmm },
				{ "VEX_Tileloadd_tmm_sibmem", Code.VEX_Tileloadd_tmm_sibmem },
				{ "VEX_Tdpbf16ps_tmm_tmm_tmm", Code.VEX_Tdpbf16ps_tmm_tmm_tmm },
				{ "VEX_Tdpbuud_tmm_tmm_tmm", Code.VEX_Tdpbuud_tmm_tmm_tmm },
				{ "VEX_Tdpbusd_tmm_tmm_tmm", Code.VEX_Tdpbusd_tmm_tmm_tmm },
				{ "VEX_Tdpbsud_tmm_tmm_tmm", Code.VEX_Tdpbsud_tmm_tmm_tmm },
				{ "VEX_Tdpbssd_tmm_tmm_tmm", Code.VEX_Tdpbssd_tmm_tmm_tmm },
				{ "Fnstdw_AX", Code.Fnstdw_AX },
				{ "Fnstsg_AX", Code.Fnstsg_AX },
				{ "Rdshr_rm32", Code.Rdshr_rm32 },
				{ "Wrshr_rm32", Code.Wrshr_rm32 },
				{ "Smint", Code.Smint },
				{ "Dmint", Code.Dmint },
				{ "Rdm", Code.Rdm },
				{ "Svdc_m80_Sreg", Code.Svdc_m80_Sreg },
				{ "Rsdc_Sreg_m80", Code.Rsdc_Sreg_m80 },
				{ "Svldt_m80", Code.Svldt_m80 },
				{ "Rsldt_m80", Code.Rsldt_m80 },
				{ "Svts_m80", Code.Svts_m80 },
				{ "Rsts_m80", Code.Rsts_m80 },
				{ "Smint_0F7E", Code.Smint_0F7E },
				{ "Bb0_reset", Code.Bb0_reset },
				{ "Bb1_reset", Code.Bb1_reset },
				{ "Cpu_write", Code.Cpu_write },
				{ "Cpu_read", Code.Cpu_read },
				{ "Altinst", Code.Altinst },
				{ "Paveb_mm_mmm64", Code.Paveb_mm_mmm64 },
				{ "Paddsiw_mm_mmm64", Code.Paddsiw_mm_mmm64 },
				{ "Pmagw_mm_mmm64", Code.Pmagw_mm_mmm64 },
				{ "Pdistib_mm_m64", Code.Pdistib_mm_m64 },
				{ "Psubsiw_mm_mmm64", Code.Psubsiw_mm_mmm64 },
				{ "Pmvzb_mm_m64", Code.Pmvzb_mm_m64 },
				{ "Pmulhrw_mm_mmm64", Code.Pmulhrw_mm_mmm64 },
				{ "Pmvnzb_mm_m64", Code.Pmvnzb_mm_m64 },
				{ "Pmvlzb_mm_m64", Code.Pmvlzb_mm_m64 },
				{ "Pmvgezb_mm_m64", Code.Pmvgezb_mm_m64 },
				{ "Pmulhriw_mm_mmm64", Code.Pmulhriw_mm_mmm64 },
				{ "Pmachriw_mm_m64", Code.Pmachriw_mm_m64 },
				{ "Cyrix_D9D7", Code.Cyrix_D9D7 },
				{ "Cyrix_D9E2", Code.Cyrix_D9E2 },
				{ "Ftstp", Code.Ftstp },
				{ "Cyrix_D9E7", Code.Cyrix_D9E7 },
				{ "Frint2", Code.Frint2 },
				{ "Frichop", Code.Frichop },
				{ "Cyrix_DED8", Code.Cyrix_DED8 },
				{ "Cyrix_DEDA", Code.Cyrix_DEDA },
				{ "Cyrix_DEDC", Code.Cyrix_DEDC },
				{ "Cyrix_DEDD", Code.Cyrix_DEDD },
				{ "Cyrix_DEDE", Code.Cyrix_DEDE },
				{ "Frinear", Code.Frinear },
				{ "Tdcall", Code.Tdcall },
				{ "Seamret", Code.Seamret },
				{ "Seamops", Code.Seamops },
				{ "Seamcall", Code.Seamcall },
				{ "Aesencwide128kl_m384", Code.Aesencwide128kl_m384 },
				{ "Aesdecwide128kl_m384", Code.Aesdecwide128kl_m384 },
				{ "Aesencwide256kl_m512", Code.Aesencwide256kl_m512 },
				{ "Aesdecwide256kl_m512", Code.Aesdecwide256kl_m512 },
				{ "Loadiwkey_xmm_xmm", Code.Loadiwkey_xmm_xmm },
				{ "Aesenc128kl_xmm_m384", Code.Aesenc128kl_xmm_m384 },
				{ "Aesdec128kl_xmm_m384", Code.Aesdec128kl_xmm_m384 },
				{ "Aesenc256kl_xmm_m512", Code.Aesenc256kl_xmm_m512 },
				{ "Aesdec256kl_xmm_m512", Code.Aesdec256kl_xmm_m512 },
				{ "Encodekey128_r32_r32", Code.Encodekey128_r32_r32 },
				{ "Encodekey256_r32_r32", Code.Encodekey256_r32_r32 },
				{ "VEX_Vbroadcastss_xmm_xmm", Code.VEX_Vbroadcastss_xmm_xmm },
				{ "VEX_Vbroadcastss_ymm_xmm", Code.VEX_Vbroadcastss_ymm_xmm },
				{ "VEX_Vbroadcastsd_ymm_xmm", Code.VEX_Vbroadcastsd_ymm_xmm },
				{ "Vmgexit_F2", Code.Vmgexit_F2 },
				{ "Uiret", Code.Uiret },
				{ "Testui", Code.Testui },
				{ "Clui", Code.Clui },
				{ "Stui", Code.Stui },
				{ "Senduipi_r64", Code.Senduipi_r64 },
				{ "Hreset_imm8", Code.Hreset_imm8 },
				{ "VEX_Vpdpbusd_xmm_xmm_xmmm128", Code.VEX_Vpdpbusd_xmm_xmm_xmmm128 },
				{ "VEX_Vpdpbusd_ymm_ymm_ymmm256", Code.VEX_Vpdpbusd_ymm_ymm_ymmm256 },
				{ "VEX_Vpdpbusds_xmm_xmm_xmmm128", Code.VEX_Vpdpbusds_xmm_xmm_xmmm128 },
				{ "VEX_Vpdpbusds_ymm_ymm_ymmm256", Code.VEX_Vpdpbusds_ymm_ymm_ymmm256 },
				{ "VEX_Vpdpwssd_xmm_xmm_xmmm128", Code.VEX_Vpdpwssd_xmm_xmm_xmmm128 },
				{ "VEX_Vpdpwssd_ymm_ymm_ymmm256", Code.VEX_Vpdpwssd_ymm_ymm_ymmm256 },
				{ "VEX_Vpdpwssds_xmm_xmm_xmmm128", Code.VEX_Vpdpwssds_xmm_xmm_xmmm128 },
				{ "VEX_Vpdpwssds_ymm_ymm_ymmm256", Code.VEX_Vpdpwssds_ymm_ymm_ymmm256 },
			};
			// GENERATOR-END: CodeHash
	}
}
